/*
 * arch/arm/mach-iop3xxx/aau.c
 *
 * Support functions for the Intel 803xx AAU channels. The AAU is
 * a HW XOR unit that is specifically designed for use with RAID5
 * applications.  This driver provides an interface that is used by
 * the Linux RAID stack.
 *
 * Original Author: Dave Jiang <dave.jiang@intel.com>
 *
 * Contributors: Deepak Saxena <dsaxena@mvista.com>
 *
 * Maintainer: Dave Jiang <dave.jiang@intel.com>
 *
 * Copyright (C) 2003 Intel Corporation
 * Copyright (C) 2003 MontaVista Software, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * TODO: Make it into a kernel device
 *      Use dma_map_xxx instead of virt_to_phy when there's support	
 * 		Replace list interface with a cache of objects?
 * 		Move interrupt handling to tasklet?
 * 		Better error handling (ERR IRQs and runtime errors)
 *
 * History:	(02/25/2003, DJ) Initial Creation
 *		(04/23/2003, DS) Cleanups, add support for iop310
 *      (10/20/2003, DJ) Moved to 2.6
 *
 */

#include <linux/types.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/mm.h>
#include <linux/module.h>

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/delay.h>
#include <asm/uaccess.h>
#include <asm/hardware.h>
#include <asm/cacheflush.h>

#include <asm/arch-iop3xx/xor.h>
/*
 * pick up local definitions
 */
#include "aau.h"

#undef DEBUG
#ifdef DEBUG
#define DPRINTK(s, args...) printk("IOP3xx AAU: " s "\n", ## args)
#define DENTER() DPRINTK("Entered...\n");
#define DEXIT() DPRINTK("Exited...\n");
#else
#define DPRINTK(s, args...)
#define DENTER()
#define DEXIT()
#endif

/* globals */
iop3xx_aau_t *aau;		/* AAU context */

/* static prototypes */
static int iop3xx_aau_chaininit(iop3xx_aau_t * aa);
static irqreturn_t iop3xx_aau_isr(int, void *, struct pt_regs *);

static inline void aau_queue_descriptor(sw_aau_t *sw_desc)
{
	unsigned long flags;

	// Set last NDAR to NULL
	sw_desc->next = NULL;
	sw_desc->status = AAU_ACTIVE;
	
//	local_fiq_disable();
	spin_lock_irqsave(&aau->process_lock, flags);
	list_add_tail(&sw_desc->link, &aau->process_q);

	aau->last_desc->aau_desc->NDAR = sw_desc->aau_phys;
	aau->last_desc = sw_desc;
	*(aau->regs.ACR) = AAU_ACR_AAU_ENABLE | AAU_ACR_CHAIN_RESUME;
	
	spin_unlock_irqrestore(&aau->process_lock, flags);
//	local_fiq_enable();

	DPRINTK("Going to sleep");
	wait_event_interruptible(sw_desc->wait,
				(sw_desc->status & (AAU_COMPLETE | AAU_ERROR)));
	DPRINTK("Woken up!");

	spin_lock_irqsave(&aau->process_lock, flags);
	list_del(&sw_desc->link);
	spin_unlock_irqrestore(&aau->process_lock, flags);

	spin_lock_irqsave(&aau->free_lock, flags);
	list_add_tail(&sw_desc->link, &aau->free_q);
	spin_unlock_irqrestore(&aau->free_lock, flags);
}

void
xor_iop3xxaau_2(unsigned int bytes, unsigned long *p1, unsigned long *p2)
{
	sw_aau_t *sw_desc = NULL;
	unsigned long flags;

	DENTER();

	spin_lock_irqsave(&aau->free_lock, flags);
	if (!list_empty(&aau->free_q)) {
		sw_desc = list_entry(aau->free_q.next, sw_aau_t, link);
		if (sw_desc->aau_phys != *aau->regs.ADAR) {
			list_del(&sw_desc->link);
		} else {
			// This should NEVER happen. We need to increase the amount
			// of descriptors if that is the case
			// TODO: put in something nice later to handle this!

			DPRINTK(KERN_ERR "out of AAU descriptors! Increase!\n");
			spin_unlock_irqrestore(&aau->free_lock, flags);
			BUG();
			return;
		}
	} else {
		// This should NEVER happen. We need to increase the amount
		// of descriptors if that is the case
		// TODO: put in something nice later to handle this!

		DPRINTK(KERN_ERR "out of AAU descriptors! Increase!\n");
		spin_unlock_irqrestore(&aau->free_lock, flags);
		BUG();
		return;
	}
	spin_unlock_irqrestore(&aau->free_lock, flags);

	// flush the cache to memory before AAU touches them
	dmac_flush_range((unsigned long) p1, (unsigned long) p1 + bytes);
	dmac_clean_range((unsigned long) p2, (unsigned long) p2 + bytes);

	// setting up AAU descriptors
	sw_desc->aau_desc->DAR = virt_to_phys(p1);
	sw_desc->aau_desc->SAR_G1[0] = virt_to_phys(p1);
	sw_desc->aau_desc->SAR_G1[1] = virt_to_phys(p2);
	sw_desc->aau_desc->BC = bytes;

	// write enable, direct fill block 1, xor the rest
	sw_desc->aau_desc->DC = AAU_ADCR_WREN | AAU_ADCR_XOR_2 | AAU_ADCR_IE;
	sw_desc->aau_desc->NDAR = 0;

	aau_queue_descriptor(sw_desc);

	if (sw_desc->status & AAU_ERROR) {
		printk(KERN_ERR "%s: IOP321 AAU operation failed!\n", __func__);
		BUG();
	}

	// invalidate the cache region to destination
//	dmac_inv_range((unsigned long) p1,
//  		       (unsigned long) p1 + bytes);
  		       
	DEXIT();
}

void
xor_iop3xxaau_x(unsigned int src_no, unsigned int bytes, void **bh_ptr)
{
	unsigned long flags;
	unsigned long *bptr = NULL;
	unsigned long DC = 0, EDC = 0;
	int i, blkcnt;
	sw_aau_t *sw_desc = NULL;

	DENTER();

	if(src_no <= 1)
	{
		printk(KERN_ERR "%s: need more than 1 src for XOR\n",
			__func__);
		BUG();
	}

	spin_lock_irqsave(&aau->free_lock, flags);
	if (!list_empty(&aau->free_q)) {
		sw_desc = list_entry(aau->free_q.next, sw_aau_t, link);
		if (sw_desc->aau_phys != *aau->regs.ADAR) {
			list_del(&sw_desc->link);
		} else {
			// This should NEVER happen. We need to increase the amount
			// of descriptors if that is the case
			// TODO: put in something nice later to handle this!

			DPRINTK(KERN_ERR "out of AAU descriptors! Increase!\n");
			spin_unlock_irqrestore(&aau->free_lock, flags);
			BUG();
			return;
		}
	} else {
		// This should NEVER happen. We need to increase the amount
		// of descriptors if that is the case
		// TODO: put in something nice later to handle this!

		DPRINTK(KERN_ERR "out of AAU descriptors! Increase!\n");
		spin_unlock_irqrestore(&aau->free_lock, flags);
		BUG();
		return;
	}
	spin_unlock_irqrestore(&aau->free_lock, flags);
	DPRINTK("XOR soft descriptor acquired\n");

	// flush the cache to memory before AAU touches them
	for(i = src_no-1; i >= 0; i--)
	{
		DPRINTK("flushing source %d\n", i);
		bptr = (unsigned long *)(bh_ptr[i]);
		dmac_clean_range((unsigned long)bptr,
				 (unsigned long)bptr + bytes);
	}

	// write enable, direct fill block 1, xor the rest
	DC = AAU_ADCR_WREN | AAU_ADCR_DIRFILL | AAU_ADCR_IE;

	DPRINTK("Setting up 0-3\n");
	blkcnt = (src_no >= 4) ? 3 : src_no - 1; 
	for(i = blkcnt; i >= 0; i--)
	{
		sw_desc->aau_desc->SAR_G1[i] = 
				virt_to_phys((unsigned long *)bh_ptr[i]);
		// set options	
		DC |= 1 << ((i * 3) + 1);
	}

	if(src_no > 4)
	{
		DPRINTK("Setting up 4-7\n");
		DC |= AAU_ADCR_4BLKS;
		blkcnt = (src_no >= 8) ? 7 : src_no - 1;
		for(i = blkcnt; i >= 4; i--)
		{
			sw_desc->aau_desc->SAR_G2[i-4] =
					virt_to_phys((unsigned long *)bh_ptr[i]);
			// set options
			DC |= 1 << ((i * 3) + 1);
		}

		if(src_no > 8)
		{	
			DPRINTK("Setting up 8-15\n");
			DC |= AAU_ADCR_12BLKS;
			EDC = 0;
			blkcnt = (src_no >= 16) ? 15 : src_no - 1;
			for(i = blkcnt; i >= 8; i--)
			{
				sw_desc->aau_desc->SAR_G3[i-8] = 
						virt_to_phys((unsigned long *)bh_ptr[i]);
				// set options
				EDC |= 1 << ((i * 3) - 23);
			}
			sw_desc->aau_desc->EDC0 = EDC;
			

			if(src_no > 16)
			{
				DPRINTK("Setting up 16-23\n");
				EDC = 0;
				DC |= AAU_ADCR_28BLKS;
				blkcnt = (src_no >= 24) ? 23 : src_no - 1;
				for(i = blkcnt; i >= 16; i--)
				{
					sw_desc->aau_desc->SAR_G4[i-16] = 
						virt_to_phys((unsigned long *)bh_ptr[i]);
					// set options
					EDC |= 1 << ((i * 3) - 47);
				}
				sw_desc->aau_desc->EDC1 = EDC;

				if(src_no > 24)
				{
					DPRINTK("Setting up 24-31\n");
					EDC = 0;
					blkcnt = src_no - 1;
					for(i = blkcnt; i >= 24; i--)
					{
						sw_desc->aau_desc->SAR_G5[i-24] =
							virt_to_phys((unsigned long *)bh_ptr[i]);
						// set options
						EDC |= 1 << ((i * 3) - 71);
					}
					sw_desc->aau_desc->EDC2 = EDC;
				} /* if > 24 */
				else
				{
					sw_desc->aau_desc->EDC2 = 0;
				}
			} /* if > 16 */
		} /* if > 8 */
	} /* if > 4*/	

	// setting up AAU descriptors
	DPRINTK("setting up rest of descriptor\n");
	sw_desc->aau_desc->NDAR = 0;
	sw_desc->aau_desc->DAR = 
			virt_to_phys((unsigned long *)bh_ptr[0]);
	sw_desc->aau_desc->BC = bytes;
	sw_desc->aau_desc->DC = DC;

	DPRINTK("queue descriptor\n");
	aau_queue_descriptor(sw_desc);

	DPRINTK("XOR complete\n");
	// check to see if AAU failed
	if (sw_desc->status & AAU_ERROR) {
		printk(KERN_ERR "%s: IOP321 AAU operation failed!\n", __func__);
		BUG();
	}

	DPRINTK("invalidate result in cache\n");
	// invalidate the cache region to destination
    bptr = (unsigned long *)(bh_ptr[0]);
	dmac_inv_range((unsigned long)bptr,
		       (unsigned long)bptr + bytes);
	DEXIT();
}
static irqreturn_t
iop3xx_aau_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	iop3xx_aau_t *aa = (iop3xx_aau_t *) dev_id;
	u32 csr = 0;
	sw_aau_t *sw_desc = NULL;
	int irq_thresh = AAU_IRQ_THRESH;
	struct list_head *entry;

	DENTER();

//	local_fiq_disable();
	
	csr = *(aa->regs.ASR);

	// make sure the interrupt is for us
	if (!(csr & (AAU_ASR_DONE_MASK | AAU_ASR_ERR_MASK))) {
		return IRQ_NONE;
	}
	// Does not support more than AAU chaining of multiple processes yet */
	while (csr && irq_thresh--) {
		// clear interrupts
		*(aa->regs.ASR) |= AAU_ASR_DONE_MASK | AAU_ASR_ERR_MASK;

		spin_lock(&aa->process_lock);

//		while (!list_empty(&aa->process_q)) {
		__list_for_each(entry, &aa->process_q) {
//			sw_desc = list_entry(aa->process_q.next, sw_aau_t, link);
			sw_desc = list_entry(entry, sw_aau_t, link);
			if ((sw_desc->aau_phys != *aa->regs.ADAR) || (!(*aa->regs.ASR & AAU_ASR_ACTIVE))) {
/* 				list_del(&sw_desc->link); */
				
				if (!(csr & AAU_ASR_ERR_MASK)) {
					sw_desc->status = AAU_COMPLETE;
				} else {
					sw_desc->status = AAU_ERROR;
				}

				wake_up_interruptible(&sw_desc->wait);

			} else {
				spin_unlock(&aa->process_lock);
//				local_fiq_enable();
				return IRQ_HANDLED;
			
			}
		}
		spin_unlock(&aa->process_lock);

		// re-read CSR to check for more interrupts
		csr = *(aa->regs.ASR);
	}			// end of while(csr)
	DEXIT();

//	local_fiq_enable();
	return IRQ_HANDLED;
}

int
iop3xx_aau_init(void)
{
	int i;
	sw_aau_t *sw_desc = NULL;
	void *desc = NULL;
	int err = 0;

	aau = kmalloc(sizeof(*aau), GFP_KERNEL);

	printk("Intel IOP3xx AAU RAID Copyright(c) 2003 Intel Corporation\n");

	/* init free stack */
	INIT_LIST_HEAD(&aau->free_q);
	INIT_LIST_HEAD(&aau->process_q);

	/* init free stack spinlock */
	spin_lock_init(&aau->free_lock);
	spin_lock_init(&aau->process_lock);

	aau->last_desc = NULL;

	/* pre-alloc AAU descriptors */
	for (i = 0; i < MAX_AAU_DESC; i++) {
		/* 
		 * we keep track of original address before alignment 
		 * adjust so we can free it later 
		 */
		sw_desc = kmalloc(sizeof (sw_aau_t), GFP_KERNEL);
		sw_desc->aau_virt = desc =
		    kmalloc((sizeof (aau_desc_t) + 0x20), GFP_KERNEL);
		memset(desc, 0, sizeof (aau_desc_t) + 0x20);

		/* 
		 * hardware descriptors must be aligned on an 
		 * 8-word boundary 
		 */
		desc = (aau_desc_t *) (((u32) desc & 0xffffffe0) + 0x20);
		// get the physical address
		sw_desc->aau_phys = (u32) virt_to_phys(desc);
		// remap it to non-cached
		sw_desc->aau_desc =
		    (aau_desc_t *) ioremap(sw_desc->aau_phys,
					   sizeof (aau_desc_t));

		sw_desc->status = AAU_FREE;
		init_waitqueue_head(&sw_desc->wait);

		/* put the descriptors on the free stack */
		list_add(&sw_desc->link, &aau->free_q);
	}

	/*
	 * Fill out AAU descriptor based on IOP type
	 */
#ifdef CONFIG_ARCH_IOP331
	if(iop_is_331()) {
                aau->regs.ACR = IOP331_AAU_ACR;
                aau->regs.ASR = IOP331_AAU_ASR;
                aau->regs.ADAR = IOP331_AAU_ADAR;
                aau->regs.ANDAR = IOP331_AAU_ANDAR;
                aau->regs.SAR[0] = IOP331_AAU_SAR1;
                aau->regs.SAR[1] = IOP331_AAU_SAR2;
                aau->regs.SAR[2] = IOP331_AAU_SAR3;
                aau->regs.SAR[3] = IOP331_AAU_SAR4;
                aau->regs.SAR[4] = IOP331_AAU_SAR5;
                aau->regs.SAR[5] = IOP331_AAU_SAR6;
                aau->regs.SAR[6] = IOP331_AAU_SAR7;
                aau->regs.SAR[7] = IOP331_AAU_SAR8;
                aau->regs.DAR = IOP331_AAU_DAR;
                aau->regs.BCR = IOP331_AAU_ABCR;
                aau->regs.DCR = IOP331_AAU_ADCR;
                aau->regs.EDCR0 = IOP331_AAU_EDCR0;
                aau->regs.EDCR1 = IOP331_AAU_EDCR1;
                aau->regs.EDCR2 = IOP331_AAU_EDCR2;

		aau->irq.EOT = IRQ_IOP331_AA_EOT;
		aau->irq.EOC = IRQ_IOP331_AA_EOC;
		aau->irq.ERR = IRQ_IOP331_AA_ERR;

		aau->device_id = "IOP331 AAU";
        aau->irq_id1 = "IOP331 AAU EOT";
        aau->irq_id2 = "IOP331 AAU EOC";
        aau->irq_id3 = "IOP331 AAU ERR";
	} 
#endif
#ifdef CONFIG_ARCH_IOP321
	if(iop_is_321()) {
                aau->regs.ACR = IOP321_AAU_ACR;
                aau->regs.ASR = IOP321_AAU_ASR;
                aau->regs.ADAR = IOP321_AAU_ADAR;
                aau->regs.ANDAR = IOP321_AAU_ANDAR;
                aau->regs.SAR[0] = IOP321_AAU_SAR1;
                aau->regs.SAR[1] = IOP321_AAU_SAR2;
                aau->regs.SAR[2] = IOP321_AAU_SAR3;
                aau->regs.SAR[3] = IOP321_AAU_SAR4;
                aau->regs.SAR[4] = IOP321_AAU_SAR5;
                aau->regs.SAR[5] = IOP321_AAU_SAR6;
                aau->regs.SAR[6] = IOP321_AAU_SAR7;
                aau->regs.SAR[7] = IOP321_AAU_SAR8;
                aau->regs.DAR = IOP321_AAU_DAR;
                aau->regs.BCR = IOP321_AAU_ABCR;
                aau->regs.DCR = IOP321_AAU_ADCR;
                aau->regs.EDCR0 = IOP321_AAU_EDCR0;
                aau->regs.EDCR1 = IOP321_AAU_EDCR1;
                aau->regs.EDCR2 = IOP321_AAU_EDCR2;

		aau->irq.EOT = IRQ_IOP321_AA_EOT;
		aau->irq.EOC = IRQ_IOP321_AA_EOC;
		aau->irq.ERR = IRQ_IOP321_AA_ERR;

		aau->device_id = "IOP321 HW XOR";
        aau->irq_id1 = "IOP321 AAU EOT";
        aau->irq_id2 = "IOP321 AAU EOC";
        aau->irq_id3 = "IOP321 AAU ERR";
	} 
#endif
	/* clear AAU control register */

	*(aau->regs.ACR) = AAU_ACR_CLEAR;
	*(aau->regs.ASR) |= (AAU_ASR_DONE_MASK | AAU_ASR_ERR_MASK);
	*(aau->regs.ANDAR) = 0;

	if(aau->irq.EOT >= 0)
		err = request_irq(aau->irq.EOT, iop3xx_aau_isr, SA_INTERRUPT,
				  aau->irq_id1, (void *)aau);
	if (err < 0) {
		printk(KERN_ERR "%s: unable to request IRQ %d for "
		       "AAU %d: %d\n", aau->device_id, aau->irq.EOT, 0, err);
		return -EBUSY;
	}

	if(aau->irq.EOC >= 0)
		err = request_irq(aau->irq.EOC, iop3xx_aau_isr, SA_INTERRUPT,
				  aau->irq_id2, (void *)aau);
	if (err < 0) {
		printk(KERN_ERR "%s: unable to request IRQ %d for "
		       "AAU %d: %d\n", aau->device_id, aau->irq.EOC, 0, err);
		return -EBUSY;
	}

	if(aau->irq.ERR >= 0)
		err = request_irq(aau->irq.ERR, iop3xx_aau_isr, SA_INTERRUPT,
				  aau->irq_id3, (void *)aau);
	if (err < 0) {
		printk(KERN_ERR "%s: unable to request IRQ %d for "
		       "AAU %d: %d\n", aau->device_id, aau->irq.ERR, 0, err);
		return -EBUSY;
	}

	err = iop3xx_aau_chaininit(aau);
	if (err < 0) {
		printk(KERN_ERR "unable to setup chaining\n");
		return -EIO;
	}

	DPRINTK("AAU init Done!\n");
	return 0;
}				/* end of iop3xx_aau_init() */

void
iop3xx_aau_exit(void)
{
	sw_aau_t *sw_desc = NULL;

	while (!list_empty(&aau->free_q)) {
		sw_desc = list_entry(aau->free_q.next, sw_aau_t, link);
		list_del(&sw_desc->link);
		iounmap(sw_desc->aau_desc);
		kfree(sw_desc->aau_virt);
		kfree((void *) sw_desc);
	}

	while (!list_empty(&aau->process_q)) {
		sw_desc = list_entry(aau->process_q.next, sw_aau_t, link);
		list_del(&sw_desc->link);
		iounmap(sw_desc->aau_desc);
		kfree(sw_desc->aau_virt);
		kfree((void *) sw_desc);
	}

	// get rid of the last dangle descriptor
	sw_desc = aau->last_desc;
	iounmap(sw_desc->aau_desc);
	kfree(sw_desc->aau_virt);
	kfree((void *) sw_desc);

	free_irq(aau->irq.EOC, (void *) &aau);
	free_irq(aau->irq.EOT, (void *) &aau);
	free_irq(aau->irq.ERR, (void *) &aau);
}

static int
iop3xx_aau_chaininit(iop3xx_aau_t * aa)
{
	unsigned long flags = 0;
	sw_aau_t *sw_desc = NULL;
	u32 csr = 0;

	spin_lock_irqsave(&aau->free_lock, flags);

	if (!list_empty(&aau->free_q)) {
		sw_desc = list_entry(aau->free_q.next, sw_aau_t, link);
		list_del(&sw_desc->link);
	} else {
		/* 
		 * This should NEVER happen. We need to increase the amount
		 * of descriptors if that is the case
		 *
		 * TODO: put in something nice later to handle this!
		 */
		spin_unlock_irqrestore(&aau->free_lock, flags);
		DPRINTK(KERN_ERR "out of AAU descriptors! Increase!\n");
		BUG();
		return -ENOMEM;
	}
	spin_unlock_irqrestore(&aau->free_lock, flags);

	DPRINTK("soft AAU desc acquired");

	/*
	 * Initialize HW chaining
	 */
	sw_desc->aau_desc->BC = 0;
	sw_desc->aau_desc->DC = 0;	// NULL command
	sw_desc->aau_desc->NDAR = 0;
	sw_desc->next = NULL;

	spin_lock_irqsave(&aau->process_lock, flags);
	list_add(&sw_desc->link, &aau->process_q);
	spin_unlock_irqrestore(&aau->process_lock, flags);

	// Do AAU stuff
	*(aau->regs.ASR) |= (AAU_ASR_DONE_MASK | AAU_ASR_ERR_MASK);
	*(aau->regs.ANDAR) = sw_desc->aau_phys;
	*(aau->regs.ACR) = AAU_ACR_AAU_ENABLE;

	DPRINTK("we are polling");
	csr = *(aau->regs.ASR);
	while (csr & AAU_ASR_ACTIVE) {
		csr = *(aau->regs.ASR);
	}

	DPRINTK("removing from processing queue");
	spin_lock_irqsave(&aau->process_lock, flags);
	list_del(&sw_desc->link);
	spin_unlock_irqrestore(&aau->process_lock, flags);

	sw_desc->status = AAU_FREE;
	spin_lock_irqsave(&aau->free_lock, flags);
	list_add_tail(&sw_desc->link, &aau->free_q);
	spin_unlock_irqrestore(&aau->free_lock, flags);

	// check to see if AAU failed
	if (!(csr & AAU_ASR_ERR_MASK)) {
		aau->last_desc = sw_desc;
	} else {
		aau->last_desc = NULL;
		return -EIO;
	}

	return 0;
}				/* end of iop3xx_aau_chaininit() */

module_init(iop3xx_aau_init);
module_exit(iop3xx_aau_exit);
MODULE_LICENSE(GPL);

