/*
 * arch/arm/mach-iop3xxx/ccnt.c
 *
 * Support functions for the Intel 803xx CCNT registers. The CCNT is
 * a cycle counter register that can be used as a high resolution 
 * time source by other applications. This code provides an interface 
 * to access this value. Further more, this code expands current 32 bit
 * cycle counter to a 64 bit value to overcome the overflow problem.
 *
 * Original Author: Ming Zhang <mingz@ele.uri.edu>
 *
 * Maintainer: Ming Zhang <mingz@ele.uri.edu>
 *
 * Copyright (C) 2004 Dragonfly Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * TODO: Integrate this code with PMU code;
 *
 * History:
 *      (02/10/2004, MZ) Initial Code, base on PMU code
 *	(02/12/2004, MZ) Export the CCNT counter via an ioctl interface
 *
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <linux/module.h>
#include <asm/atomic.h>
#include <asm/arch/irqs.h>
#include <asm/uaccess.h>

#define	PMU_ENABLE	0x001	/* Enable counters */

#ifndef CONFIG_ARCH_IOP331
#define INT_ENABLE	0x040	/* Interrupt enable */
#define CCNT_OVERFLOW	0x400	/* Clock counter overflow */
#define OVERFLOW_MASK	(CCNT_OVERFLOW)
#else	/* IOP331 */
#define CCNT_INT_ENABLE		0x01
#define INT_ENABLE	(CCNT_INT_ENABLE)
#define CCNT_OVERFLOW	0x01
#define OVERFLOW_MASK	(CCNT_OVERFLOW)
#endif

#ifdef CONFIG_ARCH_IOP321
#ifndef XSCALE_PMU_IRQ
#define XSCALE_PMU_IRQ	IRQ_IOP321_CORE_PMU
#endif
#endif
#ifdef CONFIG_ARCH_IOP331
#ifndef XSCALE_PMU_IRQ
#define XSCALE_PMU_IRQ	IRQ_IOP331_CORE_PMU
#endif
#endif

#define DEBUG
#ifdef DEBUG
#define  DBG(x...) printk(x)
#else
#define  DBG(x...)
#endif

#define CCNT_IOC_MAGIC		0xAC
#define CCNT_IOC_GETCCNT	_IOR(CCNT_IOC_MAGIC, 1, unsigned long long)

static atomic_t usage = ATOMIC_INIT(0);
volatile atomic_t ccnt_of = ATOMIC_INIT(0);

int ccnt_major = 0;

inline void __do_enable_ccnt(void)
{
	unsigned long pmnc = 0;

	/* enable intr, clear ccnt overflow bit, start CCNT
	*/	
#ifndef CONFIG_ARCH_IOP331
	pmnc = PMU_ENABLE | INT_ENABLE | OVERFLOW_MASK;
	asm ("mcr p14, 0, %0, c0, c0, 0" : : "r" (pmnc));
#else
	unsigned long flag;
	
	// disable PMON
	pmnc &= ~PMU_ENABLE;
	asm ("mcr p14, 0, %0, c0, c1, 0" : : "r" (pmnc));
	// read overflow flag register
	asm ("mrc p14, 0, %0, c5, c1, 0" : "=r" (flag));
	// clear overflow bits
	asm ("mcr p14, 0, %0, c5, c1, 0" : : "r" (flag));

	pmnc = PMU_ENABLE;
	asm ("mcr p14, 0, %0, c0, c1, 0" : : "r" (pmnc));
#endif
}
 
irqreturn_t ccnt_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	/* the only possible intr source is CCNT overflow */	
	atomic_inc(&ccnt_of);
	
	__do_enable_ccnt();
	
	return IRQ_HANDLED;
}

int enable_ccnt(void)
{
	int err = 0;

	if(atomic_read(&usage))
		return -EBUSY;

	err = request_irq(XSCALE_PMU_IRQ, ccnt_irq_handler, SA_INTERRUPT,
					  "XScale CCNT", NULL);
	if(err < 0)
	{
		printk(KERN_ERR "unable to request IRQ %d for XScale CCNT: %d\n",
			   XSCALE_PMU_IRQ, err);
		return err;
	}
	
	atomic_inc(&usage);

#ifdef CONFIG_ARCH_IOP331
	// enable interrupts for overflow
	asm ("mcr p14, 0, %0, c4, c1, 0" : : "r" (INT_ENABLE));
#endif

	__do_enable_ccnt();
	
	DBG("Enable CCNT succ\n");
	return 0;
}

unsigned long long do_getccnt(unsigned long *hi, unsigned long *lw)
{
	unsigned long long x;
	unsigned long t1, t2, of1, of2;
	
	of1 = atomic_read(&ccnt_of);	
#ifndef CONFIG_ARCH_IOP331
	asm volatile ("mrc p14, 0, %0, c1, c0, 0" : "=r" (t1));
#else
	asm volatile ("mrc p14, 0, %0, c1, c1, 0" : "=r" (t1));	
#endif

	of2 = atomic_read(&ccnt_of);
#ifndef CONFIG_ARCH_IOP331
	asm volatile ("mrc p14, 0, %0, c1, c0, 0" : "=r" (t2));
#else
	asm volatile ("mrc p14, 0, %0, c1, c1, 0" : "=r" (t2));
#endif

	/* check if overflow happen between accessing of2 and t2, if so, of1-t1 value is correct
	   else of2-t2 pair is correct and more accurate.
	 */
	if ((of2 == of1) && (t2 < t1)) {
		x = of1;	
		x = (x << 32) + t1;
		if (hi)	*hi = of1;
		if (lw)	*lw = t1;
	}
	else {
		x = of2;
		x = (x << 32) + t2;
		if (hi)	*hi = of2;
		if (lw)	*lw = t2;
	}
	return x;
}

#ifdef CONFIG_PROC_FS                                                                                                                                                                            
static int
proc_ccnt_read(char *page, char **start, off_t off, int count, int *eof,
                    void *data)
{
        char *p = page;
        int len;
	unsigned long hi, lw;
	unsigned long long x;
#ifdef DEBUG	
	struct timeval tv;
	do_gettimeofday(&tv);
#endif

	x = do_getccnt(&hi, &lw);
        p += sprintf(p, "%llu: %lu, %lu\n", x, hi, lw);
#ifdef DEBUG
	p += sprintf(p, "TV: %lu, %lu\n", tv.tv_sec, tv.tv_usec);
#endif        
	len = (p - page) - off;
        if (len < 0)
                len = 0;

        *eof = (len <= count) ? 1 : 0;
        *start = page + off;

        return len;
}
#endif

int ccnt_ioctl(struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
	int err = 0;
	unsigned long long tmp = 0;

	if (_IOC_TYPE(cmd) != CCNT_IOC_MAGIC) {
		DBG("CCNT: wrong magic\n");
		return -ENOTTY;
	}
	if (_IOC_DIR(cmd) & _IOC_READ)
		err = !access_ok(VERIFY_WRITE, (void *)arg, _IOC_SIZE(cmd));
	else if (_IOC_DIR(cmd) & _IOC_WRITE)
		err = !access_ok(VERIFY_READ, (void *)arg, _IOC_SIZE(cmd));
	if (err) {
		DBG("CCNT: invalid access\n");
		return -EFAULT;
	}

	switch (cmd) {
	case CCNT_IOC_GETCCNT:
		tmp = do_getccnt(NULL, NULL);
		err = __put_user(tmp, (unsigned long long *)arg);
		break;
	default:
		err = -EFAULT;
	}
	return err;
}

int ccnt(struct inode *inode, struct file *filp)
{
	return 0;
}

int ccnt_release(struct inode *inode, struct file *filp)
{
	return 0;
}

struct file_operations ccnt_fops = {
	llseek:     NULL,
	read:       NULL,
	write:      NULL,
	ioctl:      ccnt_ioctl,
	open:       ccnt,
	release:    ccnt_release,
};

void ccnt_clean(void)
{
	unregister_chrdev(ccnt_major, "ccnt");
	remove_proc_entry("cpu/ccnt", NULL);
	return;
}

static int __init ccnt_init(void)
{
#ifdef CONFIG_PROC_FS                                                                                                                                                        
        struct proc_dir_entry *res;
#endif
	int result;

	if (enable_ccnt())
		return -ENOMEM;

#ifdef CONFIG_PROC_FS                                                                                                                                                                            
	res = create_proc_read_entry("cpu/ccnt", S_IWUSR | S_IRUGO, NULL, proc_ccnt_read, NULL);
        if (!res) {
		printk("Fail to create ccnt\n");
                return -ENOMEM;
	}
#endif
	result = register_chrdev(ccnt_major, "ccnt", &ccnt_fops);
	if (result < 0) {
		printk(KERN_WARNING "ccnt: can't get major %d\n", ccnt_major);
		return result;
	}
	if (ccnt_major == 0) 
		ccnt_major = result;
        return 0;
}

fs_initcall(ccnt_init);

EXPORT_SYMBOL(do_getccnt);
