/**************************************************************************
 *
 * arch/arm/mach-iop3xx/dma.c
 *
 * Support functions for the Intel 803XX DMA channels.
 *
 * Authors: Dave Jiang (dave.jiang@intel.com)
 *          Carl Staelin (staelin@hpli69.hpli.hpl.hp.com)
 *          Deepak Saxena <dsaxena@mvista.com>
 *          David A. Griego (david.a.griego@intel.com)
 *
 * Copyright (C) 2003 MontaVista Software, Inc.
 * Copyright (C) 2003, 2004 Intel Corporation
 * Copyright (C) 2003 Hewlett Packard
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *
 *
 *  History:    (02/13/2003, DJ) Initial Creation
 *              (02/27/2003, DJ) Integrated Carl's DMA code
 *              (04/17/2003, DJ) Kernel memcpy seems to be working now
 *								 Reiserfs seems to act funny, ext2/3 works
 *								 fine
 *				(04/17/2003, DS) Major cleanup, full testing of boundary cases
 *				(04/18/2003, DS) Added support for user PCI mappings
 *				(06/04/2003, DJ) Merged in fixes from Deepak
 *              (02/27/2004, DJ) Ported to 2.6, add support for IOP331
 *				(04/06/2004, DG) Added copy*user functions for 2.6
 *
 *************************************************************************/
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/pagemap.h>
#include <linux/module.h>

#include <asm/io.h>
#include <asm/delay.h>
#include <asm/semaphore.h>
#include <asm/cacheflush.h>
#include <asm/uaccess.h>

#ifdef CONFIG_ARCH_IOP321
#include <asm/arch/iop321.h>
#include <asm/arch/iop321-irqs.h>
#endif

#ifdef CONFIG_ARCH_IOP331
#include <asm/arch/iop331.h>
#include <asm/arch/iop331-irqs.h>
#endif

/*
 * pick up local definitions
 */
#include <asm/arch/dma.h>
#include "dma.h"


#undef DEBUG
//#define DEBUG
#ifdef DEBUG
#define DPRINTK(s, args...) printk("IOP3XX_DMA: " s, ## args)
#else
#define DPRINTK(s, args...)
#endif

/* globals */
iop3xx_dma_t dma_chan[2];	/* DMA channels */

#ifdef CONFIG_IOP3XX_DMACOPY
int iop_memcpy = 0;
#endif

#ifdef CONFIG_ARCH_IOP321
#define IOP3XX_PCI_LOWER_MEM IOP321_PCI_LOWER_MEM_PA
#define IOP3XX_PCI_UPPER_MEM IOP321_PCI_UPPER_MEM_PA
#endif

#ifdef CONFIG_ARCH_IOP331
#define IOP3XX_PCI_LOWER_MEM IOP331_PCI_LOWER_MEM_PA
#define IOP3XX_PCI_UPPER_MEM IOP331_PCI_UPPER_MEM_PA
#endif

/* static prototypes */
static int iop3xx_dma_chaininit(iop3xx_dma_t * dma);
static irqreturn_t iop3xx_dma_isr(int, void *, struct pt_regs *);
ssize_t iop3xx_dma_cpy(void *src, int size, void *dest, int wait);

static struct proc_dir_entry *dma_proc_entry;
static int dma_hit, dma_miss, unaligned, kdma_hit;

static int dma_read_proc(char *, char **, off_t, int, int *, void *);

static inline u32 page_offset_xscale(u32 virt)
{
	return virt & (PAGE_SIZE - 1);
}

static inline u32 page_remainder(u32 virt)
{
	return PAGE_SIZE - page_offset_xscale(virt);
}

static inline int is_kernel_static(u32 virt)
{
	return ((virt >= PAGE_OFFSET) && (virt < (unsigned long)high_memory));
}

/* 
 * map a kernel virtual address or kernel logical address to a phys address
 */
static inline u32 physical_address(u32 virt)
{
	struct page *page;

	/* kernel static-mapped address */
	if (is_kernel_static((u32) virt)) {
		return __pa((u32) virt);
	}
	page = follow_page(current->mm->mmap, (u32) virt, 1);
	if (pfn_valid(page_to_pfn(page))) {
		return ((page_to_pfn(page) << PAGE_SHIFT) |
			page_offset_xscale((u32) virt & (PAGE_SIZE - 1)));
	} else {
		return 0;
	}
}

static int dma_read_proc(char *buf, char **start, off_t offset, int len,
			 int *eof, void *data)
{
	len = 0;

	len += sprintf(buf + len, "Number of DMA Hits: %d\n", dma_hit);
	len += sprintf(buf + len, "Number of DMA Misses: %d\n", dma_miss);
	len += sprintf(buf + len, "UnAligned Buffers: %d\n", unaligned);
	len += sprintf(buf + len, "Kernel DMA memcpy() hits: %d\n", kdma_hit);

	return len;
}

static irqreturn_t iop3xx_dma_isr(int irq, void *dev_id, struct pt_regs *regs)
{
	iop3xx_dma_t *dma = (iop3xx_dma_t *) dev_id;
	u32 csr = 0;
	sw_dma_t *sw_desc = NULL;

	csr = *(dma->regs.CSR);
	// **WARNING**
	// I'm only handling DMA channel 0 interrupts
	if (!(csr & (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK))) {
		return IRQ_NONE;
	}
	// Does not support more than DMA chaining of multiple processes yet */
	spin_lock(&dma->process_lock);
	sw_desc = list_entry(dma->process_q.next, sw_dma_t, link);
	sw_desc->status = csr;
	sw_desc->complete = 1;
	spin_unlock(&dma->process_lock);

	wake_up_interruptible(&dma->wait);
	// clear interrupts
	*(dma->regs.CSR) |= csr;
	return IRQ_HANDLED;
}

int iop3xx_dma_init(void)
{
	int i;
	int ctr;
	sw_dma_t *sw_desc = NULL;
	void *desc = NULL;
	int err = 0;
	unsigned long flags = 0;

	printk(KERN_INFO
	       "Intel IOP3XX DMA Copyright(c) 2004 Intel Corporation\n");
	DPRINTK("Initializing...\n");

	for (ctr = 0; ctr < MAX_IOP3XX_DMA_CHANNEL; ctr++) {
		/* init free stack */
		INIT_LIST_HEAD(&dma_chan[ctr].free_q);
		INIT_LIST_HEAD(&dma_chan[ctr].process_q);
		/* init free stack spinlock */
		dma_chan[ctr].chain_lock = RW_LOCK_UNLOCKED;
		dma_chan[ctr].free_lock = RW_LOCK_UNLOCKED;
		dma_chan[ctr].process_lock = RW_LOCK_UNLOCKED;
		spin_lock_init(&dma_chan[ctr].chan_lock);
		sema_init(&dma_chan[ctr].chan_sem, 1);

		init_waitqueue_head(&dma_chan[ctr].wait);

		/* pre-alloc DMA descriptors */
		for (i = 0; i < MAX_DMA_DESC; i++) {
			/* 
			 * we keep track of original address before alignment 
			 * adjust so we can free it later 
			 */
			sw_desc = kmalloc(sizeof(sw_dma_t), GFP_KERNEL);
			sw_desc->dma_virt = desc =
			    kmalloc((sizeof(dma_desc_t) + 0x20), GFP_KERNEL);
			memset(desc, 0, sizeof(dma_desc_t) + 0x20);

			/* 
			 * hardware descriptors must be aligned on an 
			 * 8-word boundary 
			 */
			desc =
			    (dma_desc_t *) (((u32) desc & 0xffffffe0) + 0x20);
			// get the physical address
			sw_desc->dma_phys = (u32) virt_to_phys(desc);
			// remap it to non-cached
			sw_desc->dma_desc =
			    (dma_desc_t *) ioremap(sw_desc->dma_phys,
						   sizeof(dma_desc_t));

#ifdef CONFIG_IOP3XX_CRC32C
			sw_desc->crc_src = kmalloc(sizeof(u32), GFP_KERNEL);
			sw_desc->phys_crc = (u32)virt_to_phys(sw_desc->crc_src);
#endif


			/* put the descriptors on the free stack */
			write_lock_irqsave(&dma_chan[ctr].free_lock, flags);
			list_add(&sw_desc->link, &dma_chan[ctr].free_q);
			write_unlock_irqrestore(&dma_chan[ctr].free_lock,
						flags);
		}

	}

#ifdef CONFIG_ARCH_IOP321
	// setting DMA channel 0
	dma_chan[0].regs.CCR = IOP321_DMA0_CCR;
	dma_chan[0].regs.CSR = IOP321_DMA0_CSR;
	dma_chan[0].regs.DAR = IOP321_DMA0_DAR;
	dma_chan[0].regs.NDAR = IOP321_DMA0_NDAR;
	dma_chan[0].regs.PADR = IOP321_DMA0_PADR;
	dma_chan[0].regs.PUADR = IOP321_DMA0_PUADR;
	dma_chan[0].regs.LADR = IOP321_DMA0_LADR;
	dma_chan[0].regs.BCR = IOP321_DMA0_BCR;
	dma_chan[0].regs.DCR = IOP321_DMA0_DCR;

	/* set the IRQ */
	dma_chan[0].irq.EOT = IRQ_IOP321_DMA0_EOT;
	dma_chan[0].irq.EOC = IRQ_IOP321_DMA0_EOC;
	dma_chan[0].irq.ERR = IRQ_IOP321_DMA0_ERR;

	dma_chan[0].device_id = "IOP321 DMA 0";
	dma_chan[0].irq_id1 = "IOP321 DMA 0 EOT";
	dma_chan[0].irq_id2 = "IOP321 DMA 0 E0C";
	dma_chan[0].irq_id3 = "IOP321 DMA 0 ERR";

	// setting DMA channel 1
	dma_chan[1].regs.CCR = IOP321_DMA1_CCR;
	dma_chan[1].regs.CSR = IOP321_DMA1_CSR;
	dma_chan[1].regs.DAR = IOP321_DMA1_DAR;
	dma_chan[1].regs.NDAR = IOP321_DMA1_NDAR;
	dma_chan[1].regs.PADR = IOP321_DMA1_PADR;
	dma_chan[1].regs.PUADR = IOP321_DMA1_PUADR;
	dma_chan[1].regs.LADR = IOP321_DMA1_LADR;
	dma_chan[1].regs.BCR = IOP321_DMA1_BCR;
	dma_chan[1].regs.DCR = IOP321_DMA1_DCR;

	/* set the IRQ */
	dma_chan[1].irq.EOT = IRQ_IOP321_DMA1_EOT;
	dma_chan[1].irq.EOC = IRQ_IOP321_DMA1_EOC;
	dma_chan[1].irq.ERR = IRQ_IOP321_DMA1_ERR;

	dma_chan[1].device_id = "IOP321 DMA 1";
	dma_chan[1].irq_id1 = "IOP321 DMA 1 EOT";
	dma_chan[1].irq_id2 = "IOP321 DMA 1 E0C";
	dma_chan[1].irq_id3 = "IOP321 DMA 1 ERR";

#endif

#ifdef CONFIG_ARCH_IOP331
	// setting DMA channel 0
	dma_chan[0].regs.CCR = IOP331_DMA0_CCR;
	dma_chan[0].regs.CSR = IOP331_DMA0_CSR;
	dma_chan[0].regs.DAR = IOP331_DMA0_DAR;
	dma_chan[0].regs.NDAR = IOP331_DMA0_NDAR;
	dma_chan[0].regs.PADR = IOP331_DMA0_PADR;
	dma_chan[0].regs.PUADR = IOP331_DMA0_PUADR;
	dma_chan[0].regs.LADR = IOP331_DMA0_LADR;
	dma_chan[0].regs.BCR = IOP331_DMA0_BCR;
	dma_chan[0].regs.DCR = IOP331_DMA0_DCR;

	/* set the IRQ */
	dma_chan[0].irq.EOT = IRQ_IOP331_DMA0_EOT;
	dma_chan[0].irq.EOC = IRQ_IOP331_DMA0_EOC;
	dma_chan[0].irq.ERR = IRQ_IOP331_DMA0_ERR;

	dma_chan[0].device_id = "IOP331 DMA 0";
	dma_chan[0].irq_id1 = "IOP331 DMA 0 EOT";
	dma_chan[0].irq_id2 = "IOP331 DMA 0 E0C";
	dma_chan[0].irq_id3 = "IOP331 DMA 0 ERR";

	// setting DMA channel 1
	dma_chan[1].regs.CCR = IOP331_DMA1_CCR;
	dma_chan[1].regs.CSR = IOP331_DMA1_CSR;
	dma_chan[1].regs.DAR = IOP331_DMA1_DAR;
	dma_chan[1].regs.NDAR = IOP331_DMA1_NDAR;
	dma_chan[1].regs.PADR = IOP331_DMA1_PADR;
	dma_chan[1].regs.PUADR = IOP331_DMA1_PUADR;
	dma_chan[1].regs.LADR = IOP331_DMA1_LADR;
	dma_chan[1].regs.BCR = IOP331_DMA1_BCR;
	dma_chan[1].regs.DCR = IOP331_DMA1_DCR;

	/* set the IRQ */
	dma_chan[1].irq.EOT = IRQ_IOP331_DMA1_EOT;
	dma_chan[1].irq.EOC = IRQ_IOP331_DMA1_EOC;
	dma_chan[1].irq.ERR = IRQ_IOP331_DMA1_ERR;

	dma_chan[1].device_id = "IOP331 DMA 1";
	dma_chan[1].irq_id1 = "IOP331 DMA 1 EOT";
	dma_chan[1].irq_id2 = "IOP331 DMA 1 E0C";
	dma_chan[1].irq_id3 = "IOP331 DMA 1 ERR";
#endif

	for (ctr = 0; ctr < MAX_IOP3XX_DMA_CHANNEL; ctr++) {

		/* clear DMA channel control register */

		*(dma_chan[ctr].regs.CCR) = DMA_CCR_CLEAR;
		*(dma_chan[ctr].regs.CSR) |=
		    (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK);
		*(dma_chan[ctr].regs.NDAR) = 0;

		err =
		    request_irq(dma_chan[ctr].irq.EOT, iop3xx_dma_isr,
				SA_INTERRUPT, dma_chan[ctr].irq_id1,
				(void *)&dma_chan[ctr]);
		if (err < 0) {
			printk(KERN_ERR "%s: unable to request IRQ %d for "
			       "DMA channel %d: %d\n", dma_chan[ctr].irq_id1,
			       dma_chan[ctr].irq.EOT, 0, err);
			return -EBUSY;
		}

		err =
		    request_irq(dma_chan[ctr].irq.EOC, iop3xx_dma_isr,
				SA_INTERRUPT, dma_chan[ctr].irq_id2,
				(void *)&dma_chan[ctr]);
		if (err < 0) {
			printk(KERN_ERR "%s: unable to request IRQ %d for "
			       "DMA channel %d: %d\n", dma_chan[ctr].irq_id2,
			       dma_chan[ctr].irq.EOC, 0, err);
			return -EBUSY;
		}

		err =
		    request_irq(dma_chan[ctr].irq.ERR, iop3xx_dma_isr,
				SA_INTERRUPT, dma_chan[ctr].irq_id3,
				(void *)&dma_chan[ctr]);
		if (err < 0) {
			printk(KERN_ERR "%s: unable to request IRQ %d for "
			       "DMA channel %d: %d\n", dma_chan[ctr].irq_id3,
			       dma_chan[ctr].irq.ERR, 0, err);
			return -EBUSY;
		}

		err = iop3xx_dma_chaininit(&dma_chan[ctr]);
		if (err < 0) {
			printk(KERN_ERR "unable to setup chaining\n");
			return -EIO;
		}
	}

#ifdef CONFIG_PROC_FS
	dma_proc_entry =
	    create_proc_entry("iop3xx_dmacopy", S_IFREG | S_IRUGO, 0);
	dma_proc_entry->read_proc = dma_read_proc;
	dma_proc_entry->write_proc = NULL;
	dma_proc_entry->nlink = 1;
#endif

#ifdef CONFIG_IOP3XX_DMACOPY
	iop_memcpy = 1;
#endif

	DPRINTK("DMA init Done!\n");
	return 0;
}				/* end of iop3xx_dma_init() */

void iop3xx_dma_exit(void)
{
	sw_dma_t *sw_desc = NULL;
	int ctr;

	for (ctr = 0; ctr < MAX_IOP3XX_DMA_CHANNEL; ctr++) {

		while (!list_empty(&dma_chan[ctr].free_q)) {
			sw_desc =
			    list_entry(dma_chan[ctr].free_q.next, sw_dma_t,
				       link);
			list_del(&sw_desc->link);
			iounmap(sw_desc->dma_desc);
			kfree(sw_desc->dma_virt);
#ifdef CONFIG_IOP3XX_CRC32C
			if(sw_desc->crc_src)
			{
				kfree(sw_desc->crc_src);
			}
#endif

			kfree((void *)sw_desc);
		}

		while (!list_empty(&dma_chan[ctr].process_q)) {
			sw_desc =
			    list_entry(dma_chan[ctr].process_q.next, sw_dma_t,
				       link);
			list_del(&sw_desc->link);
			iounmap(sw_desc->dma_desc);
			kfree(sw_desc->dma_virt);
#ifdef CONFIG_IOP3XX_CRC32C
			if(sw_desc->crc_src)
			{
				kfree(sw_desc->crc_src);
			}
#endif
			kfree((void *)sw_desc);
		}

		// get rid of the last dangle descriptor
		sw_desc = dma_chan[ctr].last_desc;
		iounmap(sw_desc->dma_desc);
		kfree(sw_desc->dma_virt);
		kfree((void *)sw_desc);

		free_irq(dma_chan[ctr].irq.EOC, (void *)&dma_chan[ctr]);
		free_irq(dma_chan[ctr].irq.EOT, (void *)&dma_chan[ctr]);
		free_irq(dma_chan[ctr].irq.ERR, (void *)&dma_chan[ctr]);
	}
}

static int iop3xx_dma_chaininit(iop3xx_dma_t * dma)
{
	unsigned long cflags = 0;
	sw_dma_t *sw_desc = NULL;
	u32 csr = 0;

	// this is during initialization, I will spinlock the entire block
	spin_lock_irqsave(&dma->chan_lock, cflags);
	if (!list_empty(&dma->free_q)) {
		sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
		list_del(&sw_desc->link);
	} else {
		// This should NEVER happen. We need to increase the amount
		// of descriptors if that is the case
		// XXX TODO: put in something nice later to handle this!
		spin_unlock_irqrestore(&dma->chan_lock, cflags);
		DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
		BUG();
		return -ENOMEM;
	}

	DPRINTK("soft DMA desc acquired");

	// Setting byte count to 0 doesn't do anything except loading the 
	// descriptor
	sw_desc->dma_desc->BC = 0;
	sw_desc->dma_desc->DC = DMA_DCR_MTM;
	sw_desc->dma_desc->NDAR = 0;
	sw_desc->next = NULL;

	list_add(&sw_desc->link, &dma->process_q);

	// Do DMA stuff
	*(dma->regs.CSR) |= (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK);
	*(dma->regs.NDAR) = sw_desc->dma_phys;
	*(dma->regs.CCR) = DMA_CCR_CHANNEL_ENABLE;
	DPRINTK("we are polling");
	csr = *(dma->regs.CSR);
	while (csr & DMA_CSR_CH_ACTIVE) {
		csr = *(dma->regs.CSR);
	}

	DPRINTK("removing from processing queue");
	list_del(&sw_desc->link);

	// check to see if DMA failed
	if (!(csr & DMA_CSR_ERR_MASK)) {
		dma->last_desc = sw_desc;
	} else {
		dma->last_desc = NULL;
		sw_desc->status = 0;
		sw_desc->complete = 0;
		list_add_tail(&sw_desc->link, &dma->free_q);

		spin_unlock_irqrestore(&dma->chan_lock, cflags);
		return -EIO;
	}
	spin_unlock_irqrestore(&dma->chan_lock, cflags);

	return 0;
}				/* end of iop3xx_dma_chaininit() */

#ifdef CONFIG_IOP3XX_DMACOPY
/*=======================================================================*/
/*  Procedure:  dma_memcpy()                                             */
/*                                                                       */
/*  Description:    DMA-based in-kernel memcpy.                          */
/*                                                                       */
/*  Parameters:  to: destination address                                 */
/*               from: source address                                    */
/*               n: number of bytes to transfer                          */
/*                                                                       */
/*  Returns:     void*: to                                               */
/*                                                                       */
/*  Notes/Assumptions:                                                   */
/*              Assumes that kernel physical memory is contiguous, i.e., */
/*              the physical addresses of contiguous virtual addresses   */
/*              are also contiguous.                                     */
/*              Assumes that kernel memory doesn't get paged.            */
/*              The DMA is polling                                       */
/*                                                                       */
/*  History:    Carl Staelin 1/27/03  Initial Creation                   */
/*              Dave Jiang   2/22/03  Attempted to DMA chaining          */
/*=======================================================================*/
void *dma_memcpy(void *to, const void *from, __kernel_size_t n)
{
	iop3xx_dma_t *dma;
	sw_dma_t *sw_desc = NULL;
	sw_dma_t *last_desc = NULL;
	unsigned long cflags = 0, flags = 0;
	u32 csr = 0;
	u32 unaligned_to, unaligned_from;
	void *orig_to = to;
	u32 to_pa, from_pa;
	int ua = 0;

	DPRINTK("dma_memcpy(0x%x, 0x%x, %lu): entering\n", (u32) to, (u32) from,
		(unsigned long)n);

	if (!(is_kernel_static((u32) to) && is_kernel_static((u32) from))) {
		DPRINTK("dma_memcpy(0x%x, 0x%x, %lu): falling back to memcpy\n",
			(u32) to, (u32) from, (unsigned long)n);
		dma_miss++;
		return asm_memcpy(to, from, n);
	}

	/*
	 * We can only handled completely cache-aligned transactions
	 * with the DMA engine.  Source and Dst must be cache-line
	 * aligned AND the length must be a multiple of the cache-line.
	 */

	to_pa = virt_to_phys(to);
	from_pa = virt_to_phys((void*)from);

	if (((to_pa + n > from_pa) && (to_pa < from_pa)) ||
	    ((from_pa < to_pa) && (from_pa + n > to_pa))) {
		DPRINTK
		    ("overlapping copy region (0x%x, 0x%x, %lu), falling back\n",
		     to_pa, from_pa, (unsigned long)n);
		dma_miss++;
		return asm_memcpy(to, from, n);
	}
	/*
	 * Ok, start addr is not cache line-aligned, so we need to make it so.
	 */
	unaligned_to = (u32) to & 31;
	unaligned_from = (u32) from & 31;;
	if (unaligned_to | unaligned_from) {
		ua++;
		if (unaligned_from > unaligned_to) {
			asm_memcpy(to, from, 32 - unaligned_to);
			to = (void *)((u32)to + 32 - unaligned_to);
			from = (void *)((u32)from + 32 - unaligned_to);
			n -= 32 - unaligned_to;
		} else {
			asm_memcpy(to, from, 32 - unaligned_from);
			to = (void *)((u32)to + 32 - unaligned_from);
			from = (void *)((u32)from + 32 - unaligned_from);
			n -= 32 - unaligned_from;
		}
	}

	/*
	 * Ok, we're aligned at the top, now let's check the end
	 * of the buffer and align that. After this we should have
	 * a block that is a multiple of cache line size.
	 */
	unaligned_to = ((u32) to + n) & 31;
	unaligned_from = ((u32) from + n) & 31;;
	if (unaligned_to | unaligned_from) {
		ua++;
		if (unaligned_to > unaligned_from) {
			u32 tmp_to = (u32) to + n - unaligned_to;
			u32 tmp_from = (u32) from + n - unaligned_to;

			asm_memcpy((void *)tmp_to, (void *)tmp_from,
				   unaligned_to);

			n -= unaligned_to;
		} else {
			u32 tmp_to = (u32) to + n - unaligned_from;
			u32 tmp_from = (u32) from + n - unaligned_from;

			asm_memcpy((void *)tmp_to, (void *)tmp_from,
				   unaligned_from);

			n -= unaligned_from;
		}
	}

	/*
	 * OK! We should now be fully aligned on both ends. 
	 */
	dma = &dma_chan[0];

	write_lock_irqsave(&dma->free_lock, flags);
	if (!list_empty(&dma->free_q)) {
		sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
		list_del(&sw_desc->link);
	} else {
		// This should NEVER happen. We need to increase the amount
		// of descriptors if that is the case
		// XXX TODO: put in something nice later to handle this!

		write_unlock_irqrestore(&dma->free_lock, flags);
		DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
		return asm_memcpy(to, from, n);
	}
	write_unlock_irqrestore(&dma->free_lock, flags);

	/* Ensure that the cache is clean */
	dmac_clean_range((unsigned long)from, (unsigned long)from + n);
	dmac_inv_range((unsigned long)to, (unsigned long)to + n);

	sw_desc->dma_desc->LADR = virt_to_phys(to);
	sw_desc->dma_desc->PADR = virt_to_phys((void*)from);
	sw_desc->dma_desc->PUADR = 0;
	sw_desc->buf_size = sw_desc->dma_desc->BC = n;
	sw_desc->dma_desc->DC = DMA_DCR_MTM;

	// Set last NDAR to NULL
	sw_desc->next = NULL;
	sw_desc->dma_desc->NDAR = 0;

	// this effectively prevents context switching until I'm done
	// with the DMA channel
	spin_lock_irqsave(&dma->chan_lock, cflags);
	dma->last_desc->dma_desc->NDAR = sw_desc->dma_phys;
	mb();
	*(dma->regs.CCR) = DMA_CCR_CHANNEL_ENABLE | DMA_CCR_CHAIN_RESUME;
	mb();

	csr = *(dma->regs.CSR);
	while (csr & DMA_CSR_CH_ACTIVE) {
		udelay(1);
		csr = *(dma->regs.CSR);
		rmb();
	}
	// put the last in the free queue
	last_desc = dma->last_desc;
	// hookup the new last descriptor for chaining
	dma->last_desc = sw_desc;
	spin_unlock_irqrestore(&dma->chan_lock, cflags);

	// clean up last descriptor
	last_desc->status = 0;
	last_desc->complete = 0;
	last_desc->next = 0;

	write_lock_irqsave(&dma->free_lock, flags);
	list_add_tail(&last_desc->link, &dma->free_q);
	write_unlock_irqrestore(&dma->free_lock, flags);

	// check to see if DMA failed
	if (csr & DMA_CSR_ERR_MASK) {
		*(dma->regs.CSR) |= (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK);
		printk(KERN_ERR "%s: DMA engine error\n", __func__);
		BUG();
		return asm_memcpy(to, from, n);
	}

	dma_hit++;
	kdma_hit++;
	if (ua)
		unaligned++;

	return orig_to;
}
#endif

#ifdef CONFIG_IOP3XX_DMA_COPYTOUSER
/*=======================================================================*/
/*  Procedure:  dma_copy_to_user()                                       */
/*                                                                       */
/*  Description:    DMA-based copy_to_user.                              */
/*                                                                       */
/*  Parameters:  to: destination address                                 */
/*               from: source address                                    */
/*               n: number of bytes to transfer                          */
/*                                                                       */
/*  Returns:     unsigned long: number of bytes NOT copied               */
/*                                                                       */
/*  Notes/Assumptions:                                                   */
/*              Assumes that kernel physical memory is contiguous, i.e., */
/*              the physical addresses of contiguous virtual addresses   */
/*              are also contiguous.                                     */
/*              Assumes that kernel memory doesn't get paged.            */
/*              Assumes that to/from memory regions cannot overlap       */
/*                                                                       */
/*  History:    Carl Staelin 1/27/03  Initial Creation                   */
/*              Dave Jiang   2/22/03  Attemped to DMA chaining           */
/*                                    (data corrupt with full chain)     */
/*                                    back to Carl's way for now         */
/*=======================================================================*/
unsigned long dma_copy_to_user(void *to, const void *from, unsigned long n)
{
	u32 chunk;
	u32 from_chunk;
	u32 to_chunk;
	u32 to_pa = 0;
	iop3xx_dma_t *dma;
	sw_dma_t *sw_desc = NULL;
	unsigned long cflags = 0;
	u32 csr = 0;
	u32 unaligned_to, unaligned_from;
	void *u, *k;
	int ua = 0;

	DPRINTK("dma_copy_to_user(%#10x, 0x%#10x, %lu): entering\n", (u32) to,
		(u32) from, n);
	if (((u32) to > PAGE_OFFSET) && ((u32) from > PAGE_OFFSET)) {
		DPRINTK("dma_copy_to_user(0x%x, 0x%x, %lu): kernel memcpy\n",
			(u32) to, (u32) from, (unsigned long)n);
		memcpy(to, from, n);
		return 0;
	}

	if(!is_kernel_static((u32) from))
	{
		DPRINTK("dma_copy_to_user(0x%x, 0x%x, %lu): vmalloc memcpy\n",
			(u32) to, (u32) from, (unsigned long)n);
		dma_miss++;
		return __arch_copy_to_user(to, from, n);
	}

	to_pa = physical_address((u32) to);
	DPRINTK("  physaddr(to) = %#010x\n", to_pa);
	/*
	 * We _only_ support mem -> mem and mem -> pci copies
	 */
	if (to_pa < PHYS_OFFSET &&
	    ((to_pa < IOP3XX_PCI_LOWER_MEM) || (to_pa > IOP3XX_PCI_UPPER_MEM)))
	{
		dma_miss++;
		return __arch_copy_to_user(to, from, n);
	}

	/*
	 * Walk the address range forcing page faults.  We must do writes
	 * instead of reads in case we have COW pages and need to get
	 * them re-mapped to a new address range.
	 */
	for (u = to, k = (void*)from; u < to + n; u += PAGE_SIZE, k += PAGE_SIZE)
		put_user(*(char *)k, (char *)u);

	/*
	 * Ok, start addr is not cache line-aligned, so we need to make it so.
	 */
	unaligned_to = (u32) to & 31;
	unaligned_from = (u32) from & 31;;
	if (unaligned_to | unaligned_from) {
		ua++;
		DPRINTK("Fixing up starting address\n");
		if (unaligned_from > unaligned_to) {
			DPRINTK("unaligned_from larger, copying %d bytes\n",
				32 - unaligned_to);
			__arch_copy_to_user(to, from, 32 - unaligned_to);
			to = (void *)((u32)to + 32 - unaligned_to);
			from = (void *)((u32)from + 32 - unaligned_to);
			n -= (32 - unaligned_to);
		} else {
			DPRINTK("unaligned_to larger, copying %d bytes\n",
				32 - unaligned_from);
			__arch_copy_to_user(to, from, 32 - unaligned_from);
			to = (void *)((u32)to + 32 - unaligned_from);
			from = (void *)((u32)from + 32 - unaligned_from);
			n -= (32 - unaligned_from);
		}
	}

	/*
	 * Ok, we're aligned at the top, now let's check the end
	 * of the buffer and align that. After this we should have
	 * a block that is a multiple of cache line size.
	 */
	unaligned_to = ((u32) to + n) & 31;
	unaligned_from = ((u32) from + n) & 31;;
	if (unaligned_to | unaligned_from) {
		ua++;
		DPRINTK("Fixing ending alignment\n");
		if (unaligned_to > unaligned_from) {
			u32 tmp_to = (u32) to + (n - unaligned_to);
			u32 tmp_from = (u32) from + (n - unaligned_to);

			DPRINTK("unaligned_to: Copying %d bytes\n",
				unaligned_to);

			__arch_copy_to_user((void *)tmp_to, (void *)tmp_from,
					    unaligned_to);

			n -= unaligned_to;
		} else {
			u32 tmp_to = (u32) to + (n - unaligned_from);
			u32 tmp_from = (u32) from + (n - unaligned_from);

			DPRINTK("unaligned_from: Copying %d bytes\n",
				unaligned_from);

			__arch_copy_to_user((void *)tmp_to, (void *)tmp_from,
					    unaligned_from);

			n -= unaligned_from;
		}
	}

	/*
	 * OK! We should now be fully aligned on both ends. 
	 */
	DPRINTK("Done fixup...to: %#010x from: %#010x len: %d\n", to, from, n);
	dmac_inv_range((u32) to, (u32) to + n);
	dmac_clean_range((u32) from, (u32) from + n);
	dma = &dma_chan[0];
	/* lock page tables */
	spin_lock_irqsave(&current->mm->page_table_lock, cflags);

	// TODO: !THIS DOES NOT MAP THE PAGES IN!

	from_chunk = 0;
	to_chunk = 0;

	DPRINTK("USING DMA: %#010x, %#010x, %d\n", to, from, n);

	while (n > 0) {
		if (from_chunk == 0) {
			/* kernel logical address ==> contiguous */
			if (is_kernel_static((u32) from)) {
				from_chunk = n;
			} else {
				/* virtual address */
				from_chunk = page_remainder((u32) from);
			}
		}

		if (to_chunk == 0) {
			to_chunk = page_remainder((u32) to);
			to_pa = physical_address((u32) to);

			/*
			 * This should NOT happen...
			 */
			if (!to_pa) {
				DPRINTK("dma_copy_to_user: no physical address for 0x%x,"
						" falling back! **** \n",
				     (u32) to);
				spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
				dma_miss++;
				return __arch_copy_to_user(to, from, n);
			}
		}
		chunk = ((to_chunk < from_chunk) ? to_chunk : from_chunk);
		if (n < chunk) {
			chunk = n;
		}

		if (chunk == 0) {
			break;
		}

		/*
		 * TODO: Since we don't chain or anything fancy, we're
		 * don't need the overhead of the list. Just write directly
		 * to the register and then wait on the interrupt.
		 */
		if (!list_empty(&dma->free_q)) {
			sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
			list_del(&sw_desc->link);
		} else {
			// This should NEVER happen. We need to increase the amount
			// of descriptors if that is the case
			// XXX TODO: put in something nice later to handle this!

			spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
			DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
			BUG();
			return n;
		}

		/*
		 *  Is this a mem -> mem or mem -> PCI copy?
		 */
		sw_desc->virt_src = (void*)from;
		sw_desc->virt_dest = NULL;	/* don't have it */
		sw_desc->dma_desc->PUADR = 0;
		sw_desc->dma_desc->NDAR = 0;

		if (to_pa >= PHYS_OFFSET) {
			DPRINTK("   Mem->Mem\n");
			sw_desc->dma_desc->LADR = sw_desc->phys_dest = to_pa;
			sw_desc->dma_desc->PADR = sw_desc->phys_src =
			    physical_address((u32) from);
			sw_desc->buf_size = sw_desc->dma_desc->BC = chunk;
			sw_desc->dma_desc->DC = DMA_DCR_MTM;
		} else {
			DPRINTK("   Mem->PCI\n");
			sw_desc->dma_desc->PADR = sw_desc->phys_dest = to_pa;
			sw_desc->dma_desc->LADR = sw_desc->phys_src =
			    physical_address((u32) from);
			sw_desc->buf_size = sw_desc->dma_desc->BC = chunk;
			sw_desc->dma_desc->DC = DMA_DCR_PCI_MWI;
		}

		// Set last NDAR to NULL
		sw_desc->next = NULL;

		dma->last_desc->dma_desc->NDAR = sw_desc->dma_phys;
		mb();
		*(dma->regs.CCR) = DMA_CCR_CHANNEL_ENABLE | DMA_CCR_CHAIN_RESUME;
		mb();

		// do some work while waiting
		from += chunk;
		to += chunk;
		n -= chunk;
		to_chunk -= chunk;
		from_chunk -= chunk;

		// busy wait on channel until done
		csr = *(dma->regs.CSR);
		while (csr & DMA_CSR_CH_ACTIVE) {
			udelay(1);
			rmb();
			csr = *(dma->regs.CSR);
		}

		// put the last in the free queue
		dma->last_desc->status = 0;
		dma->last_desc->complete = 0;
		dma->last_desc->next = 0;
		list_add_tail(&dma->last_desc->link, &dma->free_q);

		// hookup the new last descriptor for chaining
		sw_desc->status = 0;
		sw_desc->complete = 0;
		dma->last_desc = sw_desc;

		// check to see if DMA failed...if so, we die
		// since something is VERY wrong.
		if (csr & DMA_CSR_ERR_MASK) {
			spin_unlock_irqrestore(&current->mm->page_table_lock,
					       cflags);
			printk(KERN_ERR "copy to DMA failed\n");
			BUG();
			return n;
		}
	}

	spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
	dma_hit++;
	if(ua)
		unaligned++;

	DPRINTK("dma_copy_to_user(0x%x, 0x%x, %lu): exiting\n", (u32) to,
		(u32) from, n);
	return n;
}
#endif

#ifdef CONFIG_IOP3XX_DMA_COPYFROMUSER
/*=======================================================================*/
/*  Procedure:  dma_copy_from_user()                                     */
/*                                                                       */
/*  Description:    DMA-based copy_from_user.                            */
/*                                                                       */
/*  Parameters:  to: destination address                                 */
/*               from: source address                                    */
/*               n: number of bytes to transfer                          */
/*                                                                       */
/*  Returns:     unsigned long: number of bytes NOT copied               */
/*                                                                       */
/*  Notes/Assumptions:                                                   */
/*              Assumes that kernel virtual memory is contiguous, i.e.,  */
/*              the physical addresses of contiguous virtual addresses   */
/*              are also contiguous.                                     */
/*              Assumes that kernel memory doesn't get paged.            */
/*              Assumes that to/from memory regions cannot overlap       */
/*              XXX this one doesn't quite work right yet                */
/*                                                                       */
/*  History:    Carl Staelin 1/27/03  Initial Creation                   */
/*              Dave Jiang   2/22/03  Attempted to DMA chaining          */
/*=======================================================================*/
unsigned long dma_copy_from_user(void *to, const void *from, unsigned long n)
{
	u32 chunk;
	u32 from_chunk;
	u32 from_pa = 0;
	u32 to_chunk;
	iop3xx_dma_t *dma;
	sw_dma_t *sw_desc = NULL;
	u32 csr = 0;
	unsigned long cflags = 0;
	u32 unaligned_to, unaligned_from;
	void *u;
	volatile char f;
	int ua = 0;

	DPRINTK("dma_copy_from_user(0x%x, 0x%x, %lu): entering\n", (u32) to,
		(u32) from, n);
	if (((u32) to > PAGE_OFFSET) && ((u32) from > PAGE_OFFSET)) {
		DPRINTK("dma_copy_from_user(0x%x, 0x%x, %lu): kernel memcpy\n",
			(u32) to, (u32) from, (unsigned long)n);
		memcpy(to, from, n);
		return 0;
	}

	if(!is_kernel_static((u32) to))
	{
		DPRINTK("dma_copy_from_user(0x%x, 0x%x, %lu): vmalloc memcpy\n",
			(u32) to, (u32) from, (unsigned long)n);
		dma_miss++;
		return __arch_copy_from_user(to, from, n);
	}

	from_pa = physical_address((u32) from);
	DPRINTK("  physaddr(from) = %#010x\n", from_pa);

	/*
	 * We _only_ support mem -> mem and mem -> pci copies
	 */
	if ((from_pa < PHYS_OFFSET) && ((from_pa < IOP3XX_PCI_LOWER_MEM)
					|| (from_pa > IOP3XX_PCI_UPPER_MEM))) {
		dma_miss++;
		return __arch_copy_from_user(to, from, n);
	}

	for (u = (void*)from; (u32) u < (u32) from + n; u += PAGE_SIZE)
		get_user(f, (char *)u);

	/*
	 * Ok, start addr is not cache line-aligned, so we need to make it so.
	 */
	unaligned_to = (u32) to & 31;
	unaligned_from = (u32) from & 31;;
	if (unaligned_to | unaligned_from) {
		ua++;
		if (unaligned_from > unaligned_to) {
			__arch_copy_from_user(to, from, 32 - unaligned_to);
			to = (void *)((u32)to + 32 - unaligned_to);
			from = (void *)((u32)from + 32 - unaligned_to);
			n -= 32 - unaligned_to;
		} else {
			__arch_copy_from_user(to, from, 32 - unaligned_from);
			to = (void *)((u32)to + 32 - unaligned_from);
			from = (void *)((u32)from + 32 - unaligned_from);
			n -= 32 - unaligned_from;
		}
	}

	/*
	 * Ok, we're aligned at the top, now let's check the end
	 * of the buffer and align that. After this we should have
	 * a block that is a multiple of cache line size.
	 */
	unaligned_to = ((u32) to + n) & 31;
	unaligned_from = ((u32) from + n) & 31;;
	if (unaligned_to | unaligned_from) {
		ua++;
		if (unaligned_to > unaligned_from) {
			u32 tmp_to = (u32) to + (n - unaligned_to);
			u32 tmp_from = (u32) from + (n - unaligned_to);

			__arch_copy_from_user((void *)tmp_to, (void *)tmp_from,
					      unaligned_to);

			n -= unaligned_to;
		} else {
			u32 tmp_to = (u32) to + n - unaligned_from;
			u32 tmp_from = (u32) from + n - unaligned_from;

			__arch_copy_from_user((void *)tmp_to, (void *)tmp_from,
					      unaligned_from);

			n -= unaligned_from;
		}
	}

	/*
	 * OK! We should now be fully aligned on both ends. 
	 */
	dmac_clean_range((u32) from, (u32) from + n);
	dmac_inv_range((u32) to, (u32) to + n);

	dma = &dma_chan[0];
	/* Lock page table */
	spin_lock_irqsave(&current->mm->page_table_lock, cflags);

	from_chunk = 0;
	to_chunk = 0;

	while (n > 0) {
		if (from_chunk == 0) {
			from_chunk = page_remainder((u32) from);
			from_pa = physical_address((u32) from);
			if (!from_pa) {
				DPRINTK("dma_copy_from_user: no physical address for 0x%x,"
				     " falling back! **** \n", (u32) from);
				spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
				dma_miss++;
				return __arch_copy_from_user(to, from, n);
			}
		}

		if (to_chunk == 0) {
			/* kernel logical address ==> contiguous */
			if (is_kernel_static((u32) to)) {
				to_chunk = n;
			} else {
				/* kernel virtual address */
				to_chunk = page_remainder((u32) to);
			}
		}

		chunk = ((to_chunk < from_chunk) ? to_chunk : from_chunk);
		if (n < chunk) {
			chunk = n;
		}
		if (chunk == 0) {
			break;
		}

		DPRINTK("dma_copy_from_user(0x%x, 0x%x, %lu): to_chunk=%d, "
			"from_chunk=%d, chunk=%d\n", (u32) to, (u32) from,
			n, to_chunk, from_chunk, chunk);

		if (!list_empty(&dma->free_q)) {
			sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
			list_del(&sw_desc->link);
		} else {
			// This should NEVER happen. We need to increase the amount
			// of descriptors if that is the case
			// XXX TODO: put in something nice later to handle this!
			spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
			DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
			BUG();
			return n;
		}

		// Set last NDAR to NULL
		sw_desc->next = NULL;
		sw_desc->dma_desc->NDAR = 0;
		sw_desc->dma_desc->PUADR = 0;
		sw_desc->buf_size = sw_desc->dma_desc->BC = chunk;

		/* Is this mem -> mem or mem -> PCI? */
		if (from_pa >= PHYS_OFFSET) {
			sw_desc->dma_desc->LADR = sw_desc->phys_dest =
			    physical_address((u32) to);
			sw_desc->dma_desc->PADR = sw_desc->phys_src = from_pa;
			sw_desc->dma_desc->DC = DMA_DCR_MTM;
		} else {
			sw_desc->dma_desc->LADR = sw_desc->phys_dest =
			    physical_address((u32) to);
			sw_desc->dma_desc->PADR = sw_desc->phys_src = from_pa;
			sw_desc->dma_desc->DC = DMA_DCR_PCI_MRM;
		}

		dma->last_desc->dma_desc->NDAR = sw_desc->dma_phys;
		mb();
		*(dma->regs.CCR) = DMA_CCR_CHANNEL_ENABLE | DMA_CCR_CHAIN_RESUME;
		mb();

		/* do as much work as possible here, after starting DMA */
		from += chunk;
		to += chunk;
		n -= chunk;
		to_chunk -= chunk;
		from_chunk -= chunk;

		// busy wait on channel until done
		csr = *(dma->regs.CSR);
		while (csr & DMA_CSR_CH_ACTIVE) {
			udelay(1);
			rmb();
			csr = *(dma->regs.CSR);
		}

		// put the last in the free queue
		dma->last_desc->status = 0;
		dma->last_desc->complete = 0;
		dma->last_desc->next = 0;
		list_add_tail(&dma->last_desc->link, &dma->free_q);

		// hookup the new last descriptor for chaining
		sw_desc->status = 0;
		sw_desc->complete = 0;
		dma->last_desc = sw_desc;

		/*
		 * If the DMA failed, the user has screwed up
		 * and he needs to debug his code.
		 */
		if (csr & DMA_CSR_ERR_MASK) {
			spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
			BUG();
			return n;
		}
	}

	spin_unlock_irqrestore(&current->mm->page_table_lock, cflags);
	dma_hit++;
	if(ua)
		unaligned++;

	return n;
}
#endif

#ifdef CONFIG_IOP3XX_CRC32C
u32 iop3xx_dma_crc32c(void *buff, u32 len, int state, u32 seed) 
{
	iop3xx_dma_t *dma;
	sw_dma_t *sw_desc = NULL;
	sw_dma_t *tmp_desc = NULL;
	u32 dcr = 0;
	u32 crc = 0; 
	volatile u32 csr = 0;
	volatile u32 tmpcrc = 0;
	u32 timeout = 0xffffffff;

	DPRINTK("Entering %s\n", __func__);

	if(len <= 0)
	{
		return seed;
	}

	dma = &dma_chan[1];
   
	// we hold the semaphore on first pass
	// it's not released until the very end or error 
	if((state == CRC_FIRST_PASS) || (state == CRC_ONESHOT))
	{
		DPRINTK("acquiring semaphore\n");
		down_interruptible(&dma->chan_sem);
	}

	DPRINTK("acquired semaphore\n");

	if(!list_empty(&dma->free_q))
	{
		sw_desc = list_entry(dma->free_q.next, sw_dma_t, link);
		list_del(&sw_desc->link);
		DPRINTK("got a sw descriptor\n");
	}
	else
	{
		// This should NEVER happen. We need to increase the amount
		// of descriptors if that is the case
		// XXX TODO: put in something nice later to handle this!
		up(&dma->chan_sem);
		DPRINTK(KERN_ERR "out of DMA descriptors! Increase!\n");
		BUG();
		return 0xffffffff;
	}


	DPRINTK("setting up descriptor\n");

	/* flush buffer to mem */
	dmac_clean_range((unsigned long)buff, (unsigned long)buff+len);

	/* to addr */
	sw_desc->dma_desc->LADR = 0;
	/* from addr */
	sw_desc->dma_desc->PADR = virt_to_phys(buff);
	sw_desc->dma_desc->PUADR = 0;
	sw_desc->buf_size = sw_desc->dma_desc->BC = len;
	dcr = DMA_DCR_CRC_EN | DMA_DCR_DATA_XFER_DIS | DMA_DCR_MTM;

	if((state == CRC_IN_LOOP) || (state == CRC_DONE))
	{
		dcr |= DMA_DCR_CRCSEED_DIS;

	}
	else
	{
		*sw_desc->crc_src = (seed == 0) ? 0xffffffff : seed;
		/* flush the crc32c seend to mem */
		dmac_clean_range((unsigned long)sw_desc->crc_src,
			(unsigned long)sw_desc->crc_src + sizeof(u32));
	}

	// Set last NDAR to NULL
	sw_desc->next = NULL;
	sw_desc->dma_desc->NDAR = 0;

	/* set the descriptor control */
	sw_desc->dma_desc->DC = dcr;

	sw_desc->dma_desc->CAD = sw_desc->phys_crc;

	// add descriptor to process queue
	list_add_tail(&sw_desc->link, &dma->process_q);

	DPRINTK("dma_desc->LADR: %#08x\n", sw_desc->dma_desc->LADR);
	DPRINTK("dma_desc->PDAR: %#08x\n", sw_desc->dma_desc->PADR);
	DPRINTK("dma_desc->NDAR: %#08x\n", sw_desc->dma_desc->NDAR);
	DPRINTK("dma_desc->BC: %#08x\n", sw_desc->dma_desc->BC);
	DPRINTK("dma_desc->DC: %#08x\n", sw_desc->dma_desc->DC);
	DPRINTK("dma_desc->CAD: %#08x\n", sw_desc->dma_desc->CAD);

	DPRINTK("setting up last descriptor\n");
	dma->last_desc->dma_desc->NDAR = sw_desc->dma_phys;
	dma->last_desc = sw_desc;

	DPRINTK("starting DMA engine\n");
	// start DMA engine
	*(dma->regs.CCR) = DMA_CCR_CHANNEL_ENABLE | DMA_CCR_CHAIN_RESUME;

	// continue if we are not done chaining
	if((state != CRC_DONE) && (state != CRC_ONESHOT))
	{
		return seed;
	}

	/* now invalidate it for final value */
	dmac_inv_range((unsigned long)sw_desc->crc_src,
			(unsigned long)sw_desc->crc_src + sizeof(u32));

	// sit in polling loop to wait for it done
	mb();

	while(!(sw_desc->dma_desc->DC & DMA_DCR_CRC_DONE) && timeout)
	{
		timeout--;
		rmb();
		udelay(1);
	}

	csr = *(dma->regs.CSR);

#ifdef DEBUG
	if(csr & DMA_CSR_CH_ACTIVE)
	{
		printk(KERN_CRIT "DMA engine still active!\n");
	}
#endif

	DPRINTK("dma complete\n");

	DPRINTK("dma_desc->DC: %#08x\n", sw_desc->dma_desc->DC);
	DPRINTK("register->DC: %#08x\n", *(dma->regs.DCR));


	tmpcrc = *sw_desc->crc_src;

	// clean out the process queue
    while(!list_empty(&dma->process_q)) 
	{
    		tmp_desc = list_entry(dma->process_q.next, sw_dma_t, link);
    		list_del(&tmp_desc->link);
		list_add_tail(&tmp_desc->link, &dma->free_q);
	}

	// check to see if DMA failed
	if((csr & DMA_CSR_ERR_MASK) || (timeout == 0))
	{
		*(dma->regs.CSR) |= (DMA_CSR_DONE_MASK | DMA_CSR_ERR_MASK);
		printk(KERN_ERR "%s: DMA engine error or timed out\n", __func__);
		up(&dma->chan_sem);
		BUG();
		return 0xffffffff;
	}
	
	// release the DMA channel
	up(&dma->chan_sem);

	DPRINTK("crc dma result (byte reversed): %#08x\n", tmpcrc);

	// fix errata on CRC value backwards, IOP331
	crc = (tmpcrc & 0xff) << 24 | (tmpcrc & 0xff00) << 8 |
		(tmpcrc & 0xff0000) >> 8 | (tmpcrc & 0xff000000) >> 24;	

	return crc;
} 
#endif

module_init(iop3xx_dma_init);
module_exit(iop3xx_dma_exit);
MODULE_LICENSE(GPL);

#ifdef CONFIG_IOP3XX_DMACOPY
EXPORT_SYMBOL(iop_memcpy);
EXPORT_SYMBOL(dma_memcpy);
#endif

#ifdef CONFIG_IOP3XX_DMA_COPYTOUSER
EXPORT_SYMBOL(dma_copy_to_user);
#endif

#ifdef CONFIG_IOP3XX_DMA_COPYFROMUSER
EXPORT_SYMBOL(dma_copy_from_user);
#endif

#ifdef CONFIG_IOP3XX_CRC32C
EXPORT_SYMBOL(iop3xx_dma_crc32c);
#endif
