/*
 * pbi_cf.c
 *
 * Author: Dan Williams <dan.j.williams@gmail.com>
 * Copyright (C) 2005 Intel Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Some of this code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.
 *
 */
#include <linux/init.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/fcntl.h>	/* O_ACCMODE */
#include <linux/hdreg.h>	/* HDIO_GETGEO */
#include <linux/kdev_t.h>
#include <linux/genhd.h>
#include <linux/blkdev.h>
#include <linux/buffer_head.h>	/* invalidate_bdev */
#include <linux/bio.h>
#include <linux/delay.h>
#include <linux/ide.h>
#include <asm/arch/hardware.h>


#ifdef DEBUG
#define DEBUG_ENTRY() printk("Entered: %s\n", __FUNCTION__)
#else
#define DEBUG_ENTRY()
#endif

#define PBI_CF_MINORS	16


static int pbi_cf_major = 0;
module_param(pbi_cf_major, int, 0);
static int pbi_cf_sectors = 0;
module_param(pbi_cf_sectors, int, 0);

struct pbi_cf_dev {
	u32 ide_base;			/* IDE register base */
	u8 status;			/* IDE status bits */
	u32 size;			/* Device size in sectors */
	short users;			/* Reference count */
	spinlock_t lock;		/* For mutual exclusion */
	struct request_queue *queue;	/* The device request queue */
	struct gendisk *gd;		/* The gendisk structure */
	u8 ints_enabled;		/* Flag for interrupt jumper */
};

struct pbi_cf_dev *gDev = 0;

#define pbi_cf_port_write( reg, val ) \
	*(volatile u8 *) (dev->ide_base + (reg)) = (val)

#define pbi_cf_port_read( reg, val ) \
	(val) = *(volatile u8 *) (dev->ide_base + (reg))


/*
	Function:	pbi_cf_wait_for_drq
	Description:	Busy wait on the status register until DRQ_STAT is seen
	Parameters:	struct pbi_cf_dev *dev: driver context
	Return:		1: DRQ_STAT seen
			0: Error device reported !BUSY_STAT and !DRQ_STAT
*/
static inline int pbi_cf_wait_for_drq(struct pbi_cf_dev *dev)
{
	DEBUG_ENTRY();

    do 
    {
		/* poll because the handler may not fire again */
		pbi_cf_port_read(IDE_STATUS_OFFSET, dev->status);

		/* terminate if status has arrived */
		if (dev->status & DRQ_STAT)
		   return 1;

    } while (dev->status & BUSY_STAT);

    return 0;
}

/*
	Function: 	 pbi_cf_wait_for_ready
	Description: Busy wait on the status register until BSY_STAT & DRQ_STAT == 0
	Parameters:  struct pbi_cf_dev *dev: driver context
	Return:		 none
*/
static inline void pbi_cf_wait_for_ready(struct pbi_cf_dev *dev)
{
	DEBUG_ENTRY();
    do 
    {
		pbi_cf_port_read(IDE_STATUS_OFFSET, dev->status);
    } while (dev->status & (BUSY_STAT | DRQ_STAT));
}


/*
	Function: 	 pbi_cf_read_sectors
	Description: Setup and retrieve sectors from the device
	Parameters:  struct pbi_cf_dev *dev: driver context
				 u32 start: First sector
				 u8 count: Number of sectors to transfer
				 u8 *buf: Buffer to be filled
	Return:		 Boolean
				 1: Success
				 0: Failure
*/
int pbi_cf_read_sectors(struct pbi_cf_dev *dev, u32 start, u8 count, u8 *buf)
{
    int  i, j;
    u8 *p;

	DEBUG_ENTRY();

    pbi_cf_port_write(IDE_NSECTOR_OFFSET, count);
    pbi_cf_port_write(IDE_SECTOR_OFFSET, start & 0xff);
    pbi_cf_port_write(IDE_LCYL_OFFSET, (start >>  8) & 0xff);
    pbi_cf_port_write(IDE_HCYL_OFFSET,  (start >> 16) & 0xff);
    pbi_cf_port_write(IDE_SELECT_OFFSET, ((start >> 24) & 0xf) | 0x40);
    pbi_cf_port_write(IDE_COMMAND_OFFSET, 0x20);

    for(p = buf, i = 0; i < count; i++) {

        if (!pbi_cf_wait_for_drq(dev)) {
            printk("%s: NO DRQ\n", __FUNCTION__);
            return 0;
        }

	for (j = 0; j < SECTOR_SIZE; j++, p++)
		pbi_cf_port_read(IDE_DATA_OFFSET, *p);
    }

    return 1;
}


/*
	Function: 	 pbi_cf_write_sectors
	Description: Setup and push data to the device
	Parameters:  struct pbi_cf_dev *dev: driver context
				 u32 start: First sector
				 u8 count: Number of sectors to transfer
				 u8 *buf: Buffer to be drained
	Return:		 Boolean
				 1: Success
				 0: Failure
*/
int pbi_cf_write_sectors(struct pbi_cf_dev *dev, u32 start, u8 count, u8 *buf)
{
    int  i, j;
    u8 *p;

	DEBUG_ENTRY();

    pbi_cf_port_write(IDE_NSECTOR_OFFSET, count);
    pbi_cf_port_write(IDE_SECTOR_OFFSET, start & 0xff);
    pbi_cf_port_write(IDE_LCYL_OFFSET, (start >>  8) & 0xff);
    pbi_cf_port_write(IDE_HCYL_OFFSET,  (start >> 16) & 0xff);
    pbi_cf_port_write(IDE_SELECT_OFFSET,	((start >> 24) & 0xf) | 0x40);
    pbi_cf_port_write(IDE_COMMAND_OFFSET, 0x30);

    for(p = buf, i = 0; i < count; i++) {

        if (!pbi_cf_wait_for_drq(dev)) {
            printk("%s: NO DRQ\n", __FUNCTION__);
            return 0;
        }

        for (j = 0; j < SECTOR_SIZE; j++, p++)
            pbi_cf_port_write(IDE_DATA_OFFSET, *p);

    }
	pbi_cf_wait_for_ready(dev);

    return 1;
}


/*
	Function: 	 pbi_cf_transfer
	Description: Wrapper for the sector transfer functions
	Parameters:  struct pbi_cf_dev *dev: driver context
				 unsigned long sector: First sector
				 unsigned long nsect: Number of sectors to transfer
				 char *buffer: Buffer to be accessed
				 int write: 1=write 0=read
	Return:		 none
*/
static void pbi_cf_transfer(struct pbi_cf_dev *dev, unsigned long sector,
		unsigned long nsect, char *buffer, int write)
{
	DEBUG_ENTRY();

	if ((sector + nsect) > dev->size) {
		printk (KERN_NOTICE "Beyond-end write (%ld %ld)\n", sector, nsect);
		return;
	}
	if (write)
		pbi_cf_write_sectors(dev, sector, nsect, buffer);
	else
		pbi_cf_read_sectors(dev, sector, nsect, buffer);
}

/*
	Function: 	 pbi_cf_xfer_bio
	Description: Decodes a bio into sectors and maps a buffer
				 that is passed to pbi_cf_transfer for processing
	Parameters:  struct pbi_cf_dev *dev: driver context
				 struct bio *bio: Linux's block descriptor
	Return:		 0 = success !0 = fail
*/
static int pbi_cf_xfer_bio(struct pbi_cf_dev *dev, struct bio *bio)
{
	int i;
	struct bio_vec *bvec;
	sector_t sector = bio->bi_sector;

	DEBUG_ENTRY();

	/* Do each segment independently. */
	bio_for_each_segment(bvec, bio, i) {
		char *buffer = __bio_kmap_atomic(bio, i, KM_USER0);
		pbi_cf_transfer(dev, sector, bio_cur_sectors(bio),
				buffer, bio_data_dir(bio) == WRITE);
		sector += bio_cur_sectors(bio);
		__bio_kunmap_atomic(bio, KM_USER0);
	}
	return 0; /* Always "succeed" */
}

/*
	Function: 	 pbi_cf_make_request
	Description: Decodes a bio into sectors and maps a buffer
				 that is passed to pbi_cf_transfer for processing
	Parameters:  struct pbi_cf_dev *dev: driver context
				 struct bio *bio: Linux's block descriptor
	Return:		 0 = success !0 = fail
*/
static int pbi_cf_make_request(request_queue_t *q, struct bio *bio)
{
	struct pbi_cf_dev *dev = q->queuedata;
	int status;

	DEBUG_ENTRY();

	status = pbi_cf_xfer_bio(dev, bio);
	bio_endio(bio, bio->bi_size, status);
	return 0;
}

static int pbi_cf_ident(struct pbi_cf_dev *dev, u8 *buf)
{
    u32 i;

	DEBUG_ENTRY();

    pbi_cf_port_write(IDE_SELECT_OFFSET, 0);
    pbi_cf_port_write(IDE_COMMAND_OFFSET, 0xEC);

    if (!pbi_cf_wait_for_drq(dev))
		return 0;

    for (i = 0; i < SECTOR_SIZE; i++)
        pbi_cf_port_read(IDE_DATA_OFFSET, buf[i]);


    return 1;
}

static int pbi_cf_open(struct inode *inode, struct file *filp)
{
	struct pbi_cf_dev *dev = inode->i_bdev->bd_disk->private_data;
	DEBUG_ENTRY();

	filp->private_data = dev;
	spin_lock(&dev->lock);
	dev->users++;
	spin_unlock(&dev->lock);

	return 0;
}

static int pbi_cf_release(struct inode *inode, struct file *filp)
{
	struct pbi_cf_dev *dev = inode->i_bdev->bd_disk->private_data;
	DEBUG_ENTRY();

	spin_lock(&dev->lock);
	dev->users--;
	spin_unlock(&dev->lock);

	return 0;
}

int pbi_cf_media_changed(struct gendisk *gd)
{
	DEBUG_ENTRY();
	return 0;
}

int pbi_cf_revalidate(struct gendisk *gd)
{
	DEBUG_ENTRY();
	return 0;
}

int pbi_cf_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct hd_geometry geo;
	struct pbi_cf_dev *dev = filp->private_data;

	DEBUG_ENTRY();

	switch(cmd) {
	    case HDIO_GETGEO:
         /*
		  * Get geometry: since we are a flash device, we have to make
		  * up something plausible.  So we claim 16 sectors, four heads,
		  * and calculate the corresponding number of cylinders.  We set the
		  * start of data at sector four.
		  */
		geo.cylinders = (dev->size & ~0x3f) >> 6;
		geo.heads = 4;
		geo.sectors = 16;
		geo.start = 4;
		if (copy_to_user((void __user *) arg, &geo, sizeof(geo)))
			return -EFAULT;
		return 0;
	}

	return -ENOTTY; /* unknown command */
}

static struct block_device_operations pbi_cf_ops = {
	.owner           = THIS_MODULE,
	.open 	         = pbi_cf_open,
	.release 	 	 = pbi_cf_release,
	.media_changed   = pbi_cf_media_changed,
	.revalidate_disk = pbi_cf_revalidate,
	.ioctl	         = pbi_cf_ioctl
};

static int pbi_cf_setup(struct pbi_cf_dev *dev, int idx)
{
	/* buffer to hold ide identification response */
	struct hd_driveid id_info;

	memset(&id_info, 0, sizeof(id_info));

	memset(dev, 0, sizeof(struct pbi_cf_dev));

	/* Set the location for the ide registers */
	dev->ide_base = (u32) ioremap(PBI_CF_IDE_BASE, 16);
	spin_lock_init(&dev->lock);

 	/* retrieve the identity block */
    if (!pbi_cf_ident(dev, (u8 *) &id_info))
		return -ENODEV;

    if (id_info.config != 0x848a)
	{
		printk(KERN_NOTICE "pbi_cf: Failed to detect compact flash during identification\n");
		return -ENODEV;
	}

	/* set the sector size, default to using the auto detected value */
	if(pbi_cf_sectors)
		dev->size = pbi_cf_sectors;
	else
		dev->size = id_info.lba_capacity;

 	dev->queue = blk_alloc_queue(GFP_KERNEL);
	if (dev->queue == NULL)
		return -ENOMEM;
	
	blk_queue_make_request(dev->queue, pbi_cf_make_request);
	blk_queue_hardsect_size(dev->queue, SECTOR_SIZE);
	dev->queue->queuedata = dev;

	/* initialize the gendisk structure */
	dev->gd = alloc_disk(PBI_CF_MINORS);
	if (! dev->gd) {
		blk_put_queue(dev->queue);
		printk (KERN_NOTICE "alloc_disk failure\n");
		return -ENOMEM;
	}
	dev->gd->major = pbi_cf_major;
	dev->gd->first_minor = idx*PBI_CF_MINORS;
	dev->gd->fops = &pbi_cf_ops;
	dev->gd->queue = dev->queue;
	dev->gd->private_data = dev;
	snprintf(dev->gd->disk_name, 32, "cf%c", idx + 'a');
	set_capacity(dev->gd, dev->size);
	add_disk(dev->gd);

	printk("pbi_cf: dev: %s major: %d num_sectors: %d\n", dev->gd->disk_name, pbi_cf_major, dev->size);

	return 0;
}

static int pbi_cf_init(void)
{
	pbi_cf_major = register_blkdev(pbi_cf_major, "pbi_cf");

	if (pbi_cf_major <= 0) {
		printk(KERN_WARNING "pbi_cf: unable to get major number\n");
		return -EBUSY;
	}

	gDev = kmalloc(sizeof (struct pbi_cf_dev), GFP_KERNEL);
	if (gDev == NULL)
		goto out_unregister;

	/* for now we only support 1 compact flash device */
	return pbi_cf_setup(gDev, 0);

out_unregister:
	unregister_blkdev(pbi_cf_major, "pbi_cf");
	return -ENOMEM;
}

static void pbi_cf_exit(void)
{
	struct pbi_cf_dev *dev = gDev;

	if (dev->gd) {
		del_gendisk(dev->gd);
		put_disk(dev->gd);
	}

	if (dev->queue)
		blk_put_queue(dev->queue);

	unregister_blkdev(pbi_cf_major, "pbi_cf");
	iounmap((void *) dev->ide_base);
	kfree(gDev);
	
	printk(KERN_ALERT "pbi_cf unloading\n" );
}

module_init(pbi_cf_init);
module_exit(pbi_cf_exit);

MODULE_LICENSE("GPL");
