/*
 * pic16f8x.c
 *
 * Device driver for Microchip pic16f8x PIC Controllers.
 *
 * Author: Drew Moseley (drew.moseley@intel.com)
 * 
 * Copyright (C) 2004 Intel Corp.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/i2c.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/miscdevice.h>
#include <linux/proc_fs.h>

#include <asm/arch-iop3xx/iop321.h>
#include <asm/arch-iop3xx/iop321-irqs.h>

#include "pic16f8x.h"
#include <linux/delay.h>

#define DEBUG 1

#if DEBUG
#define DPRINT printk
#else
#define DPRINT
#endif

static int pic_mode;

static int pic16f8x_major;
static wait_queue_head_t waitq;
static struct semaphore sem;

static ssize_t pic16f8x_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
static ssize_t pic16f8x_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
static struct file_operations pic16f8x_fops = {
        owner:THIS_MODULE,
        read:pic16f8x_read,
        write:pic16f8x_write,
};

static unsigned short slave_address = PIC16F8X_I2C_SLAVE_ADDR;
static unsigned short ignore[] = { I2C_CLIENT_END };
static unsigned short normal_addr[] =
{ PIC16F8X_I2C_SLAVE_ADDR, I2C_CLIENT_END };

static struct i2c_client_address_data addr_data = {
        normal_i2c:normal_addr,
        probe:ignore,
        ignore:ignore,
};

static int pic16f8x_probe(struct i2c_adapter *adapter);
static int pic16f8x_detach(struct i2c_client *client);
struct i2c_driver pic16f8x_driver = {
        attach_adapter:pic16f8x_probe,
        detach_client:pic16f8x_detach,
};
struct i2c_client *pic16f8x_i2c_client = NULL;

static int pic16f8x_datardy = 0;
static irqreturn_t pic16f8x_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
        *IOP321_GPOD &=~ (1 << 3);        // Mask PIC interrupts
        pic16f8x_datardy = 1;
        wake_up_interruptible(&waitq);
        return IRQ_HANDLED;
}

static __u8 pic16f8x_get_command(void)
{
        int retval;

        __u8 msg_query_data[PIC16F8X_MSG_LEN] = 
                { PIC16F8X_BEGIN_MSG | PIC16F8X_CLEAR_INT, PIC16F8X_MSGID_QUERY_MSG, 1 };
        struct i2c_msg msg_query[1] = {
                {pic16f8x_i2c_client->addr, 0, PIC16F8X_MSG_LEN, msg_query_data}, 
        };
        __u8 msg_read_data[PIC16F8X_MSG_LEN];
        struct i2c_msg msg_read[1] = {
                {pic16f8x_i2c_client->addr, I2C_M_RD, PIC16F8X_MSG_LEN, msg_read_data}, 
        };

        /* Send the QUERY_MSG command. */
        retval = i2c_transfer(pic16f8x_i2c_client->adapter, msg_query, 1);
        if (retval < 0) {
                DPRINT("Error %d sending QUERY_MSG command to pic16f8x PIC\n", retval);
                return -EAGAIN;
        }

        /* Flow-control for slow PIC device */
        mdelay(2);

        /* Read the actual data. */
        retval = i2c_transfer(pic16f8x_i2c_client->adapter, msg_read, 1);
        if (retval < 0) {
                DPRINT("Error %d reading command from pic16f8x PIC\n", retval);
        }

        return msg_read_data[2];
}

static int pic16f8x_put_command(__u8 ack_nak)
{
        int retval;

        __u8 msg_ack_data[PIC16F8X_MSG_LEN] = 
                { PIC16F8X_BEGIN_MSG, PIC16F8X_MSGID_ACK, PIC16F8X_ACK };
        struct i2c_msg msg_ack[1] = {
                {pic16f8x_i2c_client->addr, 0, PIC16F8X_MSG_LEN, msg_ack_data}, 
        };
        __u8 msg_nak_data[PIC16F8X_MSG_LEN] = 
                { PIC16F8X_BEGIN_MSG, PIC16F8X_MSGID_NAK, PIC16F8X_NACK };
        struct i2c_msg msg_nak[1] = {
                {pic16f8x_i2c_client->addr, 0, PIC16F8X_MSG_LEN, msg_nak_data}, 
        };

        if (ack_nak == 0) {
                /* Send the NAK. */
                retval = i2c_transfer(pic16f8x_i2c_client->adapter, msg_nak, 1);
                if (retval < 0) {
                        DPRINT("Error %d sending IOP_NAK to pic16f8x PIC\n", retval);
                        return -EAGAIN;
                }
        } else {
                /* Send the ACK. */
                retval = i2c_transfer(pic16f8x_i2c_client->adapter, msg_ack, 1);
                if (retval < 0) {
                        DPRINT("Error %d sending IOP_ACK to pic16f8x PIC\n", retval);
                        return -EAGAIN;
                }
        }

        return retval;
}

static ssize_t pic16f8x_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
        static int this_cmd;

        if (down_interruptible(&sem))
                return -ERESTARTSYS;

        for (;;) {
                int interrupted;

                up(&sem); /* release the lock */
                if (filp->f_flags & O_NONBLOCK)
                        return -EAGAIN;

                if (pic_mode == PIC16F8X_MODE_POLLED) {
                        interrupted = wait_event_interruptible_timeout(waitq, (pic16f8x_datardy != 0), 10*HZ);
                } else {
                        *IOP321_GPOD |= (1 << 3); // Unmask PIC interrupts
                        interrupted = wait_event_interruptible(waitq, (pic16f8x_datardy != 0));
                }

                if (interrupted < 0)
                        return -ERESTARTSYS; /* signal: tell the fs layer to handle it */

                /*
                 * We either timed out or were interrupted.
                 * Query the PIC
                 */
                if (down_interruptible(&sem))
                        return -ERESTARTSYS;
                pic16f8x_datardy = 0;

                this_cmd = pic16f8x_get_command();
                if (this_cmd != PIC16F8X_QUERY_NONE)
                    break;
        }

        up (&sem); /* release the lock */

        if (copy_to_user(buf, &this_cmd, 1)) {
                return -EFAULT;
        }
  
        *f_pos++;
  
        if (pic_mode == PIC16F8X_MODE_INTERRUPT) {
                // Unmask PIC interrupts
                *IOP321_GPOD |= (1 << 3);
        }

        return 1;
}

static ssize_t pic16f8x_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
        int rc, i;

        if (down_interruptible(&sem))
                return -ERESTARTSYS;

        for (i = 0; i < count; i++) {
                static char this_cmd;

                if (copy_from_user(&this_cmd, &buf[i], 1)) {
                        up (&sem); /* release the lock */
                        return -EFAULT;
                }
                rc = pic16f8x_put_command(this_cmd);
                if (rc < 0) {
                        up (&sem); /* release the lock */
                        return rc;
                }
                *f_pos++;
        }

        up (&sem); /* release the lock */
        return i;
}

static int pic16f8x_attach(struct i2c_adapter *adapter, int addr, int kind)
{
	int ret = 0;
	struct i2c_client *c;

	if (!(c = kmalloc(sizeof(struct i2c_client), GFP_KERNEL))) {
		return -ENOMEM;
	}

	memset(c, 0, sizeof(struct i2c_client));

	strncpy(c->name, "pic16f8x", I2C_NAME_SIZE);
	
	c->addr		= addr;
	c->adapter	= adapter;
	c->driver	= &pic16f8x_driver;

	pic16f8x_i2c_client = c;

	ret = i2c_attach_client(c);

	if (ret < 0) {
		kfree(c);
	}

	return ret;
}

static int pic16f8x_probe(struct i2c_adapter *adapter)
{
        int retval;

        retval = i2c_probe(adapter, &addr_data, pic16f8x_attach);
        
        /*
         * Inform the PIC controller that we are alive.
         * This will signal to the PIC code that it is safe to interrupt us.
         */

        /* Flow-control for slow PIC device */
        mdelay(2);

        if (pic16f8x_i2c_client)
        {
                __u8 msg_data[PIC16F8X_MSG_LEN] =
                { PIC16F8X_BEGIN_MSG, PIC16F8X_MSGID_IOP_ACTIVE, 0 };
                struct i2c_msg msg_iop_active[1] = {
                        {pic16f8x_i2c_client->addr, 0, PIC16F8X_MSG_LEN, msg_data}, 
                };

                pic_mode = pic16f8x_get_command();
                if ((pic_mode != PIC16F8X_MODE_POLLED) && 
                    (pic_mode != PIC16F8X_MODE_INTERRUPT)) {
                        printk(KERN_WARNING "Querying interrupt mode from PIC unsuccessful.  Returned %x\n", pic_mode);
                        printk(KERN_WARNING "Defaulting to Polled Mode\n");
                        pic_mode = PIC16F8X_MODE_POLLED;
                }

                /* Flow-control for slow PIC device */
                mdelay(2);

                retval = i2c_transfer(pic16f8x_i2c_client->adapter, msg_iop_active, 1);
                if (retval < 0) {
                        DPRINT("iop_active unsuccessful.  Returned %d\n", retval);
                        retval = 0; // Ignore errors from the PIC.
                }

                /* Mask PIC interrupts */
                *IOP321_GPOE &=~ (1 << 3);  /* Enable Output on GPIO:3 -- used as a mask for PIC interrupts */
                *IOP321_GPOD &=~ (1 << 3);  /* Clear GPIO:3 -- mask the interrupts */
        }

        return retval;
}

static int pic16f8x_detach(struct i2c_client *client)
{
        if (client) {
                i2c_detach_client(client);
                kfree(client);
                client = pic16f8x_i2c_client = NULL;
        }
	return 0;
}

static int pic16f8x_proc_output(char *buf)
{
        char *p = buf;

        p += sprintf(p, "pic16f8x PIC Driver\n");
  
        return p - buf;
}

static int pic16f8x_read_proc(char *page, char **start, off_t off,
                              int count, int *eof, void *data)
{
        int len = pic16f8x_proc_output(page);
        if (len <= off + count)
                *eof = 1;
        *start = page + off;
        len -= off;
        if (len > count)
                len = count;
        if (len < 0)
                len = 0;
        return len;
}

static __init int pic16f8x_init(void)
{
        int retval;

		local_fiq_disable();

        normal_addr[0] = slave_address;

        retval = i2c_add_driver(&pic16f8x_driver);
        if (retval < 0) {
                printk(KERN_WARNING "pic16f8x: can't add i2c driver\n");
                return retval;
        }

        retval = register_chrdev(pic16f8x_major, "pic16f8x", &pic16f8x_fops);
        if (retval < 0) {
                i2c_del_driver(&pic16f8x_driver);
                printk(KERN_WARNING "pic16f8x: can't get major %d\n", pic16f8x_major);
                return retval;
        }

        if (pic16f8x_major == 0) pic16f8x_major = retval; /* dynamic */

        retval = request_irq(IRQ_IOP321_HPI, pic16f8x_interrupt,
                             SA_INTERRUPT, "pic16f8x", NULL);
        if (retval < 0) {
                i2c_del_driver(&pic16f8x_driver);
                unregister_chrdev(pic16f8x_major, "pic16f8x");
                printk(KERN_INFO "pic16f8x: can't get assigned irq %i\n",
                       IRQ_IOP321_HPI);
                return retval;
        }

        init_waitqueue_head(&waitq);
        sema_init(&sem, 1);
        create_proc_read_entry(PROC_PIC16F8X_NAME, 0, 0,
                               pic16f8x_read_proc, NULL);

        return 0;
}

static __exit void pic16f8x_exit(void)
{
        remove_proc_entry(PROC_PIC16F8X_NAME, NULL);
        unregister_chrdev(pic16f8x_major, "pic16f8x");
        i2c_del_driver(&pic16f8x_driver);
        free_irq(IRQ_IOP321_HPI, NULL);
}

module_init(pic16f8x_init);
module_exit(pic16f8x_exit);

module_param(slave_address, int, 0644);
MODULE_PARM_DESC(slave_address, "I2C slave address for pic16f8x PIC.");
module_param(pic16f8x_major,int, 0644);

MODULE_AUTHOR("Intel Corp.");
MODULE_LICENSE("GPL");
