/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                  I N C L U D E   S T A T E M E N T S                       */
/*                                                                            */
/******************************************************************************/

#include <gd31244.h>
#include <gd31244_lld.h>
#include <sata_debug.h>
#include <gd31244_satascsi.h>

/* Artisea Library */
#include <sata.h>

/* Linux */

/******************************************************************************/
/*                                                                            */
/*                C O N S T A N T   D E F I N I T I O N S                     */
/*                                                                            */
/******************************************************************************/
const char gd31244_version_str[] = "1.264";
const char gd31244_work_queue_name[] = "GD31244"; // Max 10 char
const char gd31244_copyright[] = "Intel(R) GD31244 SATA driver - version %s\nCopyright (c) 2002-2005 Intel Corporation.\n";


/******************************************************************************/
/*                                                                            */
/*    V A R I A B L E   A N D   S T R U C T U R E   D E F I N I T I O N S     */
/*                                                                            */
/******************************************************************************/

/* GD31244 specific */
SATA_FEATURES_T master_features_g = ARTISEA_FEATURES;

/* SCSI Specific */

static struct scsi_host_template driver_template = GD31244_TEMPLATE;
//#include "scsi_module.c"  // support for loading as module
struct scsi_host_template *gd31244_driver_template = &driver_template;
int devices_found = 0;
struct Scsi_Host *first_shp = NULL_POINTER;

/* Other */
struct workqueue_struct *gd31244_workqueue;
struct work_struct dev_issue_pending;
struct work_struct dev_change_pending;

struct proc_dir_entry proc_scsi_gd31244 ;
// = {
//     0,                             // low_ino
//     7,                             // namelen
//     "gd31244",                     // name
//     S_IFDIR | S_IRUGO | S_IXUGO,   // mode
//     2                              // nlink
// };


/******************************************************************************/
/*                                                                            */
/*                 F U N C T I O N    P R O T O T Y P E S                     */
/*                                                                            */
/******************************************************************************/
static irqreturn_t gd31244_isr(int irq, void *dev_id, struct pt_regs *regs);
static int gd31244_sync_disk(DEVICE_INFO_T *dev_info);
static void gd31244_queue_dev_change(void * adapter_port);
static int __devinit gd31244_probe_one(struct pci_dev *pdev, const struct pci_device_id *dev_id);
static void __devexit gd31244_remove_one(struct pci_dev *pdev);
static int __init gd31244_init(void);
static void __exit gd31244_exit(void);

#ifdef OLD_CODE
STATUS_CODE_T gd31244_device_init(struct scsi_device *dev);
#endif

/******************************************************************************/
/*                                                                            */
/*               D E B U G   M A C R O   D E F I N I T I O N S                */
/*                                                                            */
/******************************************************************************/


/******************************************************************************/
/*                                                                            */
/*  Util functions                                                            */
/*                                                                            */
/******************************************************************************/

/* This function verifies that the chipset is a mass storage device of SATA type*/
/* Artisea controller in DPA mode will present itself as such a device          */
/* This function will NOT verify that the controller is Artisea.                */
static int is_gd31244_in_dpa_mode(struct pci_dev *dev)
{
   unsigned int class_code_rev;
   int is_artisea_in_dpa = 0;

   DEBUG_TRACE();

   pci_read_config_dword(dev, PCI_CLASS_REVISION, &class_code_rev);

   /* are the top bits of the class code 0106h ? */
   is_artisea_in_dpa = ( (class_code_rev & 0xffff0000) == 0x01060000) ? 1 : 0;

   return is_artisea_in_dpa;
} /* end of is_gd31244_in_dpa_mode() */


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function initializes the gd31244 PCI-X device. It        */
/*              implements all the h/w initialization steps as well as all    */
/*              erratas associated with the controller itself.                */
/*                                                                            */
/* PRECONDITIONS    None.                                                     */
/*                                                                            */
/* PARAMETERS   pdev                                                          */
/*                Pointer to the PCI device to be initialized.                */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Controller not Supported/IDE mode?  */
/*              SATA_FAILURE            = Hardware errors                     */
/*                                                                            */
/******************************************************************************/

STATUS_CODE_T init_gd31244_HBA(struct pci_dev *pdev)
{
   unsigned short pcicmd, new_pcicmd;
   int bus, dev, func;

   u32 cr98 = 0;  // SU Extended Control and Status Reg 0
   u32 cra0 = 0;  // SU Bus Master Ctrl & Status Reg
   u8  cr0c = 0;  // SU Cacheline Size Register
   u32 cre4 = 0;  // SU PXI-X Status Register

   DEBUG_TRACE();

#ifdef SANITY_CHECK

   // Sanity Check on all the input parameters...
   if ( pdev == NULL )
   {
      DEBUG_ERROR("ERROR: Parameter pdev was NULL%s", "");
      return SATA_FAILURE;
   }

#endif

   DEBUG_PRINT("H/W init PCI device Name=\"%s\", Slot=\"%s\"", pdev->pretty_name, pci_name(pdev));

   // now let's enable the pci device...
   if ( pci_enable_device(pdev) )
   {
      //DEBUG_ERROR("Failed to enable PCI device Name=\"%s\", Slot=\"%s\"", pdev->pretty_name, pci_name(pdev));
      DEBUG_ERROR("Failed to enable PCI device Name=, Slot=\"%s\"", pci_name(pdev));
      return SATA_FAILURE;
   }

   // Check if the controller is in DPA mode... drop it if it is not...
   // in PCI IDE mode, the regular IDE drivers can pick it up.

   if ( is_gd31244_in_dpa_mode(pdev)== 0 )
   {
      //printk(KERN_NOTICE "%s: Chipset is NOT in DPA mode. Skipping controller \"%s\", Slot=\"%s\"", driver_template.name, pdev->pretty_name, pci_name(pdev));
      printk(KERN_NOTICE "%s: Chipset is NOT in DPA mode. Skipping controller , Slot=\"%s\"", driver_template.name, pci_name(pdev));
      return SATA_OUT_OF_RANGE_VALUE;
   }
   // Check if we need to enable mem space access...

   if ( pci_read_config_word(pdev, PCI_COMMAND, &pcicmd) )
   {
      //printk(KERN_ERR "%s: Error accessing PCI regs on Name=\"%s\", Slot=\"%s\"", driver_template.name, pdev->pretty_name, pci_name(pdev));
      printk(KERN_ERR "%s: Error accessing PCI regs on Name=, Slot=\"%s\"", driver_template.name, pci_name(pdev));
      return SATA_FAILURE;
   }

   // Set The features we want...
   new_pcicmd = pcicmd | PCI_COMMAND_MEMORY | PCI_COMMAND_IO | PCI_COMMAND_MASTER | PCI_COMMAND_PARITY | PCI_COMMAND_FAST_BACK;

   if ( pci_write_config_word(pdev, PCI_COMMAND, new_pcicmd) )
   {
      // Hmm... Failed to update the PCI command Register
      //printk(KERN_ERR "%s: Failed to update the PCI-X CMD reg. (%#0x => %#0x) for Name=\"%s\", Slot=\"%s\"",
      //       driver_template.name, pcicmd, new_pcicmd, pdev->pretty_name, pci_name(pdev));
      printk(KERN_ERR "%s: Failed to update the PCI-X CMD reg. (%#0x => %#0x) for Name=, Slot=\"%s\"",
             driver_template.name, pcicmd, new_pcicmd, pci_name(pdev));
      return SATA_FAILURE;
   }

   // Setup some of the registers on the GD31244

   // Print out the PCI-X bus info...
   pci_read_config_dword(pdev, SUPCIXSR, &cre4);

   bus = (cre4 >> 8) & 0xff;
   dev = (cre4 >> 3) & 0x1f;
   func = cre4 & 0x3;
   DEBUG_ERROR_0(KERN_INFO "%s: \"%s\", PCI-X Reg: %#08x -> %d:%d.%d", driver_template.name, pci_name(pdev), cre4, bus, dev , func);

   // Configure the Adapter for Activity LEDs per port
   pci_read_config_dword(pdev, SUECSR0, &cr98);
   cr98 = RESET_BIT(cr98, 28);  // This is the magic bit that will turn on the activity LEDs
   pci_write_config_dword(pdev, SUECSR0, cr98);

   // Non-Cache Aligned DMA Transfers Workaround.
   // issue: non-cache aligned DMA transfers can result in data corruption.
   // Solution: force GD31244 to issue cacheline aligned block cycles.

   // 1. Set the Cache line size Config Register.
   //    Read the value in SUCLSR, if that byte is not set, initialize it to default value of 08h
   pci_read_config_byte(pdev, SUCLSR, &cr0c);
   if ( cr0c == 0 )
   {
      // Set the cacheline size...
      cr0c = 0x08; // Allows Burst lengths of 80h, 40h, 20h, 10h, 8h
      pci_write_config_byte(pdev, SUCLSR, cr0c);
   }

   // 2. Enable DMA engine cacheline alignemnt
   pci_read_config_dword(pdev, SUABMCSR, &cra0);
   cra0 |= 0x03; // Turn on bit 0 and 1 to enable DMA Rd/Wr cache align.
   pci_write_config_dword(pdev, SUABMCSR, cra0);

   // We have successfully setup the GD31244 HBA hardware.
   return SATA_SUCCESS;
} // init_gd31244_HBA

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function registers the provided PCI device as a SCSI     */
/*              Host and fills in the data structures for the host. After     */
/*              this function is done, we will be ready to probe for all the  */
/*              attached SATA drives.                                         */
/*                                                                            */
/* PRECONDITIONS  init_gd31244_HBA has been called for this device.           */
/* Notes:                                                                     */
/*             it is expected that the function calling this function will    */
/*             set the host_info->adapter_number field for this HBA           */
/*                                                                            */
/* PARAMETERS   pdev                                                          */
/*                Pointer to the PCI device to be initialized.                */
/*             host                                                           */
/*                The SCSI Host Template for gd31244 HBA                      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Controller not Supported/IDE mode?  */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*              SATA_FAILURE            = Hardware errors                     */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T init_gd31244_Host(struct pci_dev *pdev)
{
   struct Scsi_Host *host              = NULL_POINTER;
   u32 bar0;
   void * remappedBar0                 = NULL_POINTER;
   HOST_INFO_T   *host_info            = NULL_POINTER;
   STATUS_CODE_T fail_code             = SATA_FAILURE; // Used to cleanup if something fails.
   STATUS_CODE_T retval                = SATA_FAILURE;
   int ctr, status;

   DEBUG_TRACE();

#ifdef SANITY_CHECK
   // Sanity Check

   if ( pdev == NULL )
   {
      DEBUG_PRINT("ERROR: Parameter pdev was NULL%s", "");
      return SATA_FAILURE;
   }

   if ( is_gd31244_in_dpa_mode(pdev)== 0 )
   {
      printk(KERN_NOTICE "%s: Chipset is NOT in DPA mode. Skipping controller \"%s\", Slot=\"%s\"",driver_template.name, pdev->pretty_name, pci_name(pdev));
      return SATA_OUT_OF_RANGE_VALUE;
   }
#endif

   DEBUG_PRINT("S/W init PCI device Name=\"%s\", Slot=\"%s\"", pdev->pretty_name, pci_name(pdev));
   // We have the PCI device... Allocate/Map the resources needed for this device...

   // Check-Request-Remap Bar0...
   /* This is the preferred method of access in 2.4 and above*/
   bar0 = pci_resource_start(pdev, 0);
   remappedBar0 = ioremap_nocache(bar0, GD31244_MEMMAP_SZ);

   //  DEBUG_PRINT("%s: Bar0 = %p to %p, len = %ld remapped at %p", pci_name(pdev),
   //              (void *)bar0Start, (void *)bar0End, bar0End - bar0Start + 1, remappedBar0 );

   // Register the host, and ask for some memory at the end for host info...
   host = scsi_host_alloc( gd31244_driver_template, sizeof(HOST_INFO_T));

   if ( host == NULL_POINTER )
   {
      DEBUG_ERROR("%s: FAILED to register the SCSI host %s. Skipping controller", driver_template.name, pci_name(pdev));
      goto cleanup;
   }

   // Record the host pointer
   if ( first_shp == NULL_POINTER )
   {
      // This is the first instance...
      first_shp = host;
   }
   else
   {
      struct Scsi_Host *shp = first_shp;
      HOST_INFO_T      *hi  = HOST_INFO(shp);

      // follow the linked list to the end...
      while ( (hi != NULL_POINTER) && (hi->next != NULL_POINTER) )
      {
         // go to the next instance...
         shp = hi->next;
         hi = HOST_INFO(shp);
      }

      // We expect to come out with hi->next == NULL and  hi != NULL
      if ( hi == NULL_POINTER )
      {
         // Error... Should not happen.
         DEBUG_ERROR("Internal Error setting up the host linked list.%s","");
         goto cleanup;
      }

      hi->next = host;
   }

   host_info = HOST_INFO(host);

   ASSERT(host_info != NULL);

   host_info->pdev = pdev;
   host_info->bar0_raw = bar0;
   host_info->bar0_remapped = remappedBar0;
   host_info->adapter_number= -1; // Set in gd31244_detect... or the function that called us.
   host_info->info_string[0] = '\0';
   host_info->next = NULL_POINTER;
   host_info->removed = 0;

   DEBUG_PRINT("A Maximum of %d devices can connect to this Adapter", MAX_DEVICES_PER_ADAPTER);

   for ( ctr = 0; ctr < MAX_DEVICES_PER_ADAPTER; ctr++ )
   {  // Initialize the dev_tags array to invalid tags...
      // these are filled in as we find new devices...
      host_info->dev_tags[ctr] = INVALID_TAG;
      host_info->dev_sdev[ctr] = NULL_POINTER;
   }
   // Initialize host fields and copy stuff over from the template...

   host->max_id = 1;  // No support for Port Multiplier yet...
   host->max_lun = 1; // We don't support LUN > 0
   host->max_channel = MAX_PORTS_PER_CONTROLLER - 1; // = 3
   host->base = bar0; // shouldn't have to use this...
   host->io_port = 0;
   host->n_io_port = 0;
   host->dma_channel = 0;
   host->irq = pdev->irq; // The IRQ we will be finally using. Not yet requested...

   host->unique_id = bar0; // that will make it unique...
   host->max_cmd_len = GD31244_MAX_CMD_LEN;
   host->this_id = gd31244_driver_template->this_id;
   host->can_queue = gd31244_driver_template->can_queue;
   host->cmd_per_lun = gd31244_driver_template->cmd_per_lun;
   host->sg_tablesize = gd31244_driver_template->sg_tablesize;

   host->max_sectors = gd31244_driver_template->max_sectors; // we might need to reduce this if we find a LBA28 drive.
   host->unchecked_isa_dma = gd31244_driver_template->unchecked_isa_dma;
   host->use_clustering = gd31244_driver_template->use_clustering;
   // host->highmem_io = gd31244_driver_template->highmem_io;

   /* Ask the adapter to allocate it's memory and return an adapter tag */
   retval = AdapterAllocatePrivateMemory( host->host_no, &(host_info->adapter_tag));
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("%s: Lib failed(%d) to allocate memory.", pci_name(pdev), retval);
      fail_code = SATA_GET_MEMORY_FAILED;
      goto cleanup;
   }

   /* Ask the adapter object to initialize itself */
   retval = AdapterInit(host_info->adapter_tag, remappedBar0);
   if ( retval == SATA_COMPLETED_WITH_ERRORS )
   {
      /* Some of the drives failed during init. */
      DEBUG_ERROR("AdapterInit returned SATA_COMPLETED_WITH_ERRORS (%d)", retval);

      // We will loop later to check which drives are present.
      // It's OK to continue with normal operations...

   }
   else if ( retval != SATA_SUCCESS )
   {
      /* Failed to initialize the library. Disable the controller */
      DEBUG_ERROR("AdapterInit Failed (%d). Skipping controller ", retval);
      goto cleanup;
   }

   /* Disable the interrpts on the adapter */
   AdapterDisableInterrupts(host_info->adapter_tag);

   // We are not interested in any interrupts geenrated by the adapter before now.
   AdapterClearInterrupts(host_info->adapter_tag);

   ASSERT(pdev->irq);

#ifndef CONFIG_ARM
   // Request and install the IRQ and ISR for this device
   DEBUG_PRINT("Requesting IRQ %d for %s", pdev->irq, pci_name(pdev));
   status = request_irq( pdev->irq, // Request the IRQ indicated by the PCI device
                         gd31244_isr, // The ISR routine
                         0 /*SA_INTERRUPT*/, // SA_INTERRUPT should be used for fast ISRs like timer isr...
                         driver_template.name, // Our name
                         host_info             // Can be changed to host if we need to...
                       );
   if ( status < 0 )
   {
      DEBUG_PRINT("Failed(%d) to get non-shared IRQ %d for %s, trying for Shared IRQ",
                  status, pdev->irq, pci_name(pdev));

      status = request_irq( pdev->irq, // Request the IRQ indicated by the PCI device
                            gd31244_isr,          // The ISR routine
                            /*SA_INTERRUPT |*/ SA_SHIRQ,
                            driver_template.name, // Our name
                            host_info             // Can be changed to host if we need to...
                          );
   }

#else
   // Request and install the IRQ and ISR for this device
   DEBUG_PRINT("Requesting IRQ %d for %s", pdev->irq, pci_name(pdev));
   status = request_irq( pdev->irq, // Request the IRQ indicated by the PCI device
                         gd31244_isr, // The ISR routine
                         SA_SHIRQ,
                         driver_template.name, // Our name
                         host_info             // Can be changed to host if we need to...
                       );
#endif

   if ( status < 0 )
   {
      DEBUG_ERROR("Failed(%d) to get an IRQ %d for %s. HBA Disabled.",
                  status, pdev->irq, pci_name(pdev));

      pdev->irq = 0;
      goto cleanup;
   }

   // Set the IRQ to what we finally got.
   host->irq = pdev->irq;

   /* Enable the interrpts on the adapter */
   AdapterEnableInterrupts(host_info->adapter_tag);

   DEBUG_PRINT("Adding Host to SCSI subsystem");
   status = scsi_add_host(host, &pdev->dev);
   if ( status < 0 )
   {
     DEBUG_ERROR("Failed(%d) to add the SCSI host %s to subsystem.",
		 status, pci_name(pdev));
     goto cleanup;
   }

   /* Set the Scsi host to the pci device */
   pci_set_drvdata(pdev, host);

   DEBUG_PRINT("Scanning for attached drives");
   scsi_scan_host(host);

   // OK... We have successfully setup the device.
   DEBUG_PRINT("Successfully initialized adapter (tag %d)", host_info->adapter_tag);

   fail_code = SATA_SUCCESS;

   cleanup:
   // Cleanup code to undo actions from the this iteration...
   if ( fail_code != SATA_SUCCESS )
   {
      DEBUG_PRINT("Cleaning up after Failed scan...%s", "");

      if ( remappedBar0 != NULL_POINTER )
      {
         DEBUG_PRINT("Unmapping GD31244 memory regions from %p", remappedBar0);
         /* unmap the GD31244 memory regions */
         iounmap( remappedBar0 );
      }
      if ( host != NULL_POINTER )
      {
         DEBUG_PRINT("unregistering SCSI host from %p", host);
         scsi_remove_host(host);
      }

      if(pdev->irq != 0)
      {
	//Release the IRQ
	free_irq(pdev->irq, host_info);
      }
   }

   return fail_code;
}

#ifdef OLD_CODE_DISABLED
STATUS_CODE_T gd31244_device_init(struct scsi_device *dev)
{
   // We will be assuming that dev != NULL
   int host;//                = dev->host->host_no;
   int channel;//             = dev->channel;
   int target;//              = dev->id;
   int lun;//                 = dev->lun;

   struct pci_dev * pdev;//    = dev->host->pci_dev;
   STATUS_CODE_T retval     = SATA_SUCCESS;
   DEVICE_INFO_T *dev_info;//  = DEVICE_INFO(dev); // should be == NULL_POINTER
   HOST_INFO_T   *host_info ;//= HOST_INFO(dev->host);
   int adapter_tag          ;//= host_info->adapter_tag; // Don't Use the Macro here.
   int dev_tag              = INVALID_TAG;
   int failed               = TRUE;
   int drive_num            = 0; // The drive number on this adapter
   int ctr;

   DEBUG_TRACE();

   // We have seen cases where dev was NULL !!!
   if ( dev == NULL_POINTER )
   {
      DEBUG_ERROR("ERROR: gd31244_device_init called with NULL dev pointer(%p)", dev);
      return SATA_FAILURE;
   }

   ASSERT(dev->host);

   host                = dev->host->host_no;
   channel             = dev->channel;
   target              = dev->id;
   lun                 = dev->lun;

   pdev        = DEVICE_PDEV(dev);
   dev_info    = DEVICE_INFO(dev); // should be == NULL_POINTER
   host_info   = HOST_INFO(dev->host);

   // Check if all the pointers are OK.
   if ( (pdev == NULL) || (host_info == NULL) )
   {
      DEBUG_ERROR("ERROR: Detected NULL pointers pdev = %p, host_info = %p", pdev, host_info);
      return SATA_FAILURE;
   }

   ASSERT(host_info);
   adapter_tag = host_info->adapter_tag; // Don't Use the Macro here.

   DEBUG_PRINT("Init Drive %d:%d:%d:%d", host, channel, target, lun);

   // Basic Sanity check...
   /* Check for valid channel, target and Lun numbers */
   if ( (channel > dev->host->max_channel ) ||
        (target  > dev->host->max_id ) ||
        (lun     > dev->host->max_lun)
      )
   {
      DEBUG_ERROR("Invalid Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host, channel, target, lun);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   // Now check if this device was already initialized...
   if ( dev_info != NULL_POINTER )
   {
      DEBUG_ERROR("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) Already initialized !", host, channel, target, lun);
      return SATA_COMPLETED_WITH_ERRORS;
   }

   // This drive is connected to an HBA with GD31244 in DPA mode right ?
   // Won't hurt to check again...
   if ( (pdev->vendor != PCI_VENDOR_ID_INTEL) ||
        (pdev->device != PCI_DEVICE_ID_INTEL_ARTISEA) ||
        ! is_gd31244_in_dpa_mode(pdev)
      )
   {

      DEBUG_ERROR("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) Not Connected to DPA mode GD31244 Controller!", host, channel, target, lun);
      return SATA_FAILURE;
   }

   // OK, things look fine... let's initialize the drive...

   // Drive number on this adapter.... NOT the global drive number.
   drive_num = channel * MAX_DEVICES_PER_PORT + target;

   ASSERT(adapter_tag != INVALID_TAG);


   DEBUG_PRINT("Calling AdapterDevicePresent %s", "");
   // Check if the drive is present...
   retval = AdapterDevicePresent( adapter_tag, drive_num, &dev_tag);
   if ( retval == SATA_FALSE )
   {
      /* Drive not present */
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) not found", host, channel, target, lun);
      return SATA_NOT_INITIALIZED;
   }
   else if ( retval != SATA_TRUE )
   {
      /* Some other Error... Should not happen. */
      DEBUG_ERROR("Error (%d) while probing for drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", retval, host, channel, target, lun);
      return SATA_FAILURE;
   }

   if ( dev_tag == INVALID_TAG )
   {
      /* Drive not present */
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) not found", host, channel, target, lun);
      return SATA_FAILURE;
   }

   DEBUG_PRINT("Found Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) - Tag %d", host, channel, target, lun, dev_tag);

   #ifdef GD31244_BIST  // Rev A0 doesn't support BIST

   DEBUG_PRINT("Starting BIST for Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), tag %d", host, channel, target, lun, dev_tag);

   retval = AdapterPortBist(adapter_tag, drive_num );
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("BIST FAILED(%d) Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), tag %d. Drive Offline.", retval, host, channel, target, lun, dev_tag);
      ERROR_DUMP(pdev, dev_tag);
      return SATA_FAILURE;
   }

   DEBUG_PRINT("BIST Completed successfully for Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), tag %d", host, channel, target, lun, dev_tag);

   #endif // GD31244_REV_A0


   // All looks good for the drive, Allocate the resources...
   if ( dev->hostdata == NULL_POINTER )
   {
      dev->hostdata = (DEVICE_INFO_T *)kmalloc(sizeof(DEVICE_INFO_T),
                                               GFP_KERNEL);
      DEBUG_PRINT("Allocating device_info %s", "");
   }
   if ( dev->hostdata == NULL_POINTER )
   {
      DEBUG_ERROR("%d:%d:%d:%d: Failed to allocate memory for device_info",
                  host, channel, target, lun);
      return SATA_GET_MEMORY_FAILED;
   }

   // If anything fails after this point, we need to do some cleanup...

   DEBUG_PRINT("Initializing dev_info %s", "");
   // initialize dev_info
   memset(dev->hostdata, 0, sizeof(DEVICE_INFO_T));

   dev_info                    = DEVICE_INFO(dev);
   dev_info->adapter_tag       = adapter_tag;
   dev_info->dev_tag           = dev_tag;
   dev_info->port_tag          = INVALID_TAG;
   dev_info->vector_table_orig = NULL_POINTER;
   dev_info->vector_table.vt   = NULL_POINTER;
   dev_info->max_ddt_entries   = AdapterGetMaxDdtEntries();
   dev_info->online = 0;

   #ifdef SATA_DEBUG_0
   dev_info->max_ddte_count_seen = 0;
   dev_info->min_ddte_count_seen = 0xfffffff; // This is large enough. The first DDT should clear this.
   dev_info->all_cmd_seen_count = 0;
   #endif

   // Now allocate the vector table.
   // Assumption: The Vector table needs to be in a 64K aligned non-paged region.
   // The vector table may not need to be aligned....
   // dev_info->vector_table_orig = kmalloc(sizeof(DMA_VT_T) * 2, GFP_KERNEL | GFP_DMA);

   dev_info->vector_table_orig =
   pci_alloc_consistent(pdev, sizeof(DMA_VT_T) * 2,
                        &dev_info->vector_table_orig_phy);

   if ( dev_info->vector_table_orig == NULL )
   {
      DEBUG_ERROR("Drive %d:%d:%d:%d: Failed to allocate memory for Vector Table",
                  host, channel, target, lun);

      retval = SATA_GET_MEMORY_FAILED;
      goto cleanup;
   }

   /* Assign the vector table to 64k boundary */
   {
      dma_addr_t orig_phys = dev_info->vector_table_orig_phy;
      dma_addr_t new_phys;
      int shift_by;

      // Check if we are already at the correct boundary
      if ( (orig_phys + sizeof(DMA_VT_T)) <=  (orig_phys | 0xffff) )
      {
         // orig_phys is on a 64K boundary
         DEBUG_PRINT("Original Vector Table is 64K aligned - %#0x", orig_phys);

         dev_info->vector_table.vt =
         (DMA_VECTORS_T *)dev_info->vector_table_orig;
         // Record the bus address of the Start of the actual vector table
         dev_info->vector_table.vtPhysicalAddress =
         dev_info->vector_table_orig_phy;
      }
      else
      {
         // orig_phys is NOT on a 64K boundary
         new_phys = (orig_phys | 0xffff) + 0x1;
         shift_by = new_phys - orig_phys;

         DEBUG_PRINT("Original Vector Table shifted from %#0x to %#0x", orig_phys, new_phys);

         dev_info->vector_table.vt =
         (DMA_VECTORS_T *)dev_info->vector_table_orig + shift_by;
         // Record the bus address of the Start of the actual vector table
         // dev_info->vector_table->vtPhysicalAddress = virt_to_bus(dev_info->vector_table->vt);
         dev_info->vector_table.vtPhysicalAddress =
         dev_info->vector_table_orig_phy + shift_by;
      }
   }

   #ifdef DEBUG
   // Verify thet the Vector table is correctly alighed.
   if ( (dev_info->vector_table.vtPhysicalAddress + sizeof(DMA_VT_T)) >
        (dev_info->vector_table.vtPhysicalAddress | 0xffff )
      )
   {
      /* The high order bytes are NOT the same. :-( */
      DEBUG_ERROR("INTERNAL ERROR: Drive %d:%d:%d:%d: The Vector Table crossed a 64K boundary.", host, channel, target, lun);
      DEBUG_ERROR("start %#0llx, end %#0llx, size %d",dev_info->vector_table.vtPhysicalAddress,
                  dev_info->vector_table.vtPhysicalAddress + sizeof(DMA_VT_T),sizeof(DMA_VT_T));

      retval = SATA_FAILURE;
      goto cleanup;
   }

   #endif

   // Now allocate the DDT's
   // We will allocate only 1 ddt at this time... We can allocate more DDTs later
   // if this drive turns out to be TCQ capable.

   retval = gd31244_allocate_ddts(dev, 1);
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("HCTL %d:%d:%d:%d - Failed(%d) to allocate DDTs", host, channel, target, lun, retval);
      goto cleanup;
   }

   DEBUG_PRINT("Initializing the Queues %s", "");

   // Now init cmd_info structures...
   for ( ctr = 0; ctr < MAX_COMMANDS; ctr++ )
   {
      dev_info->commands[ctr].slot_no = ctr;
      dev_info->commands[ctr].ddt_ptr = &dev_info->ddt[ctr];
      dev_info->commands[ctr].cmd_state = SATA_CMD_FREE;
   }

//    dev_info->commands_free = 0; // This will get incremented with every freeq_push

   // initialize the queues
   INIT_LIST_HEAD(&dev_info->tcq_free);

   // initialize the locks
   spin_lock_init(&dev_info->free_qlock);

   atomic_set(&dev_info->busy, 0);
   atomic_set(&dev_info->cmd_count, 0);

   // Move all the commands to the free queue.
   // Add them in the reverse order.
   for ( ctr = MAX_COMMANDS -1 ; ctr >= 0; ctr-- )
   {
      freeq_push(dev_info, &dev_info->commands[ctr]);
   #ifdef QDEBUG
      printk(KERN_ERR "0x%p: added\n", &dev_info->commands[ctr]);
   #endif
   }

   // Attach the Scsi_Disk to the Sata Lib...
   DEBUG_PRINT("Calling AdapterAttachDisk %s", "");
   retval = AdapterAttachDisk( adapter_tag,
                               drive_num,
                               dev,
                               &gd31244_isr,
                               &gd31244_isr,
                               &gd31244_isr,
                               &dev_info->dev_tag
                             );

   // The devTag that we are assigned in this step should be the same
   // we were handed out earlier...
   ASSERT(dev_info->dev_tag == dev_tag);

   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("AdapterAttachDisk failed for %d:%d:%d:%d, drive_num %d. retval = %d",
                  host, channel, target, lun, drive_num, retval);
      ERROR_DUMP(pdev, dev_tag);
      goto cleanup;
   }

   // Update the dev_tags array under the adapter's host_info...
   ASSERT((channel * MAX_DEVICES_PER_PORT + target) < MAX_DEVICES_PER_ADAPTER );
   host_info->dev_tags[(channel * MAX_DEVICES_PER_PORT + target)] = dev_info->dev_tag;
   host_info->dev_sdev[(channel * MAX_DEVICES_PER_PORT + target)] = dev;

   // All OK. we successfully initialized the drive...
   DEBUG_PRINT("Successfully Initialized Drive %d:%d:%d:%d, drive_num %d.",
               host, channel, target, lun, drive_num);
   failed = FALSE;

   cleanup:

   // Cleanup code to free the resources allocated...
   if ( failed )
   {
      DEBUG_PRINT("Cleaning up after Failed device init...%s", "");

      DEBUG_PRINT("Releasing Vector Table %s", "");
      if ( dev_info->vector_table_orig != NULL )
      {
         //kfree(dev_info->vector_table_orig);
         pci_free_consistent(pdev, sizeof(DMA_VT_T) * 2,
                             dev_info->vector_table_orig,
                             dev_info->vector_table_orig_phy);
      }

      DEBUG_PRINT("Releasing DDTs %s", "");
      gd31244_free_ddts(dev);

      if ( dev_info != NULL_POINTER )
      {
         DEBUG_PRINT("Releasing dev_info %p", dev_info);
         kfree(dev_info);
      }

      host_info->dev_tags[(channel * MAX_DEVICES_PER_PORT + target)] = INVALID_TAG;
      host_info->dev_sdev[(channel * MAX_DEVICES_PER_PORT + target)] = NULL_POINTER;

      // Return retval if is has the failure code...
      return( (retval == SATA_SUCCESS) ? SATA_FAILURE : retval );
   }

   return SATA_SUCCESS;

}

#endif

/******************************************************************************/
/*                                                                            */
/*  Moule entry point                                                         */
/*                                                                            */
/******************************************************************************/

MODULE_AUTHOR("Intel Corporation, Storage Component Division");
MODULE_DESCRIPTION("Intel(R) GD31244 PCI-X 4 Port SATA HBA, DPA mode driver for Linux 2.6x");
#ifdef MODULE_LICENSE
MODULE_LICENSE("Dual BSD/GPL");
#endif

/******************************************************************************/
/*                                                                            */
/*  SCSI Low level driver interface functions                                 */
/*                                                                            */
/******************************************************************************/

#ifdef OLD_CODE
/* SCSI management functions */
int gd31244_detect(struct scsi_host_template * tpnt)
{
   struct Scsi_Host *shp               = NULL_POINTER;
   HOST_INFO_T   *hi                   = NULL_POINTER;
   int devices_found                   = 0;
   int legacy_mode_devices_found       = 0;
   int ctr                             = 0;
   struct pci_dev *pdev                = NULL;
   STATUS_CODE_T retval                = SATA_FAILURE;

   DEBUG_TRACE();

   // Display the copyright.
   printk(gd31244_copyright, gd31244_version_str);

   // Initialize the Library.
   InitLib();

   // Init structure fields...
   tpnt->proc_dir  = &proc_scsi_gd31244;
   driver_template.module = THIS_MODULE;

   /* Init the Work Queue and work */
   //   gd31244_workqueue = create_workqueue(gd31244_work_queue_name);
   INIT_WORK( &dev_issue_pending, gd31244_queue_worker, NULL);
   INIT_WORK( &dev_change_pending, gd31244_queue_dev_change, NULL);

#ifdef CONFIG_PCI


   // Find Artisea HBAs
   while ( (pdev = pci_find_device(PCI_VENDOR_ID_INTEL,
                                   PCI_DEVICE_ID_INTEL_ARTISEA, pdev) ) )
   {

      DEBUG_PRINT("Found SATA Controller Name=\"%s\", Slot=\"%s\"", pdev->pretty_name, pci_name(pdev));

      // Initialize the Hardware
      retval = init_gd31244_HBA(pdev);

      if ( retval == SATA_OUT_OF_RANGE_VALUE ) // HBA not supported, IDE mode
      {
         legacy_mode_devices_found++;

         // OK... let's look at the next one...
         continue;
      }
      else if ( retval != SATA_SUCCESS ) // Something went wrong.
      {
         // Can't do anything about this one...
         printk ( KERN_ERR "%s: Failed(%#0x) to initialize PCI device %s",driver_template.name, retval, pci_name(pdev));

         // let's look at the next one...
         continue;
      }

      // Initialize the SCSI host for this HBA
      retval = init_gd31244_Host(pdev);
      if ( retval != SATA_SUCCESS )
      {
         printk (KERN_ERR "%s: Failed(%#0x) to setup SCSI host %s",driver_template.name, retval, pci_name(pdev));

         // let's look at the next one...
         continue;
      }

      // Increment the devices found ...
      devices_found++;

   } // while pci_find_device

   // Remember, we need to initialize the Adapter numbers in the Host_info structures...
   shp = first_shp;
   hi  = HOST_INFO(shp);
   ctr = 0;

   // follow the linked list and update the adapter numbers.
   while ( (shp != NULL_POINTER) && (hi != NULL_POINTER) )
   {
      // Initialize the adapter number
      hi->adapter_number = ctr;

      // Increment the Adapter number counter
      ctr++;

      // go to the next instance...
      shp = hi->next;
      hi = HOST_INFO(shp);
   }

   #ifdef SANITY_CHECK
   ASSERT(ctr == devices_found);
   #endif

   if ( legacy_mode_devices_found > 0 )
   {
      printk(KERN_NOTICE "%s: Note: %d Legacy Mode GD31244 controller(s) detected. Leaving these controllers for Legacy IDE drivers.\n",
             driver_template.name, legacy_mode_devices_found);
   }

   // We are done detecting all the controllers...
   printk(KERN_NOTICE "%s: %d Controllers Detected.\n",driver_template.name, devices_found);

   return devices_found;

#else /* CONFIG_PCI */

   #error PCI Support *Disabled* in Kernel - Cant support GD31244

   DEBUG_ERROR("Kernel doesn't support PCI. Can't probe for %s", driver_template.name );
   return 0;

#endif

} // gd31244_detect
#endif

int gd31244_slave_alloc(struct scsi_device *dev)
{

   int host, channel, target, lun;
   struct pci_dev * pdev    = NULL;
   DEVICE_INFO_T *dev_info  = NULL;
   HOST_INFO_T   *host_info = NULL;

   int adapter_tag          = INVALID_TAG;
   int dev_tag              = INVALID_TAG;
   int failed               = TRUE;
   int drive_num            = 0; // The drive number on this adapter
   STATUS_CODE_T retval     = SATA_SUCCESS;

   DEBUG_TRACE();

   ASSERT(dev->host);

   if ( dev != NULL )
   {
      DEBUG_PRINT("Allocate Resources for Drive %d:%d:%d:%d", dev->host->host_no, dev->channel, dev->id, dev->lun);
   }
   else
   {
      DEBUG_ERROR("ERROR: Invalid param: dev pointer(%p)", dev);
      return SATA_FAILURE;
   }

   dev_info    = DEVICE_INFO(dev); // should be == NULL_POINTER

   // init the HCTL numbers.
   host        = dev->host->host_no;
   channel     = dev->channel;
   target      = dev->id;
   lun         = dev->lun;

   pdev        = DEVICE_PDEV(dev);
   host_info   = HOST_INFO(dev->host);

   // Check if all the pointers are OK.
   if ( (pdev == NULL) || (host_info == NULL) )
   {
      DEBUG_ERROR("ERROR: Detected NULL pointers pdev = %p, host_info = %p", pdev, host_info);
      return SATA_FAILURE;
   }

   ASSERT(host_info);
   adapter_tag = host_info->adapter_tag; // Don't Use the Macro here.
   ASSERT(adapter_tag != INVALID_TAG);

   // Basic Sanity check...
   /* Check for valid channel, target and Lun numbers */
   if ( (channel > dev->host->max_channel ) ||
        (target  > dev->host->max_id ) ||
        (lun     > dev->host->max_lun)
      )
   {
      DEBUG_ERROR("Invalid Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host, channel, target, lun);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   // Now check if this device was already initialized...
   if ( dev_info != NULL_POINTER )
   {
      DEBUG_ERROR("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) Already initialized !", host, channel, target, lun);
      return SATA_COMPLETED_WITH_ERRORS;
   }

   // OK, things look fine... let's allocate resources for the drive...

   // Drive number on this adapter.... NOT the global drive number.
   drive_num = channel * MAX_DEVICES_PER_PORT + target;

   // Check if the drive is present... if not, don't allocate resources for the drive.
   retval = AdapterDevicePresent( adapter_tag, drive_num, &dev_tag);
   if ( retval == SATA_FALSE )
   {
      /* Drive not present */
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) not found", host, channel, target, lun);
      return SATA_SUCCESS; /* This is not an error... Report Success.*/
   }
   else if ( retval != SATA_TRUE )
   {
      /* Some other Error... Should not happen. */
      DEBUG_ERROR("Error (%d) while probing for drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", retval, host, channel, target, lun);
      return SATA_FAILURE;
   }

   if ( dev_tag == INVALID_TAG )
   {
      /* Drive not present */
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) not found, invalid tag", host, channel, target, lun);
      return SATA_FAILURE;
   }
   DEBUG_PRINT("Found Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) - Tag %d", host, channel, target, lun, dev_tag);

#ifdef GD31244_BIST  // Rev A0 doesn't support BIST

   DEBUG_PRINT("Starting BIST for Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), tag %d", host, channel, target, lun, dev_tag);

   retval = AdapterPortBist(adapter_tag, drive_num );
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("BIST FAILED(%d) Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), tag %d. Drive Offline.", retval, host, channel, target, lun, dev_tag);
      ERROR_DUMP(pdev, dev_tag);
      return SATA_FAILURE;
   }

   DEBUG_PRINT("BIST Completed successfully for Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), tag %d", host, channel, target, lun, dev_tag);

#endif // GD31244_REV_A0


   // All looks good for the drive, Allocate the resources...
   ASSERT(dev->hostdata == NULL_POINTER)

   dev->hostdata = (DEVICE_INFO_T *)kmalloc(sizeof(DEVICE_INFO_T), GFP_KERNEL); // Use GFP_NOIO ?

   if ( dev->hostdata == NULL_POINTER )
   {
      DEBUG_ERROR("%d:%d:%d:%d: Failed to allocate memory for device_info",
                  host, channel, target, lun);
      return SATA_GET_MEMORY_FAILED;
   }

   // If anything fails after this point, we need to do some cleanup...

   // initialize dev_info
   memset(dev->hostdata, 0, sizeof(DEVICE_INFO_T));

   dev_info                    = DEVICE_INFO(dev);
   dev_info->max_ddt_entries   = AdapterGetMaxDdtEntries();
   dev_info->adapter_tag       = adapter_tag;
   dev_info->dev_tag           = INVALID_TAG;
   dev_info->port_tag          = INVALID_TAG;

   // We have seen a sequence of slave_alloc, queuecommand, slave_configure
   // in this case, QueueDepth has not yet been updated... Set it to a default of 1
   dev_info->drive_features.QueueDepth = 1;

#ifdef SATA_DEBUG_0
   dev_info->max_ddte_count_seen = 0;
   dev_info->min_ddte_count_seen = 0xfffffff; // This is large enough. The first DDT should clear this.
#endif

   // Now allocate the vector table.
   // Assumption: The Vector table needs to be in a 64K aligned non-paged region.
   // The vector table may not need to be aligned....

   dev_info->vector_table_orig = pci_alloc_consistent(pdev, sizeof(DMA_VT_T) * 2, &dev_info->vector_table_orig_phy);

   if ( dev_info->vector_table_orig == NULL )
   {
      DEBUG_ERROR("Drive %d:%d:%d:%d: Failed to allocate memory for Vector Table",
                  host, channel, target, lun);

      retval = SATA_GET_MEMORY_FAILED;
      goto alloc_cleanup;
   }

   /* Assign the vector table to 64k boundary */
   {
      dma_addr_t orig_phys = dev_info->vector_table_orig_phy;
      dma_addr_t new_phys;
      int shift_by;

      // Check if we are already at the correct boundary
      if ( (orig_phys + sizeof(DMA_VT_T)) <=  (orig_phys | 0xffff) )
      {
         // orig_phys is on a 64K boundary
         dev_info->vector_table.vt = (DMA_VECTORS_T *)dev_info->vector_table_orig;
         // Record the bus address of the Start of the actual vector table
         dev_info->vector_table.vtPhysicalAddress = dev_info->vector_table_orig_phy;
      }
      else
      {
         // orig_phys is NOT on a 64K boundary
         new_phys = (orig_phys | 0xffff) + 0x1;
         shift_by = new_phys - orig_phys;

         DEBUG_PRINT("Original Vector Table shifted from %#0x to %#0x", orig_phys, new_phys);

         dev_info->vector_table.vt = (DMA_VECTORS_T *)dev_info->vector_table_orig + shift_by;
         // Record the bus address of the Start of the actual vector table
         // dev_info->vector_table->vtPhysicalAddress = virt_to_bus(dev_info->vector_table->vt);
         dev_info->vector_table.vtPhysicalAddress = dev_info->vector_table_orig_phy + shift_by;
      }
   }

#ifdef SANITY_CHECK
   // Verify thet the Vector table is correctly alighed.
   if ( (dev_info->vector_table.vtPhysicalAddress + sizeof(DMA_VT_T)) >
        (dev_info->vector_table.vtPhysicalAddress | 0xffff )
      )
   {

      /* The high order bytes are NOT the same. :-( */
      DEBUG_ERROR("INTERNAL ERROR: Drive %d:%d:%d:%d: The Vector Table crossed a 64K boundary.", host, channel, target, lun);
      DEBUG_ERROR("start %#0llx, end %#0llx, size %d",dev_info->vector_table.vtPhysicalAddress,
                  dev_info->vector_table.vtPhysicalAddress + sizeof(DMA_VT_T),sizeof(DMA_VT_T));

      retval = SATA_FAILURE;
      goto alloc_cleanup;
   }

#endif

   // Now allocate the DDT's
   // We will allocate only 1 ddt at this time... We can allocate more DDTs later
   // if this drive turns out to be TCQ capable.

   retval = gd31244_allocate_ddts(dev, 1);
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("HCTL %d:%d:%d:%d - Failed(%d) to allocate DDTs", host, channel, target, lun, retval);
      goto alloc_cleanup;
   }

   // Successfully allocated all the resources we need.
   failed = FALSE;

   alloc_cleanup:

   // Cleanup code to free the resources allocated...
   if ( failed )
   {
      DEBUG_PRINT("Cleaning up after Failed device init...%s", "");

      DEBUG_PRINT("Releasing Vector Table %s", "");
      if ( dev_info->vector_table_orig != NULL )
      {
         //kfree(dev_info->vector_table_orig);
         pci_free_consistent(pdev, sizeof(DMA_VT_T) * 2,
                             dev_info->vector_table_orig,
                             dev_info->vector_table_orig_phy);
      }

      DEBUG_PRINT("Releasing DDTs %s", "");
      gd31244_free_ddts(dev);

      if ( dev_info != NULL_POINTER )
      {
         DEBUG_PRINT("Releasing dev_info %p", dev_info);
         kfree(dev_info);
      }

      host_info->dev_tags[(channel * MAX_DEVICES_PER_PORT + target)] = INVALID_TAG;
      host_info->dev_sdev[(channel * MAX_DEVICES_PER_PORT + target)] = NULL_POINTER;

      // Return retval if is has the failure code...
      return( (retval == SATA_SUCCESS) ? SATA_FAILURE : retval );
   }
   DEBUG_PRINT("Finished allocating resources for device.");
   return SATA_SUCCESS;

} //gd31244_slave_alloc

int gd31244_slave_configure(struct scsi_device *dev)
{
   int host, channel, target, lun;
   struct pci_dev * pdev    = NULL;
   DEVICE_INFO_T *dev_info  = NULL;
   HOST_INFO_T   *host_info = NULL;

   int adapter_tag          = INVALID_TAG;
   int drive_num            = 0; // The drive number on this adapter
   int ctr;
   STATUS_CODE_T retval     = SATA_SUCCESS;

   DEBUG_TRACE();

   if ( dev != NULL )
   {
      DEBUG_PRINT("Configuring Drive %d:%d:%d:%d", dev->host->host_no, dev->channel, dev->id, dev->lun);
   }
   else
   {
      DEBUG_ERROR("ERROR: Invalid param: dev pointer(%p)", dev);
      return SATA_FAILURE;
   }

   dev_info    = DEVICE_INFO(dev); // should NOT be NULL_POINTER

   // init the HCTL numbers.
   host        = dev->host->host_no;
   channel     = dev->channel;
   target      = dev->id;
   lun         = dev->lun;

   pdev        = DEVICE_PDEV(dev);
   host_info   = HOST_INFO(dev->host);

   // Check if all the pointers are OK.
   if ( (pdev == NULL) || (host_info == NULL) )
   {
      DEBUG_ERROR("ERROR: Detected NULL pointers pdev = %p, host_info = %p", pdev, host_info);
      return SATA_FAILURE;
   }

   ASSERT(host_info);
   adapter_tag = host_info->adapter_tag; // Don't Use the Macro here.
   ASSERT(adapter_tag != INVALID_TAG);

   // Basic Sanity check...
   /* Check for valid channel, target and Lun numbers */
   if ( (channel > dev->host->max_channel ) ||
        (target  > dev->host->max_id ) ||
        (lun     > dev->host->max_lun)
      )
   {
      DEBUG_ERROR("Invalid Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host, channel, target, lun);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   // Drive number on this adapter.... NOT the global drive number.
   drive_num = channel * MAX_DEVICES_PER_PORT + target;

   // Check if the drive is present... if not, don't allocate resources for the drive.
   retval = AdapterDevicePresent( adapter_tag, drive_num, NULL);
   if ( retval != SATA_TRUE )
   {
      /* Drive not present */
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) not found", host, channel, target, lun);
      return SATA_FAILURE;
   }
   else
   {
      printk("Found SATA Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d)\n", host, channel, target, lun);
   }

   // Now check if this device was already initialized...
   if ( dev_info == NULL_POINTER )
   {
      DEBUG_ERROR("No resources allocated for Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host, channel, target, lun);

      // Ok... we can allocate the resources from here...
      retval = gd31244_slave_alloc(dev);
      if ( retval != SATA_SUCCESS )
      {
         // Well... something failed...
         return retval;
      }
   }

   // This drive is connected to an HBA with GD31244 in DPA mode right ?
   // Won't hurt to check again...
   if ( (pdev->vendor != PCI_VENDOR_ID_INTEL) ||
        (pdev->device != PCI_DEVICE_ID_INTEL_ARTISEA) ||
        ! is_gd31244_in_dpa_mode(pdev)
      )
   {

      DEBUG_ERROR("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) Not Connected to DPA mode GD31244 Controller!", host, channel, target, lun);
      return SATA_FAILURE;
   }

   // Drive number on this adapter.... NOT the global drive number.
   drive_num = channel * MAX_DEVICES_PER_PORT + target;

   ASSERT(adapter_tag != INVALID_TAG);

   // Initialize the dev_info
   dev_info = DEVICE_INFO(dev);

   // Has this device already been configured ?
   if ( dev_info->dev_tag != INVALID_TAG )
   {
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d), DevTag %d Already Configured.", host, channel, target, lun, dev_info->dev_tag);
      return SATA_SUCCESS;
   }

   // OK, things look fine... let's initialize the drive...

   // Now init cmd_info structures...
   for ( ctr = 0; ctr < MAX_COMMANDS; ctr++ )
   {
      dev_info->commands[ctr].slot_no = ctr;
      dev_info->commands[ctr].ddt_ptr = &dev_info->ddt[ctr];
      dev_info->commands[ctr].cmd_state = SATA_CMD_FREE;
   }

   // initialize the queues
   INIT_LIST_HEAD(&dev_info->tcq_free);

   // initialize the locks
   spin_lock_init(&dev_info->free_qlock);

   // Move all the commands to the free queue.
   // Add them in the reverse order.
   for ( ctr = MAX_COMMANDS -1 ; ctr >= 0; ctr-- )
   {
      freeq_push(dev_info, &dev_info->commands[ctr]);
   }

   // Attach the Scsi_Disk to the Sata Lib...
   DEBUG_PRINT("Calling AdapterAttachDisk %s", "");
   retval = AdapterAttachDisk( adapter_tag,
                               drive_num,
                               dev,
                               &gd31244_isr,
                               &gd31244_isr,
                               &gd31244_isr,
                               &dev_info->dev_tag
                             );

   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("AdapterAttachDisk failed for %d:%d:%d:%d, drive_num %d. retval = %d",
                  host, channel, target, lun, drive_num, retval);
      ERROR_DUMP(pdev, dev_info->dev_tag);
      return retval;
   }

   // Update the dev_tags array under the adapter's host_info...
   ASSERT((channel * MAX_DEVICES_PER_PORT + target) < MAX_DEVICES_PER_ADAPTER );
   host_info->dev_tags[(channel * MAX_DEVICES_PER_PORT + target)] = dev_info->dev_tag;
   host_info->dev_sdev[(channel * MAX_DEVICES_PER_PORT + target)] = dev;

   dev_info->online = 1;
   // All OK. we successfully initialized the drive...
   DEBUG_PRINT("Successfully Initialized Drive %d:%d:%d:%d, drive_num %d.",
               host, channel, target, lun, drive_num);

   return SATA_SUCCESS;

} // gd31244_slave_configure


// This function will release all the resources allocated on a per-drive basis
void gd31244_slave_destroy(struct scsi_device *dev)
{
   // We will be assuming that dev != NULL
   struct pci_dev * pdev   = DEVICE_PDEV(dev);
   int host                = dev->host->host_no;
   int channel             = dev->channel;
   int target              = dev->id;
   int lun                 = dev->lun;

   STATUS_CODE_T retval    = SATA_SUCCESS;
   DEVICE_INFO_T *dev_info  = DEVICE_INFO(dev);
   HOST_INFO_T   *host_info = HOST_INFO(dev->host);
   int adapter_tag          = ADAPTER_TAG(dev);
   int dev_tag              = DRIVE_TAG(dev);
   int drive_num            = 0; // The drive number on this adapter
   int dummy;

   DEBUG_TRACE();

#ifdef SANITY_CHECK

   if ( dev == NULL_POINTER )
   {
      DEBUG_ERROR("ERROR: gd31244_slave_destroy called with NULL dev pointer(%p)", dev);
      return;
   }

#endif

   DEBUG_PRINT("Release Resources for Drive %d:%d:%d:%d", dev->host->host_no, dev->channel, dev->id, dev->lun);

   // Basic Sanity check...
   /* Check for valid channel, target and Lun numbers */
   if ( (channel > dev->host->max_channel ) ||
        (target  > dev->host->max_id ) ||
        (lun     > dev->host->max_lun)
      )
   {
      printk(KERN_ERR "%s: Slave Destroy: Invalid HCTL (%d:%d:%d:%d)", driver_template.name, host, channel, target, lun);
      return ; // the SCSI Mid layer doesn't check the return values...
   }

   // Now check thet this device was initialized...
   if ( dev_info == NULL_POINTER )
   {
      DEBUG_PRINT( "%s: Slave Destroy: Drive HCTL-%d:%d:%d:%d. No Resources to free", driver_template.name, host, channel, target, lun);
      return ;
   }

   DEBUG_PRINT("Releasing all the resources for (%d:%d:%d:%d)", host, channel, target, lun);

   // Drive number this adapter...
   drive_num = channel * MAX_DEVICES_PER_PORT + target;
   DEBUG_PRINT("Preparing to check if drive connected.");
   retval = PortDevConnected(drive_num);
   
   // free the vector table!
   pci_free_consistent(pdev, sizeof(DMA_VT_T) * 2,
                       dev_info->vector_table_orig,
                       dev_info->vector_table_orig_phy);

   // Release the DDTs...
   gd31244_free_ddts(dev);

   if ( dev_info != NULL_POINTER )
   {

#ifdef SANITY_CHECK
      // Clear the dev_info. to be on the safe side.
      memset(dev_info, 0, sizeof(DEVICE_INFO_T));
#endif
      dev_info->dev_tag = INVALID_TAG;
      dev_info->adapter_tag = INVALID_TAG;

      DEBUG_PRINT("Releasing dev_info %p", dev_info);
      kfree(dev_info);
      dev->hostdata = NULL;
   }

   // Remove the association between the struct scsi_device and the Lib's device.

   ASSERT((channel * MAX_DEVICES_PER_PORT + target) < MAX_DEVICES_PER_ADAPTER );
   host_info->dev_tags[(channel * MAX_DEVICES_PER_PORT + target)] = INVALID_TAG;
   host_info->dev_sdev[(channel * MAX_DEVICES_PER_PORT + target)] = NULL_POINTER;

   if ( dev_tag != INVALID_TAG )
   {
      DEBUG_PRINT("Unsetting Drive Features for Tag %d (#%d)", dev_tag, drive_num);
      DevUnsetFeatures(dev_tag);

      DEBUG_PRINT("Releasing Tag %d (#%d)", dev_tag, drive_num);
      DevShutdown(dev_tag);

      AdapterAttachDisk( adapter_tag,
                         drive_num,
                         NULL_POINTER,
                         NULL_POINTER,
                         NULL_POINTER,
                         NULL_POINTER,
                         &dummy
                       );
   }
   else
   {
      DEBUG_PRINT("Drive Ho:Ch:Tgt:Lun (%d:%d:%d:%d) Had INVALID DevTag", host, channel, target, lun);
   }

}

// Function to release the resources allocated for the Host.
static void __devexit gd31244_remove_one(struct pci_dev *pdev)
{
  // DOn't forget release memory regins, device extensions, and decrement counter if ncessary
   struct scsi_device *scdev = NULL_POINTER;
   struct Scsi_Host *host  = pci_get_drvdata(pdev);
   int channel             = INVALID_TAG;
   int target              = INVALID_TAG;
   int lun                 = INVALID_TAG;
   int drive_num           = 0;
   DEVICE_INFO_T *dev_info  = NULL_POINTER;
   HOST_INFO_T   *host_info = HOST_INFO(host);
   int dev_tag              = INVALID_TAG;
   int ctr; // Loop counter

   DEBUG_TRACE();

   ASSERT(pdev);
   ASSERT(host);
   ASSERT(host_info);

   DEBUG_PRINT("Release Resources for Host %d on slot %s", host->host_no, pci_name(pdev) );

   /* These are the steps we need to take...
      1. Abort / Stop all commands issued to all drives connected to this adapter.
      2. Disable interrupts from the adapter and release the IRQ
      3. Free all the per-drive resources (DDTs & device_info)
      4. Free all the per-adapter resources (io-unmap, release-region, host_info, etc)
      NOTE: We *DONT* free the Lib... this is done when the GD31244_LIB is unloaded
   */

   flush_scheduled_work();
   //   flush_workqueue(gd31244_workqueue);

   for ( ctr = 0; ctr < MAX_DEVICES_PER_ADAPTER; ctr++ )
   {
      // For each disk under this adapter...

      // Drive number under this adapter...
      drive_num = ctr ;

      // First get the struct scsi_device Pointer...
      dev_tag = host_info->dev_tags[ctr];
      if ( dev_tag == INVALID_TAG )
      {  // This device is currently not attached...
         continue;
      }

      DEBUG_PRINT("Releasing all Resources for drive # %d, DTag %d", drive_num, dev_tag);

      scdev = (struct scsi_device *) DevReturnOsdmTag(dev_tag);
      if ( scdev == NULL_POINTER )
      {
         // Can't do anything with this one!
         DEBUG_ERROR("Scsi Drive #%d, Tag %d has invalid device_info", drive_num, dev_tag);
         continue;
      }

      dev_info = DEVICE_INFO(scdev);
      if ( gd31244_sync_disk(dev_info) == SATA_SUCCESS )
      {
         printk(KERN_INFO "SATA device #%d cache flushed\n", drive_num);
      }

// NPG - NOTE: We should release the drive related resources even if the drive is no longer present.

      channel = scdev->channel;
      target  = scdev->id;
      lun     = scdev->lun;
      dev_info = DEVICE_INFO(scdev);
      dev_tag  = DRIVE_TAG(scdev);

      // I don't expect to find any active or pending commands...
      // The SCSI mid layer should have already aborted all the commands, and
      // revoked all the drives !!!

      // Free the per-drive resources...
      gd31244_slave_destroy(scdev);

      // This drive has been blown away... on to the next drive!
   }

   // OK... now all the per-adapter resources need to be cleaned up.

   //   destroy_workqueue(gd31244_workqueue);

   // Disabel all the interrupts from the adapter...
   AdapterDisableInterrupts(host_info->adapter_tag);

   //Release the IRQ
   free_irq(pdev->irq, host_info);

   host_info->removed = 1;

   DEBUG_PRINT("Unmapping the Bar0 from %p", host_info->bar0_remapped);
   iounmap(host_info->bar0_remapped);

   // Now unregister the host...
   DEBUG_PRINT("Removing the SCSI host from the system");
   scsi_remove_host(host);

   DEBUG_PRINT("Decrementing the SCSI host counter.");
   scsi_host_put(host);

   devices_found--;

   DEBUG_PRINT("Disabling the PCI device");
   pci_disable_device(pdev);

#ifndef STANDALONE
// Free the library if we are not in 2 module mode
   FreeLib();
#endif
   // Done...
}  // gd31244_remove_one


/* SCSI Data functions */


// only return 1 if we want to retry, otherwise return 0
// or call done() with error codes!
// the function parses commands and passes them off the the appropriate
// translation functions. We support 6 and 10 byte commands on lun 0 only
int gd31244_queuecommand(struct scsi_cmnd *SCpnt,void (*done)(struct scsi_cmnd *))
{
   unsigned char *cmd      = (unsigned char *)SCpnt->cmnd;
   int host                = SCpnt->device->host->host_no;
   int channel             = SCpnt->device->channel;
   int target              = SCpnt->device->id;
   int lun                 = SCpnt->device->lun;
   int status = 0;
   STATUS_CODE_T retval     = SATA_SUCCESS;
   DEVICE_INFO_T *dev_info  = DEVICE_INFO(SCpnt->device);
   int adapter_tag          = ADAPTER_TAG(SCpnt->device);
   int dev_tag              = DRIVE_TAG(SCpnt->device);
   int drive_num;
   HOST_INFO_T  *host_info  = HOST_INFO(SCpnt->device->host);

   DEBUG_TRACE();

   //   if(dev_tag == INVALID_TAG)
   //   {
   //     DEBUG_PRINT("Found invalid tag.");
   //   }
   /*
    * If we are being notified of the mid-level reposessing a command
    * due to timeout, just return.
    */
   if ( done == NULL )
   {
      DEBUG_ERROR_0("Command with done set to NULL. Returning %d",0);
      return 0;
   }

   if(host_info == NULL)
     {
       DEBUG_ERROR_0("Host_info set to NULL for host of drive. Returning");
       SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
       done(SCpnt);
       return 0;
     }
   if(host_info->removed == 1)
     {
       DEBUG_ERROR_0("Host has been removed so commands are invalid.");
       SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
       done(SCpnt);
       return 0;
     }

   SCpnt->result = 0;
   memset(SCpnt->sense_buffer, 0, sizeof(SCpnt->sense_buffer));

   /*
    * If a command comes for the ID of the host itself,
    * just print a message and return.
    */
#ifdef SANITY_CHECK_IN_DATA_PATH
   if ( target == SCpnt->device->host->this_id )
   {
      DEBUG_ERROR_0("Command for HBA, id %d", target);
      done(SCpnt);
      return 0;
   }
#endif

   /* Check for valid channel, target and Lun numbers.*/
   /* We could see "scans" for ALL LUNS ... */
   if ( (channel > SCpnt->device->host->max_channel ) ||
        (target  > SCpnt->device->host->max_id ) ||
        (lun     > SCpnt->device->host->max_lun) )
   {
      DEBUG_ERROR("Invalid Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host,
                  channel, target, lun);
      dev_info->online = 0;
      SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
      done(SCpnt);
      return 0;
   }

   /* The following logic was added to make hotpull more reliable when
      io was active to the drive.  Essentially, we try to make sure the
      device is present before going any further.
      Note that right now the following math works since the driver only
      supports one drive per port. */
   DEBUG_PRINT("Preparing to call PortDevConnected for possibly no device attached.");
   drive_num = channel * MAX_DEVICES_PER_PORT + target;
   retval = PortDevConnected(drive_num);
   if ( retval == FALSE )
   {
      /* Drive not present */
      /* Verify dev_info is valid before dereferencing it.
         If we power-up with a drive missing it's dev_info will be NULL. */
      if ( dev_info )
      {
        dev_info->online = 0;
      }
      SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
      done(SCpnt);
      return 0;
   }

   DEBUG_PRINT_2("SCpnt = %p, SCpnt->device = %p, SCpnt->device->hostdata = %p",
                 SCpnt, SCpnt->device, SCpnt->device->hostdata);

   ASSERT(SCpnt->device->host == SCpnt->device->host);

   // OK, so the target seems to be a valid target....
   // Have we initialized our structures for it yet ?
   if ( (SCpnt->device->hostdata == NULL)  ||
	 (dev_tag == INVALID_TAG))
   {
      // Apparently not... we don't have a device_info for this drive.
     //      DEBUG_PRINT("Initializing Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host, channel, target, lun);

      //      retval = gd31244_slave_alloc(SCpnt->device);
      //      if (retval == SATA_SUCCESS)
      //      {
      retval = gd31244_slave_configure(SCpnt->device);
	  //      }
      if ( retval != SATA_SUCCESS )
      {
         // Looks like there is nothing there...
         DEBUG_PRINT("Ho:Ch:Tgt:Lun (%d:%d:%d:%d) Drive Not found", host, channel, target, lun);
         if(dev_info)
            dev_info->online = 0;
         SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
         done(SCpnt);
         return 0;
      }
      dev_info  = DEVICE_INFO(SCpnt->device);
      adapter_tag          = ADAPTER_TAG(SCpnt->device);
      dev_tag              = DRIVE_TAG(SCpnt->device);
   }

   DEBUG_PRINT("ATag %d, DTag %d, Ho:Ch:Tgt:Lun=%d:%d:%d:%d, cmd=%#02x",
               adapter_tag, dev_tag, host, channel, target, lun, (int)*cmd);

   // if we are in error mode and new commands are not request sense
   // we return SCSI status BUSY
   if ( (atomic_read(&dev_info->busy) == TRUE) &&
        (SCpnt->cmnd[0] != REQUEST_SENSE) )
   {
      DEBUG_ERROR_0("Returning Busy while in error mode. CMD %0x", SCpnt->cmnd[0]);
      SCpnt->result = SCSI_STAT(0, 0, 0, BUSY);
      done(SCpnt);
      return 0;
   }

   atomic_inc(&dev_info->cmd_count);

   if ( atomic_read(&dev_info->cmd_count) > dev_info->drive_features.QueueDepth )
   {
      atomic_dec(&dev_info->cmd_count);
      SCpnt->result = SCSI_STAT(0, 0, 0, QUEUE_FULL);
      DEBUG_ERROR_0("Dev %d: cmd queue full (%d/%d)", dev_tag,
                    dev_info->cmd_count.counter, dev_info->drive_features.QueueDepth);

      done(SCpnt);
      return 0;
   }

#if 0 // QDEBUG
   {
      int ctr, dev_fcnt, lib_fcnt;
      char astr[80] = "";
      static int cmd_ctr = 0;

      // Dump the Free and busy queues.
      // For some reason, the lib thinks this slot is in use...
      for ( ctr = 0; ctr < MAX_COMMANDS; ctr++ )
      {
         sprintf(astr + ctr, "%c",(dev_info->commands[ctr].cmd_state == 0 ? 'F':
                                   (dev_info->commands[ctr].cmd_state == 1 ? 'P':
                                    (dev_info->commands[ctr].cmd_state == 2 ? 'A': '?')))
                );
      }

      dev_fcnt = free_count(dev_info);
      lib_fcnt = DevTCQCount(dev_info->dev_tag, NULL);

      DEBUG_ERROR("cmd %d, DF %d LF %d Queue [%s]", cmd_ctr, dev_fcnt, lib_fcnt, astr);

      // We are keepinf a count of the number of commands we have seen...
      cmd_ctr++;
   }
#endif

   /* Parse the command... */
   if ( SCpnt->cmd_len == 10 )
   {
      status = gdx_translate_10bytes(SCpnt, done);

// Code moved into gdx_translate_10bytes
//        if(status == SATA_BUSY)
//        {
      // NPG: We could come here with 2 stages of completion
      // 1. - Failed to queue the command
      // 2. - Queued the command, but failed to service the command queue.
      // How do we handle these cases ?

//            atomic_dec(&dev_info->cmd_count);
//            SCpnt->result = SCSI_STAT(0, 0, 0, QUEUE_FULL);
//            DEBUG_PRINT("Dev %d: translate 10 bytes cmd queue full", dev_tag);
//            done(SCpnt);
//        }
//        else
//        {
      // TBD NPG: What do we do for other failed conditions ?
//        }

   }
   else if ( SCpnt->cmd_len == 6 )
   {
      status = gdx_translate_6bytes(SCpnt, done);
      if ( status == SATA_BUSY )
      {
         DEBUG_ERROR("gdx_translate_6bytes Failed with %d Busy", SATA_BUSY);
//            atomic_dec(&dev_info->cmd_count);
//            SCpnt->result = SCSI_STAT(0, 0, 0, QUEUE_FULL);
//            DEBUG_PRINT("Dev %d: translate 10 bytes cmd queue full", dev_tag);
//            done(SCpnt);
      }
   }
   else
   {
      // We should never get this if setup correctly!
      atomic_dec(&dev_info->cmd_count);
      DEBUG_ERROR("Ho:Ch:Tgt:Lun=%d:%d:%d:%d - Unsupported command:"
                  " opcode=%#02x", host, channel, target, lun, (int)cmd[0]);
      SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
      done(SCpnt);
   }

   DEBUG_PRINT("Exiting queuecommand %s","");
   return 0;
}

/* SCSI Error handling Functions */
int gd31244_abort(struct scsi_cmnd *SCpnt)
{
   struct scsi_cmnd  *SC         = SCpnt;
   CMD_INFO_T        *cmd_info   = NULL_POINTER;
   DEVICE_INFO_T     *devInfo    = NULL_POINTER;
   struct scsi_device *scsi_dev  = NULL_POINTER;
   void              *err_log    = NULL_POINTER;
   int               retval      = 0;
   int               abort_status = SUCCESS; // Default to Success.
   int               ctr         = 0;

   DEBUG_TRACE();

#ifdef SANITY_CHECK
   if ( SCpnt == NULL )
   {
      DEBUG_ERROR("Invalid Param %p", SCpnt);
      return FAILED;
   }
#endif

   if (SCpnt->device->hostdata == NULL )
   {
     DEBUG_ERROR("Got abort for scsi cmd %p (Host:%d Chan:%d ID:%d LUN:%d) with null hostdata.  Probably late abort from hotpull event.", 
                SCpnt,
                SCpnt->device->host->host_no,
                SCpnt->device->channel,
                SCpnt->device->id,
                SCpnt->device->lun);
      return SUCCESS;
   }

   //   cmd_info is sometimes NULL
   cmd_info = CMD_INFO(SCpnt);
   scsi_dev = SCpnt->device;
   devInfo = DEVICE_INFO(scsi_dev);

#ifdef SANITY_CHECK
//   ASSERT(cmd_info); // cmd_info is usually invalid :(
   ASSERT(scsi_dev);
   ASSERT(devInfo);
#endif

   DEBUG_ERROR_0("abort: Drive %d:%d:%d:%d dev# %d, CMD Slot %d",
                 scsi_dev->host->host_no, scsi_dev->channel, scsi_dev->id, scsi_dev->lun,
                 devInfo->dev_tag, (cmd_info == NULL ? -1 : cmd_info->slot_no));

//   ERROR_DUMP(DEVICE_PDEV(scsi_dev), devInfo->dev_tag);

#ifdef SATA_DEBUG
   {
      if ( devInfo != NULL_POINTER )
      {
         DEBUG_ERROR_0("%ld Cmds. %d commands in use", devInfo->all_cmd_seen_count, devInfo->cmd_count.counter);
         for ( ctr = 0; ctr < devInfo->drive_features.QueueDepth; ctr++ )
         {
            cmd_info = &devInfo->commands[ctr];
            if ( cmd_info == NULL_POINTER ) continue;
            if ( cmd_info->cmd_state == SLOT_FREE ) continue;
            DEBUG_ERROR_0("Slot %d, state %d, Rd/Wr %d, Num sectors %#0x, start sector %#0x",
                          ctr, cmd_info->cmd_state, cmd_info->sata_cmd, cmd_info->nr_sectors, cmd_info->start_sector );
         }
      }
   }
#endif

// NPG_TEST
#if 1

   // If there are no pending commands, return success...
   if ( devInfo->cmd_count.counter <= 0 )
   {
      return abort_status; // Default initialized to Success
   }
#endif

   if ( devInfo->drive_features.TCQ )
   {
      ERROR_DUMP(DEVICE_PDEV(SCpnt->device), devInfo->dev_tag);

      // GFP_NOIO flag did not help. we get the "might sleep" log.
      err_log = kmalloc(512, in_interrupt() ? GFP_ATOMIC : GFP_KERNEL ); 

      // Check if we got the memory
      if ( err_log == NULL_POINTER )
      {
         abort_status = SOFT_ERROR;
      }
      else
      {

         ctr = 0;

         while ( (DevIsReady(devInfo->dev_tag) != TRUE) && (ctr < WAIT_READY) )
         {
            ctr++;
            /* Wait for 1 Millisecond more for the drive to be ready */
            mdelay(1);
         }

         if ( ctr >= WAIT_READY )
         {
            /* We Timed out ... */
            DEBUG_ERROR_0("Timed out waiting for Device to be ready (%d ms)", ctr);

#if (LINUX_VERSION_CODE<KERNEL_VERSION(2,6,2))
            abort_status = SOFT_ERROR;
#else
            abort_status = TIMEOUT_ERROR;
#endif
         }
         else
         {
            retval = DevReadLogExt(devInfo->dev_tag, err_log);
            if ( retval != SATA_SUCCESS )
            {
               DEBUG_ERROR("Failed (%d) to read device log for %d", retval, devInfo->dev_tag);
               abort_status = FAILED;
            }

#ifdef SATA_DEBUG_0
            else
            {
               /* Dump out the Log for Debug */
               debug_mem_dump("SATA EXT Log, Pg 10h", err_log, 16);
            }
#endif
         } // Check for Busy
      }// Check for memory
   } // TCQ ?
   else
   {
      // Non TCQ mode.
      // cannot abort command if running non-TCQ

      ERROR_DUMP(DEVICE_PDEV(SCpnt->device), devInfo->dev_tag);

      /* Unable to abort - command is currently on the bus.  Grin and bear it. */
      abort_status = NEEDS_RETRY;
   }

   retval = DevCleanTCQ(devInfo->dev_tag);
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR("Failed to clear TCQ on %d", devInfo->dev_tag);
      abort_status = SOFT_ERROR;
   }

   for ( ctr=0; ctr < devInfo->drive_features.QueueDepth; ctr++ )
   {
      cmd_info = &devInfo->commands[ctr];
      ASSERT(cmd_info);
      if ( cmd_info->cmd_state == SATA_CMD_FREE )
      {
         continue;
      }
      SC = cmd_info->SCpnt;
      cmd_info->cmd_state = SATA_CMD_FREE;
#if (LINUX_VERSION_CODE<KERNEL_VERSION(2,6,11))
      SC->result = SCSI_STAT(0, SCpnt->abort_reason, 0, 0);
#else
      SC->result = SCSI_STAT(0, DID_ABORT, 0, 0);
#endif
      // Unmap the SGL
      if ( SC->use_sg )
      {
         pci_unmap_sg(DEVICE_PDEV(SC->device),
                      (struct scatterlist *)SC->request_buffer,
                      cmd_info->sg_nents, cmd_info->sg_dma_direction);
      }
      else
      {
         pci_unmap_single(DEVICE_PDEV(SC->device),
                          cmd_info->dma_addr, SC->request_bufflen,
                          cmd_info->sg_dma_direction);
      }

      cmd_info->scsi_done(SC);
      freeq_push(devInfo, cmd_info);
   }

   atomic_set(&devInfo->cmd_count, 0);

   if ( err_log )
   {
      kfree(err_log);
   }

#ifdef SATA_DEBUG
   if ( (abort_status != SUCCESS) && ( abort_status != SUCCESS) )
   {
      DEBUG_ERROR_0("ERROR: Abort Failed (%#0x) for Dtag %d", abort_status, devInfo->dev_tag);
   }
#endif

   return abort_status;

} /* end of gd31244_abort() */

int gd31244_bus_reset(struct scsi_cmnd *SCpnt)
{
   int host, channel, target, lun, adapter_tag, dev_tag;
   int drive_num;
   STATUS_CODE_T status;

   struct scsi_device *dev_i    = NULL_POINTER; // The Scsi Device passed down via the parameter
   struct scsi_device *dev      = NULL_POINTER; // The SCSI device info that we had recorded.

   DEVICE_INFO_T *dev_info  = NULL_POINTER;
   HOST_INFO_T   *host_info = NULL_POINTER;

   dev_i = SCpnt->device;
   ASSERT(dev_i);

   host                = dev_i->host->host_no;
   channel             = dev_i->channel;
   target              = dev_i->id;
   lun                 = dev_i->lun;

   DEBUG_ERROR("Bus Reset called for Ho:Ch:Tgt:Lun (%d:%d:%d:%d)", host, channel, target, lun);

   dev_info  = DEVICE_INFO(dev_i);
   host_info = HOST_INFO(dev_i->host);
   adapter_tag = host_info->adapter_tag; // Don't Use the Macro here.
   dev_tag     = INVALID_TAG;

   ASSERT(adapter_tag != INVALID_TAG);

   // Note: I have seen      Valid,     Valid,   NULL,        Valid
   DEBUG_ERROR_0("Bus Reset: device %p, host %p, dev_info %p, host_info %p", SCpnt->device, SCpnt->device->host, dev_info, host_info);

   /* After a hotpull, scsi_remove_device() was invoked but later
      on the kernel calls gd31244_abort() and gd31244_bus_reset().
      Below, AdapterDevicePresent() returns a valid dev_tag (don't know
      why! - the device is gone!) and this ends up allowing us to
      call gd31244_device_reset() which dies.  Instead, if we already
      know that dev_info is bad simply return now.  */
   if( dev_info == NULL)
   {
      return FAILED;
   }

   // Drive number on this adapter.... NOT the global drive number.
   drive_num = channel * MAX_DEVICES_PER_PORT + target;
   AdapterDevicePresent( adapter_tag, drive_num, &dev_tag);

   if ( dev_tag == INVALID_TAG )
   {
      // We need a valid dev_tag to be able to reset the port... return failure.
      DEBUG_ERROR("Failed to reset Host %d : Bus %d. Could not identify attached device.", host, channel);
      return FAILED;
   }

   dev = (struct scsi_device *) DevReturnOsdmTag(dev_tag);

   // Note: Seen  Valid,      NULL
   DEBUG_ERROR_0("dev_tag %d, dev_i = %p, dev = %p", dev_tag, dev_i, dev);

   SCpnt->device = dev;
   status = gd31244_device_reset(SCpnt);
   SCpnt->device = dev_i;

   return status;

} /* end of gd31244_bus_reset() */


// resets the device, actually reset the port(bus) also since there is a
// 1:1 match-up and we cannot just reset the device
int gd31244_device_reset(struct scsi_cmnd *SCpnt)
{
   int port;
   STATUS_CODE_T status;
   DEVICE_INFO_T       *dev = NULL_POINTER;
   struct scsi_device  *scd = NULL_POINTER;


   DEBUG_TRACE();

#ifdef SANITY_CHECK

   ASSERT(SCpnt);

   if ( SCpnt == NULL_POINTER )
   {
      DEBUG_ERROR("Unexpected NULL argument (%p)",SCpnt);
      return FAILED;
   }

   if ( SCpnt->device == NULL_POINTER )
   {
      DEBUG_ERROR("Failed to reset Device (invalid device pointer %p)",SCpnt->device);
      return FAILED;
   }

#endif

   scd = SCpnt->device;
   dev = DEVICE_INFO(SCpnt->device);
   port = PORT_TAG(SCpnt->device);

   if ( dev == NULL_POINTER )
   {
      DEBUG_ERROR("Failed to reset Device (invalid info %p)",SCpnt->device->hostdata);
      return FAILED;
   }

   status = DevHardReset(dev->dev_tag);
   DEBUG_ERROR("Dev Reset %d:%d:%d:%d, dev# %d: %s", scd->host->host_no, scd->channel, scd->id, scd->lun, dev->dev_tag, ((status == SATA_SUCCESS) ? "Success":"Failed"));
   ERROR_DUMP(DEVICE_PDEV(scd), dev->dev_tag);

   if ( status == SATA_SUCCESS )
   {
// #if 1
      DEBUG_ERROR_0("Reconfiguring device # %d", dev->dev_tag);
      dev->dev_identify.integrityWord = 0;
      status = gdx_config_device(SCpnt);
      if ( status == SATA_SUCCESS )
      {
          dev->online = 1;
         return SUCCESS;
//           return SCSI_RESET_SUCCESS;
      }
      else
      {
         DEBUG_ERROR("reconfigure device #%d failed", dev->dev_tag);
         if ( SCpnt->device != NULL_POINTER ) dev->online = 0;
         return FAILED;
      }
// #else
//         return SCSI_RESET_ERROR;
// #endif
   }
   else
   {
      DEBUG_ERROR("reseting device #%d failed", dev->dev_tag);
      if ( SCpnt->device != NULL_POINTER ) dev->online = 0;
      return FAILED;
   }
} /* end of gd31244_device_reset() */

/* SCSI information functions */
int gd31244_biosparam(struct scsi_device *disk, struct block_device *dev,sector_t capacity, int *info)
{
   int size;
   DEVICE_INFO_T *dev_info ;

   DEBUG_TRACE();

   dev_info = DEVICE_INFO(disk);

   if ( dev_info == NULL_POINTER )
   { // We don't have the infor from the drive. Make something up.

      size = capacity;
      /* H */ info[0] = 64; // 2 ^ 6
      /* S */ info[1] = 32; // 2 ^ 5
      /* C */ info[2] = size >> (6 + 5);
      if ( info[2] > 1024 )
      { //  big disk... Use alternage geometry
         /* H */ 
         info[0] = 255;
         /* S */ info[1] = 63;
         /* C */ info[2] = size / (255 * 63);
      }
   }
   else
   { // We have the infor from the drive. Pass it up.

      /* H */
      info[0] = dev_info->dev_identify.currentHeads;
      /* S */ info[1] = dev_info->dev_identify.currentSectorsPerTrack;
      /* C */ info[2] = dev_info->dev_identify.currentCylinders;
   }

   return 0;
} // gd31244_biosparam

/* gd31244_proc_info
 * Used if the driver currently has no own support for /proc/scsi
 * inout = 0 means read  (returns adapter infos)
 *       = 1 means write (parse user control command)
 */
int gd31244_proc_info(struct Scsi_Host *shost, char *buffer, char **start,off_t offset, int length, int inout)
{
   int len, pos, begin;
   int orig_length;
//   struct Scsi_Host *shp = first_shp;
   struct Scsi_Host *shp = shost;
   HOST_INFO_T   *hi     = HOST_INFO(shp);
   //int adapter_tag       = INVALID_TAG;
   int ctr = (shp == NULL_POINTER ? 0 : 1 );

//   char si_rev           = 0;
   int drive_num         = 0;
   int dev_tag           = INVALID_TAG;
   struct scsi_device *scdev    = NULL_POINTER;
   DEVICE_INFO_T *di     = NULL_POINTER;
   struct pci_dev *pdev  = NULL_POINTER;
   char features_on[256]  = "";
   char features_off[256] = "";
   char temp_str[40]      = ""; // NCQ info
   char fw[9], model[41], srno[21];

   DEBUG_TRACE();

   orig_length = length;

   if ( hi == NULL_POINTER )
   {
      // We did not find the host in our list...
      DEBUG_ERROR("This adapter (#%d) is not one of the adapters we know about.", hi->adapter_number);
      return -EINVAL;
   }

   if ( inout > 0 )
   {
      // Write to the proc file not yet supported.
      DEBUG_ERROR("%s #%d: Unsupported Feature", shp->hostt->name, ctr);
      return -EINVAL;
   }

   pdev = HOST_PDEV(shp);
   // pci_read_config_byte(pdev, 0x08, &si_rev);  // Read the Revision number

   begin = 0;
   pos = len = sprintf(buffer,
                       "Intel GD31244 Serial ATA driver Ver %s\n"
                       "gd31244: 1.5Gbps, %d ports PCI-X SATA Controller, Supports SATA extensions\n"
                       "         Adapter #%d (Tag %d) can Queue %d Commands total.\n"
                       "         (Max %d cmds per LUN, SG size = %d, Max Sectors = %d in current setup)\n",
                       gd31244_version_str,
                       MAX_PORTS_PER_CONTROLLER,
                       hi->adapter_number, hi->adapter_tag, shp->can_queue,
                       shp->cmd_per_lun, shp->sg_tablesize, shp->max_sectors
                      );

   for ( ctr = 0; ctr < MAX_DEVICES_PER_ADAPTER; ctr++ )
   {
      // We will be assuming that dev != NULL
      int host, channel, target, lun;
      int active_cmds = 0;
      int free_cmds = 0;
      STATUS_CODE_T drive_connected;

      // drive number under this adapter...
      drive_num = ctr ;
      drive_connected = 0;
      drive_connected = AdapterDevicePresent( hi->adapter_tag, drive_num, &dev_tag);

      dev_tag = hi->dev_tags[ctr];

      if ( dev_tag == INVALID_TAG )
      {
         if ( drive_connected == SATA_TRUE )
         {
            // This device is currently not attached...
            len += sprintf(buffer + len, "\nPort %d - Drive Connected but offline\n", ctr);
            ERROR_DUMP(pdev, INVALID_TAG);
         }
         else
         {

            // This device is currently not attached...
            len += sprintf(buffer + len, "\nPort %d - Not Connected\n", ctr);
         }
         continue;
      }
      else
      {
         len += sprintf(buffer + len, "\nPort %d - Connected at 1.5 Gbps\n", ctr);
      }

      scdev = (struct scsi_device *) DevReturnOsdmTag(dev_tag);
      if ( scdev == NULL_POINTER )
      {
         // Can't do anything with this one!
         DEBUG_ERROR("Scsi Drive #%d, Tag %d has invalid OSDM Tags",
                     drive_num, dev_tag);
         continue;
      }

      di = DEVICE_INFO(scdev);
      if ( di == NULL_POINTER )
      {
         DEBUG_ERROR("Scsi Drive #%d, Tag %d has invalid device_info",
                     drive_num, dev_tag);
         continue;
      }

      if ( di->online == 0 )
      {
         len += sprintf(buffer + len,
                        "         Drive #%d offline due to I/O errors\n", drive_num);
//            DEBUG_ERROR("Scsi Drive #%d, Tag %d is offline",
//                    drive_num, dev_tag);
         continue;
      }

      host    = scdev->host->host_no;
      channel = scdev->channel;
      target  = scdev->id;
      lun     = scdev->lun;


//      free_cmds = free_count(di); active_cmds = di->drive_features.QueueDepth - free_cmds;
      active_cmds = di->cmd_count.counter;
      free_cmds = di->drive_features.QueueDepth - active_cmds;

      strncpy(fw, di->dev_identify.firmwareRevision, 8);
      fw[8] = '\0';
      strncpy(model, di->dev_identify.modelNumber, 40);
      model[40] = '\0';
      strncpy(srno, di->dev_identify.serialNumber, 20);
      srno[20] = '\0';

      sprintf( temp_str, "\"NCQ, %d Commands\"", di->drive_features.QueueDepth);
      sprintf(features_on, "%s%s%s%s%s%s",
              di->drive_features.Atapi_6           ? "\"ATAPI6\" ": "",
              di->drive_features.Lba48             ? "\"LBA48\" ":"",
              di->drive_features.WriteCache        ? "\"Write Cache\" ":"",
              di->drive_features.LookAhead         ? "\"Look Ahead\" ":"",
              di->drive_features.SetFeaturesSpinup ? "\"Feature Spinup\" ":"",
              di->drive_features.TCQ               ? temp_str :""
             );

      sprintf(features_off, "%s%s%s%s%s%s",
              !di->drive_features.Atapi_6           ? "\"ATAPI6\" ": "",
              !di->drive_features.Lba48             ? "\"LBA48\" ":"",
              !di->drive_features.WriteCache        ? "\"Write Cache\" ":"",
              !di->drive_features.LookAhead         ? "\"Look Ahead\" ":"",
              !di->drive_features.SetFeaturesSpinup ? "\"Feature Spinup\" ":"",
              !di->drive_features.TCQ               ? "\"TCQ (NCQ)\" ":""
             );

      len += sprintf(buffer + len, "Drive number %d\n"
                     "        Model          \"%s\"\n"
                     "        Firmware       \"%s\"\n"
                     "        Serial Number  \"%s\"\n"
                     "        Size            %d MB\n"
                     "        HCTL            %d:%d:%d:%d\n"
                     "        ATag %d, DTag %d\n"
                     "        Commands: %d Active, %d Free"
#ifdef SATA_DEBUG_0
                     " Min # DDTEs %d, Max # DDTEs %d"
#endif
                     "\n        Features: \n"
                     "          ON : %s\n"
                     "          OFF: %s\n",
                     drive_num, model, fw, srno,
                     di->capacity / ( 2 /* Kb */ * 1024 /* MB*/ ),
                     host, channel, target, lun,
                     di->adapter_tag, di->dev_tag,
                     active_cmds, free_cmds,
#ifdef SATA_DEBUG_0
                     (di->min_ddte_count_seen < di->max_ddte_count_seen ? di->min_ddte_count_seen : 0),
                     di->max_ddte_count_seen,
#endif
                     features_on, features_off
                    );
#ifdef DEBUG
//      sprintf(dump_buffer, "%d:%d:%d:%d - %s,%s,%s - DevIdentify",host, channel, target, lun,
//              di->dev_identify.modelNumber, di->dev_identify.firmwareRevision,di->dev_identify.serialNumber);
//      debug_mem_dump(dump_buffer, (char *) &di->dev_identify, sizeof(di->dev_identify));
#endif
   }

   pos = len;
   if ( pos < offset )
   {
      len = 0;
      begin = pos;
   }
   *start = buffer + (offset - begin); /* Start of wanted data */
   len -= (offset - begin);
   if ( len > length )
      len = length;

   return(len);
}

int gd31244_ioctl(struct scsi_device *dev, int cmd, void * ptr)
{

   DEBUG_TRACE();

   if ( dev != NULL )
   {
      DEBUG_ERROR_0("Got unsupported ioctl(0x%x) for Drive %d:%d:%d:%d", cmd, dev->host->host_no, dev->channel, dev->id, dev->lun);
   }
   return -EPERM;
}

const char * gd31244_info(struct Scsi_Host * shp)
{
   //int host                 = shp->host_no;
   HOST_INFO_T   *host_info = HOST_INFO(shp);
   //int adapter_tag          = host_info->adapter_tag;

   DEBUG_TRACE();

   sprintf(host_info->info_string, "%s: Intel Sata Controller Driver Ver %s"
#ifdef SATA_DEBUG_0
           " Compiled on " __DATE__ ", " __TIME__
#endif
           ", Max %d Channels, Max %d devices per channel, Max %d Luns per device",
           shp->hostt->name, gd31244_version_str, shp->max_channel + 1, shp->max_id, shp->max_lun
          );

   return host_info->info_string;
}

/* Helper functions */


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function allocates the DDTs for the drive. It will free  */
/*              any DDTs that might have been already allocated by calling    */
/*              gd31244_free_ddts. if num_ddts is > 1, it will also setup the */
/*              vector table correctly.                                       */
/*                                                                            */
/* PRECONDITIONS    ge31244_device_init has been called for this drive.       */
/*                                                                            */
/* PARAMETERS   scd                                                           */
/*                  This is the pointer to the struct scsi_device structure          */
/*              num_ddts                                                      */
/*                  The number of DDT to be allocated. With TCQ disabled,     */
/*                  value will be 1, and with TCQ enabled, it will be some    */
/*                  number between 1 and 31 (both inclusive)                  */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = num_ddts is invalid                 */
/*              SATA_NOT_INITIALIZED    = Drive not initialized               */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T gd31244_allocate_ddts(struct scsi_device *scd, int num_ddts)
{

   int host                 = scd->host->host_no;
   int channel              = scd->channel;
   int target               = scd->id;
   int lun                  = scd->lun;
   int max_ddt_entries      = 0;
   int ddt_size             = 0;
   struct pci_dev * pdev    = DEVICE_PDEV(scd);
   STATUS_CODE_T retval     = SATA_SUCCESS;
   DEVICE_INFO_T *dev_info  = DEVICE_INFO(scd);
   int failed               = TRUE;
   int ctr;

   DEBUG_TRACE();

   /* Perform Basic sanity check... */
   if ( dev_info == NULL_POINTER )
   {
      DEBUG_ERROR("HCTL %d:%d:%d:%d - Not initialized !", host, channel, target, lun);
      return SATA_NOT_INITIALIZED;
   }

   if ( (num_ddts < 1) || (num_ddts > MAX_TCQ_DEPTH ) )
   {
      DEBUG_ERROR("HCTL %d:%d:%d:%d - bad number of DDTs requested - %d ", host, channel, target, lun, num_ddts);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   // Check if any DDTs were allocated. if they were, free them up and allocate a complete new set.
   if ( dev_info->ddt_orig[0].ddte != NULL_POINTER )
   {
      // Well, there were some ddt's allocated...
      gd31244_free_ddts(scd);
   }


   max_ddt_entries = AdapterGetMaxDdtEntries();
   max_ddt_entries = (LINUX_DDT_ENTRIES < max_ddt_entries) ?
                     LINUX_DDT_ENTRIES : max_ddt_entries;
   ddt_size = max_ddt_entries * DDT_ENTRY_BYTES;

   for ( ctr = 0; ctr < num_ddts; ctr++ )
   {
      // NPG: TBD: Determine if the DMA h/w needs the various DDT's to be physically contigious
      //           From What I understand, this is NOT so...

      // The DDT's have to be aligned on a 64K boundary. The buffer we allocate using
      // pci_alloc_consistent need not be aligned to 64 K boundary. The solution involves
      // allocating two times as much memory and then using the buffer starting at the
      // first 64K boundary in this region.

      /* Loop around and pre-allocate the DDT's for the drive. */
//        DEBUG_PRINT("Calling pci_alloc_consistent %s", "");
      dev_info->ddt_orig[ctr].ddte =
      pci_alloc_consistent(pdev,
                           (ddt_size * 2 ),
                           (dma_addr_t *)&dev_info->ddt_orig[ctr].ddtPhysicalAddress
                          );

      if ( dev_info->ddt_orig[ctr].ddte == NULL )
      {
         DEBUG_ERROR("Drive %d:%d:%d:%d: Failed to allocate memory for DDT",
                     host, channel, target, lun);

         retval = SATA_GET_MEMORY_FAILED;
         goto alloc_ddt_cleanup;
      }

      /* Assign correct values to ddt[ctr] */
      {
         UINT64 orig_phys = dev_info->ddt_orig[ctr].ddtPhysicalAddress;
         UINT64 new_phys;
         int shift_by;

         // Check if we are already at the correct boundary
         if ( (orig_phys + ddt_size) <=  (orig_phys | 0xffff) )
         {
            DEBUG_PRINT("Original Physical region is 64K aligned - %#0llx", orig_phys);
            // orig_phys is on a 64K boundary
            dev_info->ddt[ctr].ddte = dev_info->ddt_orig[ctr].ddte;
            dev_info->ddt[ctr].ddtPhysicalAddress = dev_info->ddt_orig[ctr].ddtPhysicalAddress;
         }
         else
         {
            // orig_phys is NOT on a 64K boundary
            new_phys = (orig_phys | 0xffff) + 0x1;
            shift_by = new_phys - orig_phys;

            DEBUG_PRINT("Physical region shifted from %#0llx to %#0llx", orig_phys, new_phys);

            dev_info->ddt[ctr].ddte = dev_info->ddt_orig[ctr].ddte + shift_by;
            dev_info->ddt[ctr].ddtPhysicalAddress = new_phys;
         }
      }

      /* DDT must not cross a 64Kb boundary... i.e. the start and the end address for the ddt */
      /* must have the same high order bits... */
      if ( (dev_info->ddt[ctr].ddtPhysicalAddress + ddt_size) >
           (dev_info->ddt[ctr].ddtPhysicalAddress | 0xffff )
         )
      {
         /* The high order bytes are NOT the same. :-( */
         DEBUG_ERROR("INTERNAL ERROR: Drive %d:%d:%d:%d: A DDT crossed a 64K boundary.", host, channel, target, lun);
         DEBUG_ERROR("start %#0llx, end %#0llx, size %d",dev_info->ddt[ctr].ddtPhysicalAddress,
                     dev_info->ddt[ctr].ddtPhysicalAddress + ddt_size,ddt_size);

         retval = SATA_FAILURE;
         goto alloc_ddt_cleanup;
      }

#ifdef QDEBUG
//		printk(KERN_INFO "DDT[%d] physical addr: %#08x\n", ctr,
//					dev_info->ddt[ctr].ddtPhysicalAddress);
#endif

#if 1  // since we change this all the time do we need to bother?
      // needs more investigation
      if ( dev_info->vector_table.vt != NULL_POINTER )
      {
         // Record the DDT address in the vector table.
         dev_info->vector_table.vt[ctr].ddtPhysicalAddress = dev_info->ddt[ctr].ddtPhysicalAddress;
         //        dev_info->vector_table.vt[ctr].ddtPhysicalAddress = 0;
      }
#endif

   } // for

   // OK. Make sure the Template indicates a Max SGLs as our max ddt entries (assuming 1:1 mapping) there is still a chance that
   // The DDT might become full... The expected value of max_ddt_entries is 512, and default value of sg_tablesize is 256

   if ( scd->host->hostt->sg_tablesize > max_ddt_entries )
   {
      scd->host->hostt->sg_tablesize = max_ddt_entries;
   }

   failed = FALSE;

   alloc_ddt_cleanup:

   // Cleanup code to free the resources allocated...
   if ( failed )
   {
      DEBUG_PRINT("Releasing DDTs %s", "");
      gd31244_free_ddts(scd);

      return(retval != SATA_SUCCESS ? retval : SATA_FAILURE );
   }

   return SATA_SUCCESS;

}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function will free all the DDT's allocated for the drive */
/*                                                                            */
/* PRECONDITIONS    ge31244_device_init has been called for this drive.       */
/*                                                                            */
/* PARAMETERS   scd                                                           */
/*                  This is the pointer to the struct scsi_device structure          */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_NOT_INITIALIZED    = Drive not initialized               */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T gd31244_free_ddts(struct scsi_device *scd)
{
   struct pci_dev * pdev    = DEVICE_PDEV(scd);
   //STATUS_CODE_T retval     = SATA_SUCCESS;
   DEVICE_INFO_T *dev_info  = DEVICE_INFO(scd);
   int ctr;
   int max_ddt_entries      = 0;
   int ddt_size             = 0;
//#ifdef DEBUG
   int host                 = scd->host->host_no;
   int channel              = scd->channel;
   int target               = scd->id;
   int lun                  = scd->lun;
//#endif

   DEBUG_TRACE();

   /* Perform Basic sanity check... */
   if ( dev_info == NULL_POINTER )
   {
      DEBUG_ERROR("HCTL %d:%d:%d:%d - Not initialized !", host, channel, target, lun);
      return SATA_NOT_INITIALIZED;
   }

   max_ddt_entries = AdapterGetMaxDdtEntries();
   max_ddt_entries = (LINUX_DDT_ENTRIES < max_ddt_entries) ?
                     LINUX_DDT_ENTRIES : max_ddt_entries;
   ddt_size = max_ddt_entries * DDT_ENTRY_BYTES;

   for ( ctr = 0; ctr < MAX_TCQ_DEPTH; ctr++ )
   {
      // The ddt_orig array was initialized to all zeros, so it it contains any values now,
      // it must be valid memory pointers.
      if ( dev_info->ddt_orig[ctr].ddte != NULL_POINTER )
      {
         pci_free_consistent(pdev,
                             ( ddt_size * 2 ),
                             dev_info->ddt_orig[ctr].ddte,
                             dev_info->ddt_orig[ctr].ddtPhysicalAddress
                            );

      }
   }

   // Now zero out the ddt arrays
   memset(dev_info->ddt_orig, 0, sizeof(DDT_T) * MAX_TCQ_DEPTH );
   memset(dev_info->ddt, 0, sizeof(DDT_T) * MAX_TCQ_DEPTH );

   return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function will call the Library API to issue all pending  */
/*              commands for a drive.                                         */
/*                                                                            */
/* PRECONDITIONS    Drive is operational.                                     */
/*                                                                            */
/* PARAMETERS   dev_info_ptr                                                  */
/*                  This is a pointer to the Drives device_info_s structure.  */
/*                                                                            */
/******************************************************************************/
void gd31244_queue_worker(void * dev_info_ptr)
{
   DEVICE_INFO_T *dev_info  = (DEVICE_INFO_T *) dev_info_ptr;
   STATUS_CODE_T status     = SATA_SUCCESS;

   DEBUG_TRACE();
#ifdef SANITY_CHECK
   if ( dev_info_ptr == NULL )
   {
      DEBUG_ERROR("Detected NULL parameter (%p)", dev_info_ptr);
      return;
   }
#endif

   status = DevServiceCommandQueue(dev_info->dev_tag);
   if ( status != SATA_SUCCESS )
   {
      // Queue the worker again for this drive...
      PREPARE_WORK(&dev_issue_pending, gd31244_queue_worker, (void *) dev_info);
      schedule_work(&dev_issue_pending);
      //      queue_work(gd31244_workqueue, &dev_issue_pending);
   }
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function will process a Device Change request after a    */
/*              hotplug event. If the indicated drive is newly added, it will */
/*              call scsi_add_device to configure it. If the drive was        */
/*              removed it will call scsi_remove_device                       */
/*                                                                            */
/* PRECONDITIONS  Adapter is operational.                                     */
/*                                                                            */
/* PARAMETERS   adapter_port                                                  */
/*                  This void pointer is treated as a 32 bit integer.         */
/*                  Top 16 bits hold the Adapter Number                       */
/*                  bottom 16 Bits hold Device num/Port num on this adapter   */
/*                                                                            */
/******************************************************************************/
static void gd31244_queue_dev_change(void * adapter_port)
{
   STATUS_CODE_T present = SATA_SUCCESS;
   int adapter_n         = (int)adapter_port >> 16;
   int dev_n             = (int)adapter_port & 0xffff;
   struct Scsi_Host *shp = first_shp;
   HOST_INFO_T      *hi  = HOST_INFO(shp);
   INT dev_tag           = INVALID_TAG;
   struct scsi_device *sdev = NULL;
   UINT32 active_commands;
   INT count;
   DEVICE_INFO_T *dev_info = NULL_POINTER;
   CMD_INFO_T *cmd_info    = NULL_POINTER;
   struct scsi_cmnd *SCpnt = NULL_POINTER;

   DEBUG_TRACE();
#ifdef SANITY_CHECK

   if(!TAG_WITHIN_RANGE(dev_n, MAX_DEVICES_PER_ADAPTER))
   {
       DEBUG_ERROR("ERROR: Invalid device number %d",dev_n);
       return ;
   }


   if(!TAG_WITHIN_RANGE(adapter_n, MAX_ADAPTERS))
   {
       DEBUG_ERROR("ERROR: Invalid Adapter number %d",adapter_n);
       return ;
   }

   if (( first_shp == NULL ) || ( hi == NULL))
   {
      DEBUG_ERROR("No known Adapters present. first_shp = %p, hi = %p", first_shp, hi);
      return;
   }
#endif

   DEBUG_ERROR("Dev Change on A# %d, D# %d (%#0x)", adapter_n, dev_n, (int) adapter_port);

   // Locate the Scsi Host structure.
   while ( (hi != NULL_POINTER) && (hi->next != NULL_POINTER) )
   {
      // Check if this is the adapter we are looking for.
      if (hi->adapter_number == adapter_n)
      {
         // Found it.
         break;
      }
      // go to the next instance...
      shp = hi->next;
      hi = HOST_INFO(shp);
   }


   if ((hi == NULL_POINTER) || (hi->adapter_number != adapter_n))
   {
      // We did not find the Requested Adapter :(
      DEBUG_ERROR("Unable to Locate A#%d (%p)", adapter_n, adapter_port);
      return;
   }

   present = AdapterDevicePresent( hi->adapter_tag, dev_n, &dev_tag);

   DEBUG_ERROR_0("A#%d D#%d %s DTag %d sdev %p", adapter_n, dev_n, 
                 present == SATA_TRUE ? "Present" : "Missing",hi->dev_tags[dev_n], hi->dev_sdev[dev_n]);

   // Determine if this is a Hot Push or a Hot pull
   if (hi->dev_tags[dev_n] == INVALID_TAG)
   {
      // Most likely a Hot Push.
      // Verify that the drive is present.
      if ( present != SATA_TRUE)
      {
         // Device physically not present, and we have no record of it !!!
         // Wonder how we got here... just ignore it.
         DEBUG_ERROR("Device A#%d:D#%d missing(Tag %d, sdev %p). Not configured", 
                     adapter_n, dev_n, hi->dev_tags[dev_n],hi->dev_sdev[dev_n] );
         return;
      }

      // Add this Device
      DEBUG_PRINT("Adding SCSI Dev on A%d:D%d", adapter_n, dev_n);
      sdev = scsi_add_device(shp, /* Channel */ dev_n, /* Id */ 0,/* LUN */ 0);
      hi->dev_sdev[dev_n] = sdev;
   }
   else
   {
      // Most likely a Hot Pull
      if ( present == SATA_TRUE)
      {
         // Device physically present, and we have a full record of it !!!
         // Wonder how we got here... just ignore it.
         DEBUG_ERROR("Device A#%d:D#%d Still connected. SDEV not released.", adapter_n, dev_n);
         return;
      }
      
      DEBUG_ERROR("Removing SCSI Dev from A%d:D%d. Sdev %p", adapter_n, dev_n, hi->dev_sdev[dev_n]);
      // Remove the Device
      if (hi->dev_sdev[dev_n] != NULL)
      {
         /* Need to abort all commands.  The call to scsi_remove_device()
            below does not appear to abort any outstanding commands.  This
            causes problems for two reasons:
              1) the programs performing i/o take a long time to time-out
              2) the scsi_remove_device() is called immediately and the
                 data structs are cleaned up but later on the scsi mid-
                 level times out the pending commands and when it attempts
                 to abort the commands and/or perform a bus reset things
                 can go crazy.
            So, the ideal solution appears to be:
              1) make sure we don't accept any additional commands
              2) abort all outstanding commands
              3) call scsi_device_remove, which should clean up our
                 data structs as well since scsi_device_remove eventually
                 ends up calling gd31244_slave_destroy().

         One more thing - gd31244_queuecommand() was also modified to 
         immediately reject commands if the device was not present. */

         sdev     = hi->dev_sdev[dev_n];
         dev_info = DEVICE_INFO(sdev);

         /* Change the device state to CANCEL (delete pending) to 
            prevent any additional commands from getting queued. */
         scsi_device_set_state(sdev, SDEV_CANCEL);
         
         /* This routine essentially returns the cached copy of sactive.*/
         DevGetActiveCommands(dev_info->dev_tag, &active_commands);

         /* Simply hack here - if we don't have TCQ but a command is 
            active just set the first bit in active commands since
            the command will be in the first slot. */
         if ( !dev_info->drive_features.TCQ )
         {
            if ( atomic_read(&dev_info->cmd_count) )
            {
              active_commands = 0x1;
            }
         }

         DEBUG_ERROR("dev_tag %d  scsidev %p  dev_info %p  cmd_count %d  active cmds %x",dev_info->dev_tag, sdev, dev_info, atomic_read(&dev_info->cmd_count), active_commands);
         if ( dev_info )
         {
            for ( count=0; count < dev_info->drive_features.QueueDepth; count++ )
            {
               DEBUG_ERROR("Checking slot %d", count);
               if ( ((active_commands >> count) & 0x1) )
               {
                  cmd_info = &dev_info->commands[count];

                  DEBUG_ERROR("Slot %d active, cmd_info: %p", count, cmd_info);
                  if ( cmd_info->cmd_state  )
                  {
                     SCpnt = cmd_info->SCpnt;

                     DEBUG_ERROR("cmd_state %d, SCpnt %p  - Calling ABORT()",cmd_info->cmd_state, SCpnt);
                     /* End the command */
                     gd31244_abort(SCpnt);
                  }
                  else
                  {
                     DEBUG_ERROR("cmd_state %d???",cmd_info->cmd_state);
                  }

               } // if(command active))

            } // for(each command slot)

         }  // if ( sanity check that dev_info pointer is valid )

         /* All commands aborted, now safe to remove device. */
         scsi_remove_device(hi->dev_sdev[dev_n]);
      }
      else
      {
         DEBUG_ERROR("Failed to remove SCSI Dev A%d:D%d. Invalid sdev pointer.",adapter_n, dev_n);
      }

   }
}

static irqreturn_t gd31244_isr(int irq, void *dev_id, struct pt_regs *regs)
{
   HOST_INFO_T *host              = NULL_POINTER;
   DEVICE_INFO_T *dev_info        = NULL_POINTER;
   struct scsi_device *scsi_dev   = NULL_POINTER;
   struct scsi_cmnd *SCpnt        = NULL_POINTER;
   int dev_tag                    = INVALID_TAG;
   UINT32 status                  = 0;
   CMD_INFO_T *cmd_info           = NULL_POINTER;
   int count;
   int atleast_one_interrupt      = FALSE;
   u32 cmd_done                   = 0;
   int cmds                       = 0;
   //int ctr;

   DEBUG_TRACE();

   ASSERT(dev_id);
   host = (HOST_INFO_T *)dev_id;

   // Check For hotplug interrupts
   status = AdapterHotPlugInterruptPending(host->adapter_tag);
   if ( status > 0 )
   {
      INT dev_tag = INVALID_TAG;
      void * vptr;

      // We have a Hotplug Interrupt on the HBA
      atleast_one_interrupt = TRUE;

      // Remember, for this new device, we won't have a DevTag under the adapters:drives array
      // till we call AdapterAttachDisk. Till then, the following while loop will not detect
      // that interrupt.

      DEBUG_ERROR("A#%d: Detected Hotplug Interrupt.(%#0x).", host->adapter_tag, status);

      // Check Which device
      for ( count = 0; count < 4; count++ )
      {
         // Bit 0= Port 0, Bit 8 = Port 1, Bit 16 = Port 2; Bit 24 = Port 3
         if ( READ_BIT(status, count * 8) )
         {
            // Phy State Change Detected...

            STATUS_CODE_T retval;
            dev_tag = INVALID_TAG;

               // Check if the device is present. It Calls AdapterServiceHotplug for new devices.
               retval = AdapterDevicePresent(host->adapter_tag, count, &dev_tag);

               DEBUG_PRINT("A#%d: Hotplug Interrupt.(%#0x). Port %d %s", host->adapter_tag, status,
                           count, retval == SATA_TRUE ? "Hot Push": "Hot Pull");

            // Schedule Device Change detection for this device.
            vptr = (void *) (( host->adapter_number << 16 ) | count);

            DEBUG_ERROR_0("Queued dev Change for A#%d, D#%d (%p)", host->adapter_number, count, vptr);
            PREPARE_WORK(&dev_change_pending, gd31244_queue_dev_change, vptr);
	    schedule_work(&dev_change_pending);
	    //            queue_work(gd31244_workqueue, &dev_change_pending);

            // Clear the Hotplug interrupt.
            DevClearInterrupts(dev_tag);
            continue;

         } // IPR bit set
      } // Check IPR bits for each port.
   } // Hotplug event ?

   /* During initial testing reducing count from 10 to 2 seemed to 
      help with hot pull. */
   //count = 10;
   count = 2;

   while ( ((dev_tag = AdapterInterruptPending(host->adapter_tag))
            != INVALID_TAG) && (count--) )
   {
      atleast_one_interrupt = TRUE;
      // pass dev tag to get dev_info
      scsi_dev = (struct scsi_device *)DevReturnOsdmTag(dev_tag);
      dev_info = DEVICE_INFO(scsi_dev);

      /* Need to disable the ASSERT() and actually enable the dev_info
         sanity checking.  That was a test case where a hot-plugged drive
         failed to initialize properly (the DevIdentify() routine failed
         because DRQ was never set) and this resulted in a failed
         initialization and the ISR getting a NULL pointer. 
         TODO: perhaps if DevIdentify() fails the device should be
               removed? */
      //ASSERT(dev_info);
      if ( dev_info == NULL_POINTER )
      {
         DEBUG_ERROR("ERROR: dev_info was null for dev_tag %d skipping...",dev_tag);
         continue;
      }
      status = DevIsr(dev_info->dev_tag, &cmd_done);
      if ( status == SATA_SUCCESS_COMPLETED )
      {
         if ( !dev_info->drive_features.TCQ )
         {
            cmd_done = 0x1;
         }

#ifdef QDEBUG
         DEBUG_PRINT("cmd_done: %#08x, queue depth: %d", cmd_done, dev_info->drive_features.QueueDepth);
#endif

#ifdef SEAGATE_NCQ_WORKAROUND
         if ( !cmd_done )
         {
            continue;
         }
#endif

         ASSERT(cmd_done > 0);

         /* XXX remove active queue */
         for ( cmds=0; cmds < dev_info->drive_features.QueueDepth; cmds++ )
         {
            // complete all done commands
            if ( !((cmd_done >> cmds) & 0x1) )
            {
               continue;
            }

#ifdef QDEBUG
            DEBUG_ERROR_0("cmd_done: %#08x, Completing Slot %d", cmd_done, cmds);
#endif
            cmd_info = &dev_info->commands[cmds];

            ASSERT(cmd_info);
            if ( cmd_info == NULL_POINTER )
            {
               DEBUG_ERROR("Unexpected interrupt,cmd_info was null for DTag %d. Ignoring the Interrupt.",dev_tag);
               continue;
            }

            if ( cmd_info->cmd_state != SATA_CMD_ACTIVE )
            {
               DEBUG_ERROR("**** ERROR: Ignoring command %d that is not active (%d)",cmds, cmd_info->cmd_state);
               continue;
            }

            cmd_info->cmd_state = SATA_CMD_FREE;
            SCpnt = cmd_info->SCpnt;
            // set all stat to good
            ASSERT(SCpnt);
            SCpnt->result = SCSI_STAT(0, 0, 0, 0);
            // Unmap the SGL
            if ( SCpnt->use_sg )
            {
               pci_unmap_sg(DEVICE_PDEV(SCpnt->device),
                            (struct scatterlist *)SCpnt->request_buffer,
                            cmd_info->sg_nents, cmd_info->sg_dma_direction);
            }
            else
            {
               pci_unmap_single(DEVICE_PDEV(SCpnt->device),
                                cmd_info->dma_addr, SCpnt->request_bufflen,
                                cmd_info->sg_dma_direction);
            }
            ASSERT(cmd_info->scsi_done);
            // call the SCSI callback
            DEBUG_PRINT_2("calling SCSI callback%s","");
//                spin_lock(&io_request_lock);
            cmd_info->scsi_done(SCpnt);
//                spin_unlock(&io_request_lock);

            freeq_push(dev_info, cmd_info);
            atomic_dec(&dev_info->cmd_count);
         } // end for()


         PREPARE_WORK(&dev_issue_pending, gd31244_queue_worker, (void *) dev_info);
	 schedule_work(&dev_issue_pending);
	 //         queue_work(gd31244_workqueue, &dev_issue_pending);

         // Try to service the queue again
//            gdx_service_queue(dev_info);

      }
      else if ( status == SATA_HOTPLUG_EVENT )
      {
         void * vptr;

         // OK. dev_tag had Hot plug event... This can only be a HotUnplug event as it was
         // associated with a DevTag (i.e. already attached)

         // Schedule a Drive Change event.
         // top 16 bits = adapter number, lower 16 bits = device number.
         vptr = (void *) (( host->adapter_number << 16 ) | scsi_dev->channel);

         DEBUG_ERROR_0("HotPull Intr. Dev Change on A#%d, D#%d (%p)", host->adapter_number, scsi_dev->channel, vptr);
         PREPARE_WORK(&dev_change_pending, gd31244_queue_dev_change, vptr);
	 schedule_work(&dev_change_pending);
	 //         queue_work(gd31244_workqueue, &dev_change_pending);

      }
      else if ( status == SATA_FAILURE )
      {
         DEBUG_PRINT("operation failure%s","");
         /* error */
         // Set device as busy so we can reject commands until
         // sense is pulled
         atomic_set(&dev_info->busy, TRUE);

         // single command support right now
         cmd_info = &dev_info->commands[0];
         ASSERT(cmd_info);
         if ( cmd_info == NULL_POINTER )
         {
            DEBUG_ERROR("Unexpected interrupt,cmd_info was null for DTag %d. Ignoring the Interrupt.",dev_tag);
            continue;
         }
         cmd_info->cmd_state = SATA_CMD_FREE;
         SCpnt = cmd_info->SCpnt;
         SCpnt->result = SCSI_STAT(0, 0, 0, CHECK_CONDITION);
         // Do I need to build sense information here?
         DEBUG_ERROR("interrupt indicated error%s","");
         ERROR_DUMP(DEVICE_PDEV(SCpnt->device), dev_info->dev_tag);
#ifdef SATA_DEBUG
         // For Debug, I have postponed interrupt clearing till this point.
         // Without the SATA_DEBUG flag, it will get cleared in DevIsr
         DevClearInterrupts(dev_info->dev_tag);
#endif
         // NPG: TBD: OK, so there was some error... We need to do something about it...
         // TBD: Implement error recovery methods other than abort and reset.

         // Unmap the SGL
         if ( SCpnt->use_sg )
         {
            pci_unmap_sg(DEVICE_PDEV(SCpnt->device),
                         (struct scatterlist *)SCpnt->request_buffer,
                         cmd_info->sg_nents, cmd_info->sg_dma_direction);
         }
         else
         {
            pci_unmap_single(DEVICE_PDEV(SCpnt->device),
                             cmd_info->dma_addr, SCpnt->request_bufflen,
                             cmd_info->sg_dma_direction);
         }
         ASSERT(cmd_info->scsi_done);
         // call the SCSI callback
//            spin_lock(&io_request_lock);
         cmd_info->scsi_done(SCpnt);
//            spin_unlock(&io_request_lock);
         freeq_push(dev_info, cmd_info);
         atomic_dec(&dev_info->cmd_count);
      }
      else
      {
         DEBUG_ERROR_0("DevIsr returned %d%s. will try %d times more.", status, (status == SATA_BUSY ? " - Busy" : ""), count);
      }

   } /* end while() */

   if ( ! atleast_one_interrupt )
   {
      // We don't expect this to be executed at all...

//       DEBUG_ERROR("ISR but no IRQ.Returning %d", IRQ_NONE);
//       ERROR_DUMP(host->pdev, INVALID_TAG); // Observed that there really was no interrupt!

      // We should correctly report if we serviced the interrupt. We could be sharing an 
      // interrupt with some other device... Linux kernel correctly handles ISR chains.
      return IRQ_NONE;
   }

   DEBUG_PRINT("exiting....%s","");

   return IRQ_HANDLED;
} /* end of gd31244_isr() */


static int gd31244_sync_disk(DEVICE_INFO_T *dev_info)
{
   int status = 0;
   SATA_TASKFILE_T taskfile;
   int dev_tag = INVALID_TAG;
   u8 ata_status;
   u32 ext_status;
   u32 error;

   DEBUG_TRACE();

//#ifdef SANITY_CHECK
   if((dev_info == NULL_POINTER) || (dev_info->dev_tag == INVALID_TAG))
   {
      DEBUG_ERROR("Ignored Flush to invalid device (%p)", dev_info);
      return SATA_FAILURE;
   }
//#endif

   dev_tag = dev_info->dev_tag;

#ifdef EXPERIMENTAL_CODE
   status = SelectDevice(dev_tag);
   if ( status != SATA_SUCCESS )
   {
      DEBUG_ERROR("Failed to select device #%d", dev_tag);
      return SATA_FAILURE;
   }
#endif

   memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));
   taskfile.command = dev_info->drive_features.Lba48 ?
                      ATA_FLUSH_CACHE_EXT : ATA_FLUSH_CACHE;

   // Issue ATA command and block wait for result (should we block??)
   status = DevExecuteRawCommand(dev_tag, TRUE, &taskfile);
   if ( status != SATA_SUCCESS )
   {
      DEBUG_ERROR("ATA command ATA_FLUSH_CACHE failed %d", status);
      // no need to check status, we failed
      return SATA_FAILURE;
   }

   // Should build sense information here
   if ( (status = DevGetStatus(dev_tag, &ata_status) == SATA_SUCCESS) )
   {
      if ( ata_status & ATA_ERR_BIT_MASK )
      {
         status = DevGetExtendedError(dev_tag, &ext_status,
                                      &error, &taskfile);
         DEBUG_ERROR("SATA extended status: %#08x", ext_status);
         DEBUG_ERROR("SATA error: %#08x", error);
         return SATA_FAILURE;
      }
   }
   else
   {
      DEBUG_ERROR("SATA status get failed %d", status);
      return SATA_FAILURE;
   }

   return SATA_SUCCESS;
} /* end of gd31244_sync_disk() */


/* This function is called once per 31244 */
static int __devinit gd31244_probe_one(struct pci_dev *pdev, const struct pci_device_id *dev_id)
{
   struct Scsi_Host *shp               = NULL_POINTER;
   HOST_INFO_T   *hi                   = NULL_POINTER;
   //   int devices_found                   = 0;
   int legacy_mode_devices_found       = 0;
   int ctr                             = 0;
   STATUS_CODE_T retval                = SATA_FAILURE;

   DEBUG_TRACE();

   // Display the copyright.
   printk(gd31244_copyright, gd31244_version_str);

   // Initialize the Library.
   InitLib();

   // Init structure fields...
   //   tpnt->proc_dir  = &proc_scsi_gd31244;
   driver_template.module = THIS_MODULE;

   /* Init the Work Queue and work */
   //   gd31244_workqueue = create_workqueue(gd31244_work_queue_name);
   INIT_WORK( &dev_issue_pending, gd31244_queue_worker, NULL);
   INIT_WORK( &dev_change_pending, gd31244_queue_dev_change, NULL);

#ifdef CONFIG_PCI

   DEBUG_PRINT("Found SATA Controller Name=\"%s\", Slot=\"%s\"", pdev->pretty_name, pci_name(pdev));

   // Initialize the Hardware
   retval = init_gd31244_HBA(pdev);

   if ( retval == SATA_OUT_OF_RANGE_VALUE ) // HBA not supported, IDE mode
   {
     legacy_mode_devices_found++;
     return retval;
   }
   else if ( retval != SATA_SUCCESS ) // Something went wrong.
   {
     // Can't do anything about this one...
     printk ( KERN_ERR "%s: Failed(%#0x) to initialize PCI device %s",driver_template.name, retval, pci_name(pdev));
     return retval;
   }

   // Initialize the SCSI host for this HBA
   retval = init_gd31244_Host(pdev);
   if ( retval != SATA_SUCCESS )
   {
     printk (KERN_ERR "%s: Failed(%#0x) to setup SCSI host %s",driver_template.name, retval, pci_name(pdev));
     return retval;
   }

   // Increment the devices found ...
   devices_found++;

   // Remember, we need to initialize the Adapter numbers in the Host_info structures...
   shp = first_shp;
   hi  = HOST_INFO(shp);
   ctr = 0;

   // follow the linked list and update the adapter numbers.
   while ( (shp != NULL_POINTER) && (hi != NULL_POINTER) )
   {
      // Initialize the adapter number
      hi->adapter_number = ctr;

      // Increment the Adapter number counter
      ctr++;

      // go to the next instance...
      shp = hi->next;
      hi = HOST_INFO(shp);
   }

#ifdef SANITY_CHECK
ASSERT(ctr == devices_found);
#endif

   if ( legacy_mode_devices_found > 0 )
   {
      printk(KERN_NOTICE "%s: Note: %d Legacy Mode GD31244 controller(s) detected. Leaving these controllers for Legacy IDE drivers.\n",
             driver_template.name, legacy_mode_devices_found);
   }

   // We are done detecting all the controllers...
   printk(KERN_NOTICE "%s: %d Controllers Detected.\n",driver_template.name, devices_found);

   return devices_found;

#else /* CONFIG_PCI */

#error PCI Support *Disabled* in Kernel - Cant support GD31244

   DEBUG_ERROR("Kernel doesn't support PCI. Can't probe for %s", driver_template.name );
   return 0;

#endif
} // gd31244_probe_one


MODULE_DEVICE_TABLE(pci, gd31244_pci_table);

static struct pci_device_id __devinitdata gd31244_pci_table[] = {
        { 0x8086, 0x3200, PCI_ANY_ID, PCI_ANY_ID, 0x10600, 0xFFFFFF, 0 },
        { 0x1725, 0x7174, PCI_ANY_ID, PCI_ANY_ID, 0x10600, 0xFFFFFF, 0 },
        { }
};

static struct pci_driver gd31244_pci_driver = {
        .name                   = "gd31244",
        .id_table               = gd31244_pci_table,
        .probe                  = gd31244_probe_one,
        .remove                 = __devexit_p(gd31244_remove_one),
};

static int __init gd31244_init(void)
{
  devices_found = 0;
  return pci_module_init(&gd31244_pci_driver);
}

static void __exit gd31244_exit(void)
{
  pci_unregister_driver(&gd31244_pci_driver);
}

module_init(gd31244_init);
module_exit(gd31244_exit);
