/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/
#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include <linux/delay.h> // for mdelay
#include <linux/types.h>
#include <linux/string.h>
#include <linux/genhd.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/proc_fs.h>
#include <linux/smp_lock.h>
#include <linux/pci.h>
#include <linux/list.h>
#include <asm/system.h>
#include <asm/scatterlist.h>
#include <asm/io.h>

#include <linux/blkdev.h> // #include <linux/blk.h> is obsolete
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_tcq.h>

#include <sata.h>
#include <sata_debug.h>

#include <gd31244.h>
#include <gd31244_satascsi.h>
#include <gd31244_lld.h>

/* Global Variables */
#ifdef DEBUG
static char dump_buffer[DUMP_BUFFER_SZ];
#endif

/* Local functions prototypes */
static int gdx_build_sync(struct scsi_cmnd *);
static int gdx_mode_sense(struct scsi_cmnd *SCpnt);
static int gdx_build_inquiry(struct scsi_cmnd *);
static int gdx_build_ddt_single(struct scsi_cmnd *SCpnt, CMD_INFO_T *cmd);
static int gdx_build_ddt(struct scsi_cmnd *, CMD_INFO_T *);
static int gdx_build_dataops10(struct scsi_cmnd *, void (*done) (struct scsi_cmnd *));
static int gdx_build_capacity(struct scsi_cmnd *);
static int gdx_set_features(DEVICE_INFO_T *dev_info, SATA_TASKFILE_T *taskfile);
//#ifdef EXPERIMENTAL_CODE
//static int gdx_identify_device(DEVICE_INFO_T *dev_info, void *buffer);
//#endif

/* Command line arg used during module install to determine whether to use NCQ */
/* NCQ should not be used due to compatibility issues with certain drives. See errata. */
static int gd31244_ncqenable = 0;
module_param(gd31244_ncqenable, int, 0644);
MODULE_PARM_DESC(gd31244_ncqenable, "Flag for en/disabling NCQ in driver(off=0, default on=1)");

//==============================================================================

inline int free_count(DEVICE_INFO_T *dev)
{
   int freecnt = 0;
   unsigned long flags;
   struct list_head *li = NULL;

   DEBUG_TRACE_2();

   if ( dev == NULL )
   {
      return 0;
   }

   spin_lock_irqsave(&dev->free_qlock, flags);

   li = dev->tcq_free.next;

   // Check for null for safety. Queue tail points back to the Head.
   while ( (li != NULL) && ( li != &dev->tcq_free) )
   {
      li = li->next;
      freecnt++;
   }

   DEBUG_PRINT("Cmd free: %d  Used %d", freecnt, dev->cmd_count.counter);

   spin_unlock_irqrestore(&dev->free_qlock, flags);

   return freecnt;
}

// Pops a command info context from the free queue
inline CMD_INFO_T * freeq_pop(DEVICE_INFO_T *dev)
{
    CMD_INFO_T *cmd_info = NULL;
    unsigned long  flags;

    DEBUG_TRACE_2();

    spin_lock_irqsave(&dev->free_qlock, flags);

#ifdef QDEBUG_DISABLED
    {
       int freecnt = 0;
       int lib_freecnt = 0;

       struct list_head *li = dev->tcq_free.next;

       // Check for null for safety. Queue tail points back to the Head.
       while ( (li != NULL) && ( li != &dev->tcq_free) )
       {
          li = li->next;
          freecnt++;
       }
       lib_freecnt = DevTCQCount(dev->dev_tag, NULL, 0);

       DEBUG_ERROR_0("Lib_Free %d, drv free: %d  Used %d",lib_freecnt, freecnt, dev->cmd_count.counter);
    }
#endif
    
    if(!queue_empty(&dev->tcq_free))
    {
        cmd_info = list_entry(dev->tcq_free.next, CMD_INFO_T, list);
        list_del(&cmd_info->list);
    }

    // Update the count of free commands.
//    dev->commands_free--;

    spin_unlock_irqrestore(&dev->free_qlock, flags);
    return cmd_info;
} /* end of freeq_pop() */

// Push a command info context onto the free queue
inline void freeq_push(DEVICE_INFO_T *dev, CMD_INFO_T *cmd_info)
{
    unsigned long  flags;

    DEBUG_TRACE_2();

    spin_lock_irqsave(&dev->free_qlock, flags);

    if (dev->drive_features.TCQ == FALSE)
    {
       // With NCQ Turned off, we re-use slot 0 for every command.
       // So stuff the last used command at the head of the queue
       list_add(&cmd_info->list, &dev->tcq_free);
    }
    else
    {
       // With NCQ turned on, we cycle through all the command slots
       // So we stuff the last used command at the end of the queue.
       list_add_tail(&cmd_info->list, &dev->tcq_free);
    }

    // Update the count of free commands.
//    dev->commands_free++;

#ifdef QDEBUG
    {
       int ctr;
       int freecnt = 0;
       int lib_freecnt = 0;
       char lq_str[80] = "";
       char astr[80] = "";

       struct list_head *li = dev->tcq_free.next;

       // Check for null for safety. Queue tail points back to the Head.
       while ( (li != NULL) && ( li != &dev->tcq_free) )
       {
          li = li->next;
          freecnt++;
       }
       lib_freecnt = DevTCQCount(dev->dev_tag, lq_str, 80);


       // Dump the Free and busy queues.
       // For some reason, the lib thinks this slot is in use...
       for ( ctr = 0; ctr < MAX_COMMANDS; ctr++ )
       {
          sprintf(astr + ctr, "%c",(dev->commands[ctr].cmd_state == 0 ? 'F':
                                     (dev->commands[ctr].cmd_state == 1 ? 'P':
                                      (dev->commands[ctr].cmd_state == 2 ? 'A': '?')))
                  );
       }


       // The Lib might complete multiple commands in one ISR... 
       // This function will get called once for each command...
       if (lib_freecnt < freecnt)
       {
          DEBUG_ERROR("ERROR: Slot %d L[%s] D[%s]", cmd_info->slot_no, lq_str, astr);
          DEBUG_ERROR_0("Lib_Free %d, drv free: %d Used %d",lib_freecnt, freecnt, dev->cmd_count.counter);
       }
    }
#endif
    
    spin_unlock_irqrestore(&dev->free_qlock, flags);
} /* end of freeq_push() */

// Code borrowed from ide.c
void fixstring (char *s, const int bytecount, const int byteswap)
{
    char *p = s, *end = &s[bytecount & ~1]; /* bytecount must be even */

    if(byteswap)
    {
        /* convert from big-endian to host byte order */
        for(p = end ; p != s;)
        {
            unsigned short *pp = (unsigned short *) (p -= 2);
            *pp = ntohs(*pp);
        }
    }

    /* strip leading blanks */
    while(s != end && *s == ' ')
        ++s;

    /* compress internal blanks and strip trailing blanks */
    while(s != end && *s)
    {
        if(*s++ != ' ' || (s != end && *s && *s != ' '))
            *p++ = *(s-1);
    }

    /* wipe out trailing garbage */
    while(p != end)
        *p++ = '\0';
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              Performs device identify and then setup the feature table     */
/*              using the dev_ident information and master table.             */
/*                                                                            */
/* PRECONDITIONS    NONE                                                      */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   DEVICE_INFO_T * - pointer to drive info context               */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_FAILURE                                                  */
/******************************************************************************/
int gdx_config_device(struct scsi_cmnd *SCpnt)
{
    PIDENTIFY_DEVICE_T dev_ident = NULL;
    PSATA_FEATURES_T features;
    PSATA_FEATURES_T master = &master_features_g;
    SATA_TASKFILE_T taskfile;
    DEVICE_INFO_T *dev_info;
    int status = SATA_SUCCESS;
    struct scsi_device *dev = SCpnt->device;

    DEBUG_TRACE();

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    dev_info = DEVICE_INFO(SCpnt->device);

    dev_ident = &dev_info->dev_identify;
    ASSERT(dev_ident);

    // Has this device already been initialized ?
    if (dev_info->dev_tag == INVALID_TAG)
    {
       // Hmm not initialized... OK. Let's initialize it here.
       DEBUG_PRINT("Drive %d:%d:%d:%d not Configured. Configuring it now...", dev->host->host_no, dev->channel, dev->id, dev->lun);

       status = gd31244_slave_configure(SCpnt->device);
       if (status != SATA_SUCCESS)
       {
          DEBUG_ERROR("Failed to initialize Drive %d:%d:%d:%d", dev->host->host_no, dev->channel, dev->id, dev->lun);
          return status;
       }
    }

    // If dev_ident not valid we do an ATA call
    if(!dev_ident->integrityWord)
    {
#ifndef IDENTDEV_IRQ
        status = DevIdentify(DRIVE_TAG(SCpnt->device), (void *)dev_ident);
#else
        status = gdx_identify_device(dev_info, (void *)dev_ident);
#endif
        if(status != SATA_SUCCESS)
        {
            DEBUG_ERROR_0("Device Identify failed(%d): Dtag %d",status, DRIVE_TAG(SCpnt->device));
            SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
// NPG: Don't clean up here...
//            gd31244_slave_destroy(SCpnt->device);
//            SCpnt->device->hostdata = NULL;
            return SATA_FAILURE;
        }
    }

    // set master feature table here
    features = &dev_info->drive_features;

    features->ReadMultiple = FALSE;

    if((master->Atapi_6 == FALSE) ||
       (dev_ident->majorVersionNumber & 0x0040))
    {
        features->Atapi_6 = FALSE;
    }
    else
    {
        features->Atapi_6 = TRUE;
    }

    features->ReadBuffer = FALSE;
    features->WriteBuffer = FALSE;
    features->HostProtectedArea = FALSE;
    features->DeviceResetCmd = FALSE;

    // ATA Write Cache configuration
    if(master->WriteCache == FALSE)
    {
        features->WriteCache = FALSE;
    }
    else
    {
        if(dev_ident->commandSetSupported0 & 0x0020)
        {
            if((dev_ident->commandSetDefault & 0xc000) && // feature valid
               (dev_ident->commandSetEnabled0 & 0x0020))
            { // w-cache enabled
                features->WriteCache = TRUE;
            }
            else
            {
                memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));

                taskfile.command = ATA_SET_FEATURES;
                taskfile.feature = ATA_ENABLE_WRITE_CACHE;

                status = gdx_set_features(dev_info, &taskfile);

                if(status == SATA_SUCCESS)
                {
                    features->WriteCache = TRUE;
                }
                else
                {
                    features->WriteCache = FALSE;
                }
            }
        }
        else
        {
            features->WriteCache = FALSE;
        }
    } /* end of WriteCache if() config */


    // Check look ahead feature
    if(master->LookAhead == FALSE)
    {
        features->LookAhead = FALSE;
    }
    else
    {
        if(dev_ident->commandSetSupported0 & 0x0040)
        {
            if((dev_ident->commandSetDefault & 0xc000) && // feature valid
               (dev_ident->commandSetEnabled0 & 0x0040))
            { // w-cache enabled
                features->LookAhead = TRUE;
            }
            else
            {
                memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));

                taskfile.command = ATA_SET_FEATURES;
                taskfile.feature = ATA_ENABLE_RLA;

                status = gdx_set_features(dev_info, &taskfile);
                if(status == SATA_SUCCESS)
                {
                    features->LookAhead = TRUE;
                }
                else
                {
                    features->LookAhead = FALSE;
                }
            }
        }
        else
        {
            features->LookAhead = FALSE;
        }
    } /* End of look ahead config if() */

    features->PwrUpStandby = FALSE;
    features->SecurityMode = FALSE;

    features->SMARTMode = FALSE;

    // power management
    if(master->MandatoryPM == FALSE)
    {
        features->MandatoryPM = FALSE;
    }
    else
    {
        features->MandatoryPM =
        (dev_ident->commandSetSupported0 & 0x0008) ? TRUE : FALSE;
    }

    if(master->MandFlushCache == FALSE)
    {
        features->MandFlushCache = FALSE;
    }
    else
    {
        features->MandFlushCache =
        (dev_ident->commandSetSupported1 & 0x1000) ? TRUE : FALSE;
    }

    if(master->FlushCacheExt == FALSE)
    {
        features->FlushCacheExt = FALSE;
    }
    else
    {
        features->FlushCacheExt =
        (dev_ident->commandSetSupported1 & 0x2000) ? TRUE : FALSE;
    }

    if(master->Lba48 == FALSE)
    {
        features->Lba48 = FALSE;
    }
    else
    {
        features->Lba48 =
        (dev_ident->commandSetSupported1 & 0x0400) ? TRUE : FALSE;
    }

    if(!features->Lba48)
    {
        /* If device is LBA28 then we can only address 0x100 sectors */
        SCpnt->device->host->max_sectors = 0x100;
	blk_queue_max_sectors(SCpnt->device->request_queue, 0x100);
    }

    features->AcousticMgt = FALSE;
    features->SetMaxSecurity = FALSE;

    // disable come up into standby
    if(master->SetFeaturesSpinup == FALSE)
    {
        features->SetFeaturesSpinup = FALSE;
    }
    else
    {
        if(dev_ident->commandSetSupported1 & 0x0200)
        {
            if((dev_ident->commandSetDefault & 0xc000) &&
               (dev_ident->commandSetEnabled1 & 0x2000))
            {
#if 0 // I don't think we care
                memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));

                taskfile.command = ATA_SET_FEATURES;
                taskfile.feature = ATA_DISABLE_PUSFS;

                status = gdx_set_features(dev_info, &taskfile);
                if(status == SATA_SUCCESS)
                {
                    features->SetFeaturesSpinup = FALSE;
                }
                else
                {
                    features->SetFeaturesSpinup = TRUE;
                }
#else
                features->SetFeaturesSpinup = TRUE;
#endif
            }
            else
            {
                features->SetFeaturesSpinup = FALSE;
            }
        }
    }

    DEBUG_PRINT("SetFeaturesSpinup: %d", features->SetFeaturesSpinup);

    features->AdvPowerMgt = FALSE;
    features->CFASupported = FALSE;
    features->DownLoadMicro = FALSE;

    // capacity setting up capacity for the device
    if(features->Lba48)
    { // LBA48 addressing
        // We can only support 32bit out of 48bit
        // Total space for this is actually defined 64bit
        dev_info->capacity = dev_ident->totalNumSectors48Bit[0] |
                             (dev_ident->totalNumSectors48Bit[1] << 16);
    }
    else
    {
        dev_info->capacity = dev_ident->totalNumSectors[0] |
                             (dev_ident->totalNumSectors[1] << 16);
    }

    memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));

    if(dev_ident->ultraDmaMode & 0x10)
    {
        taskfile.sectorCount = 0x44 /*ATA_UDMA_MODE4*/;
    }

    if(dev_ident->ultraDmaMode & 0x20)
    {
        taskfile.sectorCount = 0x45 /*ATA_UDMA_MODE5*/;
    }

    taskfile.command = ATA_SET_FEATURES;
    taskfile.feature = 3 /*ATA_SET_TRANSFER_MODE*/;

    if(taskfile.sectorCount)
    {
        DEBUG_PRINT("UDMA set sector count: %#08x", taskfile.sectorCount);

        status = gdx_set_features(dev_info, &taskfile);
        if(status != SATA_SUCCESS)
        {
            DEBUG_PRINT("Setting UDMA mode failed!%s","");
        }
    }

    // NOTE: TCQ/NCQ should NOT be used due to compatibility issues.
    // Please see spec update/Errata.
    if(features->Lba48 && master->TCQ && gd31244_ncqenable)
    {
        if(dev_ident->sataCapabilities & 0x100)
        {
            features->TCQ = TRUE;
            features->QueueDepth = dev_ident->queueDepth + 1;
        }
        else
        {
            features->TCQ = FALSE;
            features->QueueDepth = 1;
        }
    } // end TCQ if
    else
    {
        features->TCQ = FALSE;
        features->QueueDepth = 1;
    }

#ifdef DISABLE_NCQ
        features->TCQ = FALSE;
        features->QueueDepth = 1;
#endif

    /* Linux has the limitation of command per LUN configured globally at  */
    /* per adapter level and not at per device level. In order to minimize */
    /* complexity we are enforcing the rule that we will queue the         */
    /* max of the least capable a device can support out of all devices    */
    /* attached to the adapter. So if you want TCQ support all drives      */
    /* should support TCQ for maximum performance. Else the TCQ capable    */
    /* drive will only queue 1 command.                                    */
    if(features->QueueDepth <= SCpnt->device->host->cmd_per_lun)
    {
        SCpnt->device->host->cmd_per_lun = features->QueueDepth;
    }

    scsi_adjust_queue_depth(SCpnt->device,
                            (features->TCQ) ? MSG_ORDERED_TAG: 0 /* non-tagged(0) */,
                            features->QueueDepth); 
    DEBUG_ERROR_0("Drive %d:%d:%d:%d. Adjusted TCQ to %d", 
                  SCpnt->device->host->host_no, SCpnt->device->channel, SCpnt->device->id, SCpnt->device->lun, features->QueueDepth);

    // We do a new identify to update local stored information
#ifndef IDENTDEV_IRQ
    status = DevIdentify(DRIVE_TAG(SCpnt->device), (void *)dev_ident);
#else
    status = gdx_identify_device(dev_info, (void *)dev_ident);
#endif
    if(status != SATA_SUCCESS)
    {
        DEBUG_ERROR("Device Identify 2 failed%s","");
        SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
        ERROR_DUMP(DEVICE_PDEV(SCpnt->device), DRIVE_TAG(SCpnt->device));
        return SATA_FAILURE;
    }

    // Setting the feature table for the library
    DevSetFeatures(DRIVE_TAG(SCpnt->device), &dev_info->drive_features);

    if(features->TCQ)
    {
        // rebuilding DDTs
        status = gd31244_allocate_ddts(SCpnt->device, features->QueueDepth);
        if(status != SATA_SUCCESS)
        {
            DEBUG_ERROR("Reallocating DDTs failed!%s","");
            SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
            return SATA_FAILURE;
        }

        // Now we need to setup the vector table!
        status = DevSetDDTVectorTable(DRIVE_TAG(SCpnt->device),
                                      dev_info->vector_table.vtPhysicalAddress);
    }


#ifdef QDEBUG
// List out the Queue
    {
       int freecnt = 0;
       freecnt = free_count(dev_info);
       DEBUG_PRINT("Cmd free: %d  Used %d", freecnt, dev_info->cmd_count.counter);
    }
#endif

    // Fix the Model, Firmware and serial number strings...
    fixstring(dev_ident->modelNumber, sizeof(dev_ident->modelNumber), 1);
    fixstring(dev_ident->firmwareRevision,sizeof(dev_ident->firmwareRevision),1);
    fixstring(dev_ident->serialNumber,sizeof(dev_ident->serialNumber), 1);

    DEBUG_PRINT("device configuration complete%s","");
    return SATA_SUCCESS;
} /* end of gdx_config_device() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              Translates SYNCHRONIZE_CACHE to ATA_FLUSH_CACHE command.      */
/*                                                                            */
/* PRECONDITIONS    NONE                                                      */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*                                                                            */
/******************************************************************************/
static int gdx_build_sync(struct scsi_cmnd *SCpnt)
{
    int status = 0;
    DEVICE_INFO_T *dev_info;
    SATA_TASKFILE_T taskfile;
    u8 ata_status;
    u32 ext_status;
    u32 error;

    DEBUG_TRACE();

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    dev_info = DEVICE_INFO(SCpnt->device);
#ifdef EXPERIMENTAL_CODE
    status = SelectDevice(DRIVE_TAG(SCpnt->device));
    if(status != SATA_SUCCESS)
    {
        DEBUG_ERROR("Failed to select device%s","");
        return SATA_FAILURE;
    }
#endif
    memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));
    taskfile.command = dev_info->drive_features.Lba48 ?
                       ATA_FLUSH_CACHE_EXT : ATA_FLUSH_CACHE;

    // Issue ATA command and block wait for result (should we block??)
    status = DevExecuteRawCommand(DRIVE_TAG(SCpnt->device), TRUE, &taskfile);
    if(status != SATA_SUCCESS)
    {
        DEBUG_ERROR("ATA command ATA_FLUSH_CACHE failed%s","");
        // no need to check status, we failed
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    // Should build sense information here
    if((status = DevGetStatus(DRIVE_TAG(SCpnt->device),
                              &ata_status) == SATA_SUCCESS))
    {
        if(ata_status & ATA_ERR_BIT_MASK)
        {
            status = DevGetExtendedError(DRIVE_TAG(SCpnt->device), &ext_status,
                                         &error, &taskfile);
            DEBUG_ERROR("SATA extended status: %#08x", ext_status);
            DEBUG_ERROR("SATA error: %#08x", error);
            SCpnt->result = SCSI_STAT(0, 0, 0, CHECK_CONDITION);
            return SATA_FAILURE;
        }
    }
    else
    {
        DEBUG_ERROR("SATA status get failed%s","");
        SCpnt->result = SCSI_STAT(0, 0, 0, CHECK_CONDITION);
        return SATA_FAILURE;
    }

    return SATA_SUCCESS;
} /* end of gdx_build_sync() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              Builds SENSE information from status/error registers.         */
/*                                                                            */
/* PRECONDITIONS    NONE                                                      */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*              int - 0x70 current command, 0x71 previous command             */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*                                                                            */
/******************************************************************************/
int gdx_build_sense(struct scsi_cmnd *SCpnt, int which)
{
    DEVICE_INFO_T *dev_info;

    DEBUG_TRACE();

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif


    // if sense buffer already valid, don't bother building it
    if(SCpnt->sense_buffer[0] & 0x80)
    {
        return SATA_SUCCESS;
    }

    dev_info = DEVICE_INFO(SCpnt->device);

    SCpnt->sense_buffer[0] = which | 0x80; // set valid bit also (0x80)

    // Will have to collect error and do proper sense buffer sometimes
    SCpnt->sense_buffer[2] = NO_SENSE;

    // got sense, device no longer busy
    atomic_set(&dev_info->busy, FALSE);

    return SATA_SUCCESS;
} /* end of gdx_build_sense() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              Builds READ_CAPACITY return buffer                            */
/*                                                                            */
/* PRECONDITIONS    DEVICE_IDENTIFY has been performed                        */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*                                                                            */
/******************************************************************************/
static int gdx_build_capacity(struct scsi_cmnd *SCpnt)
{
    DEVICE_INFO_T * dev_info;
    u32 *temp = NULL;
    u32 tmp32;
    u8 *cmd;
    u32 block_len;
    u8 *cap = NULL;
    u8 *block_ptr;
    struct scatterlist  *sg = NULL;

    DEBUG_TRACE();

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    dev_info = DEVICE_INFO(SCpnt->device);
    cmd = SCpnt->cmnd;

    // READ CAPACITY return buffer must be 8 bytes long
    if(SCpnt->request_bufflen < 8)
    {
        DEBUG_ERROR("request buffer too small: %d!", SCpnt->request_bufflen);
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    // Quick fix for misaligned 32 bit access...    
//    temp = (u32 *)&cmd[2];
    tmp32 = cmd[5] << 24 | cmd[4] << 16 | cmd[3] << 8 | cmd[2];
    temp = &tmp32;

    // if PMI bit is 0 and LBA is not zero we have bad request
    if(!(cmd[8] & 0x01) && (*temp != 0))
    {
        // CHECK CONDITION status must be set
        // sense key must be set to ILLEGAL REQUEST
        // additional sense code set to ILLEGAL FIELD IN CDB
        SCpnt->result = SCSI_STAT(0, 0, 0, CHECK_CONDITION);
        SCpnt->sense_buffer[0] = 0x70 | 0x80;
        SCpnt->sense_buffer[2] = ILLEGAL_REQUEST;
        // SCpnt->sense_buffer[13] = ILLEGAL_FIELD_IN_CDB;
        return SATA_FAILURE;
    }

    // Need to make sure DEV IDENTIFY has been done already

    if(dev_info->capacity <= 0)
    {
        DEBUG_ERROR("INQUIRY has not been done!%s","");
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    // Setting block length in bytes to SCSI value
    block_len = 512;
    block_ptr = (u8 *)&block_len;
    if (SCpnt->use_sg) 
	{
        sg = (struct scatterlist *)SCpnt->request_buffer;
        cap = (u8 *)(page_address(sg->page) + sg->offset);
    }
    else 
	{
        cap = (u8 *)SCpnt->request_buffer;
	}

    cap[4] = block_ptr[3];
    cap[5] = block_ptr[2];
    cap[6] = block_ptr[1];
    cap[7] = block_ptr[0];

    // setting LBA
    block_ptr = (u8 *)&dev_info->capacity;
    cap[0] = block_ptr[3];
    cap[1] = block_ptr[2];
    cap[2] = block_ptr[1];
    cap[3] = block_ptr[0];

    return SATA_SUCCESS;
} /* end of gdx_build_capacity() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*            Builds the inquiry data return for the host adapter             */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_FAILURE                                                  */
/******************************************************************************/
int gdx_adapter_inquiry(struct scsi_cmnd *SCpnt)
{
    DEVICE_INFO_T * dev_info;
    int status = SATA_SUCCESS;
    u8 *inquiry;
    char vendor[] = "Intel";
    char model[] = "GD31244";
    char version[] = "A0";
    u8 *cmd;
    u8 *temp_scsi = NULL;
    struct scatterlist *sg = NULL;


    DEBUG_TRACE();

    // Any error in SCpnt->result is considered no device attached

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    dev_info = DEVICE_INFO(SCpnt->device);
    if(SCpnt->use_sg) {
        sg = (struct scatterlist *)SCpnt->request_buffer;
        inquiry = (u8 *)(page_address(sg->page) + sg->offset);
        printk(KERN_ERR "gdx_adapter_inq: sg=%p page=%p\n", sg, sg->page);
    }
    else 
	{
        inquiry = (u8 *)SCpnt->request_buffer;
    }
    
    cmd = SCpnt->cmnd;

    // queue_cmd() should have checked and make sure device valid for command

    // standard inquiry shouldn't be less than 36 bytes
    if(SCpnt->request_bufflen < 36)
    {
        DEBUG_ERROR("INQUIRY buffer less than 36 bytes%s","");
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    if(cmd[1] & 0x1)
    { // if EVDP bit is set
        DEBUG_ERROR("EVDP not supported yet!%s","");
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    // Setting up INQUIRY return data
    inquiry[0] = TYPE_PROCESSOR;   // controller processor
    inquiry[1] = 0;   // not removable
    inquiry[2] = 2;     // SCSI II(2) compliant
    inquiry[3] = 2;     // Data Format SCSI II
    inquiry[4] = 31;    // Additional Length = 30
    inquiry[7] = 0x2;     // no softreset

    // Vendor ID from DevId
    temp_scsi = &inquiry[8];
    memset(temp_scsi, 0, 8);
    memcpy(temp_scsi, vendor, strlen(vendor));

    temp_scsi = &inquiry[16];
    memset(temp_scsi, 0, 16);
    memcpy(temp_scsi, model, strlen(model));

    // Revision from DevId
    temp_scsi = &inquiry[32];
    memset(temp_scsi, 0, 4);
    memcpy(temp_scsi, version, strlen(version));

    // We don't do vendor specific stuff unless we NEED to!

    // Make sure we only handle up to 10 byte commands
    SCpnt->device->host->max_cmd_len = 10;

    DEBUG_PRINT("host INQUIRY build complete%s","");
    return status;

} /* end of gdx_adapter_inquiry() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*            Builds the inquiry data return                                  */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_FAILURE                                                  */
/******************************************************************************/
static int gdx_build_inquiry(struct scsi_cmnd *SCpnt)
{
    DEVICE_INFO_T * dev_info;
    int status = SATA_SUCCESS;
    //int i = 0;
    u8 *inquiry;
    u8 *temp_scsi = NULL;
    u8 *temp_ata = NULL;
    u8 *cmd;
    //int scount = 0;
    struct scatterlist *sg = NULL;


    DEBUG_TRACE();

    // Any error in SCpnt->result is considered no device attached

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    dev_info = DEVICE_INFO(SCpnt->device);

    if (SCpnt->use_sg) 
	{
        sg = (struct scatterlist *)SCpnt->request_buffer;
        inquiry = (u8 *)(page_address(sg->page) + sg->offset);
    }
    else 
	{
        inquiry = (u8 *)SCpnt->request_buffer;
	}

    cmd = SCpnt->cmnd;

    // queue_cmd() should have checked and make sure device valid for command

    // standard inquiry shouldn't be less than 36 bytes
    if(SCpnt->request_bufflen < 36)
    {
        DEBUG_ERROR("INQUIRY buffer less than 36 bytes%s","");
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    if(cmd[1] & 0x1)
    { // if EVDP bit is set
        DEBUG_ERROR("EVDP not supported yet!%s","");
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        return SATA_FAILURE;
    }

    // We don't support anything that's not LUN 0

    //    if((SCpnt->lun > 0) || (SCpnt->device->online == FALSE))
    // NPG-TBD: How do we know which LUN the command is intended for ?

    if(dev_info->online == 0)
    {
        SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
        return SATA_SUCCESS;
    }

#if 0
    DEBUG_PRINT("Detecting device presence...%s","");
    status = AdapterDevicePresent(ADAPTER_TAG(SCpnt->device),
                                  DRIVE_TAG(SCpnt->device), NULL);
    if(status == SATA_TRUE)
    {
        DEBUG_PRINT("Device found!%s","");
        status = 0;
        SCpnt->device->online = 1;
    }
    else
    {
        DEBUG_ERROR("Device NOT present!%s","");
        SCpnt->device->online = 0;
        SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
        status = SATA_FAILURE;
    }
#endif


#if 0
    if(status == SATA_SUCCESS)
    {
#endif
        status = gdx_config_device(SCpnt);
        if(status != SATA_SUCCESS)
        {
            dev_info->online = 0;
            SCpnt->result = SCSI_STAT(0, DID_NO_CONNECT, 0, 0);
            status = SATA_FAILURE;
            return status;
        }
#if 0
    }
#endif


    // Setting up INQUIRY return data
    inquiry[0] = TYPE_DISK;   // disk drive, device connected
    inquiry[1] = 0;   // not removable
    inquiry[2] = 2;     // SCSI II(2) compliant
    inquiry[3] = 2;     // Data Format SCSI II
    inquiry[4] = 31;    // Additional Length = 30
    inquiry[7] = 0x2;     // no softreset


    // Vendor ID from DevId
    temp_scsi = &inquiry[8];
    memset(temp_scsi, 0, 8);

    temp_ata = (u8 *)&dev_info->dev_identify.modelNumber;

    temp_scsi = &inquiry[16];
    memset(temp_scsi, 0, 16);
    memcpy(temp_scsi, temp_ata, 16);

    // Revision from DevId
    temp_scsi = &inquiry[32];
    memset(temp_scsi, 0, 4);
    temp_ata = (u8 *)&dev_info->dev_identify.firmwareRevision;
    memcpy(temp_scsi, temp_ata, 4);
    // We don't do vendor specific stuff unless we NEED to!

    // Make sure we only handle up to 10 byte commands
    SCpnt->device->host->max_cmd_len = 10;
    DEBUG_PRINT("INQUIRY build complete%s","");
    return status;

} /* end of gdx_build_inquiry() */

static int gdx_build_ddt_single(struct scsi_cmnd *SCpnt, CMD_INFO_T *cmd)
{

    DEVICE_INFO_T * drive_info;
    DDT_ENTRY_T        *ddte = NULL_POINTER;
    int count = 0;           /* number of DDT entries used. */
    int max_ddt_entries;
    DDT_T *ddt;

    DEBUG_TRACE();

    ASSERT(SCpnt);
    ASSERT(cmd);

    ddt = cmd->ddt_ptr;

    ASSERT(ddt);
    ddte = ddt->ddte;

    drive_info = DEVICE_INFO(SCpnt->device);

    if((SCpnt->cmnd[0] == READ_6) || (SCpnt->cmnd[0] == READ_10))
    {
        cmd->sg_dma_direction = PCI_DMA_FROMDEVICE;
    }
    else
    {
        cmd->sg_dma_direction = PCI_DMA_TODEVICE;
    }

    max_ddt_entries = drive_info->max_ddt_entries;
    max_ddt_entries = ( LINUX_DDT_ENTRIES < max_ddt_entries ) ?
                      LINUX_DDT_ENTRIES : max_ddt_entries;

    ASSERT(SCpnt->request_buffer);

    cmd->dma_addr = pci_map_single(DEVICE_PDEV(SCpnt->device), SCpnt->request_buffer,
                                   SCpnt->request_bufflen, cmd->sg_dma_direction);

    /* Fill Out the DDT using the sg_table. */

    if(sizeof(dma_addr_t) > 4)
    {
        /* Common Upper 32 bits of the 64 bit address */
        drive_info->ddt->ddtePage = UPPER_32_BITS((UINT64)cmd->dma_addr);
    }
    else
    {
        drive_info->ddt->ddtePage = 0;
    }

    count = 0; /* Records number of DDT entries used. */
    ddte = ddt->ddte; /* Point to start of the table... */

    {
        u32 cur_addr;
        u32 cur_len;

        /* All the buffers should be located in the same 4Gb segment */
        if(drive_info->ddt->ddtePage !=
           UPPER_32_BITS((UINT64)cmd->dma_addr))
        {
            DEBUG_PRINT("Buffers not located in the same 4Gb range.%s","");
            goto ddt_failed;
        }

        /* Take care of 64 bit addresses */
        cur_addr = cmd->dma_addr;
        cur_len  = SCpnt->request_bufflen;

        /*
         * Fill in the dma table, without crossing any 64kB boundaries.
         * Artisea requires 16-bit alignment of all blocks
         */

        while(cur_len > 0)
        {
            /* For Artisea, 0x0000 == 64Kb */
            u32 xcount;
            u32 bcount = 0x10000 - (cur_addr & 0xffff);

            if(count++ >= max_ddt_entries)
            {
                /* Used up All DDT entries. Can't use DMA for this data transfer */
                DEBUG_PRINT("DDT Full.%s","");
                goto ddt_failed;
            }

            if(bcount > cur_len)
            {
                bcount = cur_len;
            }

            xcount = bcount & 0xffff;

            ddte->baseAddress = cpu_to_le32(cur_addr);
            ddte->byteCount = cpu_to_le32(xcount);

            ddte->EOT = RESET_BIT(0, DDT_ENTRY_EOT_BIT);
            cur_addr += bcount;
            cur_len  -= bcount;

            ddte++;  /* Next DDT entry...*/
        }
    }
    /* Last entry... */
    ddte--;
    ddte->EOT = SET_BIT(0, DDT_ENTRY_EOT_BIT);

    DEBUG_PRINT("DDT has %d entries.", count);

#ifdef SATA_DEBUG_2

    printk(KERN_DEBUG "\n=======================DDT==========================\n");
    debug_mem_dump("DDT", (char *) ddt->ddte, count * sizeof(DDT_ENTRY_T));
    printk(KERN_DEBUG "\n====================================================\n");

//    DEBUG_PRINT("SG_ADDR: %#08x %#08x", cmd->dma_addr, SCpnt->request_bufflen);
#endif

    return SATA_SUCCESS;

    ddt_failed:
    DEBUG_ERROR("Failed to build DDT%s","");

    pci_unmap_single(DEVICE_PDEV(SCpnt->device), cmd->dma_addr,
                     SCpnt->request_bufflen, cmd->sg_dma_direction);

    return SATA_FAILURE; /* ddt building failed */
} /* end of gdx_build_ddt_single() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*            Translate SCSI scatterlist from OS to ATA DDT. It make sure     */
/*            memory do not cross 64k boundary                                */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*              DDT_T * - DDT to be used for translation                      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_FAILURE                                                  */
/******************************************************************************/
static int gdx_build_ddt(struct scsi_cmnd *SCpnt, CMD_INFO_T *cmd)
{
    DEVICE_INFO_T * drive_info = NULL_POINTER;
    struct scatterlist *sg     = NULL_POINTER;
    DDT_ENTRY_T        *ddte   = NULL_POINTER;
    int num_dma_buffers        = 0; /* Number of DMA buffers to transfer. May be less than nents */
    int count                  = 0;           /* number of DDT entries used. */
    int max_ddt_entries        = 0;
    DDT_T *ddt                 = NULL_POINTER;

    DEBUG_TRACE();

#ifdef SANITY_CHECK_IN_DATA_PATH
    if ((SCpnt == NULL) || (cmd == NULL))
    {
       DEBUG_ERROR("Detected Null Params. SCpnt = %p, cmd = %p", SCpnt, cmd);
       return SATA_FAILURE;
    }
#endif
    
    drive_info = DEVICE_INFO(SCpnt->device);
    sg = (struct scatterlist *)SCpnt->request_buffer;
    ddt = cmd->ddt_ptr;

#ifdef SANITY_CHECK_IN_DATA_PATH
    if ((drive_info == NULL) || (ddt == NULL))
    {
       DEBUG_ERROR("Detected Null drive_info/DDT. drive_info = %p, ddt = %p", drive_info, ddt);
       return SATA_FAILURE;
    }
#endif

    ddte = ddt->ddte;

    /****************************************************************************

    Notes: Restrictions imposed by Artisea
    1. The data buffers must be aligned on a WORD boundary
    2. Data block must not cross a 64Kb boundary
    3. The upper 32 bits of the pyhsical addresses of the buffers for all the DDT
       entries must be same.

    ******************************************************************************/

    if(!SCpnt->use_sg && (SCpnt->request_buffer != NULL))
    {
        return gdx_build_ddt_single(SCpnt, cmd);
    }

    if(sg == NULL_POINTER)
    {
        DEBUG_ERROR("Drive %d: Failed to build DDT - Missing SG table and request_buffer",
                    DRIVE_TAG(SCpnt->device));
        DEBUG_ERROR_0("use_sg %d, buffer %p, req_buffer %p, bufflen %d, req_bufflen %d", SCpnt->use_sg, SCpnt->buffer, SCpnt->request_buffer, SCpnt->bufflen, SCpnt->request_bufflen);
        return SATA_FAILURE;
    }

    /* NPG: Note: sg_dma_direction is common to all commands for a drive. Currently we */
    /* have only ONE command per drive so it's OK. This will need to be fixed for TCQ  */

    if((SCpnt->cmnd[0] == READ_6) || (SCpnt->cmnd[0] == READ_10))
    {
        cmd->sg_dma_direction = PCI_DMA_FROMDEVICE;
    }
    else
    {
        cmd->sg_dma_direction = PCI_DMA_TODEVICE;
    }

    max_ddt_entries = drive_info->max_ddt_entries; //AdapterGetMaxDdtEntries();
    max_ddt_entries = ( LINUX_DDT_ENTRIES < max_ddt_entries ) ?
                      LINUX_DDT_ENTRIES : max_ddt_entries;

    /*
    NPG: Algo note:
    For the request that we get from the block driver; the Destination/Source buffer need not
    be one single contigious memory buffer. It is a linked list of buffer elements. Each of
    these buffer elements are contigiuous "virtual" memory locations. But this contigiuous
    virtual memory need not by physically contigiuous in the RAM. We need to fill the DDT
    with physically contigious locations.... so here is the algo:
    1. Collect all the buffer elements into a scatter gather table
    2. Map the SG table into an equivalent list of physical memory locations.
       (Perform a streaming DMA mapping for the sg table)
    3. Fill the DDT with elements from the sg_table, breaking each sg_entry
       into multiple 64Kb max segments.

    */

    DEBUG_PRINT("Converting %d SG entries", SCpnt->use_sg);

    /* Perform a streaming DMA mapping for the sg table */
    /* Note: num_dma_buffers can be less than nents. */
    num_dma_buffers = pci_map_sg(DEVICE_PDEV(SCpnt->device), sg, SCpnt->use_sg,
                                 cmd->sg_dma_direction);
    // keep track of how many buffers mapped
    cmd->sg_nents = num_dma_buffers;

    /* Fill Out the DDT using the sg_table. */

    if(sizeof(dma_addr_t) > 4)
    {
        /* Common Upper 32 bits of the 64 bit address */
        drive_info->ddt->ddtePage = UPPER_32_BITS((UINT64)sg_dma_address(sg));
    }
    else
    {
        drive_info->ddt->ddtePage = 0;
    }

    count = 0; /* Records number of DDT entries used. */
    ddte = ddt->ddte; /* Point to start of the table... */
    for(; num_dma_buffers > 0; num_dma_buffers--)
    {
        /* For each sg element, break the sg entry into 64Kb chunks and append */
        /* to the DDT */

        u32 cur_addr;
        u32 cur_len;

        /* All the buffers should be located in the same 4Gb segment */
        if(drive_info->ddt->ddtePage !=
           UPPER_32_BITS((UINT64)sg_dma_address(sg)))
        {
            DEBUG_PRINT("Buffers not located in the same 4Gb range%s","");
            goto ddt_failed;
        }

        /* Take care of 64 bit addresses */
        cur_addr = LOWER_32_BITS((UINT64)sg_dma_address(sg));
        cur_len  = sg_dma_len(sg);

        /*
         * Fill in the dma table, without crossing any 64kB boundaries.
         * Artisea requires 16-bit alignment of all blocks
         */

        while(cur_len > 0)
        {
            /* For Artisea, 0x0000 == 64Kb */
            u32 xcount;
            u32 bcount = 0x10000 - (cur_addr & 0xffff);

            if(count++ >= max_ddt_entries)
            {
                /* Used up All DDT entries. Can't use DMA for this data transfer */
                DEBUG_PRINT("DDT Full.%s","");
                goto ddt_failed;
            }

            if(bcount > cur_len)
            {
                bcount = cur_len;
            }

            xcount = bcount & 0xffff;

            ddte->baseAddress = cpu_to_le32(cur_addr);
            ddte->byteCount = cpu_to_le32(xcount);

            ddte->EOT = RESET_BIT(0, DDT_ENTRY_EOT_BIT);
            cur_addr += bcount;
            cur_len  -= bcount;

            ddte++;  /* Next DDT entry...*/
        }

        /* Moving on to the next sg element ... */
        sg++;
    }

    /* Last entry... */
    ddte--;
    ddte->EOT = SET_BIT(0, DDT_ENTRY_EOT_BIT);

    DEBUG_PRINT("DDT has %d entries.", count);


#ifdef SATA_DEBUG_2
    {
        u32 cur_addr = 0;
        u32 cur_len = 0;
        int i = 0;

        printk(KERN_DEBUG "=======================DDT==========================");
        debug_mem_dump("DDT", (char *)ddt->ddte , count * sizeof(DDT_ENTRY_T) );
        printk(KERN_DEBUG "====================================================");
    #if 0

        sg = (struct scatterlist *)SCpnt->request_buffer;
        DEBUG_PRINT("Total SG Entries mapped: %d", cmd->sg_nents);
        for(i = 0; i < cmd->sg_nents; i++)
        {
            cur_addr = sg_dma_address(sg);
            cur_len  = sg_dma_len(sg);
            DEBUG_PRINT("SG_ADDR(%d): %#08x %#08x", i, cur_addr, cur_len);
            sg++;
        }
        print_command(SCpnt->cmnd);
    #endif

    }
#endif


    #ifdef SATA_DEBUG_0    
    drive_info->max_ddte_count_seen = MAX(drive_info->max_ddte_count_seen, count);
    drive_info->min_ddte_count_seen = MIN(drive_info->min_ddte_count_seen, count);
    #endif
    
    return SATA_SUCCESS;

    ddt_failed:
    DEBUG_ERROR("Failed to build DDT%s","");

    #ifdef SATA_DEBUG_0    
    drive_info->max_ddte_count_seen = MAX(drive_info->max_ddte_count_seen, count);
    drive_info->min_ddte_count_seen = MIN(drive_info->min_ddte_count_seen, count);
    #endif
    
    pci_unmap_sg(DEVICE_PDEV(SCpnt->device), sg,
                 cmd->sg_nents, cmd->sg_dma_direction);

    return SATA_FAILURE; /* ddt building failed */
} /* end of gdx_build_ddt() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*            Translates the READ/WRITE(10) commands into ATA DMA commands    */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_FAILURE                                                  */
/******************************************************************************/
static int gdx_build_dataops10(struct scsi_cmnd *SCpnt, void (*done) (struct scsi_cmnd *))
{
    DEVICE_INFO_T * drive_info = DEVICE_INFO(SCpnt->device);
    int status = SATA_SUCCESS;
    u8 *cmd = SCpnt->cmnd;
    CMD_INFO_T *cmd_info = NULL;
    CMD_INFO_T *curr_cmd = NULL;

    DEBUG_TRACE();

    // Do capacity check?


    // Get CMD from free queue
    curr_cmd = freeq_pop(drive_info);

    if(!curr_cmd)
    {
        DEBUG_PRINT("No more commands, we are full. Retry!%s","");
        return SATA_BUSY;
    }

    curr_cmd->SCpnt = SCpnt;
    curr_cmd->scsi_done = done;

    curr_cmd->start_sector = cmd[5] | (cmd[4] << 8) |
                             (cmd[3] << 16) | (cmd[2] << 24);
    curr_cmd->nr_sectors = cmd[8] | (cmd[7] << 8);

#ifdef SANITY_CHECK
    ASSERT(SCpnt);
    if(!SCpnt->use_sg && (SCpnt->request_buffer == NULL))
    {
       DEBUG_ERROR("ERR: %s%s, txsz %d Rbuf %p Rblen %d buf %p blen %d, usesg %d", 
                   (SCpnt->cmnd[0] == WRITE_10 ? "WRITE_10" : ""),
                   (SCpnt->cmnd[0] == READ_10 ? "READ_10" : ""),
                   curr_cmd->nr_sectors, SCpnt->request_buffer, SCpnt->request_bufflen,
                   SCpnt->buffer, SCpnt->bufflen, SCpnt->use_sg);
       DEBUG_ERROR("cmd: %0x %0x %0x %0x %0x %0x %0x %0x %0x %0x", 
                   cmd[0], cmd[1], cmd[2], cmd[3], cmd[4], cmd[5], cmd[6], cmd[7], cmd[8], cmd[9]);

    }
    
    if (SCpnt->request_bufflen < (curr_cmd->nr_sectors * 512) )
    {
       DEBUG_ERROR("ERROR: Detected %0x %s%s Command with req buffer smaller (%0x, %0x) than tx size (%0x) !!!", SCpnt->cmnd[0],
                   (SCpnt->cmnd[0] == WRITE_10 ? "WRITE_10" : ""),
                   (SCpnt->cmnd[0] == READ_10 ? "READ_10" : ""),
                   SCpnt->request_bufflen, SCpnt->bufflen, (curr_cmd->nr_sectors * 512) );
    }

#endif

    #if 1
    
    if (curr_cmd->nr_sectors == 0)
    {
       DEBUG_ERROR_0("%s%s Command with nr_sectors ZERO. Returning Success", 
                   (SCpnt->cmnd[0] == WRITE_10 ? "WRITE_10" : ""),
                   (SCpnt->cmnd[0] == READ_10 ? "READ_10" : ""));
       return SATA_SUCCESS;
    }

    #endif
    if(!drive_info->drive_features.Lba48)
    {
        ASSERT(curr_cmd->nr_sectors <= 0x100);
	/* Lba28 supports 256 max sector count, anymore and we have a problem */
	if(curr_cmd->nr_sectors > 0x100)
	  return SATA_FAILURE;
	curr_cmd->nr_sectors &= 0xff;
    }
    else
    {
        ASSERT(curr_cmd->nr_sectors <= 0x10000);
	if(curr_cmd->nr_sectors > 0x10000)
	  return SATA_FAILURE;
        curr_cmd->nr_sectors &= 0xffff;
    }

    curr_cmd->sata_cmd = (cmd[0] == READ_10) ? SATA_READ : SATA_WRITE;

    DEBUG_PRINT("start_sector: %d, %#08x nr_sectors: %d, %#08x", curr_cmd->start_sector,
            curr_cmd->start_sector,curr_cmd->nr_sectors, curr_cmd->nr_sectors);


    status = gdx_build_ddt(SCpnt, curr_cmd);
    if(status != SATA_SUCCESS)
    {
        SCpnt->result = SCSI_STAT(0, DID_ERROR, 0, 0);
        DEBUG_ERROR("failed to build DDT%s","");
        return SATA_FAILURE;
    }

    cmd_info = curr_cmd;

    // vector table would only exist if TCQ is enabled for the device
    if(drive_info->vector_table.vt)
    {
        // need to update vector table
//        drive_info->vector_table.vt[cmd_info->slot_no].ddtPhysicalAddress =
//            cmd_info->ddt_ptr->ddtPhysicalAddress;
    }

#ifdef SATA_DEBUG_0
    drive_info->all_cmd_seen_count++;
#endif

    status = AdapterAddCommandToQueue(ADAPTER_TAG(SCpnt->device), /* Tag for the adapter to be used             */
                                      DRIVE_TAG(SCpnt->device), /* The devTag of the target drive.            */
                                      cmd_info->start_sector,           /* Starting sector # for the transfer (LBA28) */
                                      cmd_info->nr_sectors,             /* Number of sectors to transfer.             */
                                      cmd_info->sata_cmd,  /* (Read / Write)         */
                                      FIRST_PARTY_DMA,        /* Requested mode of transfer DMA             */
                                      cmd_info->ddt_ptr[cmd_info->slot_no].ddtPhysicalAddress, /* Bus address of the DMA Descriptor Table */
                                      cmd_info->ddt_ptr[cmd_info->slot_no].ddtePage,           /* Top 32 bits of the DMA Data address in DDT */
                                      SCpnt->request_buffer, /* Virtual Address of the source/target buffer*/
                                      cmd_info->slot_no);

    if(status == SATA_SUCCESS)
    {
       int tryServiceCommandQueue = 50; // how many times should we try to service the command queue if it is busy ?
                                        // No specific reason why this number was chosen.

        // we can put command on active queue now
        cmd_info->cmd_state = SATA_CMD_ACTIVE;
        /* We could call AdapterServiceCommandQueues which will cause all pending commands for */
        /* drives attached to this HBA to be issued. */
        status = SATA_BUSY;
        while ( (status == SATA_BUSY) && (tryServiceCommandQueue-- > 0))
        {
           // We have seen cases where with NCQ, the drive is busy (returning data from another command ?)
           // and DevServiceCommandQueue returns SATA_BUSY. In this case we retry after a short delay.
           status = DevServiceCommandQueue(DRIVE_TAG(SCpnt->device));
           if (status == SATA_BUSY)
           {
              mdelay(3); // No specific reason why this number was chosen.
           }
        }
        if(status != SATA_SUCCESS)
        {
            DEBUG_PRINT("Failed(%d) to service command Queue. Will try again later.", status);

            PREPARE_WORK(&dev_issue_pending, gd31244_queue_worker, (void *) drive_info);
	    schedule_work(&dev_issue_pending);
	    //            queue_work(gd31244_workqueue, &dev_issue_pending);
        }
        DEBUG_PRINT("done servicing queue%s","");
    }
   else
   {
      char astr[80] = "";
      int free_cnt  = 0;

#ifdef QDEBUG
      {
         int ctr; 
         // Dump the Free and busy queues.
         // For some reason, the lib thinks this slot is in use...
         for ( ctr = 0; ctr < MAX_COMMANDS; ctr++ )
         {
            sprintf(astr + ctr, "%c",(drive_info->commands[ctr].cmd_state == 0 ? 'F':
                                       (drive_info->commands[ctr].cmd_state == 1 ? 'P':
                                        (drive_info->commands[ctr].cmd_state == 2 ? 'A': '?')))
                    );
         }
      }
#endif
      free_cnt = free_count(drive_info);
      DEBUG_ERROR("Failed(%d) to Queue command, Slot %d,  %d cmds on freeQ [%s]", status, cmd_info->slot_no, 
                  free_cnt, astr);
      ERROR_DUMP(DEVICE_PDEV(SCpnt->device), DRIVE_TAG(SCpnt->device));

      cmd_info->cmd_state = SATA_CMD_FREE;
      freeq_push(drive_info, cmd_info);
      return SATA_BUSY;
   }

//    ERROR_DUMP(SCpnt->device->host->pci_dev, DRIVE_TAG(SCpnt->device));

    DEBUG_PRINT("exiting...%s","");
    return SATA_SUCCESS;
} /* end of gdx_build_dataops10() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function translates all 6 byte SCSI commands to ATA      */
/*              commands.                                                     */
/*                                                                            */
/* PRECONDITIONS    NONE                                                      */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      0 on success, error code on failure                           */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*                                                                            */
/******************************************************************************/
int gdx_translate_6bytes(struct scsi_cmnd *SCpnt, void (*done) (struct scsi_cmnd *))
{
    int status = 0;
    int host;
    int channel;
    int target;
    int lun;
    DEVICE_INFO_T *dev;
    u8 *cmd;

    DEBUG_TRACE();

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    host = SCpnt->device->host->host_no;
    channel = SCpnt->device->channel;
    target = SCpnt->device->id;
    lun = SCpnt->device->lun;
    dev = DEVICE_INFO(SCpnt->device);
    cmd = SCpnt->cmnd;

    switch(cmd[0])
    {
        case INQUIRY:
            //SCSI_LOG_LLQUEUE(3, printk("Inquiry Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
            //                           host, channel, target, lun));
            DEBUG_PRINT("Inquiry Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            status = gdx_build_inquiry(SCpnt);
            DEBUG_PRINT("Callback function: 0x%p", SCpnt->done);
            break;
        case SEND_DIAGNOSTIC:
            //SCSI_LOG_LLQUEUE(3, printk("Send Diagnostic"));
            DEBUG_PRINT("Send Diagnostic for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            break;
            // case RELEASE_UNIT:
        case TEST_UNIT_READY:
            // Check unit readiness?
            DEBUG_PRINT("TEST_UNIT_READY for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            break;

        case REQUEST_SENSE: // probably want sense of previous command
            //SCSI_LOG_LLQUEUE(3, printk("Request sense for Ho:Ch:Tgt:Lun=%d:%d:"
            //                           "%d:%d", host, channel, target, lun));
            DEBUG_PRINT("Request sense for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            status = gdx_build_sense(SCpnt, 0x70);
            break;
       case MODE_SENSE:
           DEBUG_PRINT("MODE_SENSE for Ho:Ch:Tgt:Lun=%d:%d:%d:%d", host, channel, target, lun);
//           debug_mem_dump("Mode Sense CMD", SCpnt->cmnd, SCpnt->cmd_len);
           status = gdx_mode_sense(SCpnt);
           break;

            // If we are hitting here we need to support whatever command
            // that failed!
        case READ_6:
            DEBUG_PRINT("Internal Error: READ_6 for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            status = SATA_FAILURE;
            break;
        case WRITE_6:
            DEBUG_PRINT("Internal Error: WRITE_6 for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            status = SATA_FAILURE;
            break;
        case MODE_SELECT:
            DEBUG_ERROR("Internal Error: MODE_SELECT for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            status = SATA_FAILURE;
            break;
        case ALLOW_MEDIUM_REMOVAL:
            DEBUG_ERROR("Internal Error: MEDIUM_REMOVAL for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            status = SATA_FAILURE;
            break;

        default:
            DEBUG_ERROR("Internal Error: Unsupported command: %#04x", SCpnt->cmnd[0]);
            
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            status = SATA_FAILURE;
//            done(SCpnt);
//            atomic_dec(&dev->cmd_count);
    }

    done(SCpnt);
    atomic_dec(&dev->cmd_count);

    return status;
} /* end of gdx_translate_6bytes() */


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function translates all 10 byte SCSI commands to ATA     */
/*              commands.                                                     */
/*                                                                            */
/* PRECONDITIONS    NONE                                                      */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   struct scsi_cmnd * - pointer to SCSI Command context                 */
/*                                                                            */
/* RETURNS      0 on success, error code on failure                           */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*                                                                            */
/******************************************************************************/
int gdx_translate_10bytes(struct scsi_cmnd *SCpnt, void (*done) (struct scsi_cmnd *))
{
    int status = 0;
    int host;
    int channel;
    int target;
    int lun;
    u8 *cmd;
    DEVICE_INFO_T *dev;

    DEBUG_TRACE();

#ifdef SATA_DEBUG
    ASSERT(SCpnt);
#endif

    host = SCpnt->device->host->host_no;
    channel = SCpnt->device->channel;
    target = SCpnt->device->id;
    lun = SCpnt->device->lun;
    cmd = SCpnt->cmnd;
    dev = DEVICE_INFO(SCpnt->device);

    switch(cmd[0])
    {
        case READ_CAPACITY:
            //SCSI_LOG_LLQUEUE(3,
            //                 printk("READ_CAPACITY Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
            //                        host, channel, target, lun));
            DEBUG_PRINT("READ_CAPACITY for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            status = gdx_build_capacity(SCpnt);
            done(SCpnt);
            atomic_dec(&dev->cmd_count);
            break;
        case READ_10:
        case WRITE_10:
            DEBUG_PRINT("%s%s for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        (*cmd == WRITE_10 ? "WRITE_10" : ""),
                        (*cmd == READ_10 ? "READ_10" : ""),
                        host, channel, target, lun);
            status = gdx_build_dataops10(SCpnt, done);

            if(status == SATA_BUSY)
            {
               // NPG: We could come here with 2 stages of completion
               // 1. - Failed to queue the command
               // 2. - Queued the command, but failed to service the command queue.
               // How do we handle these cases ?

                atomic_dec(&dev->cmd_count);
                SCpnt->result = SCSI_STAT(0, 0, 0, QUEUE_FULL);
                DEBUG_PRINT("Dev %d: translate 10 bytes cmd queue full",dev->dev_tag);
                done(SCpnt);
            }
            break;
        case SYNCHRONIZE_CACHE:
            //SCSI_LOG_LLQUEUE(3, printk("SYNCHRONIZE_CACHE Ho:Ch:Tgt:Lun=%d:%d:"
            //                           "%d:%d", host, channel, target, lun));
            DEBUG_PRINT("SYNCHRONIZE_CACHE for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            status = gdx_build_sync(SCpnt);
            done(SCpnt);
            atomic_dec(&dev->cmd_count);
            break;
            // If we are hitting here we need to support whatever command
            // that failed!
        case MODE_SENSE_10:
            DEBUG_PRINT("MODE_SENSE_10 for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            debug_mem_dump("Mode Sense 10 CMD", SCpnt->cmnd, SCpnt->cmd_len);
            status = gdx_mode_sense(SCpnt);
            done(SCpnt);
            atomic_dec(&dev->cmd_count);
            break;
        case VERIFY:
            DEBUG_PRINT("VERIFY for Ho:Ch:Tgt:Lun=%d:%d:%d:%d",
                        host, channel, target, lun);
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            done(SCpnt);
            atomic_dec(&dev->cmd_count);
            status = SATA_FAILURE;
            break;
        default:
            DEBUG_ERROR("Ho:Ch:Tgt:Lun=%d:%d:%d:%d - Unsupported command:"
                        " opcode=%#02x", host, channel, target,
                        lun, (int)cmd[0]);
            debug_mem_dump("SCSI CMD", SCpnt->cmnd, SCpnt->cmd_len);
            SCpnt->result = SCSI_STAT(0, DID_ABORT, 0, 0);
            done(SCpnt);
            atomic_dec(&dev->cmd_count);
            status = SATA_FAILURE;
    }
    return status;
} /* end of gdx_translate_10bytes() */


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function will most likely be called by the interrupt     */
/*              handler after it completes a command. An attempt shall be     */
/*              made to service command in the TCQ, if no commands are there  */
/*              it will try to pull pending commands in the soft queue into   */
/*              the hardware queue. The command will put the command back on  */
/*              the pending soft queue if it is unable to queue to the hw q   */
/*                                                                            */
/* PRECONDITIONS    NONE                                                      */
/*                                                                            */
/* POST-CONDITIONS  NONE                                                      */
/*                                                                            */
/* PARAMETERS   DEVICE_INFO_T * - pointer to drive info context               */
/*                                                                            */
/* RETURNS      NONE                                                          */
/*                                                                            */
/******************************************************************************/
#if 0
void gdx_service_queue(DEVICE_INFO_T *dev_info)
{
    int status = SATA_SUCCESS;

    DEBUG_TRACE();

    status = DevServiceCommandQueue(dev_info->dev_tag);
    if(status != SATA_SUCCESS)
    {
        DEBUG_ERROR("Unable(%d) to service queue for Dev #%d", status, dev_info->dev_tag);
    }
} /* end of gdx_service_queue() */
#endif

static int gdx_set_features(DEVICE_INFO_T *dev_info, SATA_TASKFILE_T *taskfile)
{
    int dev_tag = dev_info->dev_tag;
    int status = SATA_SUCCESS;
    u8 ata_status = 0;
    u32 ext_status = 0;
    u32 error = 0;

    DEBUG_TRACE();

#ifdef EXPERIMENTAL_CODE
    status = SelectDevice(dev_tag);
    if(status != SATA_SUCCESS)
    {
        DEBUG_ERROR("Failed Select Device #%d", dev_tag);
        return status;
    }
#endif

    status = DevExecuteRawCommand(dev_tag, TRUE, taskfile);

    if(status != SATA_SUCCESS)
    {
        DEBUG_ERROR("Dev %d: Failed execute RAW command!", dev_tag);
        return status;
    }

    // Let's take a look at the status to see if any errors occured
    if((status = DevGetStatus(dev_tag, &ata_status)) != SATA_SUCCESS)
    {
        DEBUG_ERROR("Dev %d: Getting ATA status failed", dev_tag);
        return status;
    }

    if(ata_status & ATA_ERR_BIT_MASK)
    {
        status = DevGetExtendedError(dev_tag,
                                     &ext_status,
                                     &error,
                                     taskfile);
        DEBUG_ERROR("extended_status: %#08x", ext_status);
        DEBUG_ERROR("error: %#08x", error);
        return status;
    }

    return SATA_SUCCESS;
} /* end of gdx_set_features() */

static int gdx_mode_sense(struct scsi_cmnd *SCpnt)
{

   unsigned char * cmd     = SCpnt->cmnd;
   unsigned char * buff    = NULL;
#define ARR_SZ 256
   unsigned char arr[ARR_SZ];
   int bufflen, msense_6, alloc_len, data_length;
   int pcontrol, pcode;   
   u8 *mph                 = NULL;
   u8 *mp_cache_pg         = NULL;
   DEVICE_INFO_T *dev_info = DEVICE_INFO(SCpnt->device); 
   
   ASSERT(dev_info);

   msense_6 = (cmd[0] == MODE_SENSE); // Check if it is 0x1a. 0x5a in 10 byte mode sense.   
   alloc_len = msense_6 ? cmd[4] : ((cmd[7] << 8) | cmd[8]);
   pcontrol = (cmd[2] & 0xc0) >> 6;
   pcode = cmd[2] & 0x3f;
      
   if (SCpnt->use_sg) 
      { /* just use first element */
       struct scatterlist *sgpnt = (struct scatterlist *) SCpnt->request_buffer;
       buff = ((sgpnt != NULL) ? 
               ((unsigned char *)page_address(sgpnt[0].page) + sgpnt[0].offset) :
               (NULL)
               );
       bufflen = sgpnt[0].length;
   }
   else 
   {
       buff = (unsigned char *) SCpnt->request_buffer;
       bufflen = SCpnt->request_bufflen;
   }
   
   if (buff == NULL)
   {
      DEBUG_ERROR("Mode Sense command with Null buffer. req_buff = %p", SCpnt->request_buffer);
      return CHECK_CONDITION;
   }

//   DEBUG_ERROR("Sizes: MODE_PARAM6_HEADER_T = %d (=4?), MODE_PARAM6_CACHE_PG_T = %d (=20?)", 
//               sizeof(MODE_PARAM6_HEADER_T), sizeof(MODE_PARAM6_CACHE_PG_T));

   memset(arr, 0, sizeof(arr));
   memset(buff, 0, bufflen);
   mph = (u8 *) arr;

   if ( 0x3 == pcontrol )
   {  /* Saving values not supported */
      return CHECK_CONDITION;
   }
   
   // Fill in the Header
   // Everything to be set to 0... Already done.
   data_length = msense_6 ? 4 : 8;

   switch ( pcode )
   {
      case 0x3f:  /* Read all Mode pages */
      case 0x8:   /* Caching page, direct access */
         {
            u8 def[20] = { 0x8, 18, 0x04, 0, 
               0xff, 0xff, 0, 0, 0xff, 0xff, 0xff, 0xff, /* Prefetch data */
               0x00, /* FSW = 0, DRA = 0 */
               0x00, 0, 0, 0, 0, 0, 0};
            mp_cache_pg = mph + data_length ;
            memcpy( mp_cache_pg, def, min( 20, (int)(ARR_SZ - mph[0]) ) );

            // We will support only Read/Write cache enabled bits.
            // WCE = 1, Cache enabled
            // RCD = 0, Cache enabled. (RCD is Rd Cache DISABLED bit). Leave it at 0

            mp_cache_pg[2] = (dev_info->drive_features.WriteCache == TRUE) ? 
                             SET_BIT(mp_cache_pg[2], 2) : RESET_BIT(mp_cache_pg[2], 2);
            data_length += 20; // Adjust the Header size.
         }
         break;
      case 0x1:   /* Read-Write error recovery page, direct access */
//         break;
      case 0x2:   /* Disconnect-Reconnect page, all devices */
//         break;
      case 0x3:       /* Format device page, direct access */
//         break;
      case 0xa:   /* Control Mode page, all devices */
//         break;
      case 0x1c:  /* Informational Exceptions Mode page, all devices */
//         break;
//         break;
      default:
         DEBUG_ERROR("Unsupported Mode page %0x", pcode);
         return CHECK_CONDITION;
   }
   
   data_length -= 1; // Correct the length to the value we want to report.
   if ( msense_6 )
   {
      arr[0] = data_length - 1;
   }
   else
   {
      data_length -= 2;
      arr[0] = (data_length >> 8) & 0xff; 
      arr[1] = data_length & 0xff;
   }
   alloc_len = min(ARR_SZ,alloc_len);  // Find out what is the min size. Reuse the alloc_len variable 
   alloc_len = min(bufflen,alloc_len); // Find out what is the min size. Reuse the alloc_len variable 
   memcpy(buff, arr, alloc_len ); // Copy only the safe size of data.
   debug_mem_dump("Mode Sense Scratch pad", arr, sizeof(arr));
   debug_mem_dump("Mode Sense Data", buff, bufflen);

   return GOOD;
}

