/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                                                                            */
/*   DESCRIPTION: osl.c                                                       */
/*                                                                            */
/*         This file contains the declarations of the OS Service Layer (OSL)  */
/*		   API in prototype form.                                             */
/*         The OSL provides all the functionality that is either OS dependent */
/*         or architecture dependent as a set of utility/system APIs to the   */
/*         storage library. The OSL has a different implementation for each   */
/*         OS and possibly for different architectures, although this file    */
/*         and the interfaces contained herein should NOT change.             */
/*         IF these APIs are changed, it will require modification of the     */
/*         SATA library in order to maintain compatibility.                   */
/*		   Fundamentally, the OSL translates generic OS service requests from */
/*		   the SATA Library to OS specific functions.						  */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                  I N C L U D E   S T A T E M E N T S                       */
/*                                                                            */
/******************************************************************************/
#include <sata_debug.h>
#include <linux/ide.h>
#include <linux/slab.h>
#include <linux/string.h>        /* memset, memcpy */
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/system.h>
#include "OSL_linux.h"

/******************************************************************************/
/*                                                                            */
/*       G L O B A L   F U N C T I O N   I M P L E M E N T A T I O N S        */
/*                                                                            */
/******************************************************************************/

ide_set_handler_t *ide_set_handler_fptr;


/******************************************************************************/
/*                                                                            */
/* FUNCTION		PciMemReadDataPort()                                          */
/*                                                                            */
/* DESCRIPTION  This function is designed primarily to perform a read of      */
/*              the data register in the "shadow" IDE command register block. */
/*              Located in PCI memeory space. Thus,the source data register   */
/*              location will be read for the total number of bytes requested */
/*              and the data read will be placed into a destination buffer.   */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of bytes to read.                                  */
/*              sourceDataReg_i                                               */
/*                  Architecture independent IO address of the data register  */
/*                  to be read.                                               */
/*              * destinationBuffer_op                                        */
/*                  Pointer to the start address (Virtual memory address) of  */
/*                  the local memory buffer where the data will be stored.    */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PciMemReadDataPort (
                                 INT count_i,
                                 SATA_REG_ADDR_T sourceDataReg_i,
                                 void * destinationBuffer_op
                                 )
{
    int ctr;
    //unsigned int val;
    UINT16 * ptr = destinationBuffer_op;
//    char * cptr = destinationBuffer_op;

//    DEBUG_TRACE_4();
    DEBUG_PRINT_4("Reading %d words from %p", count_i, (void *)sourceDataReg_i);

    /* dest, src, num */
    /* memcpy_fromio( destinationBuffer_op, sourceDataReg_i, count_i); */

    for(ctr = count_i / sizeof(*ptr); ctr > 0; ctr--)
    {
        *ptr++ = readw(sourceDataReg_i);
    }

    DEBUG_PRINT_4("Done with PciMemReadDataPort %s","");
    return SATA_SUCCESS;
}
/******************************************************************************/
/*                                                                            */
/* FUNCTION		PciMrmWriteDataPort()                                         */
/*                                                                            */
/* DESCRIPTION  This function is designed primarily to perform a  write into  */
/*              the data register in the IDE command register "shadow" block  */
/*              located in PCI memory. Thus, the destination register         */
/*              location will be written for the total number of bytes        */
/*              requested and the data to be written is pointed to by the     */
/*              source buffer pointer.                                        */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of bytes to write                                  */
/*              * sourceBuffer_ip                                             */
/*                  Pointer to the start address of the local memory buffer   */
/*                  where the data to transfer is located.                    */
/*              destinationDataReg_i                                          */
/*                  Architecture independent PCI address of the data register */
/*                  to be written.                                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PciMemWriteDataPort(
                                 INT count_i,
                                 void * sourceBuffer_ip,
                                 SATA_REG_ADDR_T destinationDataReg_i
                                 )
{
    int ctr;
    char * ptr = sourceBuffer_ip;
//    DEBUG_TRACE_4();
    DEBUG_PRINT_4("Writing %d bytes to %p", count_i, (void *)destinationDataReg_i);

    /* dest, src, num */
    /*memcpy_toio( destinationDataReg_i, sourceBuffer_ip, count_i); */
    for(ctr = count_i; ctr > 0; ctr--)
    {
        writeb( *(ptr++), destinationDataReg_i);

        /* Write memory Barrier. Do we have to have this ? Slows things down...*/
        wmb();
    }

    return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* FUNCTION		SetISR()                                                      */
/*                                                                            */
/* DESCRIPTION  This function provides a method for the library to pass a     */
/*              OS specific ISR callback address into the OS. In some OS's,   */
/*              the interrupt for command completion is shared among multiple */
/*              controllers.  Thus, when the OSDM calls the SATA Library, the */
/*              library must have a way to call back into the OSDM in context.*/
/*              A pointer to an associated drive structure is passed down     */
/*              from the OSDM to the Library.  That pointer is then           */
/*              registered as a callback parameter by the library when it     */
/*              sends a command for that drive. The pointer passed down is    */
/*              *OS_Drive.  In addition, the handler pointer is sent from OSDM*/
/*              to library, and registered with the OS.  This routine may not */
/*              be used by all OS's.                                          */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   *OS_Drive                                                     */
/*                  Pointer to data structure associated to a drive by the OS */
/*              *Handler                                                      */
/*                  Pointer to interrupt handler in the OS                    */
/*                  Set the default Handler if this value is NULL_POINTER     */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#ifdef EXPERIMENTAL_CODE
STATUS_CODE_T SetISR(
                    void * OS_Drive,
                    void * Handler
                    )
{
    ide_drive_t *drive = (ide_drive_t *) OS_Drive;
    ide_handler_t *handler = (ide_handler_t *)Handler;

    DEBUG_TRACE_4();
    ASSERT(OS_Drive != NULL_POINTER);
    ASSERT(Handler != NULL_POINTER);

    if(ide_set_handler_fptr)
    {
        ide_set_handler_fptr(drive, handler, WAIT_CMD, NULL);
    }

    return SATA_SUCCESS;
}
#endif
