/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*   DESCRIPTION: osl.h                                                       */
/*                                                                            */
/*         This file contains the declarations of the OS Service Layer (OSL)  */
/*		   API in prototype form.                                             */
/*         The OSL provides all the functionality that is either OS dependent */
/*         or architecture dependent as a set of utility/system APIs to the   */
/*         storage library. The OSL has a different implementation for each   */
/*         OS and possibly for different architectures, although this file    */
/*         and the interfaces contained herein should NOT change.             */
/*         IF these APIs are changed, it will require modification of the     */
/*         SATA library in order to maintain compatibility.                   */
/*		   Fundamentally, the OSL translates generic OS service requests from */
/*		   the SATA Library to OS specific functions.						  */
/*                                                                            */
/******************************************************************************/

#ifndef OSL_HEADER
#define OSL_HEADER


/******************************************************************************/
/*                                                                            */
/*                  I N C L U D E   S T A T E M E N T S                       */
/*                                                                            */
/******************************************************************************/

#include <linux/string.h>        /* memset, memcpy */
#include <linux/types.h>
#include <linux/slab.h>  /* kmalloc, kfree */
#include <linux/delay.h>
#include <linux/interrupt.h> /*kmalloc tuning*/
#include <linux/spinlock.h>  /* spin_lock support */
#include <asm/io.h> /* PCI read/write */
#include <asm/system.h>      /* spin_lock support */
#include <sata_types.h> /* Common data types used in SATA driver */
#include <sata_debug.h>

/******************************************************************************/
/*                                                                            */
/*                C O N S T A N T   D E F I N I T I O N S                     */
/*                                                                            */
/******************************************************************************/


/******************************************************************************/
/*                                                                            */
/*                   M A C R O   D E F I N I T I O N S                        */
/*                                                                            */
/******************************************************************************/
#define UPPER_32_BITS(addr) (UINT32)(((addr) >> 32) & 0xFFFFFFFF)
#define LOWER_32_BITS(addr) (UINT32)( (addr)        & 0xFFFFFFFF)

#define OsSpinLock(lock, flags) spin_lock_irqsave(lock, flags)
#define OsSpinUnlock(lock, flags) spin_unlock_irqrestore(lock, flags)
#define OsSpinlockInit(lock) spin_lock_init(lock)

/******************************************************************************/
/*                                                                            */
/*                    T Y P E   D E F I N I T I O N S                         */
/*                                                                            */
/******************************************************************************/

typedef spinlock_t OS_SPINLOCK_T;

/******************************************************************************/
/*                                                                            */
/*    V A R I A B L E   A N D   S T R U C T U R E   D E F I N I T I O N S     */
/*                                                                            */
/******************************************************************************/

typedef void (ide_set_handler_t) (void *drive, void *handler, unsigned int timeout, void *expiry);
extern ide_set_handler_t *ide_set_handler_fptr;

/******************************************************************************/
/*                                                                            */
/*                 F U N C T I O N    P R O T O T Y P E S                     */
/*                                                                            */
/******************************************************************************/

/*-------------------------------------*/
/*-------------------------------------*/
/*    Memory Allocation Functions      */
/*-------------------------------------*/
/*-------------------------------------*/

/******************************************************************************/
/*                                                                            */
/* FUNCTION		GetMemory()                                                   */
/*                                                                            */
/* DESCRIPTION  This function gets relocatable memory                         */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   size_i                                                        */
/*                  The size in bytes of the memory block requested.          */
/*                                                                            */
/* RETURNS      Pointer to valid memory block on success,                     */
/*              NULL is returned if memory could not be allocated             */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define GetMemory(size_i) kmalloc(size_i, in_interrupt() ? GFP_ATOMIC : GFP_KERNEL)

/******************************************************************************/
/*                                                                            */
/* FUNCTION		GetLockedMemory()                                             */
/*                                                                            */
/* DESCRIPTION  This function gets non-relocatable/non-swappable memory with  */
/*              locked physical addresses.                                    */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   size_i                                                        */
/*                  The size in bytes of the memory block requested.          */
/*                                                                            */
/* RETURNS      Pointer to valid memory block on success,                     */
/*              NULL is returned if memory could not be allocated             */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define GetLockedMemory(size_i) kmalloc(size_i, GFP_KERNEL | GFP_DMA)

/******************************************************************************/
/*                                                                            */
/* FUNCTION		FreeMemory()                                                  */
/*                                                                            */
/* DESCRIPTION  This function frees previously acquired memory by either      */
/*              GetMemory() or GetLockedMemory()                              */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   * pointer_i                                                   */
/*                  Pointer to the start of the previously allocated block.   */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define FreeMemory(pointer_i) kfree(pointer_i)

/******************************************************************************/
/*                                                                            */
/* FUNCTION		VirtualToBus()                                                */
/*                                                                            */
/* DESCRIPTION  This function Returns the Bus address for the given location. */
/*              This is the address that other devices like a DMA controller  */
/*              can use the access that location.                             */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   * pointer_i                                                   */
/*                  Address of the location whose bus address is needed.      */
/*                                                                            */
/******************************************************************************/
#define VirtualToBus(pointer_i) virt_to_bus(pointer_i)

/******************************************************************************/
/*                                                                            */
/* FUNCTION		MemCopy()                                                     */
/*                                                                            */
/* DESCRIPTION  This function copies the contents from one memory region to   */
/*              another.                                                      */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              length                                                        */
/*                  The number of bytes to be copied from "from_p" to "to_p"  */
/*              from_p                                                        */
/*                  Virtual Address of the Source                             */
/*              to_p                                                          */
/*                  Virtual Address of Destination                            */
/*                                                                            */
/******************************************************************************/
#define MemCopy(length, from_ip, to_op) memcpy(to_op, from_ip, length)

/******************************************************************************/
/*                                                                            */
/* FUNCTION		MemSet()                                                      */
/*                                                                            */
/* DESCRIPTION  This function initializes the contents of a memory region with*/
/*              the provided value.                                           */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              length                                                        */
/*                  The length og the memory region to be initialized.        */
/*              value                                                         */
/*                  The Byte value the memory is to be initialized to.        */
/*              address_op                                                    */
/*                  Virtual Address of memory region                          */
/*                                                                            */
/******************************************************************************/
#define MemSet(length, value, address_op) memset(address_op, value, length)

/*-------------------------------------*/
/*-------------------------------------*/
/*    PIO Memory Read/Write Functions  */
/*-------------------------------------*/
/*-------------------------------------*/

/******************************************************************************/
/*                                                                            */
/* FUNCTION		IoReadxx()                                                    */
/*                                                                            */
/* DESCRIPTION  These functions read memory from IO space in the quantity     */
/*              designated by the xx extension where xx represents bits.      */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   sourceDataReg_i                                               */
/*                  Architecture independent IO address of the data register  */
/*                  to be read.                                               */
/*              *data_op                                                      */
/*                  Pointer to the local memory address where the data will   */
/*                  be stored.                                                */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define IoRead8
#define IoRead8
#define IoRead16
#define IoRead32
#define IoRead64

/******************************************************************************/
/*                                                                            */
/* FUNCTION		IOReadBytes()                                                 */
/*                                                                            */
/* DESCRIPTION  This function is designed primarily to perform a PIO read of  */
/*              the data register in the IDE command register block. Thus,    */
/*              the source register location will be read for the total number*/
/*              of bytes requested and the data read will be placed into a    */
/*              destination buffer.                                           */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of bytes to read.                                  */
/*              sourceDataReg_i                                               */
/*                  Architecture independent IO address of the data register  */
/*                  to be read.                                               */
/*              * destinationBuffer_op                                        */
/*                  Pointer to the start address of the local memory buffer   */
/*                  where the data will be stored.                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define IOReadBytes

/******************************************************************************/
/*                                                                            */
/* FUNCTION		IoWritexx()                                                   */
/*                                                                            */
/* DESCRIPTION  These functions write memory into IO space in the quantity    */
/*              designated by the xx extension where xx represents bits.      */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   destinationDataReg_i                                          */
/*                  Architecture independent IO address of the data register  */
/*                  to be written.                                            */
/*              value_i                                                       */
/*                  Value to write                                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define IoWrite8
#define IoWrite16
#define IoWrite32
#define IoWrite64

/******************************************************************************/
/*                                                                            */
/* FUNCTION		IOWritebytes()                                                */
/*                                                                            */
/* DESCRIPTION  This function is designed primarily to perform a PIO write to */
/*              the data register in the IDE command register block. Thus,    */
/*              the destination register location will be written for the     */
/*              total number of bytes requested and the data to be written is */
/*              pointed to by the source pointer.                             */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of bytes to write                                  */
/*              * sourceBuffer_ip                                             */
/*                  Pointer to the start address of the local memory buffer   */
/*                  where the data to transfer is located.                    */
/*              destinationDataReg_i                                          */
/*                  Architecture independent IO address of the data register  */
/*                  to be written.                                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define IOWriteBytes

/*-------------------------------------*/
/*-------------------------------------*/
/*  Memory Mapped Read/Write Functions */
/*-------------------------------------*/
/*-------------------------------------*/

/******************************************************************************/
/*                                                                            */
/* FUNCTION		PciMemReadxx()                                                */
/*                                                                            */
/* DESCRIPTION  These functions read memory mapped PCI space in the quantity  */
/*              designated by the xx extension where xx represents bits.      */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   sourceDataReg_i                                               */
/*                  Architecture independent address of the PCI memory        */
/*                  to be read.                                               */
/*              *data_op                                                      */
/*                  Pointer to the local memory address where the data will   */
/*                  be stored.                                                */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define PciMemRead8(sourceDataReg_i, data_op) \
do { \
    *data_op = readb(sourceDataReg_i); \
} while (0)

#define PciMemRead16(sourceDataReg_i, data_op) \
do { \
    *data_op = readw(sourceDataReg_i); \
} while (0)

#define PciMemRead32(sourceDataReg_i, data_op) \
do { \
    *data_op = readl(sourceDataReg_i); \
} while (0)

#define PciMemRead64(sourceDataReg_i, data_op) \
do { \
    WARN_STUB(); \
} while (0)

/******************************************************************************/
/*                                                                            */
/* FUNCTION		PciMemReadDataPort()                                          */
/*                                                                            */
/* DESCRIPTION  This function is designed primarily to perform a read of      */
/*              the data register in the "shadow" IDE command register block. */
/*              Located in PCI memeory space. Thus,the source data register   */
/*              location will be read for the total number of bytes requested */
/*              and the data read will be placed into a destination buffer.   */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of bytes to read.                                  */
/*              sourceDataReg_i                                               */
/*                  Architecture independent IO address of the data register  */
/*                  to be read.                                               */
/*              * destinationBuffer_op                                        */
/*                  Pointer to the start address of the local memory buffer   */
/*                  where the data will be stored.                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PciMemReadDataPort (
								  INT count_i,
								  SATA_REG_ADDR_T sourceDataReg_i,
								  void * destinationBuffer_op
								  );

/******************************************************************************/
/*                                                                            */
/* FUNCTION		PciMemWritexx()                                               */
/*                                                                            */
/* DESCRIPTION  These functions write memory into PCI memory space in the     */
/*              quantity designated by the xx extension where xx represents   */
/*              bits.                                                         */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   destinationDataReg_i                                          */
/*                  Architecture independent IO address of the data register  */
/*                  to be written.                                            */
/*              value_i                                                       */
/*                  Value to write                                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define PciMemWrite8(destinationDataReg_i, value_i) writeb(value_i, destinationDataReg_i)

#define PciMemWrite16(destinationDataReg_i, value_i) writew(value_i, destinationDataReg_i)

#define PciMemWrite32(destinationDataReg_i, value_i) writel(value_i, destinationDataReg_i)

#define PciMemWrite64(destinationDataReg_i, value_i) WARN_STUB()

/******************************************************************************/
/*                                                                            */
/* FUNCTION		PciMrmWriteDataPort()                                         */
/*                                                                            */
/* DESCRIPTION  This function is designed primarily to perform a  write into  */
/*              the data register in the IDE command register "shadow" block  */
/*              located in PCI memory. Thus, the destination register         */
/*              location will be written for the total number of bytes        */
/*              requested and the data to be written is pointed to by the     */
/*              source buffer pointer.                                        */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of bytes to write                                  */
/*              * sourceBuffer_ip                                             */
/*                  Pointer to the start address of the local memory buffer   */
/*                  where the data to transfer is located.                    */
/*              destinationDataReg_i                                          */
/*                  Architecture independent PCI address of the data register */
/*                  to be written.                                            */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PciMemWriteDataPort(
								  INT count_i,
								  void * sourceBuffer_ip,
								  SATA_REG_ADDR_T destinationDataReg_i
								  );

/******************************************************************************/
/*                                                                            */
/* FUNCTION		SetISR()                                                      */
/*                                                                            */
/* DESCRIPTION  This function provides a method for the library to pass a     */
/*              OS specific ISR callback address into the OS. In some OS's,   */
/*              the interrupt for command completion is shared among multiple */
/*              controllers.  Thus, when the OSDM calls the SATA Library, the */
/*              library must have a way to call back into the OSDM in context.*/
/*              A pointer to an associated drive structure is passed down     */
/*              from the OSDM to the Library.  That pointer is then           */
/*              registered as a callback parameter by the library when it     */
/*              sends a command for that drive. The pointer passed down is    */
/*              *OS_Drive.  In addition, the handler pointer is sent from OSDM*/
/*              to library, and registered with the OS.  This routine may not */
/*              be used by all OS's.                                          */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   *OS_Drive                                                     */
/*                  Pointer to data structure associated to a drive by the OS */
/*              *Handler                                                      */
/*                  Pointer to interrupt handler in the OS                    */
/*                  Set the default Handler if this value is NULL_POINTER     */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#ifdef EXPERIMENTAL_CODE
STATUS_CODE_T SetISR(
					 void * OS_Drive,
					 void * Handler
					 );
#else
#define SetISR(a,b) SATA_SUCCESS /* Nothing to do. This was for IDE driver */
#endif

/******************************************************************************/
/*                                                                            */
/* FUNCTION		MicroSecondSleep()                                            */
/*                                                                            */
/* DESCRIPTION  This function is used to intorduce smallMicrosecond level     */
/*              delays.                                                       */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of microseconds to sleep                           */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define MicrosecondSleep udelay
/******************************************************************************/
/*                                                                            */
/* FUNCTION		MillisecondSleep()                                            */
/*                                                                            */
/* DESCRIPTION  This function is used to intorduce small Millisecond level    */
/*              delays.                                                       */
/*                                                                            */
/* PRECONDITIONS    none                                                      */
/*                                                                            */
/* POST-CONDITIONS  none                                                      */
/*                                                                            */
/* PARAMETERS   count_i                                                       */
/*                  Number of Milliseconds to sleep                           */
/*                                                                            */
/* RETURNS      STATUS_CODE_T as defined in sata_types.h                      */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
#define MillisecondSleep mdelay

#endif /* OSL_HEADER */
