/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                  I N C L U D E   S T A T E M E N T S                       */
/*                                                                            */
/******************************************************************************/

#include <OSL.h>
#include <sata_controller.h>
#include <sata_globals.h>
#include <sata_lib.h>

#ifndef WINNT
    #include <sata_debug.h>
#endif

/******************************************************************************/
/*                                                                            */
/*       G L O B A L   F U N C T I O N   I M P L E M E N T A T I O N S        */
/*                                                                            */
/******************************************************************************/


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function allocates the private memory for the controller.*/
/*              it allocates a CONTROLLER_INFO_T structure and records the    */
/*              pointer in the next available slot in the controllers_gp array*/
/*              This function also initializes the allocated memory structure */
/*              and sets the adapter member to the passed adapterTag.         */
/*                                                                            */
/* PRECONDITIONS    controllers_gp array has been allocated.                  */
/*                                                                            */
/* POST-CONDITIONS  The next NULL pointer in controllers_gp is set.           */
/*                  *controllerTag_op is set to it's index number/INVALID_TAG */
/*                                                                            */
/* PARAMETERS   adapterTag_i                                                  */
/*                  The Tag of the parent adapter.                            */
/*              controllerTag_op                                              */
/*                  Address of the INT variable to hold the controllerTag.    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in controllers_gp array    */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*                                                                            */
/******************************************************************************/

STATUS_CODE_T ControllerAllocatePrivateMemory (
                                              INT adapterTag_i,
                                              INT *controllerTag_op
                                              )
{
    INT tag;
    CONTROLLER_INFO_T *newCtlrInfo;

    DEBUG_TRACE_2();

    *controllerTag_op = INVALID_TAG ; /* Will be assigned a valid value below */

    tag = getNextSlot((void**)controllers_gp, MAX_CONTROLLERS); /* NULL_POINTER safe */

    /* Check if we got an empty slot in the array */
    if(tag == INVALID_TAG)
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }

    /* Allocate the Memory */
    newCtlrInfo = GetMemory(sizeof(CONTROLLER_INFO_T));

    if(newCtlrInfo == NULL_POINTER)
    {
        /* Failed to get memory */
        return SATA_GET_MEMORY_FAILED;
    }

    /* Record the pointer in the global array */
    controllers_gp[tag] = (CONTROLLER_INFO_T *) newCtlrInfo;
    *controllerTag_op = tag ;

    /* Initialize the structure */
    MemSet(sizeof(CONTROLLER_INFO_T), 0, newCtlrInfo);
    newCtlrInfo->present = FALSE;  /* Mark as not yet completely initialized */
    newCtlrInfo->adapter = adapterTag_i;

    return SATA_SUCCESS;

}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function initializes the controller.                     */
/*                                                                            */
/* PRECONDITIONS    The private memory for the controller has been allocated. */
/*                                                                            */
/* POST-CONDITIONS  The controller and attached ports are initialized and     */
/*                  ready for I/O                                             */
/*                                                                            */
/* PARAMETERS   controllerTag_i                                               */
/*                  Tag for the controller to initialize.                     */
/*              controllerNumber_i                                            */
/*                  The controller number on the adapter represented by this  */
/*                  controller. Currently we expect only one controller per   */
/*                  adapter.                                                  */
/*              baseAddr_i                                                    */
/*                  The base address to the registers on the controller.      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in controllers_gp array    */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*              SATA_NOT_INITIALIZED    = Memory was not allocated.           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T ControllerInit (
                             INT controllerTag_i,       /* Tag for the controller to initialize.*/
                             INT controllerNumber_i,    /* The controller number on the adapter */
                             SATA_REG_ADDR_T baseAddr_i /* base addr to the registers on the controller*/
                             )
{
    INT adapter, currPort, ctr;
    STATUS_CODE_T retval = SATA_SUCCESS;
    boolean deviceInitFailed = FALSE;

    DEBUG_TRACE_2();

    /* Sanity check */
    if(!TAG_WITHIN_RANGE(controllerTag_i, MAX_CONTROLLERS))
        return SATA_OUT_OF_RANGE_VALUE ;
    if(controllers_gp[controllerTag_i] == NULL_POINTER) return SATA_NOT_INITIALIZED;

    adapter = controllers_gp[controllerTag_i]->adapter;

    /* Check if the controllerNumber_i is valid */
    if(controllerNumber_i >= MAX_CONTROLLERS_PER_ADAPTER) return SATA_OUT_OF_RANGE_VALUE;
    controllers_gp[controllerTag_i]->controllerNumber = controllerNumber_i;

    controllers_gp[controllerTag_i]->commonRegBase = baseAddr_i;

    /* NPG: TBD: Hard code the mode to PCI_DPA for now. - NPG (12/7/02) */
    controllers_gp[controllerTag_i]->mode = PCI_DPA;

    /* Initialization complete. Mark controller as ready for use. */
    controllers_gp[controllerTag_i]->present = TRUE;

    /* Any other controller initializations go here. None for Artisea */
    /* Now initialize all the ports attached to this controller. */

    /* Let's perform similar operations together rather than performing */
    /* all operations for a port at one go. Helps in Debug */

    /* ask each port to allocate it's private memory */
    for(ctr = 0; ctr < MAX_PORTS_PER_CONTROLLER; ctr++)
    {
        retval = PortAllocatePrivateMemory (controllerTag_i, &currPort);
        if(retval != SATA_SUCCESS)
        {
            /* Abort Initialization. Fatal System error */
            return retval;
        }
        /* Record the portTag in the port array */
        controllers_gp[controllerTag_i]->port[ctr] = currPort;
    }

    /* Now let's initialize each of the port */
    for(ctr = 0; ctr < MAX_PORTS_PER_CONTROLLER; ctr++)
    {
        currPort = controllers_gp[controllerTag_i]->port[ctr];
        retval = PortInit(currPort, ctr);

        if(retval == SATA_COMPLETED_WITH_ERRORS)
        {
            /* SATA_COMPLETED_WITH_ERRORS implies that a device init failed */
            /* Non fatal error */
            DEBUG_PRINT_2("A Device connected to Ctlr %d, Port %d Failed in init.", controllerNumber_i, ctr);
            deviceInitFailed = TRUE;
        }
        else if(retval != SATA_SUCCESS)
        {
            DEBUG_PRINT_2("PortInit Failed for Ctlr %d, Port %d", controllerNumber_i, ctr);
            /* Treat port failures as fatal.  */
            return retval;
        }
    }

    if(deviceInitFailed == TRUE)
    {
        /* One or more devices failed during initialization */
        return SATA_COMPLETED_WITH_ERRORS;
    }

    /* Done with controller & port initializations */

    /* Successfully initialized. Return success */
    return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* FUNCTION     ControllerIntHardReset()                                      */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function resets the SATA interface of all ports         */
/*               and bring the PHY online. Should be called in order to bring */
/*               PHY online. The BIST done by software will call this after   */
/*               test                                                         */
/*                                                                            */
/* PRECONDITIONS    The port is ready for BIST at initialization              */
/*                                                                            */
/* POST-CONDITIONS  The PHY is offline or in BIST mode                        */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T ControllerIntHardReset(INT controllerTag_i)
{
    INT ctr;
    INT currPort;
    UINT failure_counter = 0;


    STATUS_CODE_T retval = SATA_SUCCESS;

    for(ctr = 0; ctr < MAX_PORTS_PER_CONTROLLER; ctr++)
    {
        ASSERT(controllers_gp[controllerTag_i]);
        currPort = controllers_gp[controllerTag_i]->port[ctr];

        if((retval = PortIntHardReset(currPort)) != SATA_SUCCESS)
        {
            failure_counter++;
        }

    }

    switch(failure_counter)
    {
    case 0:
        // All Drives present and accounted for.
    return SATA_SUCCESS;
        break;
    case 1:
    case 2:
    case 3:
        return SATA_COMPLETED_WITH_ERRORS;
        break;

    default:
        // There is a drive connected somewhere, just not on
        // all ports.
        return SATA_FAILURE;
    }

} /* end of ControllerIntHardReset() */
