/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                  I N C L U D E   S T A T E M E N T S                       */
/*                                                                            */
/******************************************************************************/

//For KdPrint statements: #include "ntddk.h"

#include <OSL.h>
#include <sata.h>
#include <sata_port.h>
#include <sata_globals.h>
#include <sata_device.h>
#include <sata_dma.h>

#include <artisea.h>

#ifdef WINNT
   #include <stdio.h>
#else
   #include <sata_debug.h>
   #include <linux/string.h> /* sprintf */
#endif

/******************************************************************************/
/*                                                                            */
/*            L O C A L   F U N C T I O N    P R O T O T Y P E S              */
/*                                                                            */
/******************************************************************************/

#ifdef EXPERIMENTAL_CODE
static STATUS_CODE_T DevProcessPioRequest(INT devTag_i,
                                          SATA_COMMAND_T * sataCmd_iop);
#endif
static STATUS_CODE_T DevProcessDmaRequest(INT devTag_i,
                                          SATA_COMMAND_T * sataCmd_iop);
static STATUS_CODE_T DevEndCommand(INT devTag_i,
                                   SATA_COMMAND_T * sataCmd_iop);
static STATUS_CODE_T NonDataIsr(INT devTag_i, UINT32 ipr);
static void DevBuildTaskFileLba28(SATA_COMMAND_T *sataCmd_p,
                                  INT startSector, INT numSectors);
static void DevBuildTaskFileLba48(SATA_COMMAND_T *sataCmd_p,
                                  INT startSector, INT numSectors);
static void DevBuildTaskFileTCQ(SATA_COMMAND_T *sataCmd_p,
                                INT startSector, INT numSectors);

boolean DevIsReady(INT devTag_i);
boolean DevDRQSet(INT devTag_i);

#ifdef SATA_DEBUG
static void FixString(char *start, const int len);
#endif

/******************************************************************************/
/*                                                                            */
/*       G L O B A L   F U N C T I O N   I M P L E M E N T A T I O N S        */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function allocates the private memory for the device. It */
/*              allocates a DEV_INFO_T structure and records the pointer in   */
/*              the next available slot in the devices_gp array. This         */
/*              function also initializes the allocated memory structure      */
/*              and sets the port member to the passed portTag.               */
/*                                                                            */
/* PRECONDITIONS    devices_gp array has been allocated.                      */
/*                                                                            */
/* POST-CONDITIONS  The next NULL pointer in device_gp is set.                */
/*                  *devTag_op is set to it's index number or INVALID_TAG     */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  The Tag of the parent port.                               */
/*              devTag_op                                                     */
/*                  Address of the INT variable to hold the devTag.           */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in devices_gp array        */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevAllocatePrivateMemory(INT portTag_i, INT *devTag_op)
{
   INT tag;
   DEV_INFO_T *newDevInfo;

   DEBUG_TRACE_1();

#ifdef SATA_DEBUG
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(portTag_i, MAX_PORTS) )
   {
      DEBUG_ERROR("portTag_i out of range: %d", portTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devTag_op == NULL_POINTER )
   {
      DEBUG_ERROR("portTag_i %d, devTag_op was NULL", portTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   *devTag_op = INVALID_TAG ; /* Will be assigned a valid value below */

   tag = getNextSlot((void**)devices_gp, MAX_DEVICES); /* NULL_POINTER safe */

   /* Check if we got an empty slot in the array */
   if ( tag == INVALID_TAG )
   {
      DEBUG_ERROR("Dev ? on port %d: Failed to get an empty slot"
                  "in devices_gp", portTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   if ( !devices_gp[tag] )
   {
      /* Allocate the Memory */
      newDevInfo = GetMemory(sizeof(DEV_INFO_T));
   }
   else
   {
      newDevInfo = devices_gp[tag];
   }

   if ( newDevInfo == NULL_POINTER )
   {
      DEBUG_ERROR("Dev %d on port %d: Failed to get memory"
                  "for DevInfo", tag, portTag_i);
      /* Failed to get memory */
      return SATA_GET_MEMORY_FAILED;
   }

   /* Record the pointer in the global array */
   devices_gp[tag] = (DEV_INFO_T *)newDevInfo;
   *devTag_op = tag;

   /* Initialize the structure */
   MemSet(sizeof(DEV_INFO_T), 0, newDevInfo);
   newDevInfo->present = FALSE; /* Mark as not yet completely initialized */
   newDevInfo->active = FALSE;
   newDevInfo->port = portTag_i;

   DEBUG_PRINT_1("New Device was given the tag %d", tag);
   return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function initializes the device.                         */
/*                                                                            */
/* PRECONDITIONS    The private memory for the device has been allocated.     */
/*                                                                            */
/* POST-CONDITIONS  The device is initialized and ready for I/O               */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to initialize.                         */
/*              devNumber_i                                                   */
/*                  The device number on the port represented by this device  */
/*                  RSM Not implemented. This Param is currently Ingored      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in devices_gp array        */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*              SATA_NOT_INITIALIZED    = Memory was not allocated.           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevInit(INT devTag_i, INT devNumber_i)
{
   DEV_INFO_T *devp = NULL_POINTER;
   int ctr;

   DEBUG_TRACE_1();

#ifdef SATA_DEBUG
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Value out of range: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   /* Check if the devNumber_i is valid */
   /*
   RSM Not yet implemented. Ignore these values
   if (devNumber_i >= MAX_DEVICES_PER_PORT) return SATA_OUT_OF_RANGE_VALUE;
   */

   devp = devices_gp[devTag_i];

   devp->ioMode = VARIABLE; /* Use Command specified mode by default */
   for ( ctr = 0; ctr < MAX_TCQ_DEPTH; ctr++ )
   {
      devp->commandQueue[ctr].slotState = SLOT_FREE;
      devp->commandQueue[ctr].slotTag = ctr;
   }

   /* The following 4 elements are set during AdapterAttachDisk */
   devp->osDriveTag      = NULL_POINTER;
   devp->dmaHandler      = NULL_POINTER;

   /* The following 4 elements will be set during DeviceIdentify()*/
   devp->isDrive = TRUE;  /* Right now, that's all we can have */
   devp->lbaCapacity = 0;
   // devp->multisector = 0;

#ifdef SATA_DEBUG
   MemSet(sizeof(IDENTIFY_DEVICE_T), 0, &(devp->devIdentify));
#endif

   OsSpinlockInit(&devp->portLock);
   OsSpinlockInit(&devp->sactiveLock);
   OsSpinlockInit(&devp->pendingCmdLock);

   devp->sactiveCopy = 0; /* clear SActive copy */
   devp->pendingCmd = devp->pendingCmdTail = NULL_POINTER;

   /* set feature list to point to global default before OSDM sets it */
   devp->devFeatures = &devFeatures_g;

   /* Initialization complete. Mark device as ready for use. */
   devp->present = TRUE;
   devp->active = TRUE;
   devp->busyCounter =  MAX_INTERRUPT_BUSY_WAIT;

   /* initializing ATA commands to defaults */
   devp->cmd.ataReadDma = ATA_READ_DMA;
   devp->cmd.ataWriteDma = ATA_WRITE_DMA;
   devp->cmd.ataReadMultiple = ATA_READ_MULTIPLE;
   devp->cmd.ataWriteMultiple = ATA_WRITE_MULTIPLE;
   devp->cmd.ataReadSectors = ATA_READ_SECTORS;
   devp->cmd.ataWriteSectors = ATA_WRITE_SECTORS;
   devp->cmd.ataReadLog = ATA_READ_LOG_EXT;
   devp->cmd.buildTaskFile = DevBuildTaskFileLba28;

   /* Successfully initialized the device. Return success */
   return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function performs a soft reset of the device.            */
/*                                                                            */
/* PRECONDITIONS    The private memory for the device has been allocated.     */
/*                                                                            */
/* POST-CONDITIONS  The device is initialized and ready for I/O               */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to initialize.                         */
/*              devNumber_i                                                   */
/*                  The device number on the port represented by this device  */
/*                  RSM Not implemented. This Param is currently Ingored      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in devices_gp array        */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*              SATA_NOT_INITIALIZED    = Memory was not allocated.           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevSoftReset(INT devTag_i, INT devNumber_i)
{

   DEBUG_TRACE_1();

/*     For a soft reset :                                                     */
/*         1. Set SRST bit in Device Control Reg.                             */
/*         2. Wait for 5 microseconds                                         */
/*         3. Clear SRSt bit.                                                 */
/*         4. wait for 2 milliseconds                                         */
/*         5. Read status / Alt Status register                               */
/*         6. Wait for the BSY bit to be cleared.                             */
/*         7. Check the Error Register and signature (pg 280)                 */
/*           (sect cnt=01h, sect #=01h, cyl_low=0h, cyl_high=0h, dev/head=0h) */
/*         8. Done.                                                           */
/*     We should expect the next command to be an Identify device             */

   /* Not yet implemented */
   return SATA_FAILURE;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function performs a hard reset of the device by          */
/*              resetting the port it is connected to. This funxtion will     */
/*              Have to change to support RSM / Port Multiplier               */
/*                                                                            */
/* PRECONDITIONS    The private memory for the device has been allocated.     */
/*                                                                            */
/* POST-CONDITIONS  The device is initialized and ready for I/O               */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to initialize.                         */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in devices_gp array        */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*              SATA_NOT_INITIALIZED    = Memory was not allocated.           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevHardReset(INT devTag_i)
{
   DEV_INFO_T     * devp    = NULL_POINTER;
   INT              portTag = INVALID_TAG;
   STATUS_CODE_T   retval   = SATA_SUCCESS;

   FLAGS_T flags = 0; /* spinlock flags */

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SATA_DEBUG
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   // Clear out the SActive copy.
   OsSpinLock(&devp->sactiveLock, flags);
   devp->sactiveCopy = 0x0;
   OsSpinUnlock(&devp->sactiveLock, flags);

   DEBUG_PRINT_1("DevHardReset called with devTag_i = %d, Port Tag %d on port #%d",
                 devTag_i, portTag, ports_gp[portTag]->portNumber);

#ifdef PARANOID_LOCKING
   OsSpinLock(&devp->portLock, portFlags);
#endif

   PortWriteDmaCommandReg(portTag, (UINT16)DMA_STOP);
   retval = PortIntHardReset(portTag);

   PortWriteDmaCommandReg(portTag, (UINT16)DMA_STOP);
   PortClearInterrupt(portTag);
   PortClearError(portTag);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif
   return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function selects a device. Until RSM hardware exists,    */
/*                 this function simply returns SATA_SUCCESS                  */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success                                       */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/

#if 0
INLINE STATUS_CODE_T SelectDevice(INT devTag_i)
{

   DEBUG_TRACE_1();

   /* we have one device per port at this stage (SATA 1.0)   */
   /* therefore the current device is it! no need to select. */

   #ifdef SANITY_CHECK
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   #endif

   return SATA_SUCCESS;
}
#endif


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function issues the command provided in the taskfile     */
/*              without any processing if the device is not busy. It writes   */
/*              all the taskfile parameters :                                 */
/*              Feature Reg, Sector Count Reg, Sector Number Reg, Cylinder    */
/*              Low Reg, Cylinder High Reg and Device Head Reg in that order. */
/*              It then writes to the Command Register and returns.           */
/*                                                                            */
/* PRECONDITIONS   The device has been initialized, and selected (caller is   */
/*                 expected to have called SelectDevice)                      */
/*                                                                            */
/* POSTCONDITIONS  The Interrupts on the drive will be *Enabled* if           */
/*                 waitOption_i was TRUE, un-modified otherwise.              */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              waitOption_i                                                  */
/*                  Should the API wait for the command to be completed ?     */
/*                  False = No wait. Write out the taskfile contents & return.*/
/*                  True = Wait.(Disable interrupts,Issue command,WAIT for dev*/
/*                      ready, Clear interrupts, Enable interrupts and return)*/
/*              taskfile_ip                                                   */
/*                  The exact contents of the taskfile to be written to the   */
/*                  device.                                                   */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              SATA_BUSY               = Device not ready (BSY bit set)      */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevExecuteRawCommand(
                                  INT             devTag_i,       /* DevTag of the device */
                                  boolean         waitOption_i,   /* F = Issue the command. Do nothing else */
                                  /* T = Wait for the command completion    */
                                  /*     with drive Interrupts disabled     */
                                  SATA_TASKFILE_T * taskfile_ip   /* Taskfile contents for the command      */
                                  )
{
   DEV_INFO_T     * devp    = NULL_POINTER;
   INT              portTag = INVALID_TAG;
   STATUS_CODE_T   retval   = SATA_SUCCESS;
   INT             ctr;


#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   if ( taskfile_ip == NULL_POINTER )
   {
      DEBUG_ERROR("ERROR: devTag: %d, taskfile_ip is NULL", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   if ( waitOption_i == TRUE )
   { /* Wait for command completion */
      /* Disable Interrupts on the drive.*/
      retval = DevDisableInterrupts(devTag_i);
      if ( retval != SATA_SUCCESS )
      {
         DEBUG_PRINT_1("Failed(%d) to disable device interrupts. "
                       "Command not yet issued", retval);

         /* Try to enable interrupts before we return ? */
         /* DevEnableInterrupts(devTag_i) */
         return retval;
      }
   }

   ctr = 0;
   while ( (DevIsReady(devTag_i) != TRUE) && (ctr < WAIT_READY) )
   {
      ctr++;
      /* Wait for 1 Millisecond more for the drive to be ready */
      MillisecondSleep(1);
   }

   if ( ctr >= WAIT_READY )
   {
      /* We Timed out ... */
      DEBUG_ERROR_0("Timed out waiting for Device to be ready (%d ms)", ctr);
      return SATA_BUSY;
   }

#ifdef PARANOID_LOCKING
   // Lock for the entire duration of issuing the command.
   OsSpinLock(&devp->portLock, portFlags);
#endif
   retval = PortWriteTaskfileParams( portTag, taskfile_ip );

#ifdef SANITY_CHECK_IN_DATA_PATH
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_PRINT_1("Failed(%d) to write TaskfileParams. "
                    "Command not yet issued", retval);

   #ifdef PARANOID_LOCKING
      OsSpinUnlock(&devp->portLock, portFlags);
   #endif
      return retval;
   }
#endif

   /* clear the interrupt and error status of the port before issuing a command */
   PortClearInterrupt (portTag);
   PortClearError (portTag);

   retval = PortWriteCommandReg(portTag, taskfile_ip->command);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif

   /* return if we are not asked to wait...*/
   if ( waitOption_i == FALSE ) return retval;

   /* return if there was an error issuing the command... */
   if ( retval != SATA_SUCCESS )
   {
      /* We were asked to wait, so we must have disabled the interrupts */

      /* Clear the Interrupt */
      PortClearInterrupt (portTag);
      /* Enable interrupts before we return... */
      DevEnableInterrupts(devTag_i);

      return retval;
   }

   /* Wait for the drive to be ready. */

   /* Wait for some time to allow the drive to asserts Busy */
   /* spec allows drive 400ns to assert "BUSY" */

   /* Sleep for a Microsecond */
   MicrosecondSleep(1);

   /* Make sure the drive is ready... */
   for ( ctr = 0; ctr < WAIT_CMD; ctr ++ )
   {
      /* is the drive ready for I/O ? */
      if ( DevIsReady(devTag_i) == TRUE ) break;

      /* Wait for 1 Millisecond more for the drive to be ready */
      MillisecondSleep(1);
   }

   /* Clear the Interrupt */
   PortClearInterrupt (portTag);

   /* Enable interrupts before we return... */
   DevEnableInterrupts(devTag_i);

   if ( ctr >= WAIT_CMD )
   {
      /* We Timed out ... */
      DEBUG_ERROR_0("ERROR: Timed out (%d ms) waiting for device to be ready.", ctr);
      return SATA_BUSY;
   }

   return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function transfers count_i  bytes of data from the data  */
/*              port register (SUPDDR), putting them in the source/dest       */
/*              buffer (if the device is not busy)                            */
/*                                                                            */
/* PRECONDITIONS   The device has been initialized, and selected (caller is   */
/*                 expected to have called SelectDevice).                     */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              rw_i                                                          */
/*                  Read = 0, Write = 1                                       */
/*              count_i                                                       */
/*                  The number of bytes to be read from the data port reg.    */
/*              buffer_op                                                     */
/*                  Address of the source/dest buffer. Should be large enough */
/*                  to hold count_i bytes.                                    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              SATA_BUSY               = Device not ready (BSY or !DRQ )     */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevRWPioData(
                          INT devTag_i,    /* DevTag of the device                */
                          INT rw_i,        /* READ from Dev = 0, WRITE to Dev = 1 */
                          INT count_i,     /* Number of bytes.                    */
                          void * buffer_op /* pointer to the source/dest buffer   */
                          )
{
   DEV_INFO_T     *devp    = NULL_POINTER;
   INT             portTag = INVALID_TAG;
   STATUS_CODE_T   retval   = SATA_SUCCESS;
   INT             ctr = 0;


#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   if ( buffer_op == NULL_POINTER )
   {
      DEBUG_ERROR_0("ERROR: devTag: %d, buffer_op is NULL", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   if ( (rw_i != SATA_READ) && (rw_i != SATA_WRITE) )
   {
      DEBUG_ERROR_0("Invalid value for rw_i (%d)", rw_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }

   devp = devices_gp[devTag_i];
   portTag = devp->port;


   /* clear the interrupt and error status of the port before issuing a command */
   PortClearInterrupt(portTag);
   PortClearError(portTag);

   if ( count_i <= 0 )
   {
      // Hmm... No data to transfer. Return Success... We might get a SATA Protocol error later.
      return SATA_SUCCESS;
   }

   while ( (DevIsReady(devTag_i) != TRUE) && (ctr < WAIT_DRQ) )  // Max 50 ms wait.
   {
      ctr++;
      /* Wait for 1 Millisecond more for the drive to be ready */
      MillisecondSleep(1);
   }

   if ( ctr >= WAIT_DRQ )
   {
      /* We Timed out ... */
      DEBUG_ERROR_0("Timed out (%d ms) waiting for BSY to clear", ctr);
      return SATA_BUSY;
   }

   if ( !DevDRQSet(devTag_i) )
   {
      // Drive not ready for data...
      DEBUG_ERROR_0("DRQ not set on DTag %d",devTag_i);
      return SATA_BUSY;
   }

#ifdef PARANOID_LOCKING
   OsSpinLock(&devp->portLock, portFlags);
#endif

   if ( rw_i == SATA_WRITE )
   {
      retval = PortPioWriteBytes(portTag, count_i, buffer_op);
   }
   else
   {
      retval = PortPioReadBytes(portTag, count_i, buffer_op);
   }

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif

   /* Check for system errors on the transfer */
   if ( retval != SATA_SUCCESS ) return retval;

   return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function adds a command to the device queue. This API    */
/*              should not be called from the OSDM (unless you know what you  */
/*              are doing)                                                    */
/*                                                                            */
/* PRECONDITIONS   The adapter has a command ready to send to the drive       */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              sataCmd_iop                                                   */
/*                  The command to be added to the commandQueue               */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              SATA_BUSY = No unused command slots found. Try again after a  */
/*                          command completes                                 */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevAddCommandToQueue(INT devTag_i, PSATA_COMMAND_T sataCmd_iop,
                                   INT slot_i)
{
   DEV_INFO_T     * devp      = NULL_POINTER;
   SATA_COMMAND_T * sataCmd_p = NULL_POINTER;
   FLAGS_T flags = 0; /* spinlock flags */

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }

   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   if ( sataCmd_iop == NULL_POINTER )
   {
      DEBUG_ERROR_0("ERROR: devTag: %d, sataCmd_iop is NULL", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }

   devp = devices_gp[devTag_i];

   sataCmd_p = DevGetQSlot(devTag_i, slot_i);

   /* lock command queue for modification */
   if ( sataCmd_p == NULL_POINTER )
   {
      /* A command has already been queued and has not yet completed. */
      DEBUG_ERROR_0("Command Slot is in use. Returning Busy%s","");
      return SATA_BUSY;
   }
   /* Fill out the command structure */

   sataCmd_p->slotState = SLOT_OCCUPIED;
   sataCmd_p->nextCmd = NULL_POINTER;

   MemCopy(sizeof(SATA_COMMAND_T), sataCmd_iop, sataCmd_p);

   /* chaing the pending commands together */
   OsSpinLock(&devp->pendingCmdLock, flags);
   if ( devp->pendingCmdTail == NULL_POINTER )
   {
      devp->pendingCmdTail = devp->pendingCmd = sataCmd_p;
   }
   else
   {
      devp->pendingCmdTail->nextCmd = sataCmd_p;
      devp->pendingCmdTail = sataCmd_p;
   }
   OsSpinUnlock(&devp->pendingCmdLock, flags);
   return SATA_SUCCESS;
} /* End of DevAddCommandToQueue() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function services the device's command queue.            */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              SATA_BUSY               = Drive not ready                     */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevServiceCommandQueue(INT devTag_i)
{
   /* when this function is called, the device issues any pending commands */
   /* that are currently in the command queue (we expect max of 1 command) */

   SATA_COMMAND_T * curCmd    = NULL_POINTER;
   DEV_INFO_T     * devp      = NULL_POINTER;
   STATUS_CODE_T retval       = SATA_SUCCESS;
   INT portTag                = INVALID_TAG;
   FLAGS_T flags = 0; /* spinlock flags */

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif

   INT ctr = 0;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   do
   {
      OsSpinLock(&devp->pendingCmdLock, flags);
      curCmd = devp->pendingCmd;
      if ( curCmd == NULL_POINTER )
      {
         /* No commands to be issued. */
         DEBUG_PRINT_1("Dev %d: No Pending commands to be issued.",
                       devTag_i);
         OsSpinUnlock(&devp->pendingCmdLock, flags);
//            DEBUG_PRINT_1("leaving...");
         return SATA_SUCCESS; // no command, do not fail
      }

      DEBUG_PRINT_1("Processing cmd slot # %d", curCmd->slotTag);

      if ( devp->pendingCmd->nextCmd )
      {
         devp->pendingCmd = devp->pendingCmd->nextCmd;
      }
      else
      {
         devp->pendingCmd = devp->pendingCmdTail = NULL_POINTER;
      }
      OsSpinUnlock(&devp->pendingCmdLock, flags);

      ASSERT(curCmd);
      curCmd->nextCmd = NULL_POINTER;

      /* Check if the drive is ready to accept the next command */
      ctr = WAIT_READY;
      while ( (DevIsReady(devTag_i) != TRUE) &&
              (ctr-- > 0) )
      {
         MicrosecondSleep(1); // busy wait
      }

      if ( ctr <= 0 )
      {
         /* Well, the Drive is not Ready and we timedout. Return SATA_BUSY */
         DEBUG_PRINT_1("devTag %d, Drive not Ready. Returning Busy.", devTag_i);
         /* put it back! put it back! */
         OsSpinLock(&devp->pendingCmdLock, flags);
         if ( devp->pendingCmd != NULL_POINTER )
         {
            curCmd->nextCmd = devp->pendingCmd;
            devp->pendingCmd = curCmd;
         }
         else
         {
            devp->pendingCmd = devp->pendingCmdTail = curCmd;
         }
         OsSpinUnlock(&devp->pendingCmdLock, flags);
         return SATA_BUSY;
      }

#ifdef PARANOID_LOCKING
      // Lock for the entire duration of issuing the command.
      OsSpinLock(&devp->portLock, portFlags);
#endif

      /* Drive is ready... write the task file parameters */
      retval = PortWriteTaskfileParams(portTag, &curCmd->taskfile);

#ifdef SANITY_CHECK_IN_DATA_PATH
      if ( retval != SATA_SUCCESS )
      {
         curCmd->slotState = SLOT_FREE;

   #ifdef PARANOID_LOCKING
         OsSpinUnlock(&devp->portLock, portFlags);
   #endif
         return retval;
      }
#endif

      /* From here on the steps differ for PIO transfers and DMA transfers */

      if ( (curCmd->cmdType == SATA_WRITE) || (curCmd->cmdType == SATA_READ) )
      {
         /* Is this a PIO or a DMA command ? */
         if ( curCmd->txType == FIRST_PARTY_DMA )
         {

            /* Mark slotState as containing an Active Command. */
            DEBUG_PRINT_1("Changing slot state of slot # %d to ACTIVE",
                          curCmd->slotTag);
            curCmd->slotState = SLOT_ACTIVE;

            // native command queuing does not need DMA setup
            // this is performed automatically by hardware
            if ( devp->devFeatures->TCQ == TRUE )
            {
               FLAGS_T lockFlag = 0; /* spinlock flags */

               OsSpinLock(&devp->sactiveLock, lockFlag);
               {
                  PortWriteCommandReg(portTag,curCmd->taskfile.command);

                  devp->sactiveCopy |= SET_BIT(0, curCmd->slotTag);
#ifdef QDEBUG
                  DEBUG_ERROR_0("Slot %d now in use, SActive Copy now %#08x", curCmd->slotTag, devp->sactiveCopy);
#endif
               }
               OsSpinUnlock(&devp->sactiveLock, lockFlag);

            }
            else
            {
               retval = DevProcessDmaRequest(devTag_i, curCmd);
            }
         }
         else
         {
            DEBUG_ERROR("ERROR: DevTag %d, Unsupprted IO type (0x%0x)",
                        devTag_i, curCmd->txType);
            curCmd->slotState = SLOT_FREE;

#ifdef PARANOID_LOCKING
            OsSpinUnlock(&devp->portLock, portFlags);
#endif
            /* neither 1st party DMA nor PIO ... */
            /*3rd Party DMA not yet supported */
            return SATA_OUT_OF_RANGE_VALUE;
         }
      }
      else
      { /* SATA_SPECIAL */
        /* this is not a read or write command, so process this special command */
         /* But special commands should not have been queued in the first place! */
         /* TBD - Decide what is to be done for this case. */

         DEBUG_ERROR("ERROR: DevTag %d, Non Data command on Queue (0x%0x)",
                     devTag_i, curCmd->cmdType);
         curCmd->slotState = SLOT_FREE;

#ifdef PARANOID_LOCKING
         OsSpinUnlock(&devp->portLock, portFlags);
#endif
         return SATA_OUT_OF_RANGE_VALUE;
      }

#ifdef PARANOID_LOCKING
      OsSpinUnlock(&devp->portLock, portFlags);
#endif
      /* act on the return value for the DevProcess???Request function calls */
      /* Did we successfully issue the command? If not, the command is left  */
      /* in the command queue as a pending command, else it is marked as an  */
      /* active command */
      if ( retval != SATA_SUCCESS )
      {
         DEBUG_ERROR_0("DevProcessXXXRequest() failed! Returning busy (%d)", SATA_BUSY);
         curCmd->slotState = SLOT_FREE;
         // putting the command back on pending queue
         OsSpinLock(&devp->pendingCmdLock, flags);
         if ( devp->pendingCmd != NULL_POINTER )
         {
            curCmd->nextCmd = devp->pendingCmd;
            devp->pendingCmd = curCmd;
         }
         else
         {
            devp->pendingCmd = devp->pendingCmdTail = curCmd;
         }
         OsSpinUnlock(&devp->pendingCmdLock, flags);
         return SATA_BUSY;
      }

   } while ( 1 ); // Try to issue all commands

   /* NOTE: An interrupt will be generated for every completed command. The */
   /*       ISR will change the slotstate to SLOT_FREE                      */
   return SATA_SUCCESS;
} /* End of DevServiceCommandQueue() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function determines whether the device is ready.         */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      TRUE / FALSE                                                  */
/*                                                                            */
/* ON ERRORS    returns FALSE on any errors                                   */
/*                                                                            */
/******************************************************************************/
boolean DevIsReady(INT devTag_i)
{
   UINT8 altStatus = 0;
   STATUS_CODE_T retval = SATA_SUCCESS;

   DEBUG_TRACE_3();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return FALSE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return FALSE;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return FALSE;
   }
#endif

   /* check the altstatus register  */
   retval = PortReadAltStatusReg(devices_gp[devTag_i]->port, &altStatus);

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Failed to get register value. ASSUME device is NOT ready. */
   if ( retval != SATA_SUCCESS ) return FALSE;
#endif
   /*  if the BSY bit is 0, then we report that the device is ready  */
   /* The caller might need to check the DRQ bit to see if the drive */
   /* is ready for/with data                                         */

   /* What is DRDY used for ? (READ_BIT( altStatus, SUPDASR_DRDY_BIT) == 1 )*/
   if ( READ_BIT( altStatus, SUPDASR_BSY_BIT)  == 0 )
   {
      return TRUE;
   }
   else
   {
      return FALSE;
   }
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function determines whether the DRQ is set.              */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      FALSE if DRQ-Bit is set,                                      */
/*              TRUE  if DRQ-Bit is 0,                                        */
/*                                                                            */
/* ON ERRORS    returns FALSE on any errors                                   */
/*                                                                            */
/******************************************************************************/
boolean DevDRQSet(INT devTag_i)
{
   UINT8 altStatus = 0;
   STATUS_CODE_T retval = SATA_SUCCESS;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return FALSE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return FALSE;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return FALSE;
   }
#endif

   /* check the altstatus register  */
   retval = PortReadAltStatusReg(devices_gp[devTag_i]->port, &altStatus);

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Failed to get register value. ASSUME device is NOT ready. */
   if ( retval != SATA_SUCCESS ) return FALSE;
#endif

   /* if the BSY bit is 0, then we report that the device is ready   */
   /* The caller might need to check the DRQ bit to see if the drive */
   /* is readt for/with data                                         */

   if ( (READ_BIT( altStatus, SUPDASR_DRQ_BIT) == 1 ) )
   {
      DEBUG_PRINT_1("ASR = %#02x, DRQ set", altStatus);
      return TRUE;
   }
   else
   {
      DEBUG_PRINT_1("ASR = %#02x, DRQ *NOT* set", altStatus);
      return FALSE;
   }

}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function gets the status of the device by calling the    */
/*              PortReadStatusReg fcn. Calling this function will have all    */
/*              side effects of reading the taskfile status register.         */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              status_op                                                     */
/*                  pointer to a UINT8 variable that will hold the status.    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error codes on error                 */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevGetStatus(INT devTag_i, UINT8 * status_op )
{
   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   return PortReadAltStatusReg(devices_gp[devTag_i]->port, status_op);
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function gets the Alt status of the device by calling    */
/*              the PortReadAltStatusReg fcn. Calling this function WILL NOT  */
/*              automatically clear the error on the port. Call DevClearError */
/*              to clear the errors.                                          */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              status_op                                                     */
/*                  pointer to a UINT8 variable that will hold the status.    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error codes on error                 */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevGetAltStatus(INT devTag_i, UINT8 * status_op )
{
   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   return PortReadAltStatusReg(devices_gp[devTag_i]->port, status_op);
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function gets the DMA status of the device by calling    */
/*              the PortReadDmaStatusReg fcn. Calling this function WILL NOT  */
/*              automatically clear the error on the port. Call DevClearError */
/*              to clear the errors.                                          */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              status_op                                                     */
/*                  pointer to a UINT8 variable that will hold the status.    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error codes on error                 */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevGetDmaStatus(INT devTag_i, UINT8 * status_op )
{
   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   return PortReadDmaStatusReg(devices_gp[devTag_i]->port, status_op);
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function gets the contents of the taskfile error reg.    */
/*              by calling the PortReadErrorReg fcn.                          */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              error_op                                                      */
/*                  pointer to a UINT8 variable that will hold the status.    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error codes on error                 */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevGetError(INT devTag_i, UINT8 * error_op )
{
   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   return PortReadErrorReg(devices_gp[devTag_i]->port, error_op);
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function gets both the SStatus and the Serror superset   */
/*              status registers along with a the error task file for the     */
/*              currently pending task.                                       */
/*              NOTE that it is up to the OSDM to parse this data.            */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*              An outstanding task with a pending error.                     */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              sStatus_op                                                    */
/*                  pointer to a UINT32 variable that will hold the sStatus.  */
/*              sError_op 	                                                  */
/*                  pointer to a UINT32 variable containing the sError data.  */
/*              errorTaskFile_op                                              */
/*                  pointer to the error task file                            */
/*                                                                            */
/* RETURNS      status plus the filled in data pointed to by the op pointers. */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevGetExtendedError(INT devTag_i,
                                  UINT32 * sStatus_op,
                                  UINT32 * sError_op,
                                  SATA_TASKFILE_T * taskFile_op)
{
   STATUS_CODE_T retval1, retval2, retval3;
   retval1 = retval2 = retval3 = SATA_SUCCESS;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   retval1 = PortReadSStatusReg(devices_gp[devTag_i]->port, sStatus_op);

   retval2 = PortReadSErrorReg(devices_gp[devTag_i]->port, sError_op);

   retval3 = PortReadTaskfileParams(devices_gp[devTag_i]->port, taskFile_op);


   if ( (retval1 | retval2 | retval3) == SATA_SUCCESS )
   {
      return SATA_SUCCESS;
   }
   else
   {
      DEBUG_ERROR("ERROR: DevTag %d, Failed to Read Registers", devTag_i);
      return SATA_FAILURE;
   }
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function clears the errors on the port attached to the   */
/*              device                                                        */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS                                                  */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevClearError(INT devTag_i)
{
   STATUS_CODE_T retval = SATA_SUCCESS;

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

#ifdef PARANOID_LOCKING
   OsSpinLock(&devices_gp[devTag_i]->portLock, portFlags);
#endif

   retval = PortClearError(devices_gp[devTag_i]->port);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devices_gp[devTag_i]->portLock, portFlags);
#endif

   return retval;

}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function determines whether an interrupt is pending      */
/*                   for this device by calling the PortInterruptPending fcn. */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                                                                            */
/* RETURNS      TRUE / FALSE                                                  */
/*                                                                            */
/* ON ERRORS    returns FALSE on any errors (Can we do something better ?)    */
/*                                                                            */
/******************************************************************************/
INLINE boolean DevInterruptPending(INT devTag_i)
{
   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return FALSE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return FALSE;
   }
//   if ( devices_gp[devTag_i]->active == FALSE )
//   {
//      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
//      return FALSE;
//   }
#endif


#ifdef EXPERIMENTAL_CODE
   {
      STATUS_CODE_T retval = SATA_SUCCESS;

      /* Select the device, and then check the port */
      retval = SelectDevice(devTag_i);
      /* Return false for errors */
      if ( retval != SATA_SUCCESS ) return FALSE;
   }
#endif

   return PortInterruptPending(devices_gp[devTag_i]->port);
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function determines whether a DMA interrupt is pending   */
/*                   for this device.                                         */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* RETURNS      TRUE / FALSE                                                  */
/*                                                                            */
/* ON ERRORS    returns FALSE on any errors (Can we do something better ?)    */
/*                                                                            */
/******************************************************************************/
INLINE boolean DevDmaInterruptPending(INT devTag_i)
{
   STATUS_CODE_T retval = SATA_SUCCESS;
   UINT8 dsr;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return FALSE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return FALSE;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return FALSE;
   }
#endif


#ifdef EXPERIMENTAL_CODE
   /* Select the device, and then check the port */
   retval = SelectDevice(devTag_i);

   /* Return false for errors */
   if ( retval != SATA_SUCCESS ) return FALSE;
#endif

   /* Check the DMA status register for a pending DMA Interrupt */
   retval = PortReadDmaStatusReg(devices_gp[devTag_i]->port, &dsr);

#ifdef SANITY_CHECK_IN_DATA_PATH
   if ( retval != SATA_SUCCESS ) return FALSE;
#endif

   if ( READ_BIT(dsr, SUPDDSR_DMA_INTR_BIT) ) return TRUE;

   return FALSE;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function disables the interrupts for the given drive     */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to use.                                */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevDisableInterrupts (INT devTag_i)
{
   STATUS_CODE_T retval = SATA_SUCCESS;

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif


#ifdef EXPERIMENTAL_CODE
   {
      STATUS_CODE_T retval = SATA_SUCCESS;

      /* Select the device, and then check the port */
      retval = SelectDevice(devTag_i);

      /* Return false for errors */
      if ( retval != SATA_SUCCESS ) return retval;
   }
#endif

#ifdef PARANOID_LOCKING
   OsSpinLock(&devices_gp[devTag_i]->portLock, portFlags);
#endif

   retval = PortDisableInterrupts(devices_gp[devTag_i]->port);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devices_gp[devTag_i]->portLock, portFlags);
#endif

   return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function enables the interrupts for the given device.    */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to use.                                */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevEnableInterrupts(INT devTag_i)
{
   STATUS_CODE_T retval = SATA_SUCCESS;

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif


#ifdef EXPERIMENTAL_CODE
   {
      STATUS_CODE_T retval = SATA_SUCCESS;

      /* Select the device, and then check the port */
      retval = SelectDevice(devTag_i);

      /* Return false for errors */
      if ( retval != SATA_SUCCESS ) return retval;
   }
#endif

#ifdef PARANOID_LOCKING
   OsSpinLock(&devices_gp[devTag_i]->portLock, portFlags);
#endif

   retval = PortEnableInterrupts(devices_gp[devTag_i]->port);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devices_gp[devTag_i]->portLock, portFlags);
#endif

   return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function marks the given command slot as free.           */
/*              Currently, for debug purposes, the last completed command is  */
/*              left around in the device structure.                          */
/*                                                                            */
/* PRECONDITIONS  The device is ready for i/o, and the sataCmd_iop command    */
/*                has already been issued.                                    */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              sataCmd_iop                                                   */
/*                  The command to be processed                               */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              SATA_BUSY               = Not Ready for I/O                   */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
static STATUS_CODE_T DevEndCommand(INT devTag_i, SATA_COMMAND_T * sataCmd_iop)
{
   DEV_INFO_T     *devp       = NULL_POINTER;
   INT             portTag    = INVALID_TAG;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check - Could be called from the ISR */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( sataCmd_iop == NULL_POINTER )
   {
      DEBUG_ERROR("Error: devTag %d, sataCmd_iop is NULL", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   /* we are keep the last completed command in the system for some more time. */
   /* A debug function can then dump this information. This could help during  */
   /* the initial debug of the lib.                                            */
#ifdef SATA_DEBUG
   MemCopy(sizeof(SATA_COMMAND_T), sataCmd_iop, &devp->lastCommand);
#endif


#ifdef SANITY_CHECK_IN_DATA_PATH
   if ( sataCmd_iop->slotState != SLOT_ACTIVE )
   {
      DEBUG_ERROR("ERROR: EndCommand Called for slot #%d in state %d, devTag %d",
                  sataCmd_iop->slotTag, sataCmd_iop->slotState, devTag_i);
   }
#endif

   /* We will clear the slot for reuse, */
   /* irrespective of the current state of the slot. */
   sataCmd_iop->slotState = SLOT_FREE;

#ifdef SATA_DEBUG_0
   /* Record the status/error register values */
   PortReadAltStatusReg(portTag, &sataCmd_iop->status);
   PortReadSStatusReg  (portTag, &sataCmd_iop->sStatus);
   if ( sataCmd_iop->txType == FIRST_PARTY_DMA )
   {
      PortReadDmaStatusReg(portTag, &(sataCmd_iop->dma.status));
   }
   PortReadErrorReg(portTag, &sataCmd_iop->error);
   PortReadSErrorReg(portTag, &sataCmd_iop->sError);
#endif

   return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function will process a DMA request for a device.        */
/* NOTE: Note that the SATA port does not generate an interrupt when a bus    */
/*       master operation is aborted. Software should time out.               */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              sataCmd_iop                                                   */
/*                  The command to be processed                               */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
static STATUS_CODE_T DevProcessDmaRequest(INT devTag_i,
                                          SATA_COMMAND_T * sataCmd_iop)
{
   INT           portTag   = INVALID_TAG;
   STATUS_CODE_T retval    = SATA_SUCCESS;
   DEV_INFO_T   *devp      = NULL_POINTER;

   UINT32 ddtpr_lower32;
   UINT32 ddtpr_upper32;

   UINT16 cmd = 0x1;            /* DMA start */
   FLAGS_T flags = 0; /* spinlock flags */

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }

   if ( sataCmd_iop == NULL_POINTER )
   {
      DEBUG_ERROR("Error: devTag %d, sataCmd_iop is NULL", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

// The Function Calling us (DevServiceCommandQueue) will be holding this lock when they call us.
// #ifdef PARANOID_LOCKING
//    OsSpinLock(&devp->portLock, portFlags);
// #endif

   /* clear the interrupt and error status of the port */
   PortClearInterrupt(portTag);
   PortClearError(portTag);


   // We don't need to set the DDTs in NCQ/TCQ mode...
   if ( !devp->devFeatures->TCQ )
   {
      /* Calculate the lower and upper parts of the address */
      ddtpr_upper32 = UPPER_32_BITS(sataCmd_iop->dma.ddt);
      ddtpr_lower32 = LOWER_32_BITS(sataCmd_iop->dma.ddt);

      /* Load the address of the DDT into the DDTPR registers */
      PortSetDTPR(portTag, ddtpr_upper32, ddtpr_lower32);

      /* Load the upper 32 bits of the addresses in the DDT entries into SUPDUDDBPR */
      PortSetDdtePage(portTag, sataCmd_iop->dma.ddtePage);
   }

   /* write to the DMA command register to set direction, SUPDDCMDR*/
   if ( sataCmd_iop->cmdType == SATA_READ )
   {
      // NPG: TEST CODE. Commenting out the following write to DMA reg. Should be OK.
      // PortWriteDmaCommandReg(portTag, DMA_DIRECTION_DEV_TO_HOST);
      cmd = DMA_DIRECTION_DEV_TO_HOST; // DMA is started separately
   }
   else if ( sataCmd_iop->cmdType == SATA_WRITE )
   {
//        PortWriteDmaCommandReg(portTag, DMA_DIRECTION_HOST_TO_DEV);
      cmd = DMA_DIRECTION_HOST_TO_DEV; // DMA is started separately
   }

#ifdef EXPERIMENTAL_CODE  /* Currently not used under Linux or Windows */
   /* set the DMA interrupt handler */
   retval = SetISR(devp->osDriveTag, devp->dmaHandler);
   /* Check for system errors */
   if ( retval != SATA_SUCCESS ) return retval;
#endif

   // Mark the sactive Bit and issue the command. The OsSpinLock is ISR safe (i.e. interrupts disabled)
   // We mark the sactive bit and issue the command. We SHOULD not get an interrupt between these two steps
   // as it would falsely assume that the command has already completed.

   if ( devp->devFeatures->TCQ )
   {
      OsSpinLock(&devp->sactiveLock, flags);
   }

   /* issue the READ/WRITE DMA command by writing to the taskfile command reg */
   retval = PortWriteCommandReg(portTag, sataCmd_iop->taskfile.command);

   if ( devp->devFeatures->TCQ )
   {
      devp->sactiveCopy |= SET_BIT(0, sataCmd_iop->slotTag);

#ifdef QDEBUG
      DEBUG_ERROR_0("Slot %d now in use., SActive Copy now %#08x", sataCmd_iop->slotTag, devp->sactiveCopy);
#endif

      OsSpinUnlock(&devp->sactiveLock, flags);
   }

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Check for system errors on PortWrite */
   if ( retval != SATA_SUCCESS )
   {
   #ifdef EXPERIMENTAL_CODE  /* Currently not used under Linux or Windows */
      /* Set ISR to default ISR before returning. */
      SetISR(devp->osDriveTag, NULL_POINTER);
   #endif
      return retval;
   }
#endif

   /* start the DMA engine by writing to the DMA command register, SUPDDCMDR */
   retval = PortWriteDmaCommandReg(portTag, (UINT16)(DMA_START | cmd));


#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Check for system errors on PortWrite */
   if ( retval != SATA_SUCCESS )
   {
   #ifdef EXPERIMENTAL_CODE  /* Currently not used under Linux or Windows */
      /* Set ISR to default ISR before returning. */
      SetISR( devp->osDriveTag, NULL_POINTER );
   #endif
      DEBUG_ERROR("PortWriteDmaCommandReg() failed: %d", retval);

      return retval;
   }
#endif

   DEBUG_PRINT_1("exited...%s","");
   return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function sends the IDENTIFY DEVICE command to the drive  */
/*              and reads the 512 bytes of information returned by the drive. */
/*              This information is returned to the caller in buffer_op and   */
/*              a copy is saved for debug information.                        */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              buffer_op                                                     */
/*                  Pointer to a buffer of size 512 bytes to hold the         */
/*                  the reply packet.                                         */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              SATA_FAILURE            = drive not present                   */
/*              SATA_BUSY               = Timed out                           */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevIdentify(INT devTag_i, void * buffer_op)
{
   DEV_INFO_T     * devp      = NULL_POINTER;
//   STATUS_CODE_T retval       = SATA_SUCCESS;
   INT portTag                = INVALID_TAG;
   INT ctr = 0;

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }

   if ( buffer_op == NULL_POINTER )
   {
      DEBUG_ERROR("Error: DevTag %d, buffer_op is NULL", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }

#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

#ifdef SANITY_CHECK
   if ( !TAG_WITHIN_RANGE(portTag, MAX_PORTS) )
   {
      DEBUG_ERROR("Invalid portTag: %d", portTag);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
#endif

   /* Check if drive is present */
   if ( PortDevConnected(portTag) != TRUE )
   {
      DEBUG_ERROR("Error: DevTag %d, Drive Not connected !", devTag_i);
      /* Drive not connected to port ?!! This should not be happening !!!*/
      return SATA_FAILURE;
   }

   /* Wait for some time Max = WAIT_WORSTCASE, if BUSY bit set */
   /* Make sure the drive is ready... */
   /* is the drive ready for I/O ? (not Busy?) */
   ctr = 0;
   while ( (DevIsReady(devTag_i) != TRUE) && (ctr < WAIT_WORSTCASE) )
   {
      ctr++;
      /* Wait for 1 Millisecond more for the drive to be ready */

      // For Debug use. To be removed after 5/1/2003
      MillisecondSleep(1);
   }

   if ( ctr >= WAIT_WORSTCASE )
   {
      /* We Timed out ... */
      DEBUG_ERROR_0("Timed out waiting for Device to be ready (%d ms)", ctr);
      return SATA_BUSY;
   }

#ifdef PARANOID_LOCKING
   OsSpinLock(&devp->portLock, portFlags);
#endif

   /* Disable Interrupts */
   PortDisableInterrupts(portTag);

   /* Write the ATA_IDENTIFY_DEVICE command */
   PortWriteCommandReg(portTag, ATA_IDENTIFY_DEVICE);

   MicrosecondSleep(1); /* Give drive 400 nsec to set the busy flag */

   /* Wait for some time Max = WAIT_WORSTCASE, if BUSY bit set */
   /* Make sure the drive is ready... */
   /* is the drive ready for I/O ? (not Busy?) */
   while ( (DevIsReady(devTag_i) != TRUE) && (ctr < WAIT_WORSTCASE) )
   {
      ctr++;
      /* Wait for 1 Millisecond more for the drive to be ready */
      MillisecondSleep(1);
   }

   if ( ctr >= WAIT_WORSTCASE )
   {
      /* We Timed out ... */
      DEBUG_ERROR_0("Timed out waiting for Device to be ready (%d ms)", ctr);

#ifdef PARANOID_LOCKING
      OsSpinUnlock(&devp->portLock, portFlags);
#endif

      return SATA_BUSY;
   }

   /* Check for DRQ, (wait if required)*/
   if ( DevDRQSet(devTag_i) == FALSE )
   {
      /* wait 50 ms */
      MillisecondSleep(50);
      if ( DevDRQSet(devTag_i) == FALSE )
      {
         DEBUG_ERROR_0("Timed out - Dev Ready after %d ms, But no DRQ", ctr);

#ifdef PARANOID_LOCKING
         OsSpinUnlock(&devp->portLock, portFlags);
#endif
         return SATA_BUSY; /* We timed out... */
      }
   }

   /* Read PIO data, 512 bytes */
   PortPioReadBytes(portTag, SATA_SECTOR_SIZE , buffer_op);

#ifdef SATA_DEBUG
   /* Make the internal debug copy of identify info */
   MemCopy(SATA_SECTOR_SIZE, buffer_op, &(devp->devIdentify));

   /* Fix the model, f/w rev, serial no strings */
   FixString(&devp->devIdentify.modelNumber[0],
             sizeof(devp->devIdentify.modelNumber));
   FixString(&devp->devIdentify.firmwareRevision[0],
             sizeof(devp->devIdentify.firmwareRevision));
   FixString(&devp->devIdentify.serialNumber[0],
             sizeof(devp->devIdentify.serialNumber));
#endif

   /* What do we do of errors at this point ? (not that we expect any) */
   /* clear the drve interrupts. */
   PortClearInterrupt(portTag);

   /* Enable Interrupts */
   PortEnableInterrupts(portTag);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif

   /* Done. */
   return SATA_SUCCESS;
}


#ifdef SATA_DEBUG
static void FixString (char *start, const int len)
{
   char *ptr = start;
   char *end, *ptr2 ;

   DEBUG_TRACE_1();

   /* Even length for byte swapping */
   end = (len % 2) ? &start[len]: &start[len-1];

   ptr = start;

   /* strip leading space */
   while ( ptr != end && *ptr == ' ' )
   {
      ptr++;
   }

   ptr2 = start;

   /* ptr2 is at the start of the string, and ptr is at the start of*/
   /* the non-balnk part */

   /* compress internal blanks and strip trailing space */
   while ( (ptr != end) && (*ptr) != '\0' )
   {
      if ( *ptr != ' ' )
      {
         *ptr2 = *ptr;
         ptr2++;
      }
      ptr++;
   }

   /* ptr is either at the end of the string, or a null.*/
   /* ptr2 is at the end of the string we copied */

   /* wipe out trailing garbage */
   while ( ptr2 != end )
   {
      *ptr2++ = '\0';
   }
}
#endif

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function is called by the OSDM ISR to complete the data  */
/*              transfers. This function determines which command for devTag_i*/
/*              drive caused the interrupt (without TCQ, it will be the 1st   */
/*              command in the submittedQueue). For DMA transfers, It will    */
/*              check for errors, remove the command from the submitted       */
/*              commands queue, end the DMA transfer and return an appropriate*/
/*              status value. For PIO transfers, it will transfer the correct */
/*              number of bytes. If this transfer ends the ATA command, it    */
/*              will remove the command from the command queue and return     */
/*              the appropriate status. If this transfer doesn't end the ATA  */
/*              command, it will return SATA_SUCCESS_NOT_COMPLETED.           */
/*                                                                            */
/* PRECONDITIONS  In interrupt space                                          */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS_COMPLETED on success, error code on failure      */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_SUCCESS_NOT_COMPLETED = Part of ATA command completed.   */
/*                                           Command still in the Queue       */
/*              SATA_SUCCESS_COMPLETED  = ATA command completed, removed from */
/*                                        Command Queue.                      */
/*              SATA_FAILURE            = ATA Protocol level error (StatusReg)*/
/*              SATA_BUSY               = Device not Ready or device did not  */
/*                                        issue the Interrupt                 */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag / other value           */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevIsr(INT devTag_i, UINT32 *doneCmds_op)
{
   STATUS_CODE_T retval       = SATA_SUCCESS;
   DEV_INFO_T     * devp      = NULL_POINTER;
   INT portTag                = INVALID_TAG;
   SATA_COMMAND_T * currCmd   = NULL_POINTER;
   //UINT8 dmaStatReg;
   //UINT8 statReg;
   INT controller;
   SATA_REG_ADDR_T base;
   UINT32 ipr = 0;
   UINT32 sactive = 0;
   INT count = 0;
   FLAGS_T flags = 0; /* spinlock flags */
   int busy_ctr = 500;   // WAIT_READY is 30 ms... could atleast give it 500 micro sec ?
   // This should be reduced to as small a number as possible... may be seagate can fix their
   // drives so they don't stay busy after asserting interrupts.

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }

#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   // Do we even need to check if the drive is busy ?
   // Aparently, We don't need to check if the drive is busy when we do TCQ
   // With TCQ, we don't check any register other than the SActive...
   // with Non-TCQ drives, we still need to stop the DMA engine... and there fore we should
   // make sure the drive is not busy...
   if ( devp->devFeatures->TCQ != TRUE )
   {

      while ( (DevIsReady(devTag_i) != TRUE) && (busy_ctr-- > 0) )
      {
         /* Give the drive some time to clear the busy bit */
         MicrosecondSleep(1);
      }

      if ( busy_ctr <= 0 )
      {
         // TBD: NPG: What do we do if the drive died on us with the interrupt asserted and busy bit set ?
         // currently, for this case, the driver would get stuck in an interrupt storm (?)

         // Soln for above problem. maintain a counter per drive. decrement the counter everytime
         // we return busy. if the counter hits 0, we clear the pending interrupts before we return busy
         // The counter is reset to the default value after the interrupts are cleared.
         // This way, we can prevent the interrupt storm.

         devp->busyCounter--;
         if ( devp->busyCounter <=0 )
         {
            // Clear the interrupts
            PortClearInterrupt(portTag);

            // reset the counter
            devp->busyCounter = MAX_INTERRUPT_BUSY_WAIT;

            DEBUG_ERROR("Warning: Ignored an interrupt from busy drive %d. Will service later.",devTag_i );
            return SATA_BUSY;
         }

         DEBUG_ERROR_0("Drive not ready. Returning Busy (%d)", SATA_BUSY);
         return SATA_BUSY;
      }
      else
      {
         // reset the per drive busy counter cos' we were not busy !!!
         devp->busyCounter = MAX_INTERRUPT_BUSY_WAIT;
      }
   }

   /* NOTE: Locking of devp->portLock for write access to Port Registers            */
   /*       Access in ISR is for the following reasons:                             */
   /*       1. Clear Interrupt - OK to clear simultaneously from multiple locations */
   /*       2. DmaCommandReg for non NCQ drives. In this case, only one command can */
   /*          be active. Therefore we can safely assume that no new DMA will be    */
   /*          started while this command s active.                                 */

   /* Algo: This function will try to determine what caused the interrupt.     */
   /* If we think that a data transfer did not cause the interrupt, we will    */
   /* call the NonDataIsr function, which will handle all non-data interrupts  */
   /* and error interrupts                                                     */
   /* NPG: TBD: Check for errors on port Read/Write function calls             */

   /* TBD: Is there a sure-shot way to determine if the interrupt is not for a */
   /* Data Transfer ?                                                          */

   ASSERT(ports_gp[portTag]);
   controller = ports_gp[portTag]->controller;

   ASSERT(controllers_gp[controller]);
   base = controllers_gp[controller]->commonRegBase;

   // Clean out the old values in the return variable.
   *doneCmds_op = 0;

   /* Read the Interrupts pending register */
   PciMemRead32(base + SUPDIPR, &ipr);
   DEBUG_PRINT_4("IPR: %#010x", ipr);

   /* Move the interrupt bits relevant to this port in lower byte */
   ipr >>= SUPDIPR_BITS_PER_PORT * ports_gp[portTag]->portNumber;
   /* Mask off all bits not applicable to this port */
   ipr &=  SUPDIPR_BITS_PER_PORT_MASK;


   /* if it's IDE interrupt */
   if ( !(ipr & 0x7f) )
   {
      /* Clear the interrupt */
      PortClearInterrupt(portTag);

      if ( devp->devFeatures->TCQ == TRUE )
      {
         retval = PortReadSActive(portTag, &sactive);

         // if sactive bit is set and is cleared later, then we have cmd done
         // i.e., set in the sactiveCopy, but cleared in sactive
         // = (bits different) bit_AND (set in sactiveCopy)
         *doneCmds_op = (devp->sactiveCopy ^ sactive) & devp->sactiveCopy;  // sets 1 for bits turned off in sactive.

#ifdef QDEBUG
         DEBUG_ERROR_0("IPR %%04x SActive: %#08x, SActive COPY: %#08x, doneCmds: %#08x", ipr, sactive, devp->sactiveCopy, *doneCmds_op);
#endif

         // Lock the access to make it SMP safe.
         OsSpinLock(&devp->sactiveLock, flags);
         {
            // Store the new sactive to asctiveCopy.
            devp->sactiveCopy = sactive;
         }
         OsSpinUnlock(&devp->sactiveLock, flags);


         for ( count=0; count < devp->devFeatures->QueueDepth; count++ )
         {
            if ( (*doneCmds_op >> count) & 0x1 )
            {
               currCmd = &devp->commandQueue[count];

               /* End the command */
               DevEndCommand(devTag_i, currCmd);
            } // end if
         } // end for

//#if SATA_DEBUG_0
         if ( doneCmds_op <= 0 )
         {
            PciMemRead32(base + SUPDIPR, &ipr);
            DEBUG_ERROR_0("IPR %#08x, port %d SActive: %#08x, SActive COPY: %#08x, doneCmds: %#08x", ipr, ports_gp[portTag]->portNumber,
                          sactive, devp->sactiveCopy, *doneCmds_op);
         }
//#endif

      }
      else
      {
         /* Determine which command caused the Interrupt...               */
         /* without TCQ, it's the 1st command in the commandQueue */
         currCmd = &devp->commandQueue[0];

         /* Figure out if it was a PIO command, */
         /* a DMA command or is this something else ? */
         if ( currCmd->txType == FIRST_PARTY_DMA )
         {
            /* OK, so a DMA caused the interrupt. First, let's stop the DMA */

            /* stop the DMA engine by writing to the DMA command register, SUPDDCMDR */
            retval = PortWriteDmaCommandReg(portTag, DMA_STOP);

            /* End the command */
            DevEndCommand(devTag_i, currCmd);
         }
         else
         {
            // third party DMA not supported
            /* Non data commands should not be queued... */
            DEBUG_PRINT_1("SATA Command with wrong Transfer type (%d)",
                          currCmd->txType);
            return SATA_OUT_OF_RANGE_VALUE;

         }
      }
#ifdef SATA_DEBUG

      PortReadSActive(portTag, &sactive);
      DEBUG_PRINT_1("CURRENT SACTIVE: %#08x, doneCmds_op = %#08x", sactive, *doneCmds_op);
#endif

      return SATA_SUCCESS_COMPLETED;
   } /* end if(no error) */
   else if ( ipr > 0 )
   {
      /* No command submitted. Is this an interrupt from Non-data command or */
      /* Was there an error ? */

      DEBUG_ERROR_0("IPR: %#08x", ipr);
      retval = NonDataIsr(devTag_i, ipr);

#ifdef SATA_DEBUG
// NPG TEST CODE
// default action is to clear interrupt and return. For testing, we don't clear the interrupt here
// We clear it in the ISR.
      if ( retval != SATA_SUCCESS )
      {
         DEBUG_ERROR("ISR Returning Non Success Status (%d)", retval);
      }
#endif
         // clear the interrupt...
         PortClearInterrupt(portTag); /* Clear the interrupt */

      return retval;
   }
   else
   {
      DEBUG_ERROR("devTag %d, ISR called without a pending Interrupt!", devTag_i);
      /* None of the interrupt bits were set */
      return SATA_SUCCESS;
   }

   /* Should not Reach here !!! */
   DEBUG_ERROR("Internal Error. Masked IPR = %0x. Returning Failure (%d)", ipr, SATA_FAILURE);
   return SATA_FAILURE;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function handles all the non-data interrupts when        */
/*              devIsr() function is called. This includes hotplug events.    */
/*                                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to check.                              */
/*              ipr                                                           */
/*                  Interrupts asserted for processing                        */
/*                                                                            */
/* RETURNS      SATA_HOTPLUG_EVENT if hotplug event                           */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
static STATUS_CODE_T NonDataIsr(INT devTag_i, UINT32 ipr)
{
   DEV_INFO_T     * devp      = NULL_POINTER;
   INT portTag                = INVALID_TAG;
//   SATA_COMMAND_T * currCmd   = NULL_POINTER;
//   STATUS_CODE_T retval       = SATA_SUCCESS;
   ADAPTER_INFO_T *adapterp = NULL_POINTER;
   int ctr, ctlrTag, adapterTag;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   ctlrTag    = ports_gp[portTag]->controller;
   adapterTag = controllers_gp[ctlrTag]->adapter;
   adapterp   = adapters_gp[adapterTag];

   /* Looks like the ISR routine was called, but there is no submitted command */
   /* Check to verify that THIS device actually generated the interrupt */

   // PHY change, device hot-plug
   if ( READ_BIT(ipr, SUPDIPR_STATE_INTR_BIT) == 1 )
   {
      // Check if the device status changed...
      if ( devp->present && !PortDevConnected(portTag) )
      {   // Hot Unplug Event

         DevShutdown(devTag_i);

         devp->present = FALSE;

         // Make the drive go away from the drives array under the Adapter...
         for ( ctr = 0; ctr < MAX_DEVICES_PER_ADAPTER; ctr++ )
         {
            if ( adapterp->drives[ctr] == devTag_i )
            {
               adapterp->drives[ctr] = INVALID_TAG;
               break;
            }
         }

      }
      else
      {  // Hotplug Event

         // Hmm.... device is present, but we haven't configured it yet...
         // OK... let's do it now!
         AdapterServiceHotplug(adapterTag, ports_gp[portTag]->portNumber * MAX_DEVICES_PER_PORT);
         devp->present = TRUE;
      }

      return SATA_HOTPLUG_EVENT;
   }
   else
   {
      UINT8 asr, er;
      DEBUG_ERROR_0("Non Data INT on dev %d. Error ?", devTag_i);


      // Check for errors
      PortReadAltStatusReg(portTag, &asr);
      if ( READ_BIT(asr, SUPDASR_ERR_BIT) ) // if the error bit is set
      {
         PortReadErrorReg(portTag, &er);
         DEBUG_ERROR("ATA Protocol Error (ASR 0x%02X, ER 0x%02X). Aborting command.", asr, er);

         // OK there was an error. End that command (only if NCQ is off)
         if ( devp->devFeatures->TCQ != TRUE )
         {
            // Call the clean TCQ to free the single command slot
            DevCleanTCQ(devTag_i);

            // Clear the errors.
            DevClearError(devTag_i);
         }
      }
   }

   return SATA_FAILURE;
} /* End of NonDataIsr() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function verbose dumps the contents of Alt Status reg,   */
/*              Err Reg, Super Set Status reg, Super set Err reg to the       */
/*              provided buffer.                                              */
/*                                                                            */
/*                                                                            */
/* PARAMETERS   devTag_i                                                      */
/*                  Tag for the device to check.                              */
/*              limit_i                                                       */
/*                  size of buffer_op                                         */
/*              buffer_ip                                                     */
/*                  buffer to hold the dump                                   */
/*                                                                            */
/* RETURNS      Number of bytes written to the buffer                         */
/*                                                                            */
/******************************************************************************/
INT DumpErrStat(int devTag_i, INT limit_i, void * buffer_op)
{
   DEV_INFO_T     * devp    = NULL_POINTER;
   PORT_INFO_T    * portp   = NULL_POINTER;
   INT              portTag = INVALID_TAG;
   INT controller;
   SATA_REG_ADDR_T base;
   char          * out = (char *) buffer_op;
   INT             len = 0;
   UINT8           data8;
   UINT32          data32;
   UINT32 lastSR, lastASR, lastER, lastSSSR, lastSSER;
   SATA_COMMAND_T *sataCmd_p = NULL_POINTER;
   int ctr, acmd_ctr;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK

   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i); return SATA_NOT_INITIALIZED;
   }
   if ( buffer_op == NULL_POINTER )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
#endif

   /* If less than a line's length is left, truncate dump and return */
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   devp = devices_gp[devTag_i];
   portTag = devp->port;
   portp = ports_gp[portTag];

   controller = ports_gp[portTag]->controller;
   base = controllers_gp[controller]->commonRegBase;

   lastSR = portp->lastSR;
   lastASR = portp->lastASR;
   lastER = portp->lastER;
   lastSSSR = portp->lastSSSR;
   lastSSER = portp->lastSSER;

   len += sprintf(out, "A %d, Port %d\n", controllers_gp[controller]->adapter, portp->portNumber);
   out = (char *) buffer_op + len;
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   len += sprintf(out, "Last Status : 0x%02X\n",
                  portp->lastSR);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   PortReadAltStatusReg(portTag, &data8);
   out = (char *) buffer_op + len;
   len += sprintf(out, "AltStatus : 0x%02X        last: 0x%02X\n", data8,
                  lastASR);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   PortReadErrorReg(portTag, &data8);
   out = (char *) buffer_op + len;
   len += sprintf(out, "Error Reg : 0x%02X        last: 0x%02X\n", data8,
                  lastER);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   PortReadSStatusReg(portTag, &data32);
   out = (char *) buffer_op + len;
   len += sprintf(out, "SStatus   : 0x%08X        last: 0x%08X\n", data32,
                  lastSSSR);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   PortReadSErrorReg(portTag, &data32);
   out = (char *) buffer_op + len;
   len += sprintf(out, "SError    : 0x%08X        last: 0x%08X\n", data32,
                  lastSSER);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   PortReadSActive(portTag, &data32);
   out = (char *) buffer_op + len;
   len += sprintf(out, "SActive   : 0x%08X        last: 0x%08X\n", data32, devp->sactiveCopy);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   PortReadDmaStatusReg (portTag, &data8);
   out = (char *) buffer_op + len;
   len += sprintf(out, "DMAStatus : 0x%02X\n", data8);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   out = (char *) buffer_op + len;
   len += sprintf(out, "All SErrors seen:   0x%08X\n",
                  portp->cumulativeSSER);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   data32 = 0xffffffff;
   out = (char *) buffer_op + len;

   PciMemRead32(base + SUPDIMR, &data32);
   len += sprintf(out, "Interrupts Mask:    0x%08X\n",data32);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   data32 = 0xffffffff;
   out = (char *) buffer_op + len;
   PciMemRead32(base + SUPDIPR, &data32);
   len += sprintf(out, "Interrupts Pending: 0x%08X\n",data32);
   if ( limit_i - len < 80 )  goto end_of_buffer_condn;

   data32 = 0xffffffff;

   PciMemRead32( ports_gp[portTag]->portRegBase + SUPDSSSR, &data32);

   out = (char *) buffer_op + len;
/* 0 = no device, 1 = Present but NO communication, 3 = Ready, 4 = offline */
   len += sprintf(out, "Device (link) is %s%s%s%s%s%s\n",
                  ((data32 & SUPDSSSR_DEV_DET_MASK)==0) ?
                  "(0)Absent":"",
                  ((data32 & SUPDSSSR_DEV_DET_MASK)==1) ?
                  "(1)Present but NO communication":"",
                  ((data32 & SUPDSSSR_DEV_DET_MASK)==2) ?
                  "in (2)Unknown State":"",
                  ((data32 & SUPDSSSR_DEV_DET_MASK)==3) ?
                  "(3)Ready":"",
                  ((data32 & SUPDSSSR_DEV_DET_MASK)==4) ?
                  "(4)offline":"",
                  ((data32 & SUPDSSSR_DEV_DET_MASK) >4) ?
                  "in Unknown State":"");

   // Let's count and print the number of active commands
   acmd_ctr = 0;
   for ( ctr = 0; ctr <MAX_TCQ_DEPTH; ctr++ )
   {
      sataCmd_p = &devp->commandQueue[ctr];
      if ( !sataCmd_p ) break;                               // Done with all commands.
      if ( sataCmd_p->slotState == SLOT_ACTIVE ) acmd_ctr++; // CMD not free... increment counter...
   }

   if ( limit_i - len < 80 )  goto end_of_buffer_condn;
   out = (char *) buffer_op + len;
   len += sprintf(out, "NCQ %s - %d active\n", ((devp->devFeatures->TCQ == TRUE) ? "Enabled":"Disabled"), acmd_ctr);

   for ( ctr = 0; ctr <MAX_TCQ_DEPTH; ctr++ )
   {

      sataCmd_p = &devp->commandQueue[ctr];
      if ( !sataCmd_p ) break;                           // Done with all commands.
      if ( sataCmd_p->slotState == SLOT_FREE ) continue; // Free slot, goto next slot.
      if ( limit_i - len < 80 )  goto end_of_buffer_condn;
      out = (char *) buffer_op + len;
      len += sprintf(out, "slot %d %s%s", ctr,
                     (( sataCmd_p->slotState == SLOT_OCCUPIED)? "Pnd": ""),
                     (( sataCmd_p->slotState == SLOT_ACTIVE)? "Act": ""));

      if ( limit_i - len < 80 )  goto end_of_buffer_condn;
      out = (char *) buffer_op + len;
      len += sprintf(out, " f %#04x,sC %#04x,sN %#04x,cL %#04x,cH %#04x,dH %#02x,cmd %#02x\n",
                     sataCmd_p->taskfile.feature,
                     sataCmd_p->taskfile.sectorCount,
                     sataCmd_p->taskfile.sectorNumber,
                     sataCmd_p->taskfile.cylinderLow,
                     sataCmd_p->taskfile.cylinderHigh,
                     sataCmd_p->taskfile.deviceHead,
                     sataCmd_p->taskfile.command);
   }


   goto return_ok;

   end_of_buffer_condn:
   DEBUG_ERROR_0("Buffer size %d is not enough for error dump.", limit_i);
// We nearly ran out of buffer space... if possible indicate that in the output.
   if ( limit_i - len > 3 )
   {
      out = (char *) buffer_op + len;
      len += sprintf(out,"@\n");
   }

   return_ok:
   return len;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the number of FREE slots in the TCQ     */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              buffer                                                        */
/*                  char Buffer with slot states. should be at least 33 bytes */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns -1 for all errors                                     */
/*                                                                            */
/******************************************************************************/
INT DevTCQCount(INT devTag_i, char * astr, INT len)
{
   DEV_INFO_T *     devp     = NULL_POINTER;
   INT portTag                = INVALID_TAG;
   INT ctr = 0;
   FLAGS_T flags = 0; /* spinlock flags */
   int free_cmds = 0;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return -1 ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return -1;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return -1;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   OsSpinLock(&devp->pendingCmdLock, flags);
   for ( ctr = 0; ctr < MAX_GD31244_QUEUE_DEPTH; ctr++ )
   {
      if ( devp->commandQueue[ctr].slotState == SLOT_FREE )
      {
         free_cmds++;
      }

      if ( astr && (len > 0) )
      {
         len -= sprintf(astr + ctr, "%c",(devp->commandQueue[ctr].slotState == SLOT_FREE ? 'F':
                                          (devp->commandQueue[ctr].slotState == SLOT_OCCUPIED ? 'P':
                                           (devp->commandQueue[ctr].slotState == SLOT_ACTIVE ? 'A': '?')))
                       );
      }
   }

   OsSpinUnlock(&devp->pendingCmdLock, flags);

   return free_cmds;
} /* End of DevTCQCount() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function sets the pointer to the device features passed  */
/*              in from the OSDM to a per device pointer. The features will   */
/*              be used by the library to determine certain behaviors of the  */
/*              commands depending on what is set. This MUST be called after  */
/*              the OS sets the features for the drive. If features are not   */
/*              changed on the drive this still needs to be called in order   */
/*              for the library to know what features are currently set.      */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*              The device features table must be filled out and valid!       */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*              features_p                                                    */
/*                  Pointer to the device features table                      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevSetFeatures(INT devTag_i, PSATA_FEATURES_T features_p)
{
   DEV_INFO_T     *devp = NULL_POINTER;
   INT portTag;

   DEBUG_TRACE_1();


#ifdef SANITY_CHECK

   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];

   portTag = devp->port;

   if ( features_p )
   {
      devp->devFeatures = features_p;

      if ( devp->devFeatures->Lba48 )
      {

         if ( devp->devFeatures->TCQ )
         {
            devp->cmd.ataReadDma = SATA_READ_FPDMA_QUEUED;
            devp->cmd.ataWriteDma = SATA_WRITE_FPDMA_QUEUED;
            devp->cmd.buildTaskFile = DevBuildTaskFileTCQ;

            // Need to set auto service TCQ DMA
            PortEnableNatvieQDma(portTag);
         }
         else
         {
            devp->cmd.ataReadDma = ATA_READ_DMA_EXT;
            devp->cmd.ataWriteDma = ATA_WRITE_DMA_EXT;
            devp->cmd.buildTaskFile = DevBuildTaskFileLba48;
         }
         devp->cmd.ataReadMultiple = ATA_READ_MULTIPLE_EXT;
         devp->cmd.ataWriteMultiple = ATA_WRITE_MULTIPLE_EXT;
         devp->cmd.ataReadSectors = ATA_READ_SECTORS_EXT;
         devp->cmd.ataWriteSectors = ATA_WRITE_SECTORS_EXT;
         devp->cmd.ataReadLog = ATA_READ_LOG_EXT;
      }
   }
   else
   {
      { DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i); return SATA_NOT_INITIALIZED;}
   }

   return SATA_SUCCESS;
} /* End of DevSetFeatures() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function unsets the pointer to the device features       */
/*              This function should be called if device is removed when hot  */
/*              plug is being used.                                           */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevUnsetFeatures(INT devTag_i)
{
   DEV_INFO_T     *devp = NULL_POINTER;
   INT portTag;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK

   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif
   devp = devices_gp[devTag_i];
   portTag = devp->port;
   devp->devFeatures = NULL_POINTER;

   /* initializing ATA commands to defaults */
   devp->cmd.ataReadDma = ATA_READ_DMA;
   devp->cmd.ataWriteDma = ATA_WRITE_DMA;
   devp->cmd.ataReadMultiple = ATA_READ_MULTIPLE;
   devp->cmd.ataWriteMultiple = ATA_WRITE_MULTIPLE;
   devp->cmd.ataReadSectors = ATA_READ_SECTORS;
   devp->cmd.ataWriteSectors = ATA_WRITE_SECTORS;
   devp->cmd.ataReadLog = ATA_READ_LOG_EXT;
   devp->cmd.buildTaskFile = DevBuildTaskFileLba28;

   PortDisableNatvieQDma(portTag);

   return SATA_SUCCESS;
} /* End of DevUnsetFeatures() */


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function cleans up the command queue when the OS wants   */
/*              to turn off the device for hot-unplug.                        */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevShutdown(INT devTag_i)
{
   DEV_INFO_T *  devp   = NULL_POINTER;
   STATUS_CODE_T retval = SATA_SUCCESS;
   INT portTag                = INVALID_TAG;
   SATA_COMMAND_T * currCmd   = NULL_POINTER;
   int ctr = 0;

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


   DEBUG_TRACE_1();

#ifdef SANITY_CHECK

   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif
   devp    = devices_gp[devTag_i];
   portTag = devp->port;

   // Stop DMA, end commands
   retval = PortWriteDmaCommandReg(portTag, DMA_STOP);

#ifdef PARANOID_LOCKING
   OsSpinLock(&devp->portLock, portFlags);
#endif

   for ( ctr = 0; ctr < MAX_TCQ_DEPTH; ctr++ )
   {

      currCmd = &devp->commandQueue[ctr];
      if ( currCmd->slotState != SLOT_FREE )
      {
         /* End the command */
         DevEndCommand(devTag_i, currCmd);
      }
   }

   devp->active = FALSE;

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif

   return SATA_SUCCESS;
} /* End of DevShutdown() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function Returns the OsdmTag registered during           */
/*              initialization.                                               */
/*                                                                            */
/* PRECONDITIONS  The Adapter has been initialized and ready for use.         */
/*                                                                            */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      OsdmTag for the device; NULL_POINTER on Errors                */
/*                                                                            */
/******************************************************************************/
INLINE void * DevReturnOsdmTag (INT devTag_i)
{
#ifdef SANITY_CHECK


   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return NULL_POINTER ;
   }

   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return NULL_POINTER;
   }
#endif

   return devices_gp[devTag_i]->osDriveTag;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function cleans up the command queue when the OS wants   */
/*              to do a reset.                                                */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  Tag for the device to be used.                            */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = device not initialized              */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevCleanTCQ(INT devTag_i)
{
   DEV_INFO_T *     devp     = NULL_POINTER;
   SATA_COMMAND_T * currCmd   = NULL_POINTER;
   INT portTag                = INVALID_TAG;
   INT ctr = 0;


#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif

   FLAGS_T flags = 0; /* spinlock flags redefined in linux 2.6 */

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   OsSpinLock(&devp->pendingCmdLock, flags);
   currCmd = devp->pendingCmd;
   while ( currCmd )
   {
      devp->pendingCmd = devp->pendingCmd->nextCmd;
      currCmd->nextCmd = NULL_POINTER;
      currCmd->slotState = SLOT_FREE;
      currCmd = devp->pendingCmd;
   }
   devp->pendingCmdTail = NULL_POINTER;
   OsSpinUnlock(&devp->pendingCmdLock, flags);
   DEBUG_PRINT_1("Dev %d: Library pending queue cleared", devTag_i);

   for ( ctr = 0; ctr < MAX_GD31244_QUEUE_DEPTH; ctr++ )
   {
      currCmd = &devp->commandQueue[ctr];

      if ( currCmd->slotState != SLOT_FREE )
      {
         /* End the command */
         DevEndCommand(devTag_i, currCmd);
      }
   }

   // Reset the sactiveCopy ... we no longer have any active commands.
   devp->sactiveCopy = 0;

#ifdef PARANOID_LOCKING
   OsSpinLock(&devp->portLock, portFlags);
#endif
   // Stop the DMA if it was active...
   PortWriteDmaCommandReg(portTag, DMA_STOP);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif

   devp->busyCounter =  MAX_INTERRUPT_BUSY_WAIT;

   DEBUG_PRINT_1("Dev %d: All active commands cleared", devTag_i);

   return SATA_SUCCESS;
} /* End of DevCleanTCQ() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function builds the LBA28 version of ATA data command.   */
/*                                                                            */
/* PRECONDITIONS   Only to be called by DevAddCommandToQueue()                */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              sataCmd_p                                                     */
/*                  The command to be built                                   */
/*              startSector                                                   */
/*                  Starting sector for data command                          */
/*              numSectors                                                    */
/*                  Number of sectors to be accessed                          */
/*                                                                            */
/* RETURNS      N/A                                                           */
/*                                                                            */
/* ON ERRORS                                                                  */
/*                                                                            */
/******************************************************************************/
static void DevBuildTaskFileLba28(SATA_COMMAND_T *sataCmd_p,
                                  INT startSector, INT numSectors)
{
   DEBUG_TRACE_1();

   sataCmd_p->taskfile.feature = 0;
   /* Keep track of the number of sectors yet to be transfered separately. */
   /* Makes it easier to know when the Tx is over in the case when         */
   sataCmd_p->sectRemaining = numSectors;

   /* max 255, 0 == 256 */
   sataCmd_p->taskfile.sectorCount = numSectors & 0xff;

   /* compute the LBA values */
   /* All devices supporting ATA/ATAPI-5 and above support LBA */
   sataCmd_p->taskfile.sectorNumber = startSector & 0xFF; /* Lower 8 bits */
   startSector >>= 8;
   sataCmd_p->taskfile.cylinderLow  = startSector & 0xFF; /* Next 8 bits */
   startSector >>= 8;
   sataCmd_p->taskfile.cylinderHigh = startSector & 0xFF; /* Next 8 bits */
   startSector >>= 8;
   sataCmd_p->taskfile.deviceHead   = startSector & 0xF;  /* Last 4 bits */

   /* Mark the startSector address as LBA address, Dev as 0 */
   sataCmd_p->taskfile.deviceHead =
   (UINT8)SET_BIT(sataCmd_p->taskfile.deviceHead, DEVICE_HEAD_LBA_BIT);

//    DEBUG_PRINT_1("taskfile.deviceHead = %0x, Verify that Bit %d is SET.",
//                  sataCmd_p->taskfile.deviceHead, DEVICE_HEAD_LBA_BIT);
} /* End of DevBuildTaskFileLba28() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function builds the LBA48 version of ATA data command.   */
/*                                                                            */
/* PRECONDITIONS   Only to be called by DevAddCommandToQueue()                */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              sataCmd_p                                                     */
/*                  The command to be built                                   */
/*              startSector                                                   */
/*                  Starting sector for data command                          */
/*              numSectors                                                    */
/*                  Number of sectors to be accessed                          */
/*                                                                            */
/* RETURNS      N/A                                                           */
/*                                                                            */
/* ON ERRORS                                                                  */
/*                                                                            */
/******************************************************************************/
static void DevBuildTaskFileLba48(SATA_COMMAND_T *sataCmd_p,
                                  INT startSector, INT numSectors)
{
   DEBUG_TRACE_1();

   /* Fill out the taskfile */

   sataCmd_p->taskfile.feature = 0;
   /* Keep track of the number of sectors yet to be transfered separately. */
   /* Makes it easier to know when the Tx is over in the case when         */
   sataCmd_p->sectRemaining = numSectors;

   /* if we are doing LBA 48 */
   sataCmd_p->taskfile.sectorCount = numSectors & 0xffff;

   /* bits 31:24 7:0  */
   sataCmd_p->taskfile.sectorNumber =
   (startSector & 0xff) | ((startSector >> 16) & 0xff00);
   startSector >>= 8;
   sataCmd_p->taskfile.cylinderLow  = startSector & 0xff;
   startSector >>= 8;
   sataCmd_p->taskfile.cylinderHigh = startSector & 0xff;
   /* always master */
   /* deviceHead bits are reserved for Lba48 */
   //sataCmd_p->taskfile.deviceHead = 0xf;
   sataCmd_p->taskfile.deviceHead = 0;

   /* Mark the startSector address as LBA address, Dev as 0 */
   sataCmd_p->taskfile.deviceHead =
   (UINT8)SET_BIT(sataCmd_p->taskfile.deviceHead, DEVICE_HEAD_LBA_BIT);

//    DEBUG_PRINT_1("taskfile.deviceHead = %0x, Verify that Bit %d is SET.",
//                  sataCmd_p->taskfile.deviceHead, DEVICE_HEAD_LBA_BIT);

} /* End of DevBuildTaskFileLba48() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function builds the native command queuing version of    */
/*              SATA data command. This is new from SATA II                   */
/*                                                                            */
/* PRECONDITIONS   Only to be called by DevAddCommandToQueue()                */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              sataCmd_p                                                     */
/*                  The command to be built                                   */
/*              startSector                                                   */
/*                  Starting sector for data command                          */
/*              numSectors                                                    */
/*                  Number of sectors to be accessed                          */
/*                                                                            */
/* RETURNS      N/A                                                           */
/*                                                                            */
/* ON ERRORS                                                                  */
/*                                                                            */
/******************************************************************************/
static void DevBuildTaskFileTCQ(SATA_COMMAND_T *sataCmd_p,
                                INT startSector, INT numSectors)
{
   DEBUG_TRACE_1();

   /* Fill out the taskfile */
   sataCmd_p->taskfile.feature = numSectors;
   /* Keep track of the number of sectors yet to be transfered separately. */
   /* Makes it easier to know when the Tx is over in the case when         */
   sataCmd_p->sectRemaining = numSectors;

   /* if we are doing LBA 48 */
   sataCmd_p->taskfile.sectorCount = sataCmd_p->slotTag << 3; /* TCQ Tag */

   /* bits 31:24 7:0  */
   sataCmd_p->taskfile.sectorNumber =
   (startSector & 0xff) | ((startSector >> 16) & 0xff00);
   startSector >>= 8;
   sataCmd_p->taskfile.cylinderLow  = startSector & 0xff;
   startSector >>= 8;
   sataCmd_p->taskfile.cylinderHigh = startSector & 0xff;
   /* always master */
   sataCmd_p->taskfile.deviceHead = 0;

   /* Mark the startSector address as LBA address, Dev as 0 */
   sataCmd_p->taskfile.deviceHead =
   (UINT8)SET_BIT(sataCmd_p->taskfile.deviceHead, DEVICE_HEAD_LBA_BIT);

#ifdef SEAGATE_NCQ_WORKAROUND
   DEBUG_PRINT("spurious INT for NCQ workaround%s","");
   sataCmd_p->taskfile.sectorCount |= 0x1;
#endif
//    DEBUG_PRINT_1("taskfile.deviceHead = %0x, Verify that Bit %d is SET.",
//                  sataCmd_p->taskfile.deviceHead, DEVICE_HEAD_LBA_BIT);
} /* End of DevBuildTaskFileTCQ() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function attempts to acquire a free command slot for     */
/*              SATA native command queuing.                                  */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  The device tag                                            */
/*                                                                            */
/* RETURNS      pointer to free command                                       */
/*                                                                            */
/* ON ERRORS    NULL                                                          */
/*                                                                            */
/******************************************************************************/
PSATA_COMMAND_T DevGetQSlot(INT devTag_i, UINT32 slot_i)
{
   PSATA_COMMAND_T sataCmd_p = NULL_POINTER;
   DEV_INFO_T   *devp      = NULL_POINTER;
   INT portTag;
   STATUS_CODE_T retval = SATA_SUCCESS;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK_IN_DATA_PATH
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return NULL_POINTER;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return NULL_POINTER;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return NULL_POINTER;
   }

   if ( !TAG_WITHIN_RANGE(slot_i, MAX_TCQ_DEPTH) )
   {
      DEBUG_ERROR("Invalid slot_i: %d", devTag_i);
      return NULL_POINTER;
   }

#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   // acquire command
   sataCmd_p = &devp->commandQueue[slot_i];

   ASSERT(sataCmd_p);

#ifdef QDEBUG
   {
      u32 sactive = 0;

      PortReadSActive(portTag, &sactive);

      if ( READ_BIT(sactive,slot_i) )
      {
         DEBUG_ERROR("ERROR: Slot %d is %s But SActive is %#010x",slot_i,
                     ((sataCmd_p->slotState == SLOT_OCCUPIED)? "OCCUPIED" :
                      ((sataCmd_p->slotState == SLOT_FREE)? "FREE" :
                       ((sataCmd_p->slotState == SLOT_ACTIVE)? "ACTIVE" : "?"))), sactive);
      }
   }
#endif

   if ( sataCmd_p->slotState != SLOT_FREE )
   {
      /* slot busy for some reason, shouldn't be.... */
      DEBUG_ERROR("ERROR: Slot %d is NOT Free. It's %s",slot_i,
                  ((sataCmd_p->slotState == SLOT_OCCUPIED)? "OCCUPIED" :
                   ((sataCmd_p->slotState == SLOT_ACTIVE)? "ACTIVE" : "?")));

#ifdef QDEBUG
      {
         int ctr;
         char tstr[80] = "";

         for ( ctr = 0; ctr < MAX_TCQ_DEPTH; ctr++ )
         {
//               DEBUG_PRINT_1("Slot %d has tag %d", ctr,
//                             devp->commandQueue[ctr].slotTag);
            sprintf(tstr + ctr, "%c",
                    (( devp->commandQueue[ctr].slotState == SLOT_FREE)? 'F' :
                     ((devp->commandQueue[ctr].slotState == SLOT_OCCUPIED)? 'O' :
                      ((devp->commandQueue[ctr].slotState == SLOT_ACTIVE)? 'A' : '?'))));
         }

         DEBUG_ERROR_0("[%s]. slot #%d, slotTag %d, state %d, devTag %d.",tstr,
                       slot_i, devp->commandQueue[slot_i].slotTag,
                       devp->commandQueue[slot_i].slotState, devTag_i);
      }
#endif
      return NULL_POINTER;
   }

   //   /* Select the device, and then check the port */
   //   retval = SelectDevice(devTag_i);
   //   /* Return false for errors */
   //   if(retval != SATA_SUCCESS)
   //   {
   //       return NULL_POINTER;
   //   }

   if ( sataCmd_p && devp->devFeatures->TCQ )
   {
      // If we found a slot then reserve the slot to be active
      retval = PortSetSActive(portTag, sataCmd_p->slotTag);
      if ( retval != SATA_SUCCESS )
      {
         return NULL_POINTER;
      }

// Need to set it here... will be set in DevProcessDmaRequest
//        OsSpinLock(&devp->sactiveLock, lockFlag);
//        devp->sactiveCopy |= SET_BIT(0, sataCmd_p->slotTag);
//        OsSpinUnlock(&devp->sactiveLock, lockFlag);
   }

   return sataCmd_p;
} /* end of DevGetFreeQSlot() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function executes the READ_LOG_EXT SATA cmnd (Page 10h)  */
/*              This command is to read errors for a failed command during    */
/*              SATA native command queuing execution.                        */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*              The PHY is still in good condition. No SErrors.               */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*              All queued command on device is aborted. Device change to     */
/*              accept new commands state.                                    */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  The device tag                                            */
/*              buffer_op                                                     */
/*                  Log page that contain error information from device       */
/*                                                                            */
/* RETURNS      SATA_SUCCESS good status                                      */
/*                                                                            */
/* ON ERRORS    SATA error status                                             */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T DevReadLogExt(INT devTag_i, void *buffer_op)
{
   DEV_INFO_T *  devp   = NULL_POINTER;
   STATUS_CODE_T retval = SATA_SUCCESS;
   INT portTag          = INVALID_TAG;
   SATA_TASKFILE_T taskfile;
   INT ctr = 0;
   UINT32 serr = 0;

#ifdef PARANOID_LOCKING
   FLAGS_T portFlags = 0;
#endif


#ifdef SATA_DEBUG_0
   UINT32 sactive;
   UINT32 last_sactive;
#endif

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   if ( DevIsReady(devTag_i) == FALSE )
   {
      DEBUG_ERROR_0("Drive not ready. Returning Busy (%d)", SATA_BUSY);
      return SATA_BUSY;
   }

   devp = devices_gp[devTag_i];
   portTag = devp->port;

#ifdef SATA_DEBUG_0
   PortReadSActive(portTag, &last_sactive);
#endif

   retval = PortReadSErrorReg(portTag, &serr);
#ifdef SANITY_CHECK
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR_0("Failed (%d) while reading SError", retval);
      return retval;
   }
#endif

   // if there are serr then PHY not stable and we don't want to continue
   if ( serr )
   {
      DEBUG_ERROR_0("SError indicates SATA Error - %08x. Returning Failure.", serr);
      return SATA_FAILURE;
   }

   /* Disable Interrupts */
   retval = PortDisableInterrupts(portTag);
   /* Check for system errors */
   if ( retval != SATA_SUCCESS )
   {
      return retval;
   }

   /* Read the Page 10h of the Log file. */

   memset(&taskfile, 0, sizeof(SATA_TASKFILE_T));

   taskfile.command = ATA_READ_LOG_EXT;
   taskfile.sectorNumber = 0x10;  // Page 10h
   taskfile.sectorCount = 0x01;   // Only one Page

#ifdef PARANOID_LOCKING
   // Lock for the entire duration of issuing the command.
   OsSpinLock(&devp->portLock, portFlags);
#endif

   /* Write out the taskfile */
   PortWriteTaskfileParams(portTag,&taskfile);

   /* Write the ATA_READ_LOG_EXT command */
   retval = PortWriteCommandReg(portTag, ATA_READ_LOG_EXT);
#ifdef SANITY_CHECK
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR_0("Failed (%d) while writing Command Reg", retval);

   #ifdef PARANOID_LOCKING
      OsSpinUnlock(&devp->portLock, portFlags);
   #endif
      return retval;
   }
#endif

   MicrosecondSleep(1); /* Give drive 400 nsec to set the busy flag */

   /* Wait for some time Max = WAIT_CMD, if BUSY bit set */
   /* Make sure the drive is ready... */
   /* is the drive ready for I/O ? (not Busy?) */
   while ( (DevIsReady(devTag_i) != TRUE) && (ctr < WAIT_CMD) )
   {
      ctr++;
      /* Wait for 1 Millisecond more for the drive to be ready */
      MillisecondSleep(1);
   }

   if ( ctr >= WAIT_CMD )
   {
      /* We Timed out ... */
      DEBUG_ERROR_0("Timed out waiting for Device to be ready (%d ms)", ctr);
#ifdef PARANOID_LOCKING
      OsSpinUnlock(&devp->portLock, portFlags);
#endif
      return SATA_BUSY;
   }

   /* Check for DRQ, (wait if required)*/
   if ( DevDRQSet(devTag_i) == FALSE )
   {
      /* wait 50 ms */
      MillisecondSleep(50);
      if ( DevDRQSet(devTag_i) == FALSE )
      {
         DEBUG_ERROR_0("Timed out - Dev Ready after %d ms, But no DRQ", ctr);
#ifdef PARANOID_LOCKING
         OsSpinUnlock(&devp->portLock, portFlags);
#endif
         return SATA_BUSY; /* We timed out... */
      }
   }

   DEBUG_PRINT_1("devTag %d, About to read the data port...", devTag_i);
   /* Read PIO data, 512 bytes */
   retval = PortPioReadBytes(portTag, SATA_SECTOR_SIZE , buffer_op);

#ifdef PARANOID_LOCKING
   OsSpinUnlock(&devp->portLock, portFlags);
#endif

   DEBUG_PRINT_1("devTag %d, OK... Done", devTag_i);

   /* Check for system errors */
   if ( retval != SATA_SUCCESS )
   {
      DEBUG_ERROR_0("Failed (%d) to read the data port", retval);
      return retval;
   }

#ifdef SATA_DEBUG_1
   PortReadSActive(portTag, &sactive);
   DEBUG_PRINT_1("Sactive Before %#08x, After %#08x", last_sactive, sactive);
#endif

   /* Clear the Interrupt */
   PortClearInterrupt (portTag);

   /* Enable interrupts before we return... */
   DevEnableInterrupts(devTag_i);

   return SATA_SUCCESS;
} /* end of DevGetLogExt() */

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function sets the vector table address by calling        */
/*              the port function PortSetDDTVectorTable()                     */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS                                                                 */
/*              devTag_i                                                      */
/*                  The device tag                                            */
/*              vector                                                        */
/*                  The vector table address (64bit)                          */
/*                                                                            */
/* RETURNS      SATA_SUCCESS good status                                      */
/*                                                                            */
/* ON ERRORS    SATA error status                                             */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T DevSetDDTVectorTable(INT devTag_i, UINT64 vector)
{
   DEV_INFO_T *devp = NULL_POINTER;
   INT portTag                = INVALID_TAG;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check - should not be required here... */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE ;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
   if ( devices_gp[devTag_i]->active == FALSE )
   {
      DEBUG_ERROR("DevTag %d, Not Active", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   if ( DevIsReady(devTag_i) == FALSE )
   {
      DEBUG_ERROR_0("Drive not ready. Returning Busy (%d)", SATA_BUSY);
      return SATA_BUSY;
   }

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   return PortSetDDTVectorTable(portTag, vector);
} /* end of DevSetDDTVectorTable() */

INLINE STATUS_CODE_T DevClearInterrupts(INT devTag_i)
{
   DEV_INFO_T     * devp    = NULL_POINTER;
   INT              portTag = INVALID_TAG;
   STATUS_CODE_T   retval   = SATA_SUCCESS;

   DEBUG_TRACE_1();

#ifdef SANITY_CHECK
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];
   portTag = devp->port;

   PortClearInterrupt(portTag);
   //PortClearError(portTag);
   return retval;
} /* end of DevClearInterrupts() */

STATUS_CODE_T DevGetActiveCommands(INT devTag_i, UINT32 *active_commands)
{
   DEV_INFO_T     *devp     = NULL_POINTER;
   STATUS_CODE_T   retval   = SATA_SUCCESS;

#ifdef SANITY_CHECK
   /* Sanity check */
   if ( !TAG_WITHIN_RANGE(devTag_i, MAX_DEVICES) )
   {
      DEBUG_ERROR("Invalid devTag_i: %d", devTag_i);
      return SATA_OUT_OF_RANGE_VALUE;
   }
   if ( devices_gp[devTag_i] == NULL_POINTER )
   {
      DEBUG_ERROR("DevTag %d, Not Initialized", devTag_i);
      return SATA_NOT_INITIALIZED;
   }
#endif

   devp = devices_gp[devTag_i];
   *active_commands = devp->sactiveCopy;
   DEBUG_ERROR("DevGetActiveCommands Tag %d, sactive=%x pendingCmd=%p", devTag_i,devp->sactiveCopy, devp->pendingCmd);

   return(retval);
}
