/*******************************************************************************
  Copyright (c) 2003-2005, Intel Corporation
  All rights reserved.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License version 2 as published
  by the Free Software Foundation. Alternatively, this software may be
  distributed under the terms of the BSD license. Please see the accompanied
  file COPYING for copyright licensing terms.
*******************************************************************************/

/******************************************************************************/
/*                                                                            */
/*                  I N C L U D E   S T A T E M E N T S                       */
/*                                                                            */
/******************************************************************************/

#include <OSL.h>
#include <sata_port.h>
#include <sata_globals.h>
#include <sata_device.h>
#include <artisea.h>

#ifndef WINNT
    #include <sata_debug.h>
#endif

/******************************************************************************/
/*                                                                            */
/*       G L O B A L   F U N C T I O N   I M P L E M E N T A T I O N S        */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function allocates the private memory for the port. It   */
/*              allocates a PORT_INFO_T structure and records the pointer in  */
/*              the next available slot in the ports_gp array. This function  */
/*              also initializes the allocated memory structure and sets the  */
/*              controller member to the passed controllerTag.                */
/*                                                                            */
/* PRECONDITIONS    ports_gp array has been allocated.                        */
/*                                                                            */
/* POST-CONDITIONS  The next NULL pointer in ports_gp is set.                 */
/*                  *portTag_op is set to it's index number or INVALID_TAG    */
/*                                                                            */
/* PARAMETERS   controllerTag_i                                               */
/*                  The Tag of the parent controller.                         */
/*              portTag_op                                                    */
/*                  Address of the INT variable to hold the portTag.          */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in ports_gp array          */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*                                                                            */
/******************************************************************************/

STATUS_CODE_T PortAllocatePrivateMemory(INT controllerTag_i, INT *portTag_op)
{
    INT tag;
    PORT_INFO_T *newPortInfo;

    DEBUG_TRACE_3();

    *portTag_op = INVALID_TAG ; /* Will be assigned a valid value below */

    tag = getNextSlot((void**)ports_gp, MAX_PORTS); /* NULL_POINTER safe */

    /* Check if we got an empty slot in the array */
    if(tag == INVALID_TAG)
    {
        DEBUG_PRINT_3("ERROR: ports_gp FULL. Returning %d", SATA_OUT_OF_RANGE_VALUE);
        return SATA_OUT_OF_RANGE_VALUE;
    }

    /* Allocate the Memory */
    newPortInfo = GetMemory(sizeof(PORT_INFO_T));

    if(newPortInfo == NULL_POINTER)
    {
        DEBUG_PRINT_3("Error: Failed to get Memory. Returning %d",SATA_GET_MEMORY_FAILED);
        /* Failed to get memory */
        return SATA_GET_MEMORY_FAILED;
    }

    /* Record the pointer in the global array */
    ports_gp[tag] = (PORT_INFO_T *)newPortInfo;
    *portTag_op = tag ;

    /* Initialize the structure */
    MemSet(sizeof(PORT_INFO_T), 0, newPortInfo);
    newPortInfo->present = FALSE; /* Mark as not yet completely initialized */
    newPortInfo->controller = controllerTag_i;

    return SATA_SUCCESS;

}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function initializes the port.                           */
/*                                                                            */
/* PRECONDITIONS    The private memory for the port has been allocated.       */
/*                                                                            */
/* POST-CONDITIONS  The port is initialized and ready for I/O                 */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to initialize.                           */
/*              portNumber_i                                                  */
/*                  The port number on the controller represented by this port*/
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = no space in ports_gp array          */
/*              SATA_GET_MEMORY_FAILED  = Failed to allocate memory           */
/*              SATA_NOT_INITIALIZED    = Memory was not allocated.           */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortInit(INT portTag_i, INT portNumber_i)
{
    INT controller, currDev, ctr;
    SATA_REG_ADDR_T base;
    STATUS_CODE_T retval = SATA_SUCCESS;
    boolean deviceInitFailed = FALSE;
    PORT_INFO_T * currPort = NULL_POINTER;

    DEBUG_TRACE_3();

    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if(ports_gp[portTag_i] == NULL_POINTER)
    {
        DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;
    }


    /* Check if the portNumber_i is valid */
    if(portNumber_i >= MAX_PORTS_PER_CONTROLLER)
    {
        DEBUG_PRINT_3("ERROR: Invalid portNumber_i %d",portNumber_i);
        return SATA_OUT_OF_RANGE_VALUE;
    }

    currPort = ports_gp[portTag_i];
    controller = currPort->controller;
    base = controllers_gp[controller]->commonRegBase;

    currPort->portNumber = portNumber_i;

    /* Calculate our starting offset. ASSUMES contiguous addresses.*/
    currPort->portRegBase = base + COMMON_PORT_REG_SIZE
                            + portNumber_i * SATA_PORT_REG_SIZE;

    /* Other Port initializations go here. */

    /* Init the error tracking information. */
    currPort->cumulativeSSER = 0;
    currPort->lastSSER       = 0;
    currPort->lastER         = 0;
    currPort->lastASR        = 0;
    currPort->lastSR         = 0;
    currPort->lastSSSR       = 0;

    /* initialize the devTag in the devices array */
    for(ctr = 0; ctr < MAX_DEVICES_PER_PORT; ctr++)
    {
        currPort->dev[ctr] = INVALID_TAG;
    }

    /* Initialization complete. Mark port as ready for use. */
    currPort->present = TRUE;

    /* Disable all the interrupts on the port. */
    /*The OSDM will enable these when it is ready */
    PortDisableInterrupts(portTag_i);

    /* reset the SATA interface and bring it online */

    if(PortIntHardReset(portTag_i) == SATA_SUCCESS)
    {
        currPort->connected = PortDevConnected(portTag_i);
    }
    else
    {
//        currPort->present = FALSE;
        currPort->connected = FALSE;
    }

    /* Clear any interrupts that maybe hanging around(i.e. hotplug) */
    PortClearInterrupt(portTag_i);

    /* Now initialize all the devices connected to this port. */
    if(currPort->connected)
    {

        DEBUG_PRINT_3("A Device is connected to Ctlr %d, Port %d",
                      controllers_gp[controller]->controllerNumber, portNumber_i );

        /* Let's perform similar operations together rather than performing */
        /* all operations for a device at one go. Helps in Debug */

        /* ask each device to allocate it's private memory */
        for(ctr = 0; ctr < MAX_DEVICES_PER_PORT; ctr++)
        {
            retval = DevAllocatePrivateMemory(portTag_i, &currDev);
            if(retval != SATA_SUCCESS)
            {
                /* Abort Initialization. Fatal System error */
                return retval;
            }
            /* Record the devTag in the devices array */
            currPort->dev[ctr] = currDev;
        }

        /* Now let's initialize each of the device */
        for(ctr = 0; ctr < MAX_DEVICES_PER_PORT; ctr++)
        {
            retval = DevInit(currDev, ctr);
            if(retval != SATA_SUCCESS)
            {
                /* Treat device failures as non fatal. The device will be */
                /* disabled and the initialization will continue.         */
                deviceInitFailed = TRUE;
            }
        }

        if(deviceInitFailed == TRUE)
        {
            /* One or more devices failed during initialization */
            return SATA_COMPLETED_WITH_ERRORS;
        }
        /* Done with device initializations */
    }
    else
    {
        DEBUG_PRINT_3("Nothing connected to Ctlr %d, Port %d",
                      controllers_gp[controller]->controllerNumber, portNumber_i );
    }


    /* Successfully initialized the port. Return success */
    return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function checks if a device is actively connected to the */
/*              port ( Dev present & communication established ). To check    */
/*              for other connection status call PortReadSStatusReg           */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to test.                                 */
/*                                                                            */
/* RETURNS      TRUE    if the device is connected and ready                  */
/*              FALSE   if the device is not ready.                           */
/*                                                                            */
/* ON ERRORS    Returns FALSE. Error condition ignored.                       */
/*                                                                            */
/******************************************************************************/
boolean PortDevConnected(INT portTag_i)
{
    UINT32 sssr = 0;
    INT i;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS) ||
       (ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        DEBUG_PRINT_3("Error: portTag_i %d invalid or not ready. Reporting no device connected.", portTag_i);
        /* SATA_OUT_OF_RANGE_VALUE or SATA_NOT_INITIALIZED */
        return FALSE;
    }
#endif

    /* Set to not connected. Will be changed below... */
    ports_gp[portTag_i]->connected = FALSE;

    DEBUG_PRINT_3("Preparing to read the SUPDSSSR register");
    /* Read the register */
    PciMemRead32( ports_gp[portTag_i]->portRegBase + SUPDSSSR, &sssr);

    DEBUG_PRINT_3("PTag %d : sssr = regs.sssr = 0x%0x", portTag_i, sssr);
/* 0 = no device, 1 = Present but NO communication, 3 = Ready, 4 = offline */
    if((sssr & SUPDSSSR_DEV_DET_MASK) == 0x3)
    {
        ports_gp[portTag_i]->connected = TRUE;
        // mark all devices active since we are connected
        for(i=0; i < MAX_DEVICES_PER_PORT; i++)
        {
            INT devTag = ports_gp[portTag_i]->dev[i];
            DEV_INFO_T * devp = (devTag == INVALID_TAG) ? NULL_POINTER : devices_gp[devTag];

            if(devp != NULL_POINTER)
            {
                DEBUG_PRINT_3("PTag %d : Setting devTag %d active.",
                              portTag_i, devTag);
                devp->active = TRUE;
            }
        }
        return TRUE;
    }
    else
    {
        return FALSE ;
    }
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of AltStatus register      */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadAltStatusReg(INT portTag_i, UINT8 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    //DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead8(ports_gp[portTag_i]->portRegBase + SUPDASR, data_op);

    ports_gp[portTag_i]->lastASR = *data_op;

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of Status register         */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadStatusReg(INT portTag_i, UINT8 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if(ports_gp[portTag_i] == NULL_POINTER ||
       ports_gp[portTag_i]->present != TRUE)
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead8( ports_gp[portTag_i]->portRegBase + SUPDSR, data_op );

    ports_gp[portTag_i]->lastSR = *data_op;

    return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of Control register        */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadControlReg(INT portTag_i, UINT8 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS)) return SATA_OUT_OF_RANGE_VALUE;
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    /* Read the register */
    PciMemRead8(ports_gp[portTag_i]->portRegBase + SUPDDCTLR, data_op);

    return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of Error register          */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadErrorReg(INT portTag_i, UINT8 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        { DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;}
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead8( ports_gp[portTag_i]->portRegBase + SUPDER, data_op );

    ports_gp[portTag_i]->lastER = *data_op;

    return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of Super Set Status        */
/*              SStatus register                                              */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadSStatusReg(INT portTag_i, UINT32 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead32(ports_gp[portTag_i]->portRegBase + SUPDSSSR, data_op);

    ports_gp[portTag_i]->lastSSSR = *data_op;

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of SError register         */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadSErrorReg(INT portTag_i, UINT32 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead32(ports_gp[portTag_i]->portRegBase + SUPDSSER, data_op);

    ports_gp[portTag_i]->lastSSER = *data_op;
    ports_gp[portTag_i]->cumulativeSSER |= *data_op;

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of SControl register       */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadSControlReg (INT portTag_i, UINT32 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead32(ports_gp[portTag_i]->portRegBase + SUPDSSCR, data_op);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function returns the contents of Dma Status register     */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              data_op                                                       */
/*                  Address of the variable to hold the register value        */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadDmaStatusReg(INT portTag_i, UINT8 *data_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read the register */
    PciMemRead8(ports_gp[portTag_i]->portRegBase + SUPDDSR, data_op);

    //DEBUG_PRINT_3("exiting...%s","");
    return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function reads "count_io"  bytes of data from the data   */
/*              port register (SUPDDR), putting them in the destination       */
/*              buffer.                                                       */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              count_i                                                       */
/*                  The number of bytes to be read from the data port reg.    */
/*              buffer_op                                                     */
/*                  Address of the destination buffer. Should be large enough */
/*                  to hold count_i bytes.                                    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortPioReadBytes(INT portTag_i, INT count_i, void * buffer_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif
    DEBUG_PRINT_3("portTag_i %d, About to read DataPort, %d bytes", portTag_i, count_i);

    /* Call the OS Service layer to transfer the data */
    PciMemReadDataPort(count_i, ports_gp[portTag_i]->portRegBase + SUPDDR,
                       buffer_op);

    debug_mem_dump("Dev_Identify_RAW data", buffer_op, count_i);
    DEBUG_PRINT_3("portTag_i %d, Done reading DataPort", portTag_i);
    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function writes "count_io"  bytes of data to the data    */
/*              port register (SUPDDR).                                       */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              count_i                                                       */
/*                  The number of bytes to be written to the data port reg.   */
/*              buffer_ip                                                     */
/*                  Address of the Source buffer.                             */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortPioWriteBytes(INT portTag_i, INT count_i, void * buffer_ip)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Call the OS Service layer to transfer the data */
    PciMemWriteDataPort(count_i, buffer_ip,
                        ports_gp[portTag_i]->portRegBase + SUPDDR);
    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function writes the various parameters of the taskfile   */
/*              to the corresponding port specific registers. This function   */
/*              supports ONLY CHS/LBA28 addressing for now.                   */
/*              Only the lower 8 bits of sectorNumber_i, cylinderLow_i and    */
/*              cylinderHigh_i should be used at this time.                   */
/*                                                                            */
/*   T B D      Decide on parameter contents for LBA48. (not yet supported)   */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              taskfile_ip                                                   */
/*                  Contents of the task file registers. The command register */
/*                  is NOT written.                                           */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortWriteTaskfileParams(INT portTag_i,
                                      SATA_TASKFILE_T * taskfile_ip)
{

    STATUS_CODE_T retval = SATA_SUCCESS;
    SATA_REG_ADDR_T base = NULL_POINTER; // initialize AFTER sanity check

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK_IN_DATA_PATH
    /* Sanity check */
    if(taskfile_ip == NULL_POINTER)
    {
        DEBUG_PRINT_3("ERROR: Tag %d, taskfile_ip is NULL",portTag_i);
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    base = ports_gp[portTag_i]->portRegBase;

    /* Write to the register */
    PciMemWrite16(base + SUPDFR,  taskfile_ip->feature);
    PciMemWrite16(base + SUPDSCR, taskfile_ip->sectorCount);
    PciMemWrite16(base + SUPDSNR, taskfile_ip->sectorNumber);
    PciMemWrite16(base + SUPDCLR, taskfile_ip->cylinderLow);
    PciMemWrite16(base + SUPDCHR, taskfile_ip->cylinderHigh);
    PciMemWrite8(base + SUPDDHR, taskfile_ip->deviceHead);


    DEBUG_PRINT_3("FEATURE: %#08x", taskfile_ip->feature);
    DEBUG_PRINT_3("SECTOR COUNT: %#08x", taskfile_ip->sectorCount);
    DEBUG_PRINT_3("SECTOR NUMBER: %#08x", taskfile_ip->sectorNumber);
    DEBUG_PRINT_3("CYLINDER LOW: %#08x", taskfile_ip->cylinderLow);
    DEBUG_PRINT_3("CYLINDER HIGH: %#08x", taskfile_ip->cylinderHigh);
    DEBUG_PRINT_3("DEVICE HEAD: %#08x", taskfile_ip->deviceHead);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function reads the parameters of the taskfile            */
/*              from the corresponding port specific registers.               */
/*                                                                            */
/*                                                                            */
/*   T B D      Decide on parameter contents for LBA48. (not yet supported)   */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              taskfile_ip                                                   */
/*                  Location to write the task file registers.                */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortReadTaskfileParams(INT portTag_i,
                                     SATA_TASKFILE_T * taskfile_op)
{

    STATUS_CODE_T retval = SATA_SUCCESS;
    SATA_REG_ADDR_T base = ports_gp[portTag_i]->portRegBase;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(taskfile_op == NULL_POINTER)
    {
        DEBUG_PRINT_3("ERROR: Tag %d, taskfile_op is NULL",portTag_i);
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Read from the register */

    PciMemRead16(base + SUPDFR, &taskfile_op->feature);
    PciMemRead16(base + SUPDSCR, &taskfile_op->sectorCount);
    PciMemRead16(base + SUPDSNR, &taskfile_op->sectorNumber);
    PciMemRead16(base + SUPDCLR, &taskfile_op->cylinderLow);
    PciMemRead16(base + SUPDCHR, &taskfile_op->cylinderHigh);
    PciMemRead8(base + SUPDDHR, &taskfile_op->deviceHead);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function sets the value of the Command register          */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              cmd_i                                                         */
/*                  The SATA command to be written to the Command register    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortWriteCommandReg (INT portTag_i, UINT8 cmd_i)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Write to the register */
    PciMemWrite8(ports_gp[portTag_i]->portRegBase + SUPDCR, cmd_i);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function sets the value of the Control register          */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              value_i                                                       */
/*                  The new value to be written to the control register       */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortWriteControlReg (INT portTag_i, UINT8 value_i)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Write to the register */
    PciMemWrite8(ports_gp[portTag_i]->portRegBase + SUPDDCTLR, value_i);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function sets the Descriptor Table Pointer Registers     */
/*              ( Both Upper and lower DTPRs - SUPDUDDTPR, SUPDDDTPR )        */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              upperDTPR_i                                                   */
/*                   The value for SUPDUDDTPR (Upper DDTPR Register) which    */
/*                   contains the upper 32-bit PCI address of the 64-bit PCI  */
/*                   address. In PCI IDE mode, the Upper DDTPR Register is    */
/*                   not used.                                                */
/*              lowerDTPR_i                                                   */
/*                   The value for SUPDDDTPR (Lower DDTPR Register) which     */
/*                   contains the lower 32-bit PCI address.                   */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortSetDTPR(INT portTag_i, UINT32 upperDTPR_i, UINT32 lowerDTPR_i)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Write to the register */
    PciMemWrite32(ports_gp[portTag_i]->portRegBase + SUPDUDDTPR, upperDTPR_i);
    PciMemWrite32(ports_gp[portTag_i]->portRegBase + SUPDDDTPR, lowerDTPR_i);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              Sets the upper 32 bits of the addresses in the PRDT entries   */
/*              in the SUPDUDDBPR register                                    */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              value_i                                                       */
/*                   The value of SUPDUDDBPR which contains the upper 32-bit  */
/*                   PCI address of the 64-bit PCI address. All the           */
/*                   descriptors in the descriptor table share this register. */
/*                   All the DDT Entries share the same upper 32 bits of      */
/*                   their 64 bit addresses.                                  */
/*                   For example, all the data buffers must be located in the */
/*                   same 4-GByte memory space.                               */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortSetDdtePage(INT portTag_i, UINT32 value_i)
{

    STATUS_CODE_T retval = SATA_SUCCESS;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS)) return SATA_OUT_OF_RANGE_VALUE;
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    /* Write to the register */
    PciMemWrite32(ports_gp[portTag_i]->portRegBase + SUPDUDDBPR, value_i);

    return retval;
}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              writes to the DMA command register (SUPDDCMDR) for that port. */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*              command_i                                                     */
/*                  The value of SUPDDCMDR which enables/disables the DMA     */
/*                  engine (bus master capability) and also provides          */
/*                  direction control for DMA transfers.                      */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortWriteDmaCommandReg(INT portTag_i, UINT16 command_i)
{
    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    /* Write to the register */
    PciMemWrite16(ports_gp[portTag_i]->portRegBase + SUPDDCMDR, command_i);

    return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function checks if an interrupt is pending on the port.  */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*                                                                            */
/* RETURNS      TRUE / FALSE                                                  */
/*                                                                            */
/* ON ERRORS    returns FALSE                                                 */
/*                                                                            */
/******************************************************************************/
boolean PortInterruptPending(INT portTag_i)
{

    INT controller;
    SATA_REG_ADDR_T base;
    UINT32 ipr;
//  STATUS_CODE_T retval = SATA_SUCCESS;
//  UINT8 dsr;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    controller = ports_gp[portTag_i]->controller;
    base = controllers_gp[controller]->commonRegBase;

    /* Read the register */
    PciMemRead32(base + SUPDIPR, &ipr);

    /* Move the interrupt bits relevant to this port in lower byte */
    ipr >>= SUPDIPR_BITS_PER_PORT * ports_gp[portTag_i]->portNumber;
    /* Mask off all bits not applicable to this port */
    ipr &=  SUPDIPR_BITS_PER_PORT_MASK;

    if(ipr > 0)
    {
        return TRUE; /* Some interrupt bits were set */
    }

    // NPG_TEST. Should not have to check anything other than IPR
//
//    /* Check the DMA status register for a pending DMA Interrupt */
//     retval = PortReadDmaStatusReg(portTag_i, &dsr);
//     if(retval != SATA_SUCCESS) return FALSE;
//
//     DEBUG_PRINT_3("reading DMA interrupt bit%s","");
//     if(READ_BIT(dsr, SUPDDSR_DMA_INTR_BIT)) return TRUE;
//
    //DEBUG_PRINT_3("exiting...%s","");
    /* No indication of a pending interrupt.... */
    return FALSE;

}

/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function clears the interrupt for the given port.        */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortClearInterrupt(INT portTag_i)
{

    STATUS_CODE_T retval = SATA_SUCCESS;
    INT controller;
    SATA_REG_ADDR_T base;
    UINT32 serr;
    UINT8 stat;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    controller = ports_gp[portTag_i]->controller;
    base = controllers_gp[controller]->commonRegBase;

    /* Clear the SATA Port IDE Interrupt by reading the taskfile status register */
    retval = PortReadStatusReg(portTag_i, &stat);
    if(retval != SATA_SUCCESS) return retval;

#ifdef EXPERIMENTAL_CODE
// The Idea was that if after clearing the IDE interrupt, there are no pending interrupts
// then we need not access the DMA status reg and the SERR reg.
// This could have saved us a PCI_8bit write.
// Apparently, this causes a lot of performance degradation. We haven't investigated this
// throughly, but it's best not to skip writing to DDSR and SSER

// {
//     UINT32 ipr;
//     /* Read the register */
//     PciMemRead32(base + SUPDIPR, &ipr);
//
//     /* Move the interrupt bits relevant to this port in lower byte */
//     ipr >>= SUPDIPR_BITS_PER_PORT * ports_gp[portTag_i]->portNumber;
//     /* Mask off all bits not applicable to this port */
//     ipr &=  SUPDIPR_BITS_PER_PORT_MASK;
//
//     if(ipr <= 0)
//     { /* None of the interrupt bits were set */
//         return SATA_SUCCESS;
//     }
// }
#endif

    /* Reset the DMA interrupt */
    stat = 0;
    stat = SET_BIT( 0, SUPDDSR_DMA_INTR_BIT);
    PciMemWrite8(ports_gp[portTag_i]->portRegBase + SUPDDSR, stat);

    /* Clear the rest of the interrupts by writing to SError register. */
    serr = 0; /* reset all bits to 0 */
    serr = SET_BIT(serr, SUPDSSER_DIAG_N_BIT);
    serr = SET_BIT(serr, SUPDSSER_ERR_M_BIT);
    serr = SET_BIT(serr, SUPDSSER_ERR_E_BIT);
    serr = SET_BIT(serr, SUPDSSER_DIAG_H_BIT);
    serr = SET_BIT(serr, SUPDSSER_ERR_P_BIT);
    serr = SET_BIT(serr, SUPDSSER_ERR_T_BIT);
    serr = SET_BIT(serr, SUPDSSER_DIAG_C_BIT);

    PciMemWrite32(ports_gp[portTag_i]->portRegBase + SUPDSSER, serr);

    return retval;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function disables the interrupts for the given port.     */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortDisableInterrupts(INT portTag_i)
{
    PORT_INFO_T * port = NULL_POINTER;
    INT controller;
    SATA_REG_ADDR_T base;
    UINT32 mask, imr;
    UINT8 dctlr;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if(ports_gp[portTag_i] == NULL_POINTER ||
       ports_gp[portTag_i]->present != TRUE)
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    port = ports_gp[portTag_i];
    controller = ports_gp[portTag_i]->controller;
    base = controllers_gp[controller]->commonRegBase;

    /* Interrupts are enabled/disabled in a different way in PCI IDE mode       */
    /* and in DPA mode. In PCI IDE mode, you set the nIEN bit in Device Control */
    /* Register (SUPDCTLR) to mask the interrupts, and in DPA mode, you set the */
    /* correct bits in the interrupt mask registers (SUPDIMR)                   */

    if(controllers_gp[controller]->mode == PCI_DPA)
    {
        /* set the correct bits in the interrupt mask registers (SUPDIMR)      */
        /* 8 bits per port, set the bits to 0 to mask them. Set all the rest   */
        /* of the bits to 1. C-Spec Ver 0.94 page 171                          */

        /* Step 1 - Set the correct bit to 1 */
        mask = SET_BIT(0, SUPDIPR_IDE_INTR_BIT);

        /* 8 interrupt bits per port. Look at C-Spec 0.94, pg 167 for defn.    */
        /* Step 2 - Move the mask bit to the right posn for the current port   */
        mask = mask << ( 8 * port->portNumber);

        /* Step 3 - Bit inverse as 1 = enabled, 0 = disabled */
        /* mask ^= 0xffffffff; */
        mask = ~mask;

        /* Read the current values in IMR.... We don't want to enable interrupts*/
        /* for other ports by mistake !!! */

        /* Read the register */
        PciMemRead32(base + SUPDIMR, &imr);

        /* Step 4 - Modify the current values so that the 0s stay  */
        mask = mask & imr;

        /* write to the register */
        PciMemWrite32(base + SUPDIMR, mask);
    }
    else
    {
        /* PCI IDE mode / Legacy mode, write a 1 to the nIEN bit to disable the  */
        /* interrupt, and 0s to rest of the bits so that you don't change their  */
        /* values... Remember, SUPDDCTLR is a WRITE_ONLY register. You can't     */
        /* read-modify-write to it.                                              */
        dctlr = SET_BIT(0, SUPDDCTLR_nIEN_BIT);

        /* Write to the register */
        PciMemWrite8(base + SUPDDCTLR, dctlr);
    }

    /* Done.*/
    return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function enables the interrupts for the given port.      */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortEnableInterrupts(INT portTag_i)
{
    PORT_INFO_T * port   = NULL_POINTER;
    INT controller;
    SATA_REG_ADDR_T base;
    UINT32 mask, imr ;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    port = ports_gp[portTag_i];
    controller = ports_gp[portTag_i]->controller;
    base = controllers_gp[controller]->commonRegBase;

    /* Interrupts are enabled/disabled in a different way in PCI IDE mode       */
    /* and in DPA mode. In PCI IDE mode, you set the nIEN bit in Device Control */
    /* Register (SUPDCTLR) to mask the interrupts, and in DPA mode, you set the */
    /* correct bits in the interrupt mask registers (SUPDIMR)                   */

    PortClearInterrupt(portTag_i);
    PortClearError(portTag_i);

    if(controllers_gp[controller]->mode == PCI_DPA)
    {
        /* set the correct bits in the interrupt mask registers (SUPDIMR)      */
        /* 8 bits per port, set the bits to 0 to mask them. Set all the rest   */
        /* of the bits to 1. C-Spec Ver 0.94 page 171                          */
        mask = SET_BIT(0, SUPDIPR_IDE_INTR_BIT);

        /* Enable the Phy Change state interrupt. This is required for Hot plug*/
        mask = SET_BIT(mask, SUPDIPR_STATE_INTR_BIT);

        /* 8 interrupt bits per port. Look at C-Spec 0.94, pg 167 for defn.    */
        /* Move the mask bit to the right position for the current port number */
        mask = mask << ( 8 * port->portNumber);

        /* Read the current values in IMR.... We don't want to enable interrupts*/
        /* for other ports by mistake !!! */

        /* Read the register */
        PciMemRead32(base + SUPDIMR, &imr);

        /* Step 4 - Modify the current values so that the 1s stay  */
        mask = mask | imr;

        /* write to the register */
        PciMemWrite32(base + SUPDIMR, mask);
    }
    else
    {
        /* PCI IDE mode / Legacy mode, write a 0 to the nIEN bit to enable the  */
        /* interrupt again, and 0s to rest of the bits so that you don't change */
        /* their values... Remember, SUPDDCTLR is a WRITE_ONLY register. You    */
        /* can't read-modify-write to it.                                       */

        /* Write to the register */
        PciMemWrite8(base + SUPDDCTLR, 0);
    }

    /* Done.*/
    return SATA_SUCCESS;
}


/******************************************************************************/
/*                                                                            */
/* DESCRIPTION                                                                */
/*              This function clears the errors for the given port.           */
/*                                                                            */
/* PRECONDITIONS    port has been initialized.                                */
/*                                                                            */
/* POST-CONDITIONS  N/A                                                       */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to use.                                  */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/* ON ERRORS    returns the following error codes                             */
/*              SATA_OUT_OF_RANGE_VALUE = Invalid Tag                         */
/*              SATA_NOT_INITIALIZED    = Port not initialized                */
/*              Other Error codes on System errors                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortClearError(INT portTag_i)
{

    INT controllerTag;
//    SATA_REG_ADDR_T base;
    UINT32 serr;
    UINT8 stat;
    PORT_INFO_T * currPort = NULL_POINTER;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d",portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        { DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized",portTag_i);  return SATA_NOT_INITIALIZED;}
    }
#endif

    currPort = ports_gp[portTag_i];
    controllerTag = currPort->controller;
    //base = currPort->portRegBase; //  controllers_gp[controllerTag]->commonRegBase;


    /* Clear the SUPDSSER */
    /* Read the SError reg so that the Last read value is recorded. */
    PortReadSErrorReg(portTag_i,&serr);
    // Now, let's clear all the non-reserved error bits by writing 1s to them
    serr = 0;
    serr = SET_BIT(serr,SUPDSSER_DIAG_F_BIT); // Unknown FIS
    serr = SET_BIT(serr,SUPDSSER_DIAG_H_BIT); // R_ERR recieved. Handshake err
    serr = SET_BIT(serr,SUPDSSER_DIAG_C_BIT); // CRC error detected
    serr = SET_BIT(serr,SUPDSSER_DIAG_D_BIT); // Disparity error detected
    serr = SET_BIT(serr,SUPDSSER_DIAG_W_BIT); // Comm Wake detected
    serr = SET_BIT(serr,SUPDSSER_DIAG_N_BIT); // PHYRDY Change state detected
    serr = SET_BIT(serr,SUPDSSER_ERR_E_BIT);  // Inter err: FIFO over/underrun
    serr = SET_BIT(serr,SUPDSSER_ERR_P_BIT);  // Protocol err: Corrupt FIS
    serr = SET_BIT(serr,SUPDSSER_ERR_C_BIT);  // Non Recovered Comm. No signal
    serr = SET_BIT(serr,SUPDSSER_ERR_T_BIT);  // Transient Data Integrity err
    serr = SET_BIT(serr,SUPDSSER_ERR_M_BIT);  // Recovered comm err PHY Ready

    /* Write the register */
    PciMemWrite32(currPort->portRegBase + SUPDSSER, serr);

    /* Clear the SUPDDSR Bit 01 DMA Err by writing a 1 to it. */
    stat = 0;
    stat = SET_BIT(0, SUPDDSR_DMA_ERROR_BIT);

    /* Write the register */
    PciMemWrite8(currPort->portRegBase + SUPDDSR, stat);

    /* Clear SUPDSR (Bit 0 ERR ) - Auto cleared when the next command issued */
    /* Clear SUPDER - Automatically cleared when the next command is issued. */
    /* don't bother with BIST FIS CSR - SUPDBFCSR(Bit 25 )? */

    /* All errors cleared.*/
    return SATA_SUCCESS;
}

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortIntHardReset()                                            */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function resets the SATA interface and bring            */
/*               the PHY online. Should be called in order to bring PHY       */
/*               online. The BIST done by software will call this after test  */
/*                                                                            */
/* PRECONDITIONS    The port is ready for BIST at initialization              */
/*                                                                            */
/* POST-CONDITIONS  The PHY is offline or in BIST mode                        */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortIntHardReset(INT portTag_i)
{
    UINT32 reset = 0;
    INT controllerTag;
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base;
    UINT32 sssr = 0;
    INT count = 0;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        DEBUG_PRINT_3("ERROR: Invalid Tag %d", portTag_i); return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        DEBUG_PRINT_3("ERROR: Tag %d, Not Initialized", portTag_i);  return SATA_NOT_INITIALIZED;
    }
#endif

    currPort = ports_gp[portTag_i];
    controllerTag = currPort->controller;
//    base = controllers_gp[controllerTag]->commonRegBase;
    base = currPort->portRegBase; // Address of the port specific region

    /* bring PHY online and start COMRESET init sequence */
    reset = 0x1;
    PciMemWrite32(currPort->portRegBase + SUPDSSCR, reset);
    DEBUG_PRINT_3("Initiate COMRESET...%s","");
    MillisecondSleep(500);

    /*  stop initialization and bring into online state */
    reset = 0;
    PciMemWrite32(currPort->portRegBase + SUPDSSCR, reset);

    count = 2;
    do
    {
        MillisecondSleep(500);
        /* read the status */
        PciMemRead32(currPort->portRegBase + SUPDSSSR, &sssr);
    }
    while(!(sssr & SUPDSSSR_DEV_ONLINE) && --count);

    /* if device status not present and online */
    if(!(sssr & SUPDSSSR_DEV_ONLINE))
    {
        DEBUG_PRINT_3("COMRESET failed for controller %d port %d. sssr = 0x%0x ",
                      controllerTag, portTag_i, sssr);
        return SATA_NOT_INITIALIZED;
    }

    DEBUG_PRINT_3("COMRESET Succeeded for controller %d port %d",
                  controllerTag, portTag_i);

    return SATA_SUCCESS;
} /* end of PortIntHardReset() */

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortBistStatus()                                              */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function starts the built in self test (BIST) by        */
/*               doing the far end retimed loopback test using the 16bit      */
/*               counting pattern. The function puts the hardware into PHY    */
/*               online mode after doing a COMRESET                           */
/*                                                                            */
/* PRECONDITIONS    The port is ready for BIST at initialization              */
/*                                                                            */
/* POST-CONDITIONS  The PHY is online and ready                               */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortBistStatus(INT portTag_i)
{
    STATUS_CODE_T retval = SATA_SUCCESS;
//    UINT32 reset = 0;
    INT controllerTag;
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base;
    UINT32 bistCsr = 0;
    UINT32 bistErr = 0;
    UINT32 bistCount = 0;
    INT timeout = 0;

    DEBUG_TRACE_3();

    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }

    currPort = ports_gp[portTag_i];
    controllerTag = currPort->controller;
    base = currPort->portRegBase; // controllers_gp[controllerTag]->commonRegBase;

    /* set bit 25 to clear frame and error */
    bistCsr = SUPDBFCSR_CLR_BIT;

    PciMemWrite32(base + SUPDBFCSR, bistCsr);

    bistCsr = SUPDBFCSR_PATTERN |   // set the test pattern
              SUPDBFCSR_RLOOPBACK |       // set retimed loopback
              SUPDBFCSR_PATGEN |          // set pattern generator
              SUPDBFCSR_PATCHK;           // set pattern checker

    PciMemWrite32(base + SUPDBFCSR, bistCsr);

    do
    {
        MillisecondSleep(100);  // delay for 100ms to let the BIST run
        PciMemRead32(base + SUPDBFR, &bistCount);

        if(timeout++ > 10)
        {
            DEBUG_PRINT_3("BIST timed out without starting...bail%s","");
            return SATA_FAILURE;
        }

    } while(bistCount < 10); // as long as frame count < 10, we keep running

    /* read the error count to check for errors */
    PciMemRead32(base + SUPDBER, &bistErr);

    /* bist detected failure */
    if(bistErr)
    {
        DEBUG_PRINT_3("BIST failed with count: %d\n", bistErr);
        return SATA_FAILURE;
    }

    /* Is this needed if we are resetting the interface? */
    bistCsr = 0;
    PciMemWrite32(base + SUPDBFCSR, bistCsr);

    // Now we need to do a COMRESET/COMINIT sequence to conclude the test

    /* hardreset to stop BIST */
    if((retval = PortIntHardReset(portTag_i)) != SATA_SUCCESS)
    {
        return retval;
    }

    return SATA_SUCCESS;
} /* end of PortBistStatus() */

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortSetSActive()                                              */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function sets the native queuing slot/bit in the        */
/*               SActive register.                                            */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*              slot_i                                                        */
/*                  which slot to activate                                    */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/******************************************************************************/
STATUS_CODE_T PortSetSActive(INT portTag_i, UINT32 slot_i)
{
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base, bar0;
    UINT32 tx_link_stateMachine = 0xffffffff;
    UINT32 counter = 500;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    currPort = ports_gp[portTag_i];
    base = currPort->portRegBase;
    bar0 = controllers_gp[currPort->controller]->commonRegBase;

    //
    // Work around for the Silicon errata to correct the timing problem
    // between the Set Device Bits (SDB) FIS and a PCI/PCI-X Write of
    // the SDB FIS reigster.
    //
    // Check the trasport/link state machine to make sure that it is
    // in an idle state before issuing the command.  The trasport/link
    // state machine register is located at BAR0+1d8H.  The lower 8 bits
    // of this register need to be 00 before the write.  This is a
    // reserved fied.
    //
    while(((tx_link_stateMachine & RX_LINK_STATEMACHINE_MASK) != 0) &&
          (counter-- > 0))
    {
        PciMemRead32(bar0+SUPDTR2, &tx_link_stateMachine);
    }

    PciMemWrite32(base + SUPDSAR, SET_BIT(0, slot_i));

    return SATA_SUCCESS;
} /* end of PortSetSActive() */

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortReadSActive()                                             */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function reads the native queuing SActive register.     */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*              regVal_op                                                     */
/*                  SActive register value to be written back                 */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortReadSActive(INT portTag_i, UINT32 *regVal_op)
{
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    currPort = ports_gp[portTag_i];
    base = currPort->portRegBase;

    PciMemRead32(base + SUPDSAR, regVal_op);

    return SATA_SUCCESS;
} /* end of PortReadSActive() */

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortEnableNativeQDma()                                        */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function enables First Party DMA auto processing        */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortEnableNatvieQDma(INT portTag_i)
{
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    currPort = ports_gp[portTag_i];
    base = currPort->portRegBase;

    // BIT 27 is the "Enable first party DMA auto processing for CQ
    PciMemWrite32(base + SUPDDSFCSR, SET_BIT(0, SUPDDSFCSR_CQDA_BIT));

    return SATA_SUCCESS;
} /* End of PortEnableNativeQDma() */

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortDisableNativeQDma()                                       */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function disables First Party DMA auto processing       */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortDisableNatvieQDma(INT portTag_i)
{
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base;
    UINT32 csr;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    currPort = ports_gp[portTag_i];
    base = currPort->portRegBase;

    PciMemRead32(base + SUPDDSFCSR, &csr);

    if(csr & SET_BIT(0, SUPDDSFCSR_CQDA_BIT))
    {
        // Unset the BIT 27 to disable first party DMA auto processing
        csr &= ~SET_BIT(0, SUPDDSFCSR_CQDA_BIT);
        PciMemWrite32(base + SUPDDSFCSR, csr);
    }

    return SATA_SUCCESS;
} /* End of PortDisableNativeQDma() */

/******************************************************************************/
/*                                                                            */
/* FUNCTION     PortSetDDTVectorTable()                                       */
/*                                                                            */
/* DESCRIPTION                                                                */
/*               This function sets the DDT vector table address to the       */
/*               GD31244 SUPDLVTBAR                                           */
/*                                                                            */
/* PRECONDITIONS                                                              */
/*                                                                            */
/* POST-CONDITIONS                                                            */
/*                                                                            */
/* PARAMETERS   portTag_i                                                     */
/*                  Tag for the port to be used.                              */
/*              vector                                                        */
/*                  Address for the vector table                              */
/*                                                                            */
/* RETURNS      SATA_SUCCESS on success, error code on failure                */
/*                                                                            */
/******************************************************************************/
INLINE STATUS_CODE_T PortSetDDTVectorTable(INT portTag_i, UINT64 vector)
{
    UINT32 csr = 0;
    PORT_INFO_T * currPort = NULL_POINTER;
    SATA_REG_ADDR_T base;

    DEBUG_TRACE_3();

#ifdef SANITY_CHECK
    /* Sanity check */
    if(!TAG_WITHIN_RANGE(portTag_i, MAX_PORTS))
    {
        return SATA_OUT_OF_RANGE_VALUE;
    }
    if((ports_gp[portTag_i] == NULL_POINTER) ||
       (ports_gp[portTag_i]->present != TRUE))
    {
        return SATA_NOT_INITIALIZED;
    }
#endif

    currPort = ports_gp[portTag_i];
    base = currPort->portRegBase;

    DEBUG_PRINT("vector passed in: %#08x", (UINT32)vector);
    DEBUG_PRINT("lower 32: %#08x", LOWER_32_BITS(vector));
    DEBUG_PRINT("upper 32: %#08x", UPPER_32_BITS(vector));

    PciMemWrite32(base + SUPDLVTBAR, LOWER_32_BITS(vector));
    PciMemWrite32(base + SUPDUVTBAR, UPPER_32_BITS(vector));

    // I might as well enable queuing auto processing here
    PciMemRead32(base + SUPDDSFCSR, &csr);

    csr = SET_BIT(csr, 27);

    PciMemWrite32(base + SUPDDSFCSR, csr);
    return SATA_SUCCESS;
} /* end of PortSetDDTVectorTable() */
